<?php
/** @license MIT
 * Copyright 2018 J. King et al.
 * See LICENSE and AUTHORS files for details */

declare(strict_types=1);
namespace MensBeam\Intl\TestCase\Encoding;

use MensBeam\Intl\Encoding\Big5;
use MensBeam\Intl\Encoding\Encoding;
use MensBeam\Intl\Encoding\EncoderException;

class TestBig5 extends \MensBeam\Intl\Test\CoderDecoderTest {
    protected $testedClass = Big5::class;
    /*
        Char 0 U+007A   (1 byte)  Offset 0
        Char 1 U+86CC   (2 bytes) Offset 1
        Char 2 U+6C34   (2 bytes) Offset 3
        Char 3 U+00CA   (0 bytes) Offset 5
        Char 4 U+0304   (2 bytes) Offset 5
        Char 5 U+00EA   (0 bytes) Offset 7
        Char 6 U+030C   (2 bytes) Offset 7
        End of string at char 7, offset 9
    */
    protected $seekString = "7A D7AA A4F4 8862 88A5";
    protected $seekCodes = [0x7A, 0x86CC, 0x6C34, 0xCA, 0x304, 0xEA, 0x30C];
    protected $seekOffsets = [0, 1, 3, 5, 5, 7, 7, 9];
    /* This string contains an invalid character sequence sandwiched between two null characters */
    protected $brokenChar = "00 FF 00";

    public function provideCodePoints() {
        return [
            'U+0064 (HTML)'    => [false, 0x64, "64"],
            'U+0064 (fatal)'   => [true,  0x64, "64"],
            'U+00CA (HTML)'    => [false, 0xCA, bin2hex("&#202;")],
            'U+00CA (fatal)'   => [true,  0xCA, new EncoderException("", Encoding::E_UNAVAILABLE_CODE_POINT)],
            'U+3007 (HTML)'    => [false, 0x3007, "C6 E2"],
            'U+3007 (fatal)'   => [true,  0x3007, "C6 E2"],
            'U+5341 (HTML)'    => [false, 0x5341, "A4 51"],
            'U+5341 (fatal)'   => [true,  0x5341, "A4 51"],
            'U+2561 (HTML)'    => [false, 0x2561, "F9 EB"],
            'U+2561 (fatal)'   => [true,  0x2561, "F9 EB"],
            'U+256D (HTML)'    => [false, 0x256D, "A2 7E"],
            'U+256D (fatal)'   => [true,  0x256D, "A2 7E"],
            '-1 (HTML)'        => [false, -1, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
            '-1 (fatal)'       => [true,  -1, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
            '0x110000 (HTML)'  => [false, 0x110000, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
            '0x110000 (fatal)' => [true,  0x110000, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
        ];
    }

    public function provideStrings() {
        return [
            'empty string' => ["", []],
            'sanity check' => ["40", [64]],
            'two-byte character' => ["D7 D7", [36290]],
            'EOF after first byte' => ["D7", [65533]],
            'low byte after first byte' => ["D7 39", [65533, 57]],
            '0x80 as first byte' => ["80 D7 00", [65533, 65533, 0]],
            '0xFF as first byte' => ["FF D7 00", [65533, 65533, 0]],
            'invalid high byte as first byte' => ["81 D7 00", [65533, 0]],
            '0x7F after first byte' => ["D7 7F", [65533, 127]],
            '0xFF after first byte' => ["D7 FF", [65533]],
            'invalid high byte after first byte' => ["D7 81", [65533]],
            'broken string' => ["00 FF 00", [0, 65533, 0]],
            'double-characters low' => ["88 62 88 64", [202, 772, 202, 780]],
            'double-characters high' => ["88 A3 88 A5", [234, 772, 234, 780]],
            'mixed string' => ["7A D7 AA A4 F4 88 62 88 A5", [122, 34508, 27700, 202, 772, 234, 780]],
            'mixed string 2' => ["62 D7 D7 D7 D7 62", [98, 36290, 36290, 98]],
        ];
    }

    /**
     * @dataProvider provideCodePoints
     * @covers MensBeam\Intl\Encoding\Encoder
     * @covers MensBeam\Intl\Encoding\Big5::encode
     * @covers MensBeam\Intl\Encoding\Big5::errEnc
     */
    public function testEncodeCodePoints(bool $fatal, $input, $exp) {
        return parent::testEncodeCodePoints($fatal, $input, $exp);
    }

    /**
     * @dataProvider provideCodePoints
     * @covers MensBeam\Intl\Encoding\Big5::encode
     * @covers MensBeam\Intl\Encoding\Big5::errEnc
     */
    public function testEncodeCodePointsStatically(bool $fatal, $input, $exp) {
        return parent::testEncodeCodePointsStatically($fatal, $input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\Big5::__construct
     * @covers MensBeam\Intl\Encoding\Big5::nextCode
     */
    public function testDecodeMultipleCharactersAsCodePoints(string $input, array $exp) {
        return parent::testDecodeMultipleCharactersAsCodePoints($input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\Big5::__construct
     * @covers MensBeam\Intl\Encoding\Big5::nextChar
     */
    public function testDecodeMultipleCharactersAsStrings(string $input, array $exp) {
        return parent::testDecodeMultipleCharactersAsStrings($input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\Big5::seekBack
     */
    public function testSTepBackThroughAString(string $input, array $exp) {
        return parent::testSTepBackThroughAString($input, $exp);
    }

    /**
     * @covers MensBeam\Intl\Encoding\Big5::seek
     * @covers MensBeam\Intl\Encoding\Big5::posChar
     * @covers MensBeam\Intl\Encoding\Big5::posByte
     * @covers MensBeam\Intl\Encoding\Big5::rewind
     */
    public function testSeekThroughAString() {
        return parent::testSeekThroughAString();
    }

    /**
     * @covers MensBeam\Intl\Encoding\Big5::posChar
     * @covers MensBeam\Intl\Encoding\Big5::posByte
     * @covers MensBeam\Intl\Encoding\Big5::eof
     */
    public function testTraversePastTheEndOfAString() {
        return parent::testTraversePastTheEndOfAString();
    }

    /**
     * @covers MensBeam\Intl\Encoding\Big5::peekChar
     * @covers MensBeam\Intl\Encoding\Big5::stateSave
     * @covers MensBeam\Intl\Encoding\Big5::stateApply
     */
    public function testPeekAtCharacters() {
        return parent::testPeekAtCharacters();
    }

    /**
     * @covers MensBeam\Intl\Encoding\Big5::peekCode
     * @covers MensBeam\Intl\Encoding\Big5::stateSave
     * @covers MensBeam\Intl\Encoding\Big5::stateApply
     */
    public function testPeekAtCodePoints() {
        return parent::testPeekAtCodePoints();
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\Big5::lenChar
     * @covers MensBeam\Intl\Encoding\Big5::lenByte
     * @covers MensBeam\Intl\Encoding\Big5::stateSave
     * @covers MensBeam\Intl\Encoding\Big5::stateApply
     */
    public function testGetStringLength(string $input, array $points) {
        return parent::testGetStringLength($input, $points);
    }

    /**
     * @covers MensBeam\Intl\Encoding\Big5::errDec
     */
    public function testReplacementModes() {
        return parent::testReplacementModes();
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\Big5::rewind
     * @covers MensBeam\Intl\Encoding\Big5::chars
     * @covers MensBeam\Intl\Encoding\Big5::codes
     */
    public function testIterateThroughAString(string $input, array $exp) {
        return parent::testIterateThroughAString($input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @coversNothing
     */
    public function testIterateThroughAStringAllowingSurrogates(string $input, array $strictExp, array $relaxedExp = null) {
        return parent::testIterateThroughAStringAllowingSurrogates($input, $strictExp, $relaxedExp);
    }

    /**
     * @covers MensBeam\Intl\Encoding\Big5::seekBack
     */
    public function testSeekBackOverRandomData() {
        return parent::testSeekBackOverRandomData();
    }

    /**
     * @group optional
     */
    public function testPedanticallyDecodeSingleCharactersAsCodePoint() {
        $series = [
            'standard characters' => [["A1B1","C6D8","A1C2","A258","A1D3","A150","A1D1","A1D2","C8FB","C8FC","C8FA","C8F6","C8F8","C8F7","C8FE","C8F9","C8F5","C8FD","C6D9","A3BE","A3BC","A3BD","A3BF","A1C5","A3BB","A344","A345","A346","A347","A348","A349","A34A","A34B","A34C","A34D","A34E","A34F","A350","A351","A352","A353","A354","A355","A356","A357","A358","A359","A35A","A35B","A35C","A35D","A35E","A35F","A360","A361","A362","A363","A364","A365","A366","A367","A368","A369","A36A","A36B","A36C","A36D","A36E","A36F","A370","A371","A372","A373","C7F9","C7F3","C7F4","C7F5","C7F6","C7F7","C7F8","C7FA","C7FB","C7FC","C7FD","C7FE","C840","C841","C842","C843","C844","C845","C846","C847","C848","C849","C84A","C84B","C84C","C84D","C84E","C84F","C850","C851","C852","C853","C854","C855","C856","C857","C858","C859","C85A","C85C","C85D","C85E","C85F","C860","C861","C862","C863","C864","C865","C866","C867","C868","C869","C86A","C86B","C86C","C86D","C86E","C86F","C870","C871","C872","C873","C874","C875","C85B","A156","A158","A1A5","A1A6","A1A7","A1A8","A14C","A14B","A145","A1AC","A1AB","A1B0","A3E1","A24A","A1C1","A24B","C8D2","C8D3","A2B9","A2BA","A2BB","A2BC","A2BD","A2BE","A2BF","A2C0","A2C1","A2C2","C6B5","C6B6","C6B7","C6B8","C6B9","C6BA","C6BB","C6BC","C6BD","C6BE","A1F6","A1F4","A1F7","A1F5","A1F8","A1F9","A1FB","A1FA","C877","C878","C876","A241","A1D4","A1DB","A1E8","A1E7","A1FD","A1FC","A1E4","A1E5","A1EC","A1ED","A1EF","A1EE","A1DC","A1DA","A1DD","A1D8","A1D9","A1F2","A1F3","A1E6","A1E9","A3C0","A3C1","A3C2","A3C3","A3C4","A3C5","A3C6","A3C7","A3C8","A3C9","A3CA","A3CB","A3CC","A3CD","A3CE","A3CF","A3D0","A3D1","A3D2","A3D3","A3D4","A3D5","A3D6","A3D7","A3D8","A3D9","A3DA","A3DB","A3DC","A3DD","A3DE","A3DF","A3E0","C6A1","C6A2","C6A3","C6A4","C6A5","C6A6","C6A7","C6A8","C6A9","C6AA","C6AB","C6AC","C6AD","C6AE","C6AF","C6B0","C6B1","C6B2","C6B3","C6B4","A277","A278","A27A","A27B","A27C","A27D","A275","A274","A273","A272","A271","F9F9","F9F8","F9E6","F9EF","F9DD","F9E8","F9F1","F9DF","F9EC","F9F5","F9E3","F9EE","F9F7","F9E5","F9E9","F9F2","F9E0","F9EB","F9F4","F9E2","F9E7","F9F0","F9DE","F9ED","F9F6","F9E4","F9EA","F9F3","F9E1","A27E","A2A1","A2A3","A2A2","A2AC","A2AD","A2AE","A15A","A262","A263","A264","A265","A266","A267","A268","A269","A270","A26F","A26E","A26D","A26C","A26B","A26A","A276","A279","A1BD","A1BC","A1B6","A1B5","A1BF","A1BE","A1BB","A1BA","A1B3","A1B7","A1B4","A2A8","A2A9","A2AB","A2AA","A1B9","A1B8","A1F0","A1F1","C6E6","C8D6","C8D7","C8D8","C8D9","C8DA","C8DB","C8DC","C8DD","C8DE","C8DF","C8E0","C8E1","C8E2","C8E3","C8E4","C8E5","C8E6","C8E7","C8E8","C8E9","C8EA","C8EB","C8EC","C8ED","C8EE","C8EF","C8F0","C8F1","C6CD","A140","A142","A143","A1B2","C6E0","C6E1","C6E2","A171","A172","A16D","A16E","A175","A176","A179","A17A","A169","A16A","A245","A165","A166","A1A9","A1AA","A2C3","A2C4","A2C5","A2C6","A2C7","A2C8","A2C9","A2CA","A2CB","C6E7","C6E8","C6E9","C6EA","C6EB","C6EC","C6ED","C6EE","C6EF","C6F0","C6F1","C6F2","C6F3","C6F4","C6F5","C6F6","C6F7","C6F8","C6F9","C6FA","C6FB","C6FC","C6FD","C6FE","C740","C741","C742","C743","C744","C745","C746","C747","C748","C749","C74A","C74B","C74C","C74D","C74E","C74F","C750","C751","C752","C753","C754","C755","C756","C757","C758","C759","C75A","C75B","C75C","C75D","C75E","C75F","C760","C761","C762","C763","C764","C765","C766","C767","C768","C769","C76A","C76B","C76C","C76D","C76E","C76F","C770","C771","C772","C773","C774","C775","C776","C777","C778","C779","C77A","C8D4","C8D5","C6DC","C6DD","C77B","C77C","C77D","C77E","C7A1","C7A2","C7A3","C7A4","C7A5","C7A6","C7A7","C7A8","C7A9","C7AA","C7AB","C7AC","C7AD","C7AE","C7AF","C7B0","C7B1","C7B2","C7B3","C7B4","C7B5","C7B6","C7B7","C7B8","C7B9","C7BA","C7BB","C7BC","C7BD","C7BE","C7BF","C7C0","C7C1","C7C2","C7C3","C7C4","C7C5","C7C6","C7C7","C7C8","C7C9","C7CA","C7CB","C7CC","C7CD","C7CE","C7CF","C7D0","C7D1","C7D2","C7D3","C7D4","C7D5","C7D6","C7D7","C7D8","C7D9","C7DA","C7DB","C7DC","C7DD","C7DE","C7DF","C7E0","C7E1","C7E2","C7E3","C7E4","C7E5","C7E6","C7E7","C7E8","C7E9","C7EA","C7EB","C7EC","C7ED","C7EE","C7EF","C7F0","C7F1","C7F2","C6E3","C6DA","C6DB","A374","A375","A376","A377","A378","A379","A37A","A37B","A37C","A37D","A37E","A3A1","A3A2","A3A3","A3A4","A3A5","A3A6","A3A7","A3A8","A3A9","A3AA","A3AB","A3AC","A3AD","A3AE","A3AF","A3B0","A3B1","A3B2","A3B3","A3B4","A3B5","A3B6","A3B7","A3B8","A3B9","A3BA","C879","C8D1","A1C0","A255","A256","A250","A251","A252","A254","A257","A253","A1EB","A1EA","A24F","FA68","FB70","FB7A","FB5E","FC66","FA48","FB65","FB7D","FE64","FDA2","FBB6","FBB9","FDC4","FE78","FBC1","FD6C","FA6B","FEC5","FBEA","FBEF","FC42","FACD","FCA6","FC64","FD4A","FCB4","FCC4","FCC2","FAE8","FC49","FCEE","FD4C","FD53","FBF0","FD5D","FD61","FAA5","FDA3","FD49","FDD1","FDCE","FC59","FBBF","FAFD","FC65","FDE9","FC6A","FE4D","FEFB","FACC","FE6E","FEB6","FAE0","FBC9","FCD1","FDE8","FA45","FD50","C87E","FAD0","FD7A","FCF7","FB4B","FDCA","FAFA","FBCA","FC4A","FBDC","FE56","FE7D","FCCF","FB43","FBD3","A440","A442","A443","C945","A456","A454","A457","A455","C946","A4A3","C94F","C94D","A4A2","A4A1","A542","A541","A540","A543","A4FE","A5E0","A5E1","A8C3","A458","A4A4","C950","A4A5","C963","A6EA","CBB1","C6BF","A459","A4A6","A544","C964","C6C0","C940","A444","A45B","C947","A45C","FAE5","A4A7","A545","A547","A546","A5E2","A5E3","A8C4","ADBC","A441","C87B","C941","A445","A45E","A45D","A5E4","A8C5","B0AE","D44B","B6C3","DCB1","DCB2","C6C1","A446","A4A9","A8C6","A447","C948","A45F","A4AA","A4AC","C951","A4AD","A4AB","A5E5","A8C7","A8C8","AB45","C6C2","A460","A4AE","A5E6","A5E8","A5E7","A6EB","A8C9","A8CA","AB46","AB47","ADBD","DCB3","FBF8","F6D6","A448","A4B0","A4AF","C952","A4B1","A4B7","A4B2","A4B3","C954","C953","A4B5","A4B6","A4B4","A54A","A54B","A54C","A54D","A549","A550","C96A","C966","C6DF","A551","A561","C968","A54E","A54F","A548","C965","C967","A5F5","C9B0","A5F2","A5F6","C9BA","C9AE","A5F3","C9B2","A5F4","A5F7","A5E9","C9B1","A5F8","C9B5","C9B9","C9B6","C9B3","A5EA","A5EC","A5F9","A5EE","C9AB","A5F1","A5EF","A5F0","C9BB","C9B8","C9AF","A5ED","C9AC","A5EB","C9B4","C9B7","C9AD","CA66","A742","A6F4","CA67","A6F1","A744","A6F9","A6F8","CA5B","A6FC","A6F7","CA60","CA68","CA64","A6FA","A6FD","A6EE","A747","CA5D","CBBD","A6EC","A743","A6ED","A6F5","A6F6","CA62","CA5E","A6FB","A6F3","CA5A","A6EF","CA65","A745","A748","A6F2","A740","A746","A6F0","CA63","A741","CA69","CA5C","A6FE","CA5F","CA61","A8D8","CBBF","CBCB","A8D0","CBCC","A8CB","A8D5","A8CE","CBB9","A8D6","CBB8","CBBC","CBC3","CBC1","A8DE","A8D9","CBB3","CBB5","A8DB","A8CF","CBB6","CBC2","CBC9","A8D4","CBBB","CBB4","A8D3","CBB7","A8D7","CBBA","A8D2","A8CD","A8DC","CBC4","A8DD","CBC8","CBC6","CBCA","A8DA","CBBE","CBB2","CBC0","A8D1","CBC5","A8CC","CBC7","FA57","AB56","AB4A","CDE0","CDE8","AB49","AB51","AB5D","CDEE","CDEC","CDE7","AB4B","CDED","CDE3","AB59","AB50","AB58","CDDE","CDEA","CDE1","AB54","CDE2","CDDD","AB5B","AB4E","AB57","AB4D","CDDF","CDE4","CDEB","AB55","AB52","CDE6","AB5A","CDE9","CDE5","AB4F","AB5C","AB53","AB4C","AB48","CDEF","ADD7","ADC1","ADD1","ADD6","D0D0","D0CF","D0D4","D0D5","ADC4","ADCD","ADDA","ADCE","D0C9","ADC7","D0CA","FA59","ADDC","ADD3","ADBE","ADBF","D0DD","B0BF","ADCC","ADCB","D0CB","ADCF","D45B","ADC6","D0D6","ADD5","ADD4","ADCA","D0CE","D0D7","D0C8","ADC9","D0D8","ADD2","D0CC","ADC0","ADC3","ADC2","D0D9","ADD0","ADC5","ADD9","ADDB","D0D3","ADD8","D0DB","D0CD","D0DC","D0D1","D0DA","D0D2","ADC8","D463","D457","B0B3","D45C","D462","B0B2","D455","B0B6","D459","D452","B0B4","D456","B0B9","B0BE","D467","D451","B0BA","D466","B0B5","D458","B0B1","D453","D44F","D45D","D450","D44E","D45A","D460","D461","B0B7","D85B","D45E","D44D","D45F","B0C1","D464","B0C0","D44C","D454","D465","B0BC","B0BB","B0B8","B0BD","B0AF","B0B0","B3C8","D85E","D857","B3C5","D85F","D855","D858","B3C4","D859","FD56","B3C7","D85D","D853","D852","B3C9","B3CA","B3C6","B3CB","D851","D85C","D85A","D854","B3C3","D856","B6CA","B6C4","DCB7","B6CD","DCBD","DCC0","B6C6","B6C7","DCBA","B6C5","DCC3","B6CB","DCC4","DCBF","B6CC","DCB4","B6C9","DCB5","DCBE","DCBC","DCB8","B6C8","DCB6","B6CE","DCBB","DCC2","DCB9","DCC1","B9B6","B9B3","B9B4","E0F9","E0F1","B9B2","B9AF","E0F2","B9B1","E0F5","E0F7","E0FE","FC72","E0FD","E0F8","B9AE","E0F0","B9AC","E0F3","B9B7","E0F6","E0FA","B9B0","B9AD","E0FC","E0FB","B9B5","E0F4","BBF8","E4EC","E4E9","BBF9","BBF7","E4F0","E4ED","E4E6","BBF6","FA67","BBFA","E4E7","BBF5","BBFD","E4EA","E4EB","BBFB","BBFC","E4F1","E4EE","E4EF","FA69","BEAA","E8F8","BEA7","E8F5","BEA9","BEAB","E8F6","BEA8","E8F7","E8F4","C076","ECBD","C077","ECBB","ECBC","ECBA","ECB9","ECBE","C075","EFB8","EFB9","E4E8","EFB7","C078","C35F","F1EB","F1EC","C4D7","C4D8","F5C1","F5C0","C56C","C56B","F7D0","A449","A461","A4B9","A4B8","A553","A552","A5FC","A5FB","A5FD","A5FA","A74A","A749","A74B","A8E0","A8DF","A8E1","AB5E","A259","D0DE","A25A","B0C2","A25C","A25B","D860","FA6F","A25D","B9B8","A25E","A44A","A4BA","A5FE","A8E2","FA71","A44B","A4BD","A4BB","A4BC","A640","A74C","A8E4","A8E3","A8E5","ADDD","BEAC","C6C3","C94E","C8A2","A554","A555","A641","CA6A","AB60","AB5F","D0E0","D0DF","B0C3","C6C4","A4BE","C955","CBCD","AB61","ADE0","ADDE","ADDF","BEAD","C6C5","A556","A642","C9BC","FA7D","FAA8","FA47","A74D","A74E","FA7E","CA6B","CBCE","A8E6","CBCF","D0E2","D0E3","ADE3","FDB6","D0E4","FAA2","D0E1","ADE4","ADE2","ADE1","D0E5","FAA3","D468","FAA4","FAA6","D861","DCC5","E140","BBFE","BEAE","E8F9","FDDB","A44C","A45A","FAA9","FAAB","B0C4","B3CD","B9B9","FC7A","C942","A4BF","A559","A557","A558","A8E7","A44D","A44E","C87D","A462","A4C0","A4C1","A4C2","C9BE","A55A","FAB0","C96B","A646","C9BF","A644","A645","C9BD","A647","A643","CA6C","AAEC","CA6D","CA6E","A750","A74F","FAB1","A753","A751","A752","A8ED","A8EC","CBD4","CBD1","CBD2","CBD0","A8EE","A8EA","A8E9","A8EB","A8E8","FAB2","A8EF","AB63","CDF0","CBD3","AB68","CDF1","AB64","AB67","AB66","AB65","AB62","D0E8","ADE7","D0EB","ADE5","FAB4","D0E7","ADE8","ADE6","ADE9","D0E9","D0EA","D0E6","D0EC","B3D1","B0C5","D469","D46B","D46A","D46C","B0C6","B3CE","B3CF","B3D0","B6D0","DCC7","DCC6","DCC8","DCC9","B6D1","B6CF","E141","E142","B9BB","B9BA","E35A","BC40","BC41","BC42","BC44","E4F2","E4F3","BC43","BEAF","BEB0","FAB5","F1ED","F5C3","F5C2","F7D1","A44F","A55C","A55B","A648","C9C0","A755","A756","A754","A757","CA6F","CA70","FAB3","FAB6","A8F1","CBD5","A8F0","CDF2","AB6C","CDF3","AB6B","FAB7","AB69","AB6A","D0ED","FBC4","B0C7","D46E","B0CA","D46D","B1E5","B0C9","B0C8","B3D4","B3D3","B3D2","B6D2","FABA","B6D5","B6D6","B6D4","B6D3","E143","E144","E4F5","BC45","E4F4","BEB1","ECBF","C079","F1EE","C455","C6C6","A463","A4C3","C956","A4C4","A4C5","A55D","A55E","A649","CA71","CBD6","CBD7","AB6D","D0EE","B0CC","B0CB","D863","D862","A450","A4C6","A55F","B0CD","C943","C96C","A560","C9C2","A64B","A64A","C9C1","A758","ADEA","D46F","B6D7","E145","B9BC","FAC4","E8FA","F3FD","C6C7","A4C7","CBD8","CDF4","B0D0","B0CE","B0CF","A451","FAAA","A464","A2CD","A4CA","A4C9","A4C8","A563","A562","C96D","C9C3","A8F5","A8F2","A8F4","A8F3","AB6E","B3D5","A452","A4CB","A565","A564","CA72","A8F6","C6C8","C957","FAD1","A567","A566","A64C","A64D","CA73","A759","FAD2","A75A","A8F7","A8F8","A8F9","AB6F","CDF5","FAD4","ADEB","C944","A4CC","C9C4","CA74","CA75","CBD9","FAD9","CBDA","CDF7","CDF6","CDF9","CDF8","AB70","D470","ADED","D0EF","ADEC","FADB","D864","B3D6","FBF7","D865","FBFA","FADC","E146","B9BD","FADD","BC46","F1EF","C6C9","C958","A568","FAE2","B0D1","FAE3","A453","A465","A4CE","A4CD","A4CF","A8FB","A8FA","A8FC","FAE7","AB71","ADEE","FAEA","E8FB","C24F","A466","A56A","A579","A574","A56F","A56E","A575","A573","A56C","A57A","A56D","A569","A578","A577","A576","A56B","A572","FAED","A571","A57B","A570","FB59","A653","A659","A655","A65B","C9C5","A658","A64E","A651","A654","A650","A657","A65A","A64F","A652","A656","A65C","FAEF","CA7E","CA7B","A767","CA7C","A75B","A75D","A775","A770","FD6D","CAA5","CA7D","A75F","A761","CAA4","A768","CA78","A774","A776","A75C","A76D","FB44","CA76","A773","A764","A76E","A76F","CA77","A76C","A76A","A76B","A771","CAA1","A75E","A772","CAA3","A766","A763","CA7A","A762","CAA6","A765","A769","A760","CAA2","CA79","CBEB","CBEA","A94F","CBED","CBEF","CBE4","CBE7","CBEE","A950","CBE1","CBE5","FAF4","CBE9","CE49","A94B","CE4D","A8FD","CBE6","A8FE","A94C","A945","A941","CBE2","A944","A949","A952","CBE3","CBDC","A943","CBDD","CBDF","A946","A948","CBDB","CBE0","A951","A94D","CBE8","A953","FAF8","A94A","CBDE","A947","A942","A940","CBEC","A94E","CE48","CDFB","CE4B","FAF9","CDFD","AB78","ABA8","AB74","ABA7","AB7D","ABA4","AB72","CDFC","CE43","ABA3","CE4F","ABA5","AB79","CE45","CE42","AB77","CDFA","ABA6","CE4A","AB7C","CE4C","ABA9","AB73","AB7E","AB7B","CE40","ABA1","CE46","CE47","AB7A","ABA2","AB76","AB75","CDFE","CE44","CE4E","D144","ADFB","D0F1","D0F6","ADF4","AE40","D0F4","ADEF","ADF9","ADFE","D0FB","ADFA","ADFD","D0FE","ADF5","D0F5","D142","D143","ADF7","D141","ADF3","AE43","D0F8","ADF1","D146","D0F9","D0FD","ADF6","AE42","D0FA","ADFC","D140","D147","D4A1","D145","AE44","ADF0","D0FC","D0F3","ADF8","D0F2","D0F7","D0F0","AE41","D477","FAF1","B0E4","D4A7","B0E2","B0DF","D47C","B0DB","D4A2","B0E6","D476","D47B","D47A","ADF2","B0E1","D4A5","D4A8","D473","B3E8","D4A9","B0E7","B0D9","B0D6","D47E","B0D3","FB42","D4A6","FABF","B0DA","D4AA","D474","D4A4","B0DD","D475","D478","D47D","FBA3","B0DE","B0DC","B0E8","B0E3","FAF7","B0D7","B1D2","B0D8","D479","B0E5","B0E0","D4A3","B0D5","B0D4","D471","D472","D86A","B3D7","B3DA","D875","B3EE","D878","B3D8","D871","B3DE","B3E4","B5BD","FB46","B3E2","D86E","B3EF","B3DB","B3E3","D876","DCD7","D87B","D86F","D866","D873","D86D","B3E1","D879","B3DD","B3F1","B3EA","B3DF","B3DC","B3E7","D87A","D86C","D872","D874","D868","D877","B3D9","D867","FB47","B3E0","B3F0","B3EC","D869","B3E6","B3ED","B3E9","B3E5","D870","B3EB","DCD5","DCD1","DCE0","DCCA","DCD3","B6E5","B6E6","B6DE","DCDC","B6E8","DCCF","DCCE","DCCC","DCDE","B6DC","DCD8","DCCD","B6DF","DCD6","B6DA","DCD2","DCD9","DCDB","DCDF","B6E3","DCCB","B6DD","DCD0","B6D8","B6E4","DCDA","B6E0","B6E1","B6E7","B6DB","A25F","B6D9","DCD4","B6E2","DCDD","B9CD","B9C8","E155","E151","E14B","B9C2","B9BE","E154","B9BF","E14E","E150","E153","FB48","B9C4","B9CB","B9C5","E149","B9C6","B9C7","E14C","B9CC","E14A","E14F","B9C3","E148","B9C9","B9C1","B9C0","E14D","E152","B9CA","E147","BC4D","E547","E544","BC47","BC53","BC54","BC4A","E542","BC4C","E4F9","BC52","FB4F","E546","BC49","E548","BC48","E543","E545","BC4B","E541","E4FA","E4F7","D86B","E4FD","E4F6","E4FC","E4FB","E4F8","FB54","BC4F","FB55","BC4E","BC50","E4FE","BEB2","E540","E945","E8FD","BEBE","E942","BEB6","BEBA","E941","BEB9","BEB5","BEB8","BEB3","BEBD","E943","E8FE","BEBC","E8FC","BEBB","E944","E940","BC51","BEBF","E946","BEB7","BEB4","ECC6","ECC8","C07B","ECC9","ECC7","ECC5","ECC4","C07D","ECC3","C07E","ECC1","ECC2","C07A","C0A1","C07C","ECC0","C250","EFBC","EFBA","EFBF","EFBD","EFBB","EFBE","C360","F1F2","F1F3","C456","F1F4","F1F0","F1F5","F1F1","C251","F3FE","F441","C459","F440","C458","C457","C45A","F5C5","F5C6","C4DA","C4D9","C4DB","F5C4","F6D8","F6D7","C56D","C56F","C56E","F6D9","C5C8","F8A6","C5F1","F8A5","F8EE","C949","A57D","A57C","A65F","A65E","C9C7","A65D","C9C6","A779","CAA9","CAA8","A777","A77A","FB5C","CAA7","FB5B","A778","FB57","CBF0","CBF1","A954","ABAA","FB5A","D148","D149","AE45","AE46","D4AC","B0E9","B0EB","D4AB","B0EA","D87C","B3F2","B6E9","B6EA","DCE1","B9CF","B9CE","E549","E948","E947","F96B","A467","C959","C96E","C96F","A662","A666","C9C9","A664","A663","C9C8","A665","A661","A660","C9CA","A7A6","A7A3","A77D","CAAA","FB64","FB76","CAAB","FB60","A7A1","CAAD","A77B","CAAE","CAAC","A77E","A7A2","A7A5","A7A4","A77C","CAAF","A959","CBFE","A95B","A95A","CC40","A958","A957","CBF5","CBF4","CBF2","CBF7","CBF6","CBF3","CBFC","CBFD","CBFA","CBF8","A956","CBFB","A95C","CC41","CBF9","ABAB","A955","ABAC","CE54","CE5A","FC67","ABB2","CE58","CE5E","CE55","CE59","CE5B","CE5D","CE57","CE56","CE51","CE52","ABAD","ABAF","ABAE","CE53","CE5C","ABB1","CE50","D153","D152","D157","D14E","D151","D150","D154","D158","AE47","AE4A","D14F","D155","AE49","D14A","ABB0","D4BA","D156","D14D","AE48","D14C","D4B1","B0EC","B0F0","D4C1","D4AF","D4BD","B0F1","D4BF","FB67","D4C5","D4C9","D4C0","D4B4","D4BC","D4CA","D4C8","D4BE","D4B9","D4B2","D8A6","D4B0","B0F5","D4B7","B0F6","B0F2","D4AD","D4C3","D4B5","FAE6","D4B3","D4C6","B0F3","FB69","D4CC","B0ED","B0EF","D4BB","D4B6","AE4B","B0EE","D4B8","D4C7","D4CB","D4C2","D4C4","D4AE","D8A1","D8AA","D8A9","B3FA","D8A2","B3FB","B3F9","D8A4","B3F6","D8A8","FB6C","D8A3","D8A5","D87D","B3F4","D8B2","D8B1","D8AE","B3F3","B3F7","B3F8","D14B","D8AB","B3F5","B0F4","D8AD","D87E","D8B0","D8AF","D8B3","DCEF","D8AC","D8A7","DCE7","B6F4","B6F7","B6F2","DCE6","DCEA","DCE5","B6EC","B6F6","DCE2","B6F0","DCE9","B6EE","B6ED","DCEC","B6EF","DCEE","FB6E","DCEB","B6EB","B6F5","DCF0","DCE4","DCED","DCE3","B6F1","B6F3","DCE8","DCF1","E15D","B9D0","E163","B9D5","E15F","E166","E157","B9D7","B9D1","E15C","BC55","E15B","E164","B9D2","B9D6","E15A","E160","E165","E156","B9D4","E15E","E162","E168","E158","E161","B9D3","E167","E159","BC59","E54B","BC57","BC56","E54D","E552","E54E","E551","BC5C","BEA5","BC5B","FB6F","E54A","E550","BC5A","E54F","E54C","BC58","E94D","F9D9","E94F","E94A","BEC1","E94C","BEC0","E94E","BEC3","E950","BEC2","E949","E94B","C0A5","ECCC","C0A4","ECCD","C0A3","ECCB","C0A2","ECCA","C253","C252","F1F6","F1F8","FB72","F1F7","C361","C362","FB71","C363","F442","C45B","F7D3","F7D2","C5F2","A468","A4D0","A7A7","CE5F","B3FC","B3FD","FB74","DCF2","B9D8","E169","E553","C95A","CAB0","C6CA","CC42","CE60","D159","AE4C","FE42","F1F9","C4DC","A469","A57E","C970","A667","A668","A95D","FB7B","B0F7","B9DA","B9DB","B9D9","A46A","A4D1","A4D3","A4D2","C95B","A4D4","A5A1","C971","A5A2","A669","A66A","C9CB","A7A8","CAB1","A961","CC43","A95F","A960","A95E","D15A","ABB6","ABB5","ABB7","ABB4","CE61","A962","ABB3","AE4D","AE4E","AE4F","D4CD","B3FE","D8B4","B0F8","B6F8","B9DD","B9DC","E16A","BC5D","BEC4","EFC0","F6DA","F7D4","A46B","A5A3","A5A4","C9D1","A66C","A66F","C9CF","C9CD","A66E","C9D0","C9D2","C9CC","A671","A670","A66D","A66B","C9CE","A7B3","A7B0","CAB6","CAB9","CAB8","A7AA","A7B2","A7AF","CAB5","CAB3","A7AE","A7A9","A7AC","CAB4","CABB","CAB7","A7AD","A7B1","A7B4","CAB2","CABA","A7AB","A967","A96F","CC4F","CC48","A970","CC53","CC44","CC4B","A966","CC45","A964","CC4C","CC50","A963","CC51","CC4A","CC4D","A972","A969","CC54","CC52","FBA6","A96E","A96C","CC49","A96B","CC47","CC46","A96A","A968","A971","A96D","A965","CC4E","ABB9","FBAB","ABC0","CE6F","ABB8","CE67","CE63","CE73","CE62","ABBB","CE6C","ABBE","ABC1","ABBC","CE70","ABBF","AE56","CE76","CE64","CE66","CE6D","CE71","CE75","CE72","CE6B","CE6E","FBB2","CE68","ABC3","CE6A","CE69","CE74","ABBA","CE65","ABC2","ABBD","AE5C","D162","AE5B","D160","AE50","AE55","D15F","D15C","D161","AE51","D15B","AE54","AE52","D163","AE53","AE57","AE58","FBA2","AE5A","AE59","D15D","D15E","D164","D4D4","B0F9","D8C2","D4D3","D4E6","B140","D4E4","B0FE","B0FA","D4ED","D4DD","D4E0","B143","D4EA","D4E2","B0FB","B144","D4E7","D4E5","D4D6","D4EB","D4DF","D4DA","D4D0","D4EC","D4DC","D4CF","B142","D4E1","D4EE","D4DE","D4D2","D4D7","D4CE","B141","FBB5","D4DB","D4D8","B0FC","D4D1","D4E9","B0FD","D4D9","D4D5","D4E8","B440","D8BB","D8B8","D8C9","D8BD","D8CA","B442","D8C6","D8C3","FDEF","D8C4","D8C7","D8CB","D4E3","D8CD","DD47","FDC1","B443","D8CE","D8B6","D8C0","FBBA","D8C5","B441","B444","D8CC","D8CF","D8BA","D8B7","FC73","D8B9","D8BE","D8BC","B445","D8C8","FBB4","D8BF","D8C1","D8B5","DCFA","DCF8","B742","B740","DD43","DCF9","DD44","DD40","DCF7","DD46","DCF6","DCFD","B6FE","B6FD","B6FC","DCFB","DD41","B6F9","B741","DCF4","DCFE","DCF3","DCFC","B6FA","DD42","DCF5","B6FB","DD45","FBBC","E16E","B9E2","B9E1","B9E3","E17A","E170","E176","E16B","E179","E178","E17C","E175","B9DE","E174","B9E4","E16D","B9DF","E17B","B9E0","E16F","E172","E177","E171","E16C","E173","E555","BC61","E558","E557","E55A","E55C","F9DC","BC5F","E556","E554","E55D","E55B","E559","E55F","E55E","BC63","BC5E","BC60","BC62","E560","E957","E956","E955","E958","E951","E952","E95A","E953","BEC5","E95C","E95B","E954","ECD1","C0A8","ECCF","ECD4","ECD3","E959","C0A7","ECD2","ECCE","ECD6","ECD5","C0A6","ECD0","BEC6","C254","EFC1","F1FA","F1FB","F1FC","C45C","C45D","F443","FEA4","F5C8","F5C7","F6DB","F6DC","F7D5","F8A7","A46C","A46D","A46E","A4D5","A5A5","C9D3","A672","A673","A7B7","A7B8","A7B6","A7B5","A973","CC55","A975","A974","CC56","ABC4","AE5D","D165","D4F0","B145","B447","D4EF","B446","B9E5","FBC5","E17D","BEC7","C0A9","ECD7","FBC7","C45E","C570","C6CB","C972","FA79","A5A6","C973","A676","A674","A675","A677","A7BA","A7B9","CABC","A7BB","CABD","CC57","CC58","A976","A978","A97A","A977","A97B","A979","FBD2","ABC8","ABC5","ABC7","ABC9","ABC6","D166","CE77","FC7D","D168","D167","AE63","AE5F","AE60","AE62","AE64","AE61","AE66","AE65","B14A","D4F2","D4F1","B149","B148","B147","B14B","B146","D8D5","D8D2","B449","D8D1","D8D6","B44B","D8D4","B448","B44A","D8D3","FBCC","DD48","FEAE","DD49","DD4A","B9E6","B9EE","E17E","B9E8","B9EC","E1A1","B9ED","B9E9","B9EA","B9E7","B9EB","BC66","D8D0","BC67","BC65","BC64","E95D","BEC8","ECD8","ECD9","FBD1","C364","C45F","A46F","A678","FB75","ABCA","D169","AE67","FBD4","B14E","B14D","B14C","B44C","B44D","D8D7","B9EF","BEC9","A470","C95C","A4D6","C974","FBD6","FBD8","C9D4","A679","A97C","DD4B","A471","A4D7","C9D5","CABE","CABF","A7BC","D8D8","B44E","DD4C","C0AA","A472","A4A8","A4D8","C975","A5A7","A7C0","A7BF","A7BD","A7BE","CC59","A97E","A9A1","CC5A","A97D","FBDB","ABCE","CE78","ABCD","ABCB","ABCC","AE6A","AE68","D16B","AE69","D16A","AE5E","D4F3","B150","B151","B14F","B9F0","E1A2","BC68","BC69","E561","C0AB","EFC2","EFC3","C4DD","F8A8","C94B","A4D9","A473","C977","C976","A67A","C9D7","C9D8","C9D6","C9D9","FBDD","CAC7","CAC2","CAC4","CAC6","CAC3","A7C4","CAC0","CAC1","A7C1","A7C2","CAC5","CAC8","A7C3","CAC9","FDF2","CC68","CC62","CC5D","A9A3","CC65","CC63","CC5C","CC69","CC6C","CC67","CC60","A9A5","CC66","A9A6","CC61","CC64","CC5B","CC5F","CC6B","A9A7","A9A8","CC5E","CC6A","A9A2","A9A4","FBE7","CEAB","CEA4","CEAA","CEA3","CEA5","CE7D","CE7B","CEAC","CEA9","CE79","ABD0","CEA7","CEA8","CEA6","CE7C","CE7A","ABCF","CEA2","CE7E","CEA1","CEAD","AE6F","FBDE","AE6E","D16C","AE6B","D16E","FBDF","AE70","D16F","AE73","AE71","D170","CEAE","D172","AE6D","AE6C","D16D","D171","AE72","B153","B152","D4F5","D4F9","D4FB","B154","D4FE","FBE3","B158","D541","B15A","B156","B15E","FBE4","B15B","D4F7","B155","D4F6","D4F4","D543","D4F8","B157","D542","B15C","D4FD","D4FC","B15D","D4FA","B159","D544","D540","D8E7","D8EE","D8E3","B451","D8DF","D8EF","D8D9","D8EC","D8EA","D8E4","D8ED","D8E6","D8DE","D8F0","D8DC","D8E9","D8DA","D8F1","FBE5","B452","D8EB","DD4F","D8DD","B44F","D8E1","B450","D8E0","D8E5","D8E2","D8E8","DD53","DD56","DD4E","DD50","DD55","DD54","B743","D8DB","DD52","B744","DD4D","DD51","E1A9","E1B0","E1A7","E1AE","E1A5","E1AD","E1B1","E1A4","E1A8","E1A3","B9F1","E1A6","B9F2","E1AC","E1AB","E1AA","FBE0","E1AF","E565","E567","BC6B","E568","E563","E562","E56C","E56A","BC6A","E56D","E564","E569","E56B","E566","E961","E966","E960","E965","E95E","E968","E964","E969","E963","E95F","E967","E96A","E962","FC58","ECDA","C0AF","C0AD","C0AC","C0AE","EFC4","F172","F1FD","F444","F445","C460","F5C9","C4DE","F5CA","F6DE","C572","C571","F6DD","C5C9","FBE8","F7D6","C6CC","A474","A67B","C9DA","CACA","A8B5","B15F","A475","A5AA","A5A9","A5A8","A7C5","AE74","DD57","A476","A477","A478","A4DA","ABD1","CEAF","B453","A479","C95D","A5AB","A5AC","C978","A67C","FBFC","CACB","A7C6","CACC","A9AE","CC6E","A9AC","A9AB","CC6D","A9A9","CC6F","A9AA","A9AD","ABD2","ABD4","CEB3","CEB0","CEB1","CEB2","CEB4","ABD3","D174","D173","AE76","AE75","FBF1","B162","D546","B161","B163","B160","B455","D545","B456","D8F3","B457","D8F2","B454","DD5A","DD5C","B745","DD5B","DD59","DD58","E1B4","B9F7","B9F5","B9F6","E1B2","E1B3","B9F3","E571","E56F","BC6D","E570","BC6E","BC6C","B9F4","E96D","E96B","E96C","E56E","ECDC","C0B0","ECDB","EFC5","EFC6","E96E","F1FE","A47A","A5AD","A67E","C9DB","A67D","A9AF","B746","FBF4","A4DB","A5AE","ABD5","B458","C6CE","C979","C97A","FBC3","C9DC","A7C8","CAD0","CACE","A7C9","CACD","CACF","CAD1","A7C7","A9B3","A9B4","A9B1","A9B0","CEB8","A9B2","ABD6","CEB7","CEB9","CEB6","CEBA","ABD7","AE79","D175","D177","AE77","D178","AE78","D176","CEB5","D547","D54A","D54B","D548","B167","B166","B164","B165","D549","B168","B45A","B45B","B45C","DD5D","DD5F","DD61","B748","B747","B459","DD60","DD5E","E1B8","FBF9","E1B6","E1BC","B9F8","E1BD","E1BA","B9F9","E1B7","E1B5","E1BB","BC70","E573","E1B9","BC72","E574","BC71","BC74","E575","BC6F","BC73","E973","E971","E970","E972","E96F","C366","F446","F447","F5CB","F6DF","C655","C6CF","A9B5","A7CA","FC40","ABD8","FC41","FC43","A47B","A4DC","A5AF","C9DD","A7CB","CAD2","CEBB","ABD9","B9FA","A47C","FC46","A6A1","B749","A47D","A4DD","A4DE","A5B1","A5B0","C9DE","A6A2","CAD3","A7CC","CC71","CC72","CC73","A9B6","A9B7","CC70","A9B8","ABDA","CEBC","D17A","AE7A","D179","B169","D54C","B16A","D54D","FC4C","B45D","DD62","E1BF","E1BE","B9FB","BC75","E576","BECA","E974","C0B1","C573","F7D8","C6D0","CC74","CEBD","B16B","D8F4","B74A","C255","C6D1","A7CE","FC51","A7CD","ABDB","D17B","B16D","B343","B16E","B16C","B45E","E1C0","B9FC","BC76","FC54","C94C","C9DF","CAD5","A7CF","CAD4","A7D0","FAAF","A9BC","CC77","CC76","A9BB","A9B9","A9BA","CC75","ABDD","CEBE","ABE0","ABDC","ABE2","ABDE","ABDF","ABE1","AE7D","AE7C","AE7B","D54F","B16F","B172","B170","D54E","B175","B171","D550","B174","B173","FA61","D8F6","D8F5","FC57","B461","B45F","B460","D8F7","B74B","DD64","B74C","DD63","E577","BC78","E1C1","BC77","B9FD","ECDE","E975","C0B2","ECDD","F240","F448","F449","A4DF","A5B2","C97B","A7D2","A7D4","C9E2","CAD8","CAD7","CAD6","C9E1","C9E0","A6A4","A7D3","A7D1","A6A3","A9BD","CC78","FCD5","A9BE","CADD","CADF","CADE","CC79","CADA","A7D8","A7D6","CAD9","CADB","CAE1","A7D5","CADC","CAE5","A9C0","CAE2","A7D7","CAE0","CAE3","A9BF","A9C1","CAE4","CCAF","CCA2","CC7E","CCAE","CCA9","ABE7","A9C2","CCAA","CCAD","ABE3","CCAC","A9C3","A9C8","A9C6","CCA3","CC7C","CCA5","A9CD","CCB0","ABE4","CCA6","ABE5","A9C9","CCA8","FCA9","CECD","ABE6","CC7B","A9CA","ABE8","A9CB","A9C7","A9CC","CCA7","CC7A","CCAB","A9C4","FC61","CC7D","CCA4","CCA1","A9C5","CEBF","CEC0","CECA","D1A1","CECB","ABEE","CECE","CEC4","ABED","CEC6","CEC7","FACB","CEC9","ABE9","AEA3","F9DA","CEC5","CEC1","AEA4","CECF","AE7E","D17D","CEC8","D17C","CEC3","CECC","ABEC","AEA1","ABF2","AEA2","CED0","D17E","ABEB","AEA6","ABF1","ABF0","ABEF","AEA5","CED1","AEA7","ABEA","CEC2","B176","D1A4","D1A6","D1A8","AEA8","AEAE","D553","D1AC","D1A3","B178","D551","AEAD","AEAB","D1AE","D552","D1A5","AEAC","D1A9","AEAF","D1AB","AEAA","D1AA","D1AD","D1A7","FC6B","AEA9","B179","D1A2","B177","FC6C","B17A","D555","D55E","B464","FC6D","B17C","B1A3","B465","D560","B1AA","D8F9","D556","B1A2","B1A5","B17E","D554","D562","D565","D949","D563","D8FD","B1A1","B1A8","B1AC","D55D","D8F8","D561","B17B","D8FA","D564","D8FC","D559","B462","D557","D558","B1A7","B1A6","D55B","B1AB","D55F","B1A4","D55C","FD64","B1A9","B466","B463","D8FB","D55A","B17D","B46B","B46F","D940","B751","B46D","D944","B471","DD65","D946","B753","B469","B46C","D947","D948","D94E","B473","B754","D94A","D94F","D943","B75E","B755","B472","D941","D950","B75D","B470","B74E","D94D","B474","D945","D8FE","B46A","D942","D94B","B74D","B752","B467","D94C","B750","B468","B75C","E1C3","DD70","DD68","E1C2","DD6C","DD6E","DD6B","B75B","DD6A","B75F","E1D2","B75A","BA40","DD71","E1C4","FC76","B758","DD69","DD6D","B9FE","B74F","DD66","DD67","BA41","B757","B759","B756","DD6F","E1C8","E1C9","E1CE","BC7D","E1D5","BA47","BA46","E1D0","FCAA","BC7C","E1C5","BA45","FBCD","E1D4","BA43","BA44","FC74","E1D1","E5AA","BC7A","B46E","E1D3","BCA3","E1CB","BC7B","BCA2","E1C6","E1CA","E1C7","E1CD","BA48","BC79","BA42","E57A","E1CF","BCA1","BCA4","E1CC","FC79","BC7E","E579","FC7C","E57E","BECE","E578","E9A3","E5A9","BCA8","BCA6","BECC","E5A6","E5A2","BCAC","E978","BCAA","E5A1","E976","E5A5","E5A8","E57D","BCAB","BCA5","E977","BECD","E5A7","BCA7","BCA9","E5A4","BCAD","E5A3","E57C","E57B","BECB","E5AB","E97A","ECE0","BED0","E9A2","E97E","ECE1","BED1","E9A1","E97C","C0B4","ECDF","E979","E97B","C0B5","BED3","C0B3","BED2","C0B7","E97D","BECF","FCA5","FCA2","EFCF","EFC7","ECE7","EFC8","ECE3","C256","ECE5","ECE4","C0B6","ECE2","ECE6","EFD0","EFCC","EFCE","EFC9","EFCA","EFCD","EFCB","C367","C36A","C369","C368","C461","F44A","C462","F241","C4DF","F5CC","C4E0","C574","C5CA","F7D9","F7DA","F7DB","F9BA","A4E0","C97C","A5B3","A6A6","A6A7","A6A5","A6A8","A7DA","A7D9","CCB1","A9CF","A9CE","D1AF","B1AD","B1AE","B475","DD72","B760","B761","DD74","DD76","DD75","E1D7","E1D6","BA49","E1D8","E5AC","BCAE","BED4","C0B8","C257","C0B9","A4E1","CAE6","CCB2","A9D1","A9D0","A9D2","ABF3","CED2","CED3","D1B0","AEB0","B1AF","B476","D951","A4E2","A47E","A4E3","C97D","A5B7","A5B6","A5B4","A5B5","A6AB","C9E9","C9EB","A6AA","C9E3","C9E4","C9EA","C9E6","C9E8","A6A9","C9E5","C9EC","C9E7","A7E1","A7EA","A7E8","CAF0","CAED","CAF5","A7E6","CAF6","A7DF","CAF3","A7E5","CAEF","CAEE","A7E3","CAF4","A7E4","A9D3","A7DE","CAF1","CAE7","A7DB","A7EE","CAEC","CAF2","A7E0","A7E2","CAE8","CAE9","CAEA","A7ED","A7E7","A7EC","CAEB","A7EB","A7DD","A7DC","A7E9","A9E1","CCBE","CCB7","A9DC","A9EF","CCB3","CCBA","CCBC","CCBF","A9EA","CCBB","CCB4","A9E8","CCB8","CCC0","A9D9","CCBD","A9E3","A9E2","CCB6","A9D7","A9D8","A9D6","FCAE","A9EE","A9E6","A9E0","A9D4","CCB9","A9DF","A9D5","A9E7","A9F0","CED4","A9E4","CCB5","A9DA","A9DD","A9DE","FCB0","A9EC","A9ED","A9EB","A9E5","A9E9","A9DB","ABF4","FA51","CEDA","AC41","ABF8","ABFA","AC40","CEE6","ABFD","D1B1","AEB1","AC43","CED7","CEDF","ABFE","CEDE","CEDB","CEE3","CEE5","ABF7","ABFB","AC42","AEB3","CEE0","ABF9","AC45","CED9","ABFC","AEB2","ABF6","CED6","CEDD","CED5","CED8","CEDC","D1B2","AC44","CEE1","CEE2","CEE4","ABF5","AEC1","D1BE","AEBF","AEC0","D1B4","D1C4","AEB6","D566","D1C6","D1C0","D1B7","D1C9","D1BA","AEBC","D57D","D1BD","AEBE","AEB5","D1CB","D1BF","AEB8","D1B8","D1B5","D1B6","AEB9","D1C5","D1CC","AEBB","D1BC","D1BB","AEC3","AEC2","AEB4","AEBA","AEBD","D1C8","D1C2","AEB7","D1B3","D1CA","D1C1","D1C3","D1C7","D567","B1B7","B1CB","B1CA","B1BF","FCB2","D579","D575","D572","D5A6","B1BA","B1B2","D577","B4A8","B1B6","D5A1","B1CC","B1C9","D57B","D56A","B1C8","D5A3","D569","B1BD","B1C1","D5A2","D573","B1C2","B1BC","D568","FCAC","B478","D5A5","D571","B1C7","D574","D5A4","B1C6","D952","B1B3","D56F","B1B8","B1C3","B1BE","D578","D56E","D56C","D57E","B1B0","B1C4","B1B4","B477","D57C","B1B5","B1B1","B1C0","B1BB","B1B9","D570","B1C5","D56D","D57A","D576","D954","D953","D56B","D964","B47A","D96A","D959","D967","DD77","B47D","D96B","D96E","B47C","D95C","D96D","D96C","B47E","D955","B479","B4A3","B4A1","D969","D95F","B4A5","D970","D968","D971","B4AD","B4AB","D966","D965","D963","D95D","B4A4","B4A2","D1B9","D956","DDB7","D957","B47B","B4AA","DD79","B4A6","B4A7","D958","D96F","DD78","D960","D95B","B4A9","D961","D95E","FCB6","B4AE","B770","DD7C","DDB1","DDB6","DDAA","B76C","DDBB","B769","DD7A","DD7B","B762","B76B","DDA4","B76E","B76F","DDA5","DDB2","DDB8","B76A","B764","DDA3","DD7D","DDBA","DDA8","DDA9","DD7E","DDB4","DDAB","DDB5","DDAD","B765","E1D9","B768","B766","DDB9","DDB0","DDAC","DDA1","BA53","DDAF","B76D","DDA7","FCB5","DDA6","FCC3","B767","B763","E1EE","DDB3","DDAE","DDA2","E1E9","E1DA","E1E5","E1EC","BA51","B4AC","E1EA","BA4C","BA4B","E1F1","E1DB","E1E8","E1DC","E1E7","BA4F","E1EB","D962","E1F2","E1E3","BA52","E5BA","BCAF","E1F0","E1EF","BA54","E5AD","BCB0","E5AE","E1DF","E1E0","E1DD","E1E2","E1DE","E1F3","BA4E","BCB1","BA50","BA55","E1E1","E1ED","E1E6","E5B1","BA4A","BCB4","E9AA","E5B6","E5B5","E5B7","E5B4","BCB5","BCBB","BCB8","BCB9","E5AF","E5B2","E5BC","BCC1","BCBF","E5B3","D95A","BCB2","E5B9","E5B0","BCC2","E5B8","BA4D","BCB7","E1E4","BCBA","BCBE","BCC0","BCBD","BCBC","FED4","BCB6","E5BB","BCB3","BCC3","BED8","BED9","E9A9","BEE2","BEDF","BED6","BEDD","E9AB","BEDB","BED5","BEDC","E9A8","C0BB","BED7","BEDE","C0BA","E9A7","E9A6","BEE0","BEE1","E9A5","E9A4","C0BC","E9AE","BEDA","E9AC","C0BD","FCBF","C0C2","ECEA","ECEC","FCC0","C0BF","ECED","ECE9","ECEB","C0C0","C0C3","ECE8","C0BE","C0C1","C259","E9AD","C258","C25E","EFD4","C25C","C25D","EFD7","EFD3","C25A","EFD1","C36B","EFD5","EFD6","EFD2","C25B","F242","F245","F246","F244","F247","C36C","F243","F44E","C464","F44D","F44C","F44B","C463","C465","F5CD","C4E2","C4E1","FCAB","F6E1","F6E0","F6E3","C5CB","C575","F7DD","F6E2","F7DC","C5CD","C5CC","C5F3","F8A9","F8EF","A4E4","D972","E9AF","C6D2","A6AC","CAF7","A7F1","A7EF","A7F0","CCC1","A9F1","AC46","CEE7","CEE8","AC47","D1CE","AEC4","AEC5","D1CD","FCC5","B1D3","B1CF","D5A7","B1D6","B1D5","B1CE","B1D1","B1D4","B1D0","D976","B1CD","B4AF","FCCB","B4B1","B4B2","D975","D978","B4B0","D973","D977","D974","B771","FCCA","DDBC","BA56","E1F4","BEE3","BCC4","E5BD","BCC5","BCC6","E5BF","E5BE","E5C0","E9B1","E9B0","ECEF","ECEE","C0C4","C0C5","F248","FCC9","A4E5","FBC6","D979","B4B4","B4B3","DDBD","EFD8","C4E3","F7DE","A4E6","AEC6","B1D8","B1D7","D97A","D97B","B772","E1F5","BA57","E9B2","A4E7","A5B8","A9F2","CCC2","CEE9","AC48","B1D9","D97C","B4B5","B773","E5C1","E5C2","FCCD","ECF0","C25F","F8F0","A4E8","CCC3","A9F3","AC49","CEEA","AEC7","D1D2","D1D0","D1D1","AEC8","D1CF","B1DB","B1DC","D5A8","B1DD","B1DA","D97D","FCD0","D97E","DDBE","BA59","BA58","ECF1","EFD9","F24A","F249","F44F","C6D3","C95E","AC4A","FCD4","A4E9","A5B9","A6AE","A6AD","A6AF","A6B0","C9EE","C9ED","CAF8","A7F2","CAFB","CAFA","CAF9","CAFC","A9F4","CCC9","CCC5","CCCE","A9FB","A9F9","CCCA","CCC6","CCCD","A9F8","AA40","CCC8","CCC4","A9FE","CCCB","A9F7","CCCC","A9FA","A9FC","CCD0","CCCF","CCC7","A9F6","A9F5","A9FD","FCD7","CEEF","CEF5","AC50","AC4D","CEEC","CEF1","FE63","AC53","AC4B","CEF0","AC4E","AC51","CEF3","AC4C","CEF8","AC4F","AC52","CEED","CEF2","CEF6","CEEE","CEEB","CEF7","CEF4","AED0","AEC9","AECC","FCDA","AECF","D1D5","AECA","D1D3","FCDD","AECE","AECB","D1D6","AECD","FAF2","D5AC","B1DF","D5AB","D5AD","B1DE","B1E3","D1D4","D5AA","D5AE","B1E0","D5A9","B1E2","FCDF","B1E1","D9A7","D9A2","B4B6","B4BA","B4B7","D9A5","D9A8","FCE1","B4B8","B4B9","B4BE","DDC7","D9A6","B4BC","D9A3","D9A1","B4BD","D9A4","B779","FC62","DDBF","B776","B777","B775","DDC4","DDC3","DDC0","B77B","DDC2","B4BB","DDC6","DDC1","B778","B774","B77A","DDC5","BA5C","E1F8","E1F7","E1F6","BA5A","FB52","BA5B","E5C5","E5C8","BCC8","FB53","BCC7","E5C9","E5C4","BCCA","E5C6","FB4D","BCC9","E5C3","E5C7","BEE9","BEE6","E9BB","E9BA","E9B9","E9B4","E9B5","BEE7","BEE4","BEE8","E9B3","BEE5","E9B6","E9B7","E9BC","FB50","E9B8","ECF2","C0C7","EFDC","C0C6","EFDA","EFDB","C260","C36E","F24B","C36D","F451","F452","C466","F450","C4E4","F7DF","C5CE","F8AA","F8AB","A4EA","A6B1","A6B2","A7F3","CCD1","AC54","AED1","B1E4","B0D2","B4BF","B4C0","B3CC","D9A9","FCEB","B77C","E1FA","E1F9","A4EB","A6B3","CCD2","AA42","AA41","CEF9","CEFA","D1D7","D1D8","AED2","AED3","AED4","D5AF","B1E6","B4C2","B4C1","DDC8","DF7A","E1FB","E9BD","C261","C467","A4EC","A5BC","A5BD","A5BB","A5BE","A5BA","A6B6","C9F6","A6B5","A6B7","C9F1","C9F0","C9F3","C9F2","C9F5","A6B4","C9EF","C9F4","FA50","CAFD","A7FD","CAFE","CB43","A7FC","CB47","CB42","CB45","A7F5","A7F6","A7F7","A7F8","A840","CB41","A7FA","A841","CB40","CB46","A7F9","CB44","A7FB","A7F4","A7FE","FCF3","FCF2","AA57","CCD4","AA43","AA4D","AA4E","AA46","AA58","AA48","CCDC","AA53","CCD7","AA49","CCE6","CCE7","CCDF","CCD8","AA56","CCE4","AA51","AA4F","CCE5","CCE3","CCDB","CCD3","CCDA","AA4A","AA50","AA44","CCDE","CCDD","CCD5","AA52","CCE1","CCD6","AA55","CCE8","AA45","AA4C","CCD9","CCE2","AA54","AA47","AA4B","CCE0","FD4D","CF5B","AC5C","AC69","FD5E","CF56","CF4C","AC62","CF4A","AC5B","CF45","AC65","CF52","CEFE","CF41","CF44","CEFB","CF51","CF61","AC60","CF46","CF58","CEFD","CF5F","CF60","CF63","CF5A","CF4B","CF53","AC66","AC59","AC61","AC6D","AC56","AC58","FCF6","CF43","AC6A","AC63","CF5D","CF40","AC6C","AC67","CF49","AC6B","CF50","CF48","AC64","CF5C","CF54","AC5E","CF62","CF47","AC5A","CF59","CF4F","AC5F","CF55","AC57","CEFC","AC68","AEE3","AC5D","CF4E","CF4D","CF42","CF5E","CF57","AC55","FCFB","FE4F","D1EC","AEEA","D1ED","D1E1","AEDF","AEEB","D1DA","FAC9","D1E3","D1EB","D1D9","D1F4","AED5","FCF8","D1F3","D1EE","D1EF","AEDD","AEE8","D1E5","D1E6","D1F0","D1E7","D1E2","D1DC","D1DD","D1EA","D1E4","FDA9","AED6","AEDA","D1F2","D1DE","AEE6","AEE2","FC44","AEE5","AEEC","AEDB","AEE7","D1E9","AEE9","AED8","AED7","D1DB","D1DF","AEE0","D1F1","D1E8","D1E0","AEE4","AEE1","AED9","AEDC","FCFE","D5C4","D5B4","D5B5","D5B9","D5C8","D5C5","D5BE","D5BD","B1ED","D5C1","D5D0","D5B0","D5D1","D5C3","D5D5","D5C9","B1EC","D5C7","B1E7","B1FC","B1F2","B1F6","B1F5","D5B1","D5CE","D5D4","D5CC","D5D3","D5C0","D5B2","D5D2","D5C2","B1EA","B1F7","D5CB","B1F0","D5CA","D5B3","B1F8","B1FA","D5CD","B1FB","B1E9","D5BA","D5CF","FB7C","B1EF","B1F9","D5BC","D5C6","D5B7","D5BB","B1F4","D5B6","B1E8","B1F1","B1EE","D5BF","AEDE","D9C0","B1EB","FE4A","FD45","B1F3","D9C3","D9D9","D9CE","B4D6","FEE0","B4D1","D9BD","B4D2","D9CD","D9C6","D9D3","B4CE","D9AB","D9D5","B4C4","D9B3","B4C7","B4C6","B4D7","D9AD","D9CF","D9D0","B4C9","B4C5","D9BB","B4D0","D9B6","D9D1","B4CC","D9C9","D9D6","D9B0","D9B5","D9AF","B4CB","D9C2","DDDE","D9B1","B4CF","D9BA","D9D2","B4CA","D9B7","D9B4","D9C5","B4CD","B4C3","B4D9","D9C8","D9C7","FD48","FD47","FEF2","FE6A","D9AC","B4C8","D9D4","D9BC","D9BE","D9CB","D9CA","D9AA","B4D3","B4D5","D9B2","D9B9","D9C1","B4D4","D9B8","D9C4","D9D7","D9CC","D9D8","D9AE","DDF2","B7A6","DDF0","DDDB","DDE0","DDD9","FD51","DDEC","DDCB","DDD2","DDEA","DDF4","DDDC","FAAD","DDCF","DDE2","DDE7","DDD3","DDE4","DDD0","DDD7","DDD8","B7A8","DDEB","DDE9","DDCC","DDEE","DDEF","DDF1","B7AC","B7A4","D5B8","DDD4","DDE6","DDD5","B7A1","B7B1","DDED","B7AF","B7AB","DDCA","B7A3","FD4E","DDCD","B7B0","DDDD","DDC9","B7A9","DDE1","DDD1","B7AA","DDDA","B77E","B4D8","DDE3","D9BF","DDCE","FD44","DDE8","B7A5","DDE5","B7A2","DDDF","B7AD","DDD6","DDF3","B7A7","DEC6","B7AE","E24A","E248","E25E","E246","E258","B77D","BA5F","E242","E25D","FD52","E247","E255","BA64","BA5D","E25B","E240","E25A","BA6F","E251","E261","BA6D","E249","BA5E","E24B","E259","BA67","E244","BA6B","BA61","E24D","E243","E1FC","E257","BA68","E260","E1FD","BA65","E253","BA66","E245","E250","E24C","E24E","BA60","E25F","BA6E","E24F","E262","E1FE","E254","BA63","BA6C","BA6A","E241","E256","BA69","BA62","E252","E25C","FD41","E5D5","E5D1","E5CD","E5E1","E5DE","BCCD","E5E5","E5D4","BCD8","E5DB","E5D0","E5DA","BCD5","E5EE","E5EB","E5DD","E5CE","FD57","FCEF","E5E2","E5E4","BCD1","E5D8","E5D3","E5CA","BCCE","BCD6","E5E7","BCD7","E5CB","E5ED","E5E0","E5E6","BCD4","FD42","E5E3","E5EA","BCD9","BCD3","E5DC","E5CF","E5EF","E5CC","E5E8","BCD0","E5D6","E5D7","BCCF","BCCC","E5D2","BCD2","BCCB","E5E9","E5EC","E5D9","E9CA","FE7B","E9C2","E9BE","BEF6","BEEB","BEF0","BEEC","E9CC","E9D7","BEEA","E9C4","E9CD","E5DF","E9CE","BEF1","FD5A","E9DD","BEF5","BEF8","E9C0","BEF4","E9DB","E9DC","E9D2","E9D1","E9C9","E9D3","E9DA","E9D9","BEEF","BEED","E9CB","E9C8","E9C5","E9D8","BEF7","E9D6","BEF3","BEF2","E9D0","E9BF","E9C1","E9C3","E9D5","E9CF","BEEE","E9C6","E9D4","E9C7","C0CF","ED45","C0C8","ECF5","ED41","C0CA","ED48","ECFC","ECF7","FBF2","ED49","ECF3","ECFE","C0D1","ED44","ED4A","ECFD","C0C9","ED40","ECF4","C0D0","ED47","ECF9","C0CC","FD5C","ECFB","ECF8","C0D2","ECFA","C0CB","C0CE","ED43","ECF6","ED46","ED42","C263","EFE7","C268","C269","C262","EFE6","EFE3","EFE4","C266","EFDE","EFE2","C265","EFDF","C267","C264","EFDD","EFE1","EFE5","FD5F","F251","F24E","F257","F256","F254","F24F","C372","F250","C371","C0CD","F253","C370","F258","F252","F24D","EFE0","C36F","F24C","F456","F455","F255","C468","F459","F45A","F454","F458","F453","F5D1","F457","C4E7","C4E5","F5CF","F5D2","F5CE","F5D0","C4E6","F6E5","F6E6","C576","F6E4","F7E2","C5CF","F7E0","F7E1","F8AC","C656","F8F3","F8F1","F8F2","F8F4","FD62","F9BB","A4ED","A6B8","AA59","CCE9","CF64","D1F5","D1F7","D1F6","D1F8","B1FD","D5D7","D1F9","FD65","D5D6","D5D8","D5D9","D9DA","B4DB","D9DB","D9DD","B4DC","B4DA","D9DC","DDFA","DDF8","DDF7","DDF6","DDF5","B7B2","DDF9","BA70","E263","E265","BA71","E264","BCDB","BCDA","E5F0","E9DF","E9DE","E9E0","BEF9","ED4B","C0D3","EFE8","C26A","F259","C577","A4EE","A5BF","A6B9","A842","AA5A","AA5B","AC6E","D1FA","B7B3","FD66","E6D1","BEFA","C26B","A4EF","A6BA","CCEB","AA5C","CCEA","CF65","AC6F","CF66","AC70","D1FC","AEEE","AEED","D5DE","D5DC","D5DD","D5DB","D5DA","D9DE","D9E1","B4DE","D9DF","B4DD","D9E0","DDFB","E266","E267","E268","E5F3","E5F2","BCDC","E5F1","E5F4","E9E1","E9E2","E9E3","ED4C","C0D4","C26C","F25A","C4E8","C95F","AC71","CF67","AEEF","B1FE","B4DF","D9E2","B7B5","B7B4","E269","E26A","BCDD","BCDE","E9E5","E9E4","EFE9","F7E3","A4F0","C960","A5C0","A843","CB48","AC72","B7B6","A4F1","CF68","AC73","CF69","C0D5","A4F2","FD71","CCEC","CF6A","FD6F","D242","D241","D1FE","D1FD","D243","D240","B240","B241","B4E0","D9E3","D9E4","D9E5","DE41","DE42","DE40","DDFD","DDFE","B7B7","E26B","E5F7","E5F6","E5F5","E5F8","E9E7","E9E6","BEFB","E9E8","C0D6","ED4D","EFEA","F25B","F6E7","A4F3","A5C2","A5C1","AA5D","C961","C97E","A6BB","C9F7","CB49","CB4A","AA5E","CCED","AC74","CF6B","CF6C","AEF0","AEF4","D244","AEF3","AEF1","AEF2","D5DF","B242","B4E3","B4E1","B4E2","D9E6","BA72","A4F4","C9A1","FD72","A5C3","C9A4","A5C6","C9A3","A5C5","A5C4","A844","C9A2","C9F8","FAE4","C9FC","C9FE","CA40","A6C5","A6C6","C9FB","A6C1","C9F9","C9FD","A6C2","A6BD","A6BE","FD76","A6C4","C9FA","A6BC","A845","A6BF","A6C0","A6C3","CB5B","CB59","CB4C","A851","CB53","A84C","CB4D","CB55","FB62","CB52","A84F","CB51","A856","CB5A","A858","A85A","CB4B","FD78","A84D","CB5C","A854","A857","CD45","A847","A85E","A855","CB4E","A84A","A859","CB56","A848","A849","CD43","CB4F","A850","A85B","CB5D","CB50","A84E","A853","CCEE","A85C","CB57","A852","A85D","A846","CB54","A84B","CB58","CD44","AA6A","AA7A","CCF5","AA71","CD4B","AA62","AA65","CD42","CCF3","CCF7","AA6D","AA6F","CCFA","AA76","AA68","AA66","AA67","AA75","CD47","AA70","CCF9","CCFB","AA6E","AA73","CCFC","CD4A","AC75","AA79","FAC7","AA63","CD49","CD4D","CCF8","CD4F","CD40","AA6C","CCF4","AA6B","AA7D","AA72","CCF2","CF75","AA78","AA7C","CD41","CD46","AA7E","AA77","AA69","AA5F","AA64","CCF6","AA60","CD4E","CCF0","CCEF","CCFD","CCF1","AA7B","AEF5","AA74","CCFE","AA61","ACA6","CD4C","CF7C","CFA1","CFA4","CF77","CFA7","CFAA","CFAC","CF74","AC76","AC7B","D249","ACAD","CFA5","CFAD","CF7B","CF73","D264","AC7E","CFA2","CF78","CF7A","ACA5","CF7D","AC7D","CF70","CFA8","CFAB","AC7A","ACA8","CF6D","ACAA","AC78","ACAE","CFA9","CF6F","ACAB","D25E","CD48","AC7C","AC77","CF76","CF6E","ACAC","ACA4","CFA3","ACA9","ACA7","CF79","ACA1","CF71","ACA2","ACA3","CF72","CFA6","AC79","CF7E","D24C","AEFD","AF43","FAF3","FDAE","D255","D25B","D257","D24A","D24D","D246","D247","AF4A","AEFA","D256","D25F","AF45","AEF6","AF40","D24E","AF42","D24F","D259","FBAF","AF44","D268","D248","AEFC","AEFB","AF48","D245","D266","D25A","D267","D261","D253","D262","D25C","D265","D263","AF49","D254","AEF9","AEF8","AF41","AF47","D260","AF46","D251","B243","D269","D250","D24B","AEFE","AF4B","AEF7","FDAD","D258","D25D","B265","D5E1","D5E5","B252","B250","B247","D5E3","D5E2","B25B","D5E8","B255","D5FA","D647","B244","D5F7","D5F0","B267","D5E0","D5FC","B264","B258","B263","B24E","D5EC","D5FE","D5F6","B24F","B249","D645","D5FD","D640","B251","B259","D642","D5EA","D5FB","D5EF","D644","B25E","B246","B25C","D5F4","D5F2","D5F3","B253","D5EE","D5ED","B248","D5E7","D646","B24A","D5F1","B268","B262","D5E6","B25F","B25D","B266","D5F8","B261","D252","D5F9","B260","D641","B245","D5F5","B257","D5E9","B256","B254","B24C","B24B","D9E7","D643","D5EB","D9FC","B24D","B541","B25A","B4EE","D9F6","B4FC","D9EA","B4EB","B4E7","DA49","B4ED","B4F1","B4EC","B4F5","DA4D","DA44","FEF9","D9F1","B4FA","B4F4","D9FD","B4E4","DA4A","DA43","B4E8","D9F7","B4F7","DA55","DA56","B4E5","DA48","B4F9","D9FB","D9ED","D9EE","B4FD","D9F2","D9F9","D9F3","B4FB","B544","D9EF","D9E8","D9E9","D9EB","B4EA","D9F8","B4F8","B542","FDC0","FCF9","D9FA","DA53","DA4B","B4E6","DA51","B4F2","B4F0","FB7E","DA57","B4EF","DA41","D9F4","D9FE","B547","DA45","DA42","D9F0","B543","DA4F","DA4C","DA54","B4E9","DA40","B546","DA47","B4F3","B4F6","DA46","B545","D9F5","D5E4","DA50","DA4E","DA52","FDAF","D9EC","B540","DE61","DE60","DE46","B7BD","DE5F","DE49","DE4A","B7C7","DE68","B7C2","DE5E","DE43","B7C8","B7BE","DE52","DE48","DE4B","DE63","B7B8","DE6A","DE62","B7C1","DE57","B7CC","B7CB","B7C5","DE69","B7B9","DE55","DE4C","DE59","DE65","B7CD","FD68","B7BB","DE54","DE4D","B7C4","B7C3","DE50","DE5A","DE64","DE47","DE51","B7BC","DE5B","B7C9","B7C0","DE4E","B7BF","DE45","DE53","DE67","B4FE","BAB0","DE56","E26C","DE58","DE66","B7C6","DE4F","B7BA","B7CA","BCF0","DE44","DE5D","FAC0","FA64","DE5C","E2AA","BAAD","E27D","E2A4","BAA2","E26E","BAAF","BA77","E26D","E2B0","BAB1","E271","E2A3","FDC7","E273","E2B3","E2AF","BA75","BAA1","E653","BAAE","BA7D","E26F","FDB0","E2AE","BAA3","E2AB","E2B8","E275","E27E","E2B6","E2AC","BA7C","E27C","BA76","BA74","BAA8","FCC6","E27A","E277","E278","E2B2","E2B7","E2B5","BA7A","E2B9","BA7E","BAA7","E270","E5FA","E279","BA78","BAAC","BAA9","BA7B","E2A5","E274","BAAA","E2A7","BAA4","BAA6","BA73","E2A9","E2A1","E272","BAA5","E2B1","E2B4","E27B","E2A8","FE50","BA79","BCDF","E2A6","E5F9","E2AD","FDCC","E276","E644","E64E","BCE2","E64D","E659","BCE4","E64B","E64F","BCEF","E646","BCE7","FDCD","E652","E9F0","BCF3","BCF2","E654","E643","E65E","BCED","BCE3","E657","E65B","E660","E655","E649","BCE6","BCE9","BCF1","BCEC","E64C","E2A2","FDCF","E648","E65F","BCE8","BCEB","E661","BCE0","E656","E5FB","E65C","C0DF","E64A","BCE1","E645","BCE5","E5FC","BAAB","E641","FCBA","E65A","E642","E640","BCEA","E658","E5FE","E651","E650","E65D","E647","BCEE","FDC5","E9F3","FDD2","BF49","BEFE","EA40","E9EB","BF41","E9F7","BF48","BF43","E9F5","ED4F","E9FB","EA42","E9FA","E9E9","E9F8","EA44","EA46","BEFD","EA45","BF44","BF4A","BF47","E9FE","BF46","E9F9","E9ED","E9F2","E9FD","BF45","BF42","BEFC","BF40","E9F1","E5FD","E9EC","E9EF","EA41","E9F4","E9EA","ED4E","EA43","E9EE","E9FC","FDD4","ED51","C0E3","C0D7","C0DB","ED53","ED59","ED57","C0D9","C0DA","C0E1","ED5A","ED52","C0DC","ED56","ED55","ED5B","C0E2","C0DD","C0E0","ED54","C0E4","C0DE","C0E5","C0D8","ED58","ED50","EFF7","FDC3","C271","EFF4","EFF6","C26F","EFF2","EFF3","EFEE","E9F6","EFEF","C270","EFEB","C26D","EFF8","C26E","EFEC","EFED","EFF1","C273","C272","EFF0","C378","F25F","F265","C379","F25C","C376","C373","F267","C377","C374","F25E","F261","F262","F263","F266","EFF5","F25D","C375","F264","F268","F260","F45D","C46A","F460","C46B","F468","F45F","F45C","F45E","F462","F465","F464","F467","F45B","C469","F463","F466","F469","F461","F5D3","F5D4","F5D8","F5D9","F5D6","F5D7","F5D5","FDE0","C4E9","C578","F6EB","F6E8","F6E9","F6EA","C579","F7E5","F7E4","F8AF","C5F4","F8AD","F8B0","F8AE","F8F5","C657","C665","F9A3","F96C","F9A2","F9D0","F9D1","A4F5","A6C7","CA41","CB5E","A85F","A862","FAF0","CB5F","A860","A861","FDE1","FDE3","CD58","CD5A","CD55","CD52","CD54","AAA4","FB63","AAA2","CD56","AAA3","CD53","CD50","AAA1","CD57","CD51","AAA5","CD59","CFAF","CFB3","ACB7","FDE2","CFB6","ACAF","ACB2","ACB4","ACB6","ACB3","CFB2","CFB1","ACB1","CFB4","CFB5","CFAE","ACB5","ACB0","FDFD","CFB0","D277","D278","D279","AF50","AF4C","D26E","FDE4","D276","D27B","AF51","D26C","D272","D26B","D275","FDE5","FDE6","D271","AF4D","AF4F","D27A","D26A","D26D","D273","FDE7","D274","D27C","D270","AF4E","B26D","D64E","D650","D64C","D658","D64A","D657","B269","D648","DA5B","D652","B26C","D653","D656","D65A","D64F","D654","B26A","B26B","D659","D64D","D649","D65B","D651","D655","D64B","B548","B549","DA65","B54F","DA59","DA62","DA58","B54C","DA60","DA5E","DA5F","B54A","DA63","FDED","FDF7","DA5C","DA5A","B54B","DA5D","DA61","B54D","DA64","DE70","DE77","DE79","DEA1","FDEE","B7DA","DE6B","B7D2","FDF0","DE7A","B7D7","DEA2","B7CE","FDF4","DE7D","DE6D","DE7E","DE6C","B7DC","DE78","B7CF","DEA3","B7D4","DE71","B7D9","DE7C","DE6F","DE76","DE72","DE6E","B7D1","B7D8","B7D6","B7D3","B7DB","B7D0","DE75","B7D5","B54E","DE7B","DE73","DE74","E2C1","BAB4","E2BD","E2C3","E2BF","BAB6","E2BE","E2C2","E2BA","E2BC","BAB5","E2C0","E2BB","BAB7","BAB2","FDEB","E2C4","BAB3","E667","E664","E670","E66A","E66C","BCF4","E666","E66E","E66D","E66B","E671","BCF7","E668","E66F","BCF5","E663","E665","BCF6","E662","E672","FDEA","E669","EA4A","BF51","FDFB","EA55","EA53","BF4B","EA49","EA4C","EA4D","EA48","BF55","BF56","EA47","EA56","EA51","BF4F","BF4C","EA50","EA4E","BF52","EA52","BF4D","BF4E","EA4F","BF50","EA4B","EA54","BF53","EA57","EA58","BF54","FACF","C0E7","C0EE","ED5C","ED62","ED60","C0EA","C0E9","C0E6","ED5E","C0EC","C0EB","C0E8","ED61","ED5D","ED5F","C0ED","C277","EFFB","C274","C275","EFFD","C276","EFFA","EFF9","F26C","EFFC","F26D","C37A","F26B","F26A","F269","C37B","FDFE","C46C","F46A","F46B","FE41","F5DC","F5DB","C4EA","F5DA","F6EC","F6ED","F7E6","F8B1","FE44","F8F6","F9BC","C679","F9C6","A4F6","AAA6","AAA7","FE47","ACB8","C0EF","A4F7","AAA8","AF52","B7DD","A4F8","B26E","BAB8","C962","FE48","CFB7","D27D","E2C5","C0F0","A4F9","AAA9","CFB8","CFB9","DA66","B550","DEA4","B7DE","E2C6","FE4B","BCF8","FE4C","C37C","A4FA","DA67","A4FB","A6C9","CA42","A6C8","A865","A864","A863","CB60","AAAA","AAAB","CD5B","CFBA","CFBD","ACBA","CFBB","ACB9","CFBC","ACBB","D2A2","D2A1","D27E","AF53","D65D","D65E","B26F","D65C","D65F","B552","B270","FE51","B551","DA6B","DA6A","DA68","DA69","DA6C","DEA6","DEA5","DEA9","DEA8","DEA7","BAB9","E2C9","E2C8","BABA","E2C7","E673","E674","BCF9","EA59","EA5A","F272","C37D","F271","F270","F26E","F26F","C4EB","F46C","F6EE","F8F7","A4FC","C9A5","A5C7","C9A6","CA43","CA44","CB66","CB62","CB61","AAAC","CB65","A867","CB63","A866","CB67","CB64","CD5F","CFBE","CD5D","CD64","AAAD","AAB0","CD65","CD61","CD62","CD5C","AAAF","CD5E","AAAE","CD63","CD60","CFC2","ACBD","ACBE","CFC5","CFBF","CFC4","CFC0","ACBC","CFC3","CFC1","D2A8","D2A5","D2A7","AF58","AF57","AF55","D2A4","D2A9","AF54","AF56","D2A6","D667","D2A3","D2AA","D662","D666","D665","DA6E","DA79","D668","D663","DA6D","B274","B273","D661","D664","B275","B272","B271","D660","D669","DA70","DA77","B554","DA76","DA73","FE58","B556","FE52","FE53","DA75","FE59","DA6F","DA71","DA74","DA72","B555","DA78","B553","B7DF","DEAD","DEAC","DEAA","B7E2","B7E1","DEAE","DEAB","E2CA","BABB","B7E0","DEB0","DEAF","E2CD","E2CB","BCFA","BABC","E2CC","E676","BCFB","E675","E67E","E67D","E67B","E67A","E677","E678","E679","E67C","E6A1","EA5F","EA5C","EA5D","BF57","EA5B","EA61","EA60","EA5E","ED64","ED65","C0F1","C0F2","ED63","C279","EFFE","C278","C37E","C3A1","C46D","F46E","F46D","F5DD","F6EF","C57A","F7E8","F7E7","F7E9","A5C8","CFC6","AF59","B276","D66A","A5C9","C9A7","A4FD","CA45","CB6C","CB6A","CB6B","CB68","A868","CB69","FAE1","CD6D","AAB3","CD6B","CD67","CD6A","CD66","AAB5","CD69","FADE","AAB2","AAB1","FE5B","AAB4","CD6C","CD68","ACC2","ACC5","CFCE","CFCD","CFCC","ACBF","CFD5","CFCB","ACC1","D2AF","CFD2","CFD0","ACC4","CFC8","CFD3","CFCA","CFD4","CFD1","CFC9","FE5E","ACC0","CFD6","CFC7","ACC3","FBD7","FE5A","D2B4","D2AB","D2B6","FACA","D2AE","D2B9","D2BA","D2AC","D2B8","D2B5","D2B3","D2B7","AF5F","AF5D","D2B1","FE74","D2AD","D2B0","D2BB","D2B2","AF5E","CFCF","AF5A","AF5C","FA46","D678","D66D","D66B","FE68","D66C","D673","D674","D670","B27B","D675","D672","D66F","B279","D66E","B277","B27A","D671","D679","AF5B","B278","D677","D676","B27C","FE69","DA7E","FB45","DAA1","B560","DAA7","DAA9","DAA2","B55A","DAA6","DAA5","B55B","B561","B562","DAA8","B558","DA7D","DA7B","DAA3","DA7A","B55F","DA7C","DAA4","DAAA","B559","B55E","B55C","B55D","FE6C","B557","B7E9","DEB7","B7E8","DEBB","DEB1","DEBC","FE73","FE5F","DEB2","DEB3","DEBD","DEBA","DEB8","DEB9","DEB5","DEB4","FDBD","DEBE","B7E5","DEB6","B7EA","B7E4","B7EB","B7EC","FEB9","B7E7","B7E6","FE71","E2CE","BABE","BABD","FBBB","E2D3","BCFC","BABF","FE77","BAC1","E2D4","B7E3","BAC0","E2D0","E2D2","E2CF","FE79","E2D1","FE75","E6AB","E6AA","E6A7","BD40","EA62","BD41","E6A6","FE7C","BCFE","E6A8","E6A5","E6A2","E6A9","E6A3","E6A4","BCFD","ED69","EA66","EA65","EA67","ED66","BF5A","EA63","BF58","BF5C","BF5B","EA64","EA68","BF59","FC71","ED6D","C0F5","C27A","C0F6","C0F3","ED6A","ED68","ED6B","ED6E","C0F4","ED6C","ED67","F042","F045","F275","F040","F46F","F046","C3A2","F044","C27B","F041","F043","F047","F276","F274","FEA7","C3A3","F273","C46E","C4ED","F6F1","C4EC","F6F3","F6F0","F6F2","C5D0","F8B2","A5CA","CD6E","D2BC","D2BD","B27D","DEBF","BF5D","C3A4","C57B","F8B3","A5CB","CD6F","A260","CFD7","CFD8","D2BE","D2BF","B27E","B2A1","DAAB","DEC2","DEC1","DEC0","E2D5","E2D6","E2D7","BAC2","E6AD","E6AC","EA69","BF5E","BF5F","FEA9","ED72","ED6F","ED70","ED71","F049","F048","C27C","F277","F5DE","A5CC","ACC6","B2A2","DEC3","FEAB","A5CD","D2C0","B2A3","B563","B564","A5CE","A5CF","CA46","A86A","A869","ACC7","CFD9","DAAC","A5D0","A5D1","A5D2","A5D3","A86B","A86C","CB6E","CB6D","AAB6","CD72","CD70","CD71","CFDA","CFDB","FEB2","ACCB","ACC9","FEB1","ACCA","ACC8","AF60","AF64","AF63","D2C1","AF62","AF61","D2C2","B2A6","D67B","D67A","B2A4","B2A5","FEB3","B566","B565","DAAE","FEB4","DAAD","B2A7","B7ED","DEC5","B7EE","DEC4","E2D8","E6AE","BD42","EA6A","ED73","C3A6","C3A5","C57C","A5D4","CD73","FEB8","B2A8","E2D9","BAC3","C6D4","CB6F","CB70","CD74","AAB8","AAB9","AAB7","FEBA","ACCF","ACD0","ACCD","ACCE","CFDC","CFDD","ACCC","D2C3","AF68","AF69","FEBB","B2AB","D2C9","AF6E","AF6C","D2CA","D2C5","AF6B","AF6A","AF65","D2C8","D2C7","D2C4","AF6D","D2C6","AF66","AF67","B2AC","D6A1","D6A2","B2AD","D67C","D67E","D6A4","D6A3","D67D","B2A9","B2AA","DAB6","B56B","B56A","DAB0","B568","DAB3","B56C","DAB4","B56D","DAB1","B567","B569","DAB5","DAB2","DAAF","DED2","DEC7","B7F0","B7F3","B7F2","B7F7","B7F6","DED3","DED1","DECA","DECE","DECD","B7F4","DED0","DECC","DED4","DECB","B7F5","B7EF","B7F1","FEBC","DEC9","E2DB","BAC7","E2DF","BAC6","E2DC","BAC5","DEC8","DECF","E2DE","BAC8","E2E0","E2DD","E2DA","E6B1","E6B5","E6B7","E6B3","E6B2","E6B0","BD45","BD43","BD48","BD49","E6B4","BD46","E6AF","BD47","BAC4","E6B6","BD44","FEBD","EA6C","EA6B","EA73","EA6D","EA72","EA6F","BF60","EA71","BF61","BF62","EA70","EA6E","C0F8","ED74","C0F7","ED77","ED75","ED76","C0F9","FEBF","F04D","FEBE","C2A1","F04E","C27D","F04F","C27E","F04C","F050","F04A","C3A7","F278","C3A8","C46F","F04B","C470","C4EE","F5DF","C57E","F6F4","C57D","FEC0","F7EA","C5F5","C5F6","F9CC","C6D5","ACD1","CFDE","B56E","B56F","A5D5","A6CA","CA47","CB71","A86D","AABA","ACD2","ACD3","ACD4","D6A6","D2CB","AF6F","B2AE","D6A5","FEC3","DAB8","B571","DAB7","B570","DED5","BD4A","E6BB","E6B8","E6B9","E6BA","FEC8","ED78","FEC9","F051","F471","F470","F6F5","A5D6","CD75","AF70","B572","DED6","FECA","E2E1","BD4B","EA74","F052","F472","A5D7","AABB","ACD7","CFDF","ACD8","ACD6","ACD5","D2CC","AF71","FECB","AF72","AF73","B2B0","D6A7","B2AF","DAB9","B2B1","B573","DED7","B7F8","B7F9","BAC9","BACA","BD4C","BF64","EA75","BF63","ED79","C0FA","F053","F473","A5D8","A86E","CD78","CD77","AABC","CD76","AABD","CD79","CFE5","ACDB","ACDA","CFE7","CFE6","ACDF","ACDE","ACD9","CFE1","CFE2","CFE3","ACE0","CFE0","ACDC","CFE4","ACDD","D2CF","D2D3","D2D1","D2D0","D2D4","D2D5","D2D6","D2CE","D2CD","FED1","AF75","AF76","D2D7","D2D2","D6B0","FED2","D2D8","AF77","AF74","D6AA","D6A9","D6AB","D6AC","D6AE","D6AD","D6B2","B2B5","B2B2","B2B6","D6A8","B2B7","D6B1","B2B4","D6AF","B2B3","FED3","DABC","DABE","DABA","DABB","DABF","DAC1","DAC2","DABD","DAC0","B574","DEDB","DEE0","DED8","DEDC","FED6","DEE1","DEDD","B7FA","B843","B7FD","DED9","DEDA","BACE","B846","B7FE","B844","B7FC","DEDF","B845","DEDE","B841","B7FB","B842","DEE2","E2E6","E2E8","B840","E2E3","BACC","E2E9","BACD","E2E7","E2E2","E2E5","E2EA","BACB","E2E4","BD4E","E6BF","E6BE","BD51","BD4F","E6BC","BD4D","E6BD","BD50","EA7D","EAA1","EA7E","EA76","EA7A","EA79","EA77","BF66","BF67","BF65","EA78","EA7B","EA7C","BF68","C140","EDA3","C0FC","ED7B","C0FE","C141","FED8","C0FD","EDA2","ED7C","C0FB","EDA1","ED7A","ED7E","ED7D","F055","C2A4","C2A5","C2A2","C2A3","F054","F27B","FCE8","C3A9","F279","F27A","F474","F477","F475","F476","F5E0","C4EF","F7EB","F8B4","C5F7","F8F8","F8F9","C666","A5D9","ACE1","DAC3","DEE3","A5DA","A86F","AABE","FAD8","CFE8","CFE9","AF78","DAC4","B575","B847","C142","EDA4","F27C","F478","A5DB","FEDC","CDA1","CD7A","CD7C","CD7E","CD7D","CD7B","AABF","ACE2","CFF2","CFED","CFEA","CFF1","ACE4","ACE5","CFF0","CFEF","CFEE","CFEB","CFEC","CFF3","ACE3","AF7C","AFA4","AFA3","D2E1","D2DB","D2D9","AFA1","D6B9","AF7A","D2DE","D2E2","D2E4","D2E0","D2DA","AFA2","D2DF","D2DD","AF79","D2E5","AFA5","D2E3","AF7D","D2DC","AF7E","AF7B","FA4F","B2B9","D6BA","D6B3","D6B5","D6B7","D6B8","D6B6","B2BA","D6BB","D6B4","DAC8","B576","DAD0","DAC5","DAD1","DAC6","DAC7","DACF","DACE","DACB","B2B8","B577","DAC9","DACC","B578","DACD","DACA","DEEE","DEF2","B84E","E2F0","B851","DEF0","F9D6","DEED","DEE8","DEEA","DEEB","DEE4","B84D","B84C","B848","DEE7","B84F","B850","DEE6","DEE9","DEF1","B84A","B84B","DEEF","DEE5","E2F2","BAD0","E2F4","DEEC","E2F6","BAD4","E2F7","E2F3","BAD1","E2EF","BAD3","E2EC","E2F1","E2F5","E2EE","FEE1","B849","FEE9","E2EB","BAD2","E2ED","BD54","E6C1","BD58","BD56","BACF","E6C8","E6C9","BD53","FEE2","E6C7","E6CA","BD55","BD52","E6C3","E6C0","E6C5","E6C2","BD59","E6C4","FEE3","E6C6","BD57","FEE7","BF6A","EAA8","EAA2","EAA6","EAAC","EAAD","EAA9","EAAA","EAA7","EAA4","BF6C","BF69","EAA3","EAA5","BF6B","EAAB","C146","FB56","EDAA","EDA5","C145","C143","EDAC","C144","EDA8","EDA9","EDA6","EDAD","F056","C147","EDA7","EDAE","EDAB","F05A","F057","C2A6","F05B","F05D","F05C","F058","F059","F2A3","C3AA","F27E","F2A2","F27D","F2A4","F2A1","F47A","F47D","F479","C471","F47B","F47C","F47E","C472","C474","C473","F5E1","FEE5","F5E3","F5E2","FEE8","F6F6","F8B5","F8FA","A5DC","FEF7","CB72","AAC0","CDA3","AAC1","AAC2","CDA2","CFF8","CFF7","ACE6","ACE9","ACE8","ACE7","CFF4","CFF6","CFF5","D2E8","AFA7","D2EC","D2EB","D2EA","D2E6","AFA6","AFAA","AFAD","AFAE","D2E7","D2E9","AFAC","AFAB","AFA9","AFA8","D6C2","D6C0","D6BC","B2BB","D6BD","B2BC","D6BE","D6BF","D6C1","B2BD","DAD5","FC69","DAD4","DAD3","DAD2","DEF6","B852","DEF3","DEF5","B853","FEF3","B854","DEF4","FEF0","E341","E2F9","E2FA","BAD7","BAD5","BAD6","E343","E342","E2FE","E2FD","E2FC","E2FB","E340","E2F8","E6CB","E6D0","E6CE","FEF5","E6CD","E6CC","E6CF","EAAE","BF6D","C148","EDB0","FEF8","C149","EDAF","F05F","F05E","C2A7","F2A5","C3AB","F4A1","C5A1","F6F7","F8B7","F8B6","C9A8","ACEA","ACEB","D6C3","B856","A5DD","A872","A871","A870","CDA4","FEFC","AAC4","AAC3","ACEE","FDBF","CFFA","CFFD","CFFB","ACEC","ACED","FEFE","CFF9","CFFC","AFB5","D2F3","D2F5","D2F4","AFB2","D2EF","AFB0","AFAF","AFB3","AFB1","AFB4","D2F2","D2ED","D2EE","D2F1","D2F0","D6C6","D6C7","D6C5","D6C4","B2BE","B57D","DAD6","DAD8","DADA","B57C","B57A","DAD7","B57B","DAD9","B579","DF41","DEF7","DEFA","DEFE","B85A","DEFC","DEFB","DEF8","DEF9","B858","DF40","B857","B85C","B85B","B859","DEFD","E349","E348","E344","BAD8","E347","E346","BAD9","BD5E","E6D2","BD5F","BD5B","BD5D","BD5A","BD5C","EAAF","BF70","EAB1","EAB0","E345","BF72","BF71","BF6E","BF6F","EDB5","EDB3","C14A","EDB4","EDB6","EDB2","EDB1","F060","C2AA","C2A8","C2A9","F2A6","F2A7","C3AD","C3AC","F4A3","F4A4","F4A2","F6F8","F6F9","A5DE","CA48","A873","CDA5","AAC6","AAC5","CDA6","D040","ACEF","CFFE","ACF0","AFB6","D2F8","D2F6","D2FC","AFB7","D2F7","D2FB","D2F9","D2FA","D6C8","D6CA","B2BF","D6C9","B2C0","B5A2","B5A1","B57E","DADB","DF44","B85D","B85E","DF43","DF42","E34A","BADB","BADA","E34B","E34C","BD61","BD60","EAB5","E6D3","E6D5","E6D4","EAB4","EAB2","EAB6","EAB3","BF73","EDB7","C14B","EDB8","EDB9","C2AB","C2AC","C475","C5D1","A5DF","D041","D2FD","AFB8","B3BA","B3B9","B5A4","DADD","B5A3","DADC","DF45","BADC","E34D","BADD","C476","F4A5","A6CB","AAC7","CDA7","ACF2","ACF1","D042","D043","D340","D342","AFB9","D344","D347","D345","D346","D343","D2FE","AFBA","D348","D341","D6D3","B2C6","D6DC","B2C3","D6D5","B2C7","B2C1","D6D0","D6DD","D6D1","D6CE","B2C5","B2C2","D6D4","D6D7","B2C4","D6D8","B2C8","D6D9","D6CF","D6D6","D6DA","D6D2","D6CD","D6CB","D6DB","DADF","DAE4","DAE0","DAE6","B5A7","D6CC","DAE1","B5A5","DADE","B5AC","DAE2","B5AB","DAE3","B5AD","B5A8","B5AE","B5A9","B5AA","B5A6","DAE5","B861","DF50","DF53","DF47","DF4C","DF46","B863","DF4A","DF48","B862","DF4F","DF4E","DF4B","DF4D","DF49","BAE1","DF52","B85F","DF51","E35D","BAE8","E358","BAE7","E34E","E350","BAE0","E355","E354","E357","BAE5","E352","E351","BAE4","BADF","E353","BAE2","E359","E35B","E356","E34F","BAE3","BD69","BADE","E35C","E6D9","BD62","E6DB","BD63","BD65","E6DE","E6D6","BAE6","E6DC","E6D8","B860","BD68","BD64","BD66","BD67","BF76","E6DD","E6D7","BD6A","E6DA","EAC0","EABB","EAC5","BF74","EABD","BF78","EAC3","EABA","EAB7","EAC6","C151","BF79","EAC2","EAB8","BF77","EABC","BF7B","EAB9","EABE","BF7A","EAC1","EAC4","EDCB","EDCC","EDBC","EDC3","EDC1","C14F","EDC8","EABF","EDBF","EDC9","C14E","EDBE","EDBD","EDC7","EDC4","EDC6","EDBA","EDCA","C14C","EDC5","EDCE","EDC2","C150","C14D","EDC0","EDBB","EDCD","BF75","FAB8","F063","F061","F067","C2B0","F065","F064","C2B2","F06A","C2B1","F06B","F068","C2AE","F069","F062","C2AF","C2AD","F2AB","F066","F06C","F2A8","C3B2","C3B0","F2AA","F2AC","F2A9","C3B1","C3AE","C3AF","C3B3","C478","F4AA","F4A9","F4A7","F4A6","F4A8","C477","C479","C4F0","F5E5","F5E4","F6FA","F6FC","F6FE","F6FD","F6FB","C5A3","C5A2","C5D3","C5D2","C5D4","F7ED","F7EC","F8FB","F8B8","F8FC","C658","C659","F96D","C67E","A6CC","CDA8","D045","D046","D044","ACF3","D047","D048","D049","D349","D34F","D34D","AFBB","D34B","D34C","D34E","D34A","B2C9","D6DE","B2CB","D6E0","B2CA","D6DF","DAE8","B5AF","DAEA","DAE7","D6E1","B5B0","F9DB","DAE9","DF56","B864","DF54","B865","DF55","B866","BAE9","E361","E35E","E360","BAEA","BAEB","E35F","E6DF","E6E0","BD6B","E6E2","E6E1","A261","EACA","EACB","EAC7","EAC8","BF7C","BF7D","EAC9","C157","C153","C158","C154","C156","C152","C155","C2B3","EDCF","F2AE","F2AD","F4AB","C47A","C47B","F741","F5E6","F740","F8FD","F9A4","A6CD","A874","CDA9","AAC8","ACF6","D04C","ACF4","D04A","ACF9","ACF5","ACFA","ACF8","D04B","ACF7","AFBF","AFBE","D35A","AFC7","D353","D359","AFC3","D352","D358","D356","AFC2","AFC4","D355","AFBD","D354","AFC8","AFC5","AFC9","AFC6","D351","D350","D357","AFC0","AFBC","AFC1","D6F0","D6E9","B5B5","D6E8","B2CF","B2D6","B2D3","B2D9","B2D8","B2D4","D6E2","D6E5","D6E4","B2D0","D6E6","D6EF","B2D1","D6E3","D6EC","D6ED","B2D2","D6EA","B2D7","B2CD","B2D5","D6E7","B2CC","D6EB","D6EE","DAFB","DAF2","B5B2","DAF9","DAF6","DAEE","DAF7","B5B4","DAEF","DAEB","B86C","DAF4","B5B1","DAFA","B5B8","B5BA","DAED","B5B9","DAF0","B5B3","DAF8","DAF1","DAF5","DAF3","B5B6","DAEC","B5BB","B2CE","B5B7","B5BC","B868","DF5D","DF5F","DF61","DF65","DF5B","DF59","B86A","DF60","DF64","DF5C","DF58","DF57","DF62","DF5A","DF5E","B86B","B869","DF66","B867","DF63","E372","BAEE","E36A","BD78","E374","BAF1","E378","BAF7","E365","E375","E362","E377","E366","BAFE","BAFB","E376","E370","BAED","BAF5","BAF4","BAF3","BAF9","E363","BAFA","E371","BAF6","BAEC","E373","BAEF","BAF0","BAF8","E368","E367","E364","E36C","E369","E36D","BAFD","E379","BAF2","E36E","E36F","E36B","BAFC","E6E7","BD70","BD79","BD75","E6E4","BD72","BD76","E6F0","BD6C","E6E8","BD74","E6EB","E6E6","BD73","BD77","E6E5","BD71","E6EF","BD6E","E6EE","E6ED","BD7A","E572","BD6D","E6EC","E6E3","BD7B","E6EA","BD6F","E6E9","BFA2","BFA7","BF7E","EAD8","EACF","EADB","EAD3","EAD9","BFA8","BFA1","EACC","EAD2","EADC","EAD5","EADA","EACE","EAD6","BFA3","EAD4","BFA6","BFA5","EAD0","EAD1","EACD","EAD7","BFA4","EADE","EADD","EDDA","EDD6","C15F","EDD0","C159","C169","EDDC","C161","C15D","EDD3","C164","C167","EDDE","C15C","EDD5","C165","EDE0","EDDD","EDD1","C160","C15A","C168","EDD8","C163","EDD2","C15E","EDDF","C162","C15B","EDD9","C166","EDD7","EDDB","F06E","F074","C2B9","F077","C2B4","C2B5","F06F","F076","F071","C2BA","C2B7","F06D","C2B6","F073","F075","C2B8","F072","F070","F2B8","C3B7","C3B8","C3B4","C3B5","F2B4","F2B2","F2B6","C3BA","F2B7","F2B0","F2AF","F2B3","F2B1","C3B6","F2B5","F4AC","C47E","C47D","F4AD","F4AF","F4AE","C4A1","F5EB","F5E8","F5E9","F5E7","F5EA","C4F2","F5EC","C4F1","F742","C5D5","C5D7","F7EE","C5D6","F8B9","F940","F942","F8FE","F941","C66C","A6CE","ACFB","D26F","AFCA","B2DA","DAFC","DAFD","EADF","C16A","EDE1","C2BB","F2BA","F2B9","C4A2","F5ED","F743","C5F8","CA49","AAC9","A875","D04D","D360","D35B","D35F","D35D","AFCB","D35E","D35C","D6F1","DAFE","DB40","DF69","DF6A","B86E","B86F","DF68","DF6B","DF67","B86D","BB40","B870","E37A","BD7C","E6F1","BD7D","BFA9","EAE2","EAE0","EAE1","EDE4","EDE3","EDE2","F2BB","C3B9","F2BC","F744","C5F9","F8BA","A6CF","AACB","AACA","D04F","ACFC","FDA8","D04E","D362","AFCC","D6F2","D361","B2DC","D6F5","D6F3","D6F4","B2DB","DB42","DB43","DB41","B873","DF6D","DF6C","DF6E","B872","B871","E6F2","E6F4","BD7E","E6F3","EAE3","BFAA","F079","F078","C3BB","F2BD","C3BD","C3BC","F4B0","F5EE","C4F3","A6D0","D050","ACFD","D365","AFCE","D364","D363","AFCD","D6FB","D6FD","D6F6","D6F7","B2DD","D6F8","B2DE","D6FC","D6F9","D6FA","B2DF","B5BE","B5BF","DB44","DF6F","DF70","E37E","BB43","BB41","BB42","E37B","E37C","E37D","E6F9","E6FA","BDA1","E6F7","E6F6","E6F8","E6F5","BFAD","EAE4","BFAB","BFAC","EDE6","C16B","EDE5","EFA8","F07A","F07B","C2BC","C2BD","C16C","F2BE","F2BF","F4B1","C4A3","A6D1","A6D2","ACFE","AACC","AFCF","D051","B5C0","A6D3","AD41","D052","D053","AD40","AD42","A6D4","D054","AFD1","D366","AFD3","AFD0","AFD2","D741","B2E0","D740","D6FE","DF71","E3A1","BDA2","BFAE","EAE6","EAE5","EDE7","F5EF","A6D5","CB73","CDAA","AD43","D055","D368","AFD4","D367","AFD5","D743","B2E2","D742","D744","B2E1","DB46","DB47","DB45","B5C1","B874","B875","BB45","E3A3","E3A2","BB44","E6FB","E6FC","EAE7","C170","C16F","C16D","C16E","C171","F07C","C2BF","C2BE","F2C0","F4B2","C5A5","C5A4","A6D6","D1FB","B877","B5C2","B876","BB46","A6D7","C9A9","A6D8","A6D9","CDAB","CB76","CB77","A877","CB74","A876","A879","CB75","A87B","A87A","CB78","A878","AAD1","AACF","CDAD","AACE","AAD3","AAD5","AAD2","CDB0","CDAC","AAD6","AAD0","A87C","AAD4","CDAF","CDAE","AACD","D05B","AD47","AD48","D05D","D057","D05A","D063","D061","AD49","D067","AD4C","D064","D05C","D059","DB49","D062","AD44","D065","D056","D05F","AD46","AD4B","D060","AD4F","AD4D","D058","AD4A","D05E","AD4E","AD45","D066","AFDA","AFE3","AFD8","AFD6","D36A","AFDE","AFDB","D36C","AFDD","D36B","D369","D36E","AFE2","AFE0","DB48","D36F","D36D","AFD7","AFD9","AFDC","AFDF","AFE1","D74E","B2E4","D745","D747","D748","D750","D74C","D74A","D74D","D751","B2E5","B2E9","D746","D74F","B2E7","B2E6","D74B","D749","B2E3","B2E8","B5C8","DB51","DB4F","B5CA","DB4A","DFA1","B5C9","DB4E","DB4B","B5C5","B5CB","DB50","B5C7","DB4D","BB47","B5C6","DB4C","B5CC","B5C4","B5C3","DF77","DF75","DF7B","DF73","DFA2","DF78","DF72","B87B","B8A3","DF7D","DF76","B87E","B87C","DF7E","B879","B878","DF79","B87D","B5CD","DF7C","DF74","B87A","B8A1","B8A2","BB4C","BB48","BB4D","E3A6","E3A5","E3A7","BB4A","E3A4","BB4B","E3AA","E3A9","E3A8","BB49","E741","E744","BDA8","E743","BDA7","BDA3","BDA4","BDA5","E740","E6FE","BDA6","E742","E6FD","EAE9","EAF3","BFB1","BFB0","EAED","EAEF","EAEA","EAEE","EAE8","EAF1","BFAF","EAF0","EAEC","EAF2","EAEB","C174","EDE8","EDEE","C178","C17A","C177","C176","C175","C173","EDE9","EDEC","C172","EDED","C179","EDEB","EDEA","C2C0","C2C1","F0A1","F07D","F07E","F2C2","F2C1","C3BE","F4B4","C4A4","F4B3","F5F0","F745","C5A6","F943","F944","C5D8","A6DA","AAD7","DB52","BB4E","C17B","EDEF","A6DB","AFE5","AFE4","DB53","FEC4","EAF4","A6DC","AD50","DB54","DB55","DB56","BB4F","BFB2","A6DD","AAD8","D068","AFE6","D370","B2EA","DB57","B8A4","BB50","BFB3","C17C","C2C2","F4B5","A6DE","AAD9","AFE7","D752","B5CE","BB51","E3AB","E745","A6DF","B5CF","DFA3","BB52","A6E0","CDB1","D069","AD51","D372","FD77","AFEA","AFE8","AFE9","AFEB","D371","D757","D754","D756","B2EB","B2ED","B2EC","D753","B2EE","D755","DB58","DB59","DB5A","DFA6","DFA7","DFA5","DFA8","B8A5","DFA4","BB53","E74A","E746","E749","E74B","E748","E747","EAF5","EAF6","EAF7","BFB4","BFB5","EDF1","EDF0","EDF2","F0A3","F0A2","F2C4","F2C5","F2C3","C4A5","F4B6","F4B7","F746","F7EF","F8BB","A6E1","A87D","C17D","A6E2","D758","DB5B","C641","CA4A","CA4B","CA4D","A6E3","CA4E","CA4C","CBA2","CBA3","CB7B","FBEE","CBA1","A8A1","A8A2","CB7C","CB7A","CB79","CB7D","A87E","CB7E","D06A","CDB6","AADC","CDB5","CDB7","AADB","CDBC","AADF","CDB2","CDC0","CDC6","AAE6","CDC3","AAE3","CDB9","CDBF","CDC1","CDB4","AAE2","AADD","CDBA","AAE4","AAE7","AAE1","AADA","CDBE","CDB8","CDC5","AAE9","AAE5","AAE0","CDBD","AFEC","CDBB","AADE","AAE8","CDB3","CDC2","CDC4","AD62","AD5C","AD64","AD61","D071","D074","AD5D","D06B","AD56","AD60","AD63","AD65","D0A2","D077","AD55","D0A1","AD59","AD57","AD52","D06F","D07E","D073","D076","D0A5","FA4D","AD66","D07D","AD5E","D078","D0A4","D075","D079","D07C","D06D","D0A3","D07B","FBE9","D06C","D070","AD5F","AD5A","AD53","AD58","AD54","AD67","D06E","D3A5","AD5B","D07A","CE41","D3A8","AFFA","D376","D3A3","D37D","D3B2","D3AA","D37E","D3A9","D378","D37C","D3B5","AFFD","D3AD","D3A4","AFED","D3B3","D374","D3AC","AFFC","AFF7","D373","AFF5","AFF4","AFF9","D3AB","AFF1","AFF8","D072","DB5C","D3A6","D37A","AFFB","D37B","D3A1","AFFE","D375","D3AF","D3AE","D3B6","AFF3","AFF0","D3B4","D3B0","D3A7","D3A2","AFF6","AFF2","D377","AFEE","D3B1","AFEF","D379","FD55","D75E","D760","D765","D779","B2FC","B2F2","D75D","B2FD","B2FE","D768","D76F","D775","D762","D769","B340","D777","D772","B2FA","B2F8","D76E","D76A","D75C","B2EF","D761","D759","B2F7","B2F9","D766","D763","B2F4","D773","B2F1","D764","D77A","D76C","D76B","B2F0","B2FB","B2F3","D75A","D75F","D770","D776","B341","D75B","D767","D76D","B2F6","D778","D771","D774","FE76","B2F5","DB6C","DB60","B5D7","DB7D","DBA7","DBAA","B5D5","DB68","DBA3","DB69","DB77","B5E2","DB73","B5DF","FAAC","DB74","DB5D","DBA4","B5E8","DBA1","DB75","DBAC","DB70","DFC8","DBAF","B5E6","DB6E","DB7A","B5E9","B5D4","DB72","DBAD","DB6B","DB64","DB6F","DB63","DB61","B5D0","DBA5","DB6A","DBA8","DBA9","B5D8","B5DD","B5D9","B5E1","DB7E","B5DA","DB76","DB66","B5D2","DB5E","DBA2","DBAB","DB65","B5E0","DBB0","DB71","DB6D","B5D1","B5E5","DB7C","B5E7","DB78","B5DC","B5D6","B5DE","B5D3","B5E4","DB79","DB67","DB7B","DB62","DBA6","FA6C","DBAE","DB5F","FC75","DFC7","DFDD","B855","DFCC","FDB9","DFCA","DFB5","B8A9","DFC5","DFD9","DFC1","B8B1","DFD8","DFBF","B5E3","DFCF","DFC0","DFD6","B8B0","B8A8","DFAA","DFB2","DFCB","DFC3","DFDC","DFC6","B8B6","DFD7","B8AD","DFC9","DFD1","DFB6","DFD0","DFE1","DFB1","DFD2","DFDF","DFAB","B5DB","DFB9","DFB8","B8AF","DFBC","DFBE","DFCD","DFDE","B8B2","FECD","B8B3","DFB0","B8AB","DFB4","DFDA","B8B4","B8AC","B8AE","B8B5","DFE0","DFD3","DFCE","DFBB","DFBA","B8AA","DFAC","B8A7","DFC4","DFAD","DFC2","DFB7","DFDB","B8A6","DFB3","DFAF","DFD5","DFAE","BB60","E3D3","E3C2","E3AC","E3CA","BB58","E3BB","E3C5","BB5B","E3BE","BB59","E3AF","E3CD","E3AE","E3C1","E3AD","E3BF","E3C8","E3C6","E3BA","E3B5","E3B3","E3B4","E3C7","E3D2","E3BC","BB5A","E3B7","E3CB","BB5D","E3B6","E3B0","E3C0","BB61","BB55","BB5E","E3B8","E3B2","BB57","DFD4","BB56","E3C3","BB54","BB63","BB5C","E3C4","E3B9","E3B1","E3CC","E3BD","BB62","E3D0","BB5F","E3CF","E3C9","E3CE","E3D1","E773","E774","E767","E766","E762","BDB4","BDAC","E776","E775","DFA9","E75F","E763","E75D","E770","E761","E777","E75A","E758","E764","E76E","E769","BDB6","E74F","E76D","FBA5","BDB7","DFBD","E75B","E752","E755","E77B","E75C","E753","E751","E74E","BDB0","E765","BDAF","BDB3","E760","E768","BDA9","E778","E77C","BDAB","E757","E76B","E76F","E754","E779","BDB2","BDB1","E74C","BDB5","E772","E756","E76A","E750","E75E","E759","BDAD","BDAE","E76C","E77D","E77A","E771","FDB4","E74D","BDAA","EB49","EB40","EB43","FAB9","BFBB","EB45","EAF9","EB41","EB47","BFB8","BFBC","BFB6","FA44","EAFB","EB4C","EB46","EAFC","EB55","EB4F","EAF8","EE46","EAFE","BFB7","EB4A","EB54","BFBF","EB51","EAFD","EB44","EB48","EB42","EB56","EB53","EB50","BFB9","BFBA","BFBE","EAFA","EB57","BFBD","EB4D","EB4B","EB4E","EE53","EE40","EE45","EE52","EE44","EDFB","EE41","C1A2","EDF4","EE4D","EE4F","EDF3","C1A1","EE51","EE49","C1A8","EE50","EE42","C1AA","EDF9","EB52","EE4A","EE47","EDF5","EE55","C1A4","C1A5","EDF7","EE48","EE54","EE4B","EDFD","C1A7","C1A3","EE4C","EDFE","EE56","EDF8","EE43","EE4E","EDFA","EDFC","C2CB","EDF6","C1A9","C2C4","C17E","C1A6","C2C8","F0B3","F0A9","F0A4","F0AA","F0B4","F0B8","F0B7","C2CA","C2C9","F0AB","F0B9","F0AE","F0A6","F0A8","F0A7","F0AD","F0B2","F0A5","F0AC","F0B1","C2C7","F0AF","C2C5","F0B0","C2C3","C2C6","F2D5","F0B5","C3C2","F2CD","F2D1","F2C9","F2CC","F2D4","C3C0","F2D9","F2D2","F2CA","F2DA","F2D3","C3C3","C3C4","F2D7","F2CB","C3BF","C3C1","F2C6","F2CE","F2C8","F2D8","F2D6","F2C7","F2CF","F4BE","C3C5","F2D0","C4A7","C4A9","C4A6","F4C3","F4BB","F4B9","F4BD","F4BA","F4BF","F4C1","C4AA","C4AC","F4C0","C4AD","C4AB","F4C2","FABB","C4A8","C4F4","F5F1","F5F7","C4F6","F4BC","F5F6","F5FD","F5F4","F5FB","F5FA","F4B8","F5F5","F0B6","F5FE","F5F3","F5F8","F5FC","F5F2","F74A","C4F5","F5F9","F7F4","F74B","F749","F747","F748","F74C","C5D9","F7F2","F7F0","F7F5","F7F3","F7F6","C5DA","F7F1","F8BC","F945","F946","F947","F9C7","F9BD","CA4F","AAEA","AD68","D3B8","D3B7","B040","B342","D77C","D77B","B5EA","B8B8","B8B7","B8B9","E3D4","E77E","EB58","EB5A","EB59","C1AB","EE57","F0BA","F9A5","A6E4","CDC9","CDCA","CDC8","CDC7","AAEB","D0A9","D0A7","D0A6","AD69","AD6B","AD6A","D0A8","D3C4","D3C1","D3BF","B041","D3C2","B046","D3BC","D3CB","D3CD","D3BD","B043","D3CE","D3C9","D3BB","D3C0","D3CA","D3C6","D3C3","B048","D3CC","D3BE","D3C7","D3B9","B047","B044","D3C5","D3C8","D3BA","B045","B042","B34C","D7A5","B34B","D7A8","D7AB","B348","B346","D77E","D7A9","D7A7","D7A4","D7AC","D7AD","D7AF","D7B0","D77D","B345","D7A2","D7A1","D7AE","B347","D7A3","B349","B344","D7A6","B34D","B34A","D7AA","B5F1","DBBF","DBB4","B5EE","DFE7","DBBD","DBB1","B5EC","DBB6","B5EF","DBBA","DBB8","B5F2","B5EB","DBB2","DBB5","B5F0","DBB3","DBBE","DBBC","DBB7","DBB9","DBBB","B5ED","DFE8","DFEE","DFE4","DFEA","B8BA","DFE6","B8C0","B8BF","B8BE","DFED","B8C1","B8C2","DFE3","DFF0","B8C3","B8BD","B8BC","DFEC","B8C4","DFE2","DFE5","DFEF","DFEB","E3F4","E3E9","B8BB","BB6A","E3DD","E3F2","E3DE","BB65","E3DB","E3E4","E3DC","BB67","E3D6","E3F1","BB68","E3EE","E3EF","E3D7","BB6D","E3E6","E3E0","E3E7","E3DA","E3F3","E3EB","E3E5","E3D5","BB69","E3EC","BB6C","E3F0","E3EA","BB66","E3E8","E3E2","BB64","E3D9","E3E1","E3ED","E3DF","E3E3","BDC1","DFE9","E7B2","E7BB","E7B1","E7AD","E7AA","BDC2","E7A8","BB6B","E7A1","BDC0","E7A7","BDBF","E7AC","E7A9","E7B9","E7B4","E7AE","E7B3","BDBB","E7AB","E7BE","E7A2","E7A3","E7BA","BDBC","E7BF","BDBE","E7C0","E7B0","E3D8","E7B6","E7AF","E7B8","E7B5","E7A6","BDB9","E7BD","BDBA","E7A4","BDBD","EB64","E7B7","E7BC","FA7A","EB61","BDB8","BFC0","EB6B","EB67","EB65","EB60","EB6F","BFC4","EB5C","EB68","EB69","EB5F","EB5E","EB6C","EB62","EB5D","EB63","EB6E","EB5B","EB6D","EB6A","BFC2","BFC1","BFC3","EB66","F0CB","EE59","C1B1","EE5D","EE5A","EE61","EE67","EE5C","EE70","C1AE","EE6A","EE5F","EE6B","EE66","EE6D","EE5E","C1B3","C1B2","EE60","EE6E","EE58","EE6C","C1AC","EE64","EE63","EE68","EE5B","C1B0","C1B4","EE62","EE69","C1B5","EE65","C1AD","C1AF","F0C7","F0C5","F0CC","F0C9","F0CD","F0BE","F0C6","F0D1","EE6F","F0C2","C2CF","E7A5","F0BD","F0CA","F0C4","F0C1","F0BC","F0BB","F0D0","F0C0","F0BF","C2CD","F0C8","C2CC","C2CE","F0C3","F0CF","F2DE","F2DF","C3C9","F2DC","C3C6","F2E4","C3CA","F2E6","F2DB","F0CE","F2E8","F2DD","C3C7","F2E3","F2E5","F2E0","F2E7","F2E2","F2E1","C3C8","F4C5","F4C6","F4C8","C4AE","C4AF","F4C9","F4C7","F4C4","F642","F645","F641","C4FA","F643","C4F9","C4F8","C4F7","F644","F751","F74F","F74E","F640","F750","F646","F74D","F7F9","F7D7","F7F7","C5DB","F7F8","F7FA","F8BF","C5FA","F8BE","F8BD","C5FB","C65A","F96E","F9A7","F9A6","F9A8","A6E5","D0AA","D3CF","D3D0","DBC0","F647","F8C0","A6E6","AD6C","D0AB","D7B1","B34E","DBC2","DBC1","B5F3","B8C5","E7C1","BDC3","BDC4","BFC5","C5FC","A6E7","D0AC","AAED","D0AE","D0AD","AD6D","D3D1","D3D8","B049","D3D6","D3D4","D3DB","D3D2","D3D3","B04A","B04E","D3DC","B04D","D3DA","D3D7","D3D5","B04B","B04C","D3D9","FEEC","B350","D7B2","B355","D7C2","B354","D7C4","D7B8","B352","D7C3","D7B3","B353","D7BF","D7BB","D7BD","D7B7","D7BE","B34F","D7BA","D7B9","D7B5","D7C0","D7BC","D7B4","D7B6","B351","D7C1","B5F6","DBCD","DBC9","DBCB","DBC6","DBC5","DBC3","DBCA","DBCC","DBC8","DBC7","B5F4","B5F5","DBCF","B8CD","DFF2","DFF8","DFF3","DFF4","F9D8","DFF9","B8CF","B8C7","B8CE","DFF1","DBC4","B8CA","B8C8","DFF7","DFF6","B8C9","B8CB","DFF5","B8C6","B8CC","E3F6","BB74","E442","E441","E3FB","BB76","E440","E3F7","E3F8","BB6E","BB70","E3FD","E3F5","BB72","BB71","E3F9","E3FE","E3FC","BB73","E3FA","FEF1","DBCE","BB6F","E7C2","E7C9","BDC6","E7CD","BDCA","E7C5","E7C3","E7CC","BDC5","E7CB","BDC7","BDC8","E7C4","BDC9","E7CA","E7C6","E7C7","E7C8","BB75","EB70","EB7C","BFCA","EB77","EB79","BFC8","EB71","EB75","EB78","BFC6","BFC9","EB7B","EB73","EB74","EB7A","EB72","EB76","BFC7","EE72","EE71","C1B7","EE77","C1B9","C1B6","EE73","C1BA","EE74","EE75","EE78","C1B8","F0D6","F0D9","F0D3","F0D5","F0D4","F0D7","F0D8","EE76","F0D2","C3CD","F2EC","F2EF","F2F1","F2EA","F2EB","F2EE","F2F0","C3CE","C3CC","C3CB","F2ED","F2E9","F4CA","C4B0","F4CB","F649","C4FB","F64B","C4FC","F648","F64A","C5A8","F752","C5A7","F7FD","F7FC","F7FB","F948","F949","F94B","F94A","CA50","A6E8","AD6E","D7C5","B5F7","DFFA","C2D0","F2F2","A8A3","B357","B356","DBD0","B5F8","DBD2","DBD1","DFFB","B8D0","E443","E446","E445","E444","E7CE","E7D0","E7CF","BFCC","BFCB","C1BB","EE79","EE7B","EE7A","C2D1","F2F4","F2F3","F4CC","C4B1","C4FD","F754","F753","C65B","A8A4","D0AF","AD6F","D7C8","D7C6","D7C7","DBD4","DBD5","E043","DBD3","DFFC","E041","E040","E042","B8D1","DFFE","DFFD","E044","E449","E447","E448","E7D3","E7D1","E7D2","EB7D","EE7C","EE7D","C2D2","F2F5","F4CD","C4B2","F64C","F755","C5A9","F7FE","F94C","A8A5","AD71","AD72","D0B0","D0B1","AD70","B054","B052","B051","B058","B050","B059","D3DD","B056","B053","B057","B055","B04F","B35F","B359","D7CC","B35E","B360","B35A","B35B","D7CA","B358","D7CB","B35D","D7C9","B35C","B644","B646","DBD8","B645","B5F9","B5FD","B8E4","E049","DBDA","B5FE","DBDD","DBDE","B643","DBE0","DBE2","DBE3","DBD7","DBD6","DBE4","B642","DBE1","DBDF","B640","B5FB","B647","DBDB","DBDC","DBD9","B641","B5FC","B5FA","E048","B8DF","B8DA","B8D5","B8E5","B8D6","B8D2","B8E1","B8DE","B8E0","B8D7","B8DC","B8D3","B8D4","E050","E04D","E045","E04A","B8E2","E051","B8E3","B8D9","E047","E04F","E04B","E04E","E04C","B8DD","E046","B8D8","E44C","BB78","BB7B","E44E","BBA5","E44D","BB7D","BDCF","E44F","BBA4","E44B","BBA6","BB79","B8DB","BB7C","BB7A","BB7E","BBA2","BB77","BBA7","BBA3","BBA1","E44A","BDD6","BDD2","BDD9","E7D6","BDDA","E7E2","E7DB","BDCB","E7E3","E7DD","BDD5","E7DE","BDD4","E7E1","BDCE","E7DF","E7D5","BDCD","EBAA","BDD3","BDD0","BDD8","E7D4","E7D8","BDCC","E7D7","E7D9","E7DA","BDD7","E7DC","E7E0","E7E4","BDDB","BFD2","EBA5","EBAB","EBA8","EB7E","EBAC","EBA1","EBA7","BFCD","BFD3","EBAD","BFCF","BFD9","BFD4","EBAF","EBA9","BFD0","EBA2","BFDA","EBA3","EBA4","BFDB","BFD8","BDD1","BFCE","EBB0","BFDC","BFD5","EBAE","BFD1","BFD6","BFD7","C1C3","EEA4","EEAD","EEAA","EEAC","C1C0","EEA5","EEAB","C1BC","EEA7","C1C4","EEA3","EEA8","EEAF","EBA6","EEA9","EEA2","C1BD","EEA1","C1BE","EEB0","C1BF","EEAE","C1C2","EE7E","C1C1","EEA6","F0DC","F0EA","F0E5","F0E7","F0DB","C2D3","F0DA","C2D6","C2D5","F0E9","F0E1","F0DE","F0E4","F0DD","F0DF","F0E8","F0E6","C2D4","F0ED","F0EB","F0E2","F0EC","F0E3","F2F9","C3CF","F341","F64F","C3D6","F0E0","F2F7","C3D2","F2F8","F2FD","C3D4","C3D5","F2F6","F340","F342","F2FA","F2FC","F2FE","F2FB","F343","C3D1","C3D7","C3D3","C3D0","F4D0","C4B7","F4CE","F4D2","F4D3","C4B5","F4D4","F4D1","F4CF","C4B8","C4B4","F4D5","C4B6","C4B3","C4FE","C540","F64E","F64D","F650","F651","C541","F756","F75B","C5AA","F758","F757","F75A","F759","F843","C5DC","F842","F840","F841","C5FE","C5FD","F8C1","F8C2","C640","F94D","F94E","C667","C66D","F9A9","F9C8","A8A6","D7CD","D7CE","E052","E450","E7E5","C1C6","C1C5","F0EE","F344","F844","A8A7","D3DE","B05A","B361","E054","E053","BDDC","E7E6","BDDD","EEB1","C2D7","C676","A8A8","CDCB","D3DF","B362","D7CF","D7D0","DBE5","B648","B8E6","E056","E055","E057","E451","E452","BBA8","BFDD","BDDE","BFDE","EEB5","EEB2","EEB4","EEB3","C1C7","F0EF","F346","F345","CBA4","B05C","B05B","D3E0","D7D1","DBE7","DBE6","B649","E059","E05A","E058","B8E8","B8E7","BBAA","BBA9","E7E7","EBB3","EBB1","EBB2","BFDF","EEB7","EEB6","F0F2","F0F1","F0F0","F347","F9AA","A8A9","AD73","AD74","B05D","B05E","D3E2","D3E1","D7D2","B368","B366","B363","B367","B365","B364","B64A","DBEA","B8ED","B64C","B651","DBEC","B653","B652","B655","DBEB","DBE8","B64F","B64B","B64D","DBE9","B654","B650","B64E","B8EF","B8EE","B8EC","B8F0","B8EA","B8EB","B8E9","E05B","E454","BBAC","BBAD","BBAB","E453","E455","E7EA","E7EC","BDE7","E7ED","BDE0","E7E9","BDDF","BDE9","BDE5","BDE6","BDE2","E7E8","BDE1","E7EE","E7EB","BDE8","BDE3","BDE4","EBB5","EBB7","EBB6","EBB8","BFE0","EBB4","C1CB","EEB8","C1C8","C1CC","C1CA","C1C9","F0F3","F0F6","F0F5","F0F4","C2D8","F348","F349","C3D8","F34A","C3D9","C4BA","C4B9","F652","C542","F653","F75C","C5AB","C5AC","F845","C642","A8AA","B36A","B369","E05C","E05D","BBAE","EBB9","BDEA","EBBA","EEB9","A8AB","D0B2","AD76","AD75","D3E3","B05F","D3E4","D7D5","D7D4","D7D3","DBEE","B658","DBED","B657","DBEF","B656","E05F","E062","E060","E061","E065","E05E","E066","E063","E064","BBB0","E456","BBAF","E7F2","E7F0","BDEB","E7EF","E7F1","BDEC","EBBB","EBBC","C1CD","F34C","F34E","F34B","F34D","F4D6","F654","F96F","A8AC","AD77","D3E5","D3E7","D3E6","D7D8","B36C","D7D6","B36B","D7D9","D7DA","D7D7","DBFB","B660","DBF3","DBF9","B65B","B65E","DBF2","B659","DBF6","E06C","B65D","DBF1","DBF7","DBF4","DBFA","DBF0","DBF8","B65C","B65F","DBF5","B65A","B8F2","E068","B8F1","E06F","E06E","B8F8","B8F9","E070","B8F3","E06D","B8F7","E072","E069","E06B","B8F4","E067","E06A","E071","B8F5","E073","B8F6","BBB1","E45B","E461","E459","E462","E458","E45D","E463","E460","E45F","E45E","E457","E45C","E45A","BDF1","BDEE","E7FB","E841","E843","E840","E7F8","E7FA","E845","E842","E7FC","E846","E7F9","E844","BDEF","BDF5","BDF3","E7F3","BDF4","BDF0","E7F4","E7F6","E7F5","E7FD","E7FE","BDF2","BDED","E7F7","EBC6","BFE2","EBBD","BFE3","BFE6","EBC2","EBBF","BFE5","EBC3","EBC4","EBBE","EBC7","EBC0","EBC5","BFE4","BFE1","EBC1","EEBF","C1D0","C1CE","C1D1","C1CF","EEBE","EEBB","EEBA","EEBD","EEBC","F145","C2DE","F0FB","F0FA","C2D9","F141","F140","F0F7","F143","F0FC","C2DD","F0F9","F142","F0F8","C2DA","C2DC","F0FD","C2DB","F0FE","F144","F352","C3DE","F34F","F353","C3DB","F351","C3E0","C3DD","F350","C3DF","F354","C3DA","C4BC","C4BE","F4D9","C4BD","F4D7","C3DC","F4D8","C4BB","C543","C545","F656","C544","F655","F761","C5AD","F760","C5AE","F75E","F75D","F762","F763","F846","F75F","F8C6","F8C3","F8C4","F8C5","C65C","F951","F950","F94F","F970","F9BE","F9AB","C66E","A8AD","B060","B8FA","BDF6","EBC8","C2DF","F355","F9AC","A8AE","AAEE","AD79","AD78","B063","D3E8","B061","D3E9","B062","D7DF","D7DB","B36D","D7DE","D7DD","D7DC","B36E","D7E0","D7E1","DC43","DC41","DC45","DC46","DC4C","DC48","DC4A","DC42","DBFC","DC49","DC4B","DC44","DC47","DBFD","B662","DC40","DBFE","B661","B663","B8FD","E075","E077","E076","E07B","B8FB","E078","E074","E079","E07A","B8FC","B8FE","E07C","E467","E466","E464","E465","BBB3","BBB5","BBB2","BBB4","E84D","E84E","E849","E84A","BDF8","BDFD","BDF7","BDFE","BDF9","E84B","E84C","E848","BE40","BDFB","BDFA","BDFC","E847","EBCA","BFE8","EBCC","BFEA","EBCF","EBCB","EBC9","EBCE","BFE9","EBCD","BFE7","C1D3","C1D6","EEC1","C1D4","EEC0","C1D2","C1D5","F146","F147","F148","C2E0","F149","C2E1","C3E2","F358","F359","F357","F356","F35A","C3E1","F4DD","F4DB","F4DC","F4DE","F4DA","F4DF","F658","F659","F657","C546","F764","C5AF","F765","F848","F847","A8AF","B664","B940","BBB6","BFEC","BFEB","C3E3","C47C","C547","A8B0","B064","B941","F35B","C6D6","FEEB","CBA6","A8B1","A8B4","A8B3","A8B2","CBA5","CDCD","CDCF","AAEF","AAF1","CDCC","CDCE","AAF0","CDD1","CDD0","CDD2","D0B6","D0B4","AD7C","D0B3","ADA3","AD7E","AD7B","ADA4","AD7D","ADA2","ADA1","D0B5","AD7A","B06A","D3EB","D3F1","B067","B06E","B069","D3EE","D3F0","B06C","D3EA","D3ED","B068","B065","D3EC","B06B","D3EF","B06D","B066","D7E3","D7E6","B370","B37A","B376","D7E4","B37E","B377","B37C","B372","B36F","B371","B37D","D7E5","B375","B378","B374","B379","D7E7","B37B","B373","D7E2","DC4D","B665","DC4F","B667","B669","DC4E","B666","B66A","B668","B947","E0A3","B94F","E07E","B950","B945","E0A1","B94A","E0A2","B943","B942","B94D","B94C","B94B","B949","B94E","E07D","B944","B946","B948","BBB8","BBBB","BBBF","BBB9","BBBE","BBBC","BBB7","BBBD","BBBA","E852","BE43","BE41","E853","BE44","BE42","E851","E850","BFF0","E84F","BFEE","BFED","EBD0","BE45","BFEF","EBD1","BFF2","EBD2","BFF1","C1D8","EEC3","C1D7","C1DC","C1DA","C1DB","C2E3","C1D9","EEC2","EBD3","C2E2","C2E4","C3E4","C3E5","F4E0","C5DE","C5DD","A8B6","CA55","B06F","CA52","CA53","CA51","CA54","CBAA","CBA7","CBAC","CBA8","A8B7","A8BA","CBA9","A8B9","CBAB","A8B8","CDD5","CDD7","AAF4","CDD3","CDD6","CDD4","AAF2","AAF5","AAF3","D0B8","D0BC","D0B9","ADA7","ADA8","D0BB","D0BD","D0BF","ADA5","D0BE","ADA6","D7EE","D0BA","D3F2","D3FB","D3F9","D3F4","D3F5","D3FA","D3FC","B071","D3F7","D3F3","B070","B072","D3F6","D3FD","D3F8","B3A1","D7F1","D7E9","D7EF","D7F0","B3A2","D7E8","D7EA","D0B7","D7EC","D7ED","D7EB","B66C","DC56","EBD4","DC57","DC54","B3A3","B66E","DC53","DC59","DC58","B66B","DC5C","DC52","DC5B","DC50","DC5A","DC55","B66D","E0AA","E0A5","E0AB","E0A6","E0A4","E0A7","B951","E0A9","E0A8","B952","BBC1","BBC0","E46E","E471","E469","E46D","BBC2","E46C","E46A","E470","E46B","E468","E46F","E859","BE48","F14A","E856","E857","E855","DC51","BE47","E85A","E854","BE46","BE49","E858","EBD5","BFF3","EBD6","EBD7","EEC4","C1DD","F14B","F14C","F14D","F35D","F35C","F4E2","F4E1","F65B","F65C","F65A","F766","C5B0","A8BB","ADAA","ADA9","B075","B074","D440","D441","D3FE","B073","D7F5","D7F6","D7F2","B3A4","D7F3","D7F4","DC5F","DC61","DC5D","DC60","B66F","DC5E","B670","DD73","B955","B954","B953","E0AC","E0AD","E473","E475","BBC6","BBC3","BBC5","BBC4","E474","E472","E861","E85E","E85F","BE4D","E860","E85B","E85C","BE4A","BE4B","E85D","BE4C","EBDB","EBDC","EBD9","EBDA","BFF4","EBD8","EEC8","EEC5","EEC7","C1E0","EECB","C1DF","EEC9","EECC","EECA","EEC6","C1DE","F14F","F150","F14E","F152","C2E5","C2E6","F35F","C3E7","F151","F35E","C3E6","F4E5","F4E6","C4BF","F4E4","F4E3","F65D","C548","F849","F8C8","F8C7","C643","C65D","F8C9","F971","C66F","A8BC","AAF6","B956","C4C0","A8BD","ADAB","B3A5","B671","C2E7","AAF7","D0C1","D0C0","D442","FC5E","B078","B076","B07A","D444","B079","B077","D443","B3A8","D7FC","B3A7","B3A9","D842","B3AB","D7FE","D840","D7F7","B3AA","D843","D7F9","D7FA","D7F8","B3A6","D841","D7FB","D7FD","DC6D","DC6C","DC6A","DC62","DC71","DC65","DC6F","DC76","DC6E","B679","B675","DC63","DC69","B677","DC68","B678","B67A","DC6B","B672","B673","DC77","DC75","DC74","DC66","DC72","B676","B674","DC73","DC64","DC67","DC70","E4BA","E0B7","E0B0","E0C3","E0CC","E0B3","B961","E0C0","B957","B959","B965","E0B1","FCFA","B95A","B95C","B966","B95B","B964","E0B9","E0AE","B962","E0B8","B95E","E0CA","B963","E0C8","E0BC","E0C6","B960","E0AF","E0C9","E0C4","E0CB","B958","B967","B95D","E0B5","E0BD","E0C1","E0C5","B95F","E0B4","E0B2","E0BE","E0BB","E0BA","E0BF","E0C2","E0C7","E478","BBC7","E4A4","E47A","BBCC","BBD0","E4AD","E4B5","E4A6","BBC8","E4AA","E0B6","BBC9","E4B1","E4B6","E4AE","E4B0","E4B9","E4B2","E47E","E4A9","BBD1","BBCD","E47C","E4AB","BBCB","E4A5","BBCA","E4B3","E4A2","E479","BBCE","E4B8","E47B","E4AF","E4AC","E4A7","E477","E476","E4A1","E4B4","BBCF","E4B7","E47D","E4A3","BE52","BE5A","BE55","E8A4","E8A1","E867","BE50","F9D7","BE4F","BE56","E865","BE54","E871","E863","E864","BE4E","E8A3","BE58","E874","E879","E873","EBEE","E86F","E877","E875","E868","E862","E87D","BE57","E87E","E878","E86D","E86B","E866","FA41","E86E","E87B","E86A","E87A","E8A2","BE53","E876","E87C","E872","E86C","BE51","E4A8","E870","BE59","E869","EBF4","BFF7","EBF3","EBF0","EC44","BFFB","EC41","EBF8","EC43","EBE9","EBF6","BFFD","EBE1","EBDF","EC42","EC40","EBFE","EBED","EBEC","EBE2","C040","EBE8","EBF2","EBFD","C043","EC45","C1E8","C045","BFFE","EBE6","EBEF","EBDE","EBE0","BFF5","C042","BFFA","EBE7","EBF7","EBF1","C041","EBDD","C1E3","EBF9","EBFC","BFFC","EBEB","C044","BFF9","BFF8","EBF5","EBFB","BFF6","EBE4","EBFA","EBE5","FC55","FE45","FA4B","EBEA","EED2","EED7","C1E5","C1E7","EEDD","C1E1","EEEC","EEE3","EED8","EED9","EEE2","C1EE","EEE1","EED1","EEE0","EED4","EEED","C1ED","C1EB","EED5","EEE8","EEDA","EEE7","FDF5","EEE9","EED0","C1E6","EEEA","EEDE","C1EA","EEDB","C1EC","EEE4","C1E4","EED6","EEE5","EEDF","EBE3","EEE6","EED3","C1E9","EEEB","C1E2","EECE","FEB0","F160","F159","C2E9","F154","F163","F15B","EEDC","F165","F155","C2E8","F15F","C2EA","C2F2","C2F0","F161","C2F1","F157","F158","F15D","F162","EECD","C2EB","F16A","F167","F16B","F15E","F15A","F168","F36A","F15C","C2EE","C2ED","EECF","C2EF","F164","F166","C2EC","F169","F153","F156","F373","F363","C3EB","F371","F361","C3EC","F36C","F368","C3F1","F372","F362","F365","C3E9","F374","FB79","F36D","F370","C3EF","C3F4","C3F2","F369","F364","C3ED","C3EE","F360","C3EA","C3E8","C3F0","F36F","C3F3","F36B","F375","C3F5","F367","F36E","FDCB","FE7A","F4F3","F542","F4F5","F4FC","F366","F4FA","F4E9","F540","C4C3","F4ED","F4FE","F4F4","C4C2","F544","F4F6","F4FB","F4FD","F4E7","F541","F4F2","F4F7","F4EB","F4EF","F543","F4F9","F4E8","F4EC","F4EE","F4F8","C4C1","F4F1","FC45","F4EA","F4F0","F661","F666","C54F","F668","C549","F664","F66A","C54E","C54A","C54B","F660","F667","C54D","F665","C54C","F65F","F663","F662","F65E","F669","FE40","FE43","C5B1","F76D","F770","F76C","F76E","F76F","F769","F76A","F767","F76B","F768","C5B2","C5B3","F84B","F84D","F84C","F84E","C5E0","F84A","C5DF","C5E1","F8CB","F8CC","C644","F8CA","F953","F952","F954","C65F","F955","C65E","F956","F972","F975","F974","C668","F973","FCC1","C672","C670","C671","C677","F9C0","F9C1","F9BF","F9C9","AAF8","D844","DC78","E8A5","F376","AAF9","ADAC","B07B","D845","D846","B3AC","B67D","DC7A","DC79","B6A3","B67C","DC7B","B67E","B6A2","B6A1","B67B","B968","E0D0","E0CE","E0CF","E0CD","BBD2","BBD5","BBD7","BBD6","BBD3","BBD4","E8A7","E8A6","BE5B","E8A8","E8A9","BE5C","EC4D","EC4B","EEF3","EC49","EC4A","C046","EC46","EC4E","EC48","EC4C","EEEF","EEF1","EEF2","C1F3","EEEE","C1F2","EEF0","C1EF","C1F0","C1F1","EC47","C2F5","F16E","F16C","F16D","C2F3","C2F6","C2F4","F377","F378","C3F6","F545","F547","F546","C4C4","C550","F66D","F66C","F66B","AAFA","C9AA","CA58","A6E9","CA56","CA59","CA57","CBAE","A8C1","A8C2","CBB0","A8BF","CBAF","CBAD","A8C0","A8BE","CDD8","CDDB","AAFD","CDDA","CDD9","AAFC","AAFB","AB40","CDDC","AAFE","D0C6","ADAE","ADAF","ADB0","D0C7","D0C3","ADAD","D0C4","D0C5","D0C2","B0A4","B0A1","D445","B0A2","B0A5","D446","B07E","B07C","B07D","B0A3","B3AD","D849","B3B5","D848","D84B","B3B1","D84A","B6AB","B3AF","B3B2","B3AE","B3B3","B3B4","B3B0","D847","B6A7","DC7D","DCA3","DCA2","B6AC","B6A8","B6A9","DC7C","DC7E","DCA1","B6A4","B6A6","B6AA","B6A5","E0D3","E0D1","E0D2","B96A","B96B","E0D4","B969","BBD8","BBDA","BBD9","E4BB","E4BC","E8AB","E8AA","FEE4","C047","C048","EC4F","C049","EEF6","EEF4","EEF5","C1F4","F16F","C3F7","C6D7","C1F5","AB41","B0A6","D447","D84C","B3B6","B6AD","DCA4","DCA6","B6AF","B6AE","B6B0","B6B1","DCA5","B96E","B96F","B96D","BBDB","B96C","E0D5","BBDC","E8AC","EC50","C04A","C1F6","F170","F174","C2F9","F171","C2FA","C2F8","F175","C2FB","F173","F379","C2F7","C3F8","F8CD","AB42","B3B8","B3B7","B6B2","DCA8","DCA7","B6B3","E0D9","B973","B970","E0D8","B972","E0D6","B971","E0D7","E4BD","BBDD","E8AF","BE5D","E8AD","BE5E","BE5F","E8AE","BE60","EC51","C04E","C04B","C050","EC53","C04C","EC52","C04F","C04D","EEF9","EEFB","C1F7","EEFA","C1F8","EEF8","EEF7","F177","F176","C2FC","F178","F37E","C3FA","F37D","F37A","C3F9","F37B","F37C","F548","F549","C4C5","C553","F66E","C551","C552","F66F","C5B4","C5B5","F771","C645","F8CF","C647","F8CE","F8D0","C646","F957","F9AD","AB43","B974","E4BE","E8B0","C051","C052","AB44","BE61","C3FB","ADB1","C053","C5E2","ADB2","D84D","DCA9","DCAB","DCAA","E0DD","E0DA","B975","B976","E0DB","E0DC","E4C0","E4C5","BBDE","E4BF","E4C1","E4C8","E4C3","E4C7","E4C4","E4C2","E4C6","BBDF","FB58","E8B3","E8B1","BE63","BE62","E8B2","BE64","EC56","EC55","C054","EC54","EEFC","EEFE","EF41","EF40","C1F9","EEFD","F1A1","C2FD","F17D","F1A2","C2FE","F17B","F17E","F17C","F179","C340","F17A","F3A1","F3A3","F3A2","F54A","F54B","FC52","F670","C5B7","C5B6","F84F","F850","C648","F8D1","C669","ADB3","B6B4","E4CA","E4C9","E8B5","E8B4","C1FA","EF43","EF42","F1A5","F1A3","F1A6","F1A4","C3FC","F3A4","F3A5","F3A6","F671","F772","F8D2","ADB4","EC57","EF44","ADB5","BBE0","EC58","C341","F1A7","C3FD","F54C","F54D","C554","F851","ADB6","B3BB","B3BC","D84E","B6B5","B6B6","DCAC","B6B7","B97A","B97C","E0DF","E0E0","E0DE","B977","B978","B97B","B979","FCBC","E4CB","BBE1","BBE2","E8BC","BE67","E8B7","E8B6","E8BB","BE65","C05B","E8B8","E8BD","E8BA","E8B9","BE66","C059","EC5A","C055","EC5B","EC59","C058","C056","C05A","C057","EF45","EF4A","EF46","EF49","C1FB","EDD4","EF48","EF47","C344","C342","C345","C343","F1A8","F1A9","F1AA","C346","F3AA","C440","F3A8","C441","F3A7","F3A9","C3FE","F551","F54E","F54F","F550","F672","C556","C555","F774","F773","C5B8","FA6A","C5E3","C649","C660","F958","F9AE","F9AF","ADB7","DCAD","E0E1","E4CC","E4CD","BBE3","BBE4","E8BE","BE68","C1FC","F1AB","C347","F3AD","C442","F3AC","F3AE","F3AB","F675","F552","F553","C4C6","F674","F673","F775","F9B0","ADB8","ADB9","B0A7","D448","D84F","B6B8","B6BB","B6B9","DCAE","B6BD","B6BA","B6BC","B97E","E0E2","E0E3","E8C0","B97D","B9A1","B9A2","E4CF","E4CE","BBE5","BBE6","E4D0","E8BF","BBE8","BE69","BBE7","C05C","E8C1","BE6B","BE6A","E8C2","E8C5","E8C3","E8C4","BE6C","C061","C05F","C05E","EC5D","C060","EC5C","EF4B","EC5E","C05D","EC5F","EF4E","EF4C","EF4D","EF52","C34B","EF51","EF54","EF53","EF50","EF4F","C1FD","F1AE","F1AD","C34A","C348","C349","F1AC","F3B1","C443","F3B0","F3AF","C444","F558","F557","F555","F554","C4C8","C4C7","F559","F776","C5B9","F677","C557","F676","F556","F777","C5E4","C661","F959","F9B1","ADBA","D850","EF55","ADBB","E4D2","E4D1","EC60","EF57","EF56","FCEA","C34C","F3B2","F3B3","C4C9","F9B2","B0A8","B6BF","B6BE","E0E4","E0E6","B9A4","E0E5","B9A3","B9A5","E0E7","E4D4","E4D6","E4D5","E4D8","BBE9","E4D7","E4D3","E4D9","E8CC","E8CF","E8D1","E8C7","E8CB","E8C8","BE6E","BE71","BE73","E8C9","E8CA","BE72","E8CD","E8D0","E8CE","BE74","BE70","E8C6","BE6D","BE6F","C063","EC66","EC64","EC63","EC69","EC68","EC67","EC62","C062","EC61","EC65","C064","EF5A","EF5E","EF5B","EF5D","EF5C","EF59","EF5F","EF62","EF60","EF61","C240","C1FE","EF58","EF63","F1B3","F1B6","F1B8","F1B7","F1B1","F1B5","F1B0","F1B2","C34D","F1AF","F1B4","F3C0","F3B5","C445","C446","F3B4","F3B9","F3BF","F3B7","F3BE","F3BB","F3BA","F3BD","F3B8","F3B6","F3BC","F560","F55E","C4CA","F55D","F563","F561","C4CB","F55C","F55A","F55B","C4CD","F55F","C4CC","F562","F678","F67E","F679","C55B","F6A1","C55A","F67D","F67C","C559","F67B","C558","F67A","F77D","F7A1","F77E","F77B","C5BB","F778","F77C","F7A3","F7A2","F779","F77A","C5BA","F852","C5E7","F853","C5E5","C5E6","F8D3","C64A","F976","C66A","F9B3","C66B","F9B4","F9B5","F9C3","F9C2","C67A","F9CD","B0A9","E0E9","E0E8","BBEA","BBEB","E4DA","E8D2","EC6C","BE75","C065","EC6A","EC6D","C066","EF64","EC6B","F1B9","C34E","F3C1","F566","F564","F565","F6A2","C55C","F7A4","C5EA","C5BC","C5E8","C5E9","F8D4","C662","B0AA","F1BA","D449","B9A6","E4DB","BBEC","E4DC","E8D4","E8D3","C068","BE76","BE77","E8D7","E8D6","E8D5","EC6E","EC71","EC70","EC6F","C067","EF68","EF66","EF65","EF67","C34F","F1BC","F1BD","C350","F1BB","F3C3","F3C2","F3C5","C447","F3C4","F567","F569","F568","F6A3","F6A6","F6A4","F6A5","F7A5","C5BD","F854","F855","F856","C64B","C663","F9B6","B0AB","BE78","C069","F1BE","F7A6","F9C4","D44A","C67B","B0AC","EC72","F1BF","F3C6","F6A7","F7A7","B0AD","E4DD","E4DE","BBED","BBEE","E8D9","BE7A","BE79","E8D8","EF69","F1C0","F1C2","F1C1","C353","C352","C351","C55E","F6A8","C55D","F7A9","F7A8","C64C","F8D5","B3BD","E0EA","E4E1","E4DF","E4E0","E8E2","E8DD","E8DA","E8E1","E8E3","BE7C","E8E0","E8DC","E8DB","E8DF","E8DE","BE7B","EC7D","EC78","EC76","ECA1","EC77","EC73","EC79","FDA5","EC74","EF72","EC75","ECA2","EC7C","C06A","EC7B","EC7A","EC7E","EF6A","EF6D","EF6C","EF74","EF6F","EF73","EF71","EF70","EF6E","EF6B","C243","C242","C244","C241","EF75","F1C8","F1CB","F1C9","F1CD","F1CE","F1C6","C358","F1C7","F1C5","F1CC","F1C4","F1C3","C357","C355","C354","F1CA","F3CF","F3D5","C44A","F3D0","F3D3","F3D7","C44B","F3D2","F3CA","F3C9","F3D6","F3CD","F3CB","F3D4","F3CC","C449","C448","F3C7","F3C8","F3D1","F3CE","F56C","F56F","C356","F56D","F573","F571","F56B","F576","F56A","C4CF","F572","F56E","C4CE","F575","F574","F6AB","F6AA","F6B1","F6AD","F6B0","C560","F6AE","F6AF","F6A9","F6AC","C55F","C5BF","F7B4","F7AF","F7B3","F7B6","F7B2","F7AE","C5C1","F7B1","F7B5","C5C0","F7AC","F570","F7B0","F7AD","F7AA","F7AB","C5BE","F85A","F85C","F85F","F85B","F860","F859","F857","C5EB","F85D","C5ED","C5EC","F858","F85E","F8DA","C64D","F8DB","F8D9","F8D6","F8D8","F8D7","F95A","F95C","F95B","F979","F978","F977","F97A","C673","C674","F9CA","F9CE","B3BE","DCAF","E0ED","B9A7","E0EB","E0EC","E4E2","E4E3","BBF1","BBEF","E4E4","BBF0","E8E8","E8EB","E8E5","E8EC","E8E4","E8E6","E8E7","E8EA","BEA1","E8EF","E8EE","BE7D","E8E9","E8ED","BE7E","ECAC","C06F","ECA7","C06B","ECA4","ECAA","ECAD","C070","ECA9","ECA6","ECAE","ECA5","ECAB","C06C","ECA3","C06D","C06E","ECA8","EFA9","EF7A","EF7B","EF7E","EF7C","EF76","FAA1","EF79","EFA5","EF7D","C245","EFA7","EFA4","C246","EFA6","EF77","EFA2","EFA3","EFA1","F1D2","F1D4","F1D7","F1D1","C359","F1D9","F1D0","F1DA","F1D6","F1D8","F1DC","F1D5","F1DD","F1D3","F1CF","C35A","F1DB","C35B","C44D","EF78","F3F1","F3E8","C44F","F3E4","C450","F3ED","F3E7","F3DD","C44E","F3EA","F3E5","F3E6","F3D8","F3DF","F3EE","F3EB","F3E3","F3EF","F3DE","F3D9","F3EC","F3DB","F3E9","F3E0","F3F0","F3DC","C44C","F3DA","F3E1","F3E2","F57D","F57B","F5A2","F5AE","F5A5","F57C","F578","F5A7","F57E","F5A3","F57A","F5AA","F577","F5A1","F5A6","F5A8","F5AB","F579","F5AF","F5B0","F5A9","F5AD","F5A4","F6C1","F6C4","C561","F6C3","F6C8","F6C6","C562","F6BD","F6B3","F6B2","C564","F6BF","F6C0","F6BC","F6B4","F6B9","F5AC","F6B5","C563","F6BB","F6BA","F6B6","F6C2","F6B7","F7BB","F6C5","F6C7","F6BE","F6B8","F7BC","F7BE","F7B8","C5C2","F7C5","F7C3","C5C3","F7C2","F7C1","F7BA","F7B7","F7BD","F7C6","F7B9","F7BF","F869","F86E","F864","F867","C5EE","F86B","F872","F7C0","F865","F86F","F873","F86A","F863","F86D","F86C","F871","F870","F7C4","F868","F862","F866","C64E","C64F","F861","F8E6","F8DD","F8E5","F8E2","F8E3","F8DC","F8DF","F8E7","F8E1","F8E0","F8DE","F8E4","F95D","F95E","F960","F95F","F962","F961","F97C","F97B","F9B7","F9B8","F9C5","C678","C67C","F9CF","C67D","B3BF","C4D0","F6C9","C650","C651","B3C0","E0EE","B9A8","E8F0","ECB0","ECB1","ECAF","EFAB","EFAA","C247","F1DF","EFAC","F1DE","F3F3","C451","C453","F3F2","C452","F5B1","F5B3","F5B2","F6CA","C565","C5EF","F8E8","F963","F9D2","B3C1","E4E5","BEA2","ECB3","ECB2","EFAD","C454","C4D1","F7C7","F9CB","B3C2","BBF2","BEA3","F3F4","F874","B6C0","EFAE","C664","B6C1","BEA4","C248","F875","B6C2","E8F1","C072","ECB4","ECB5","C071","EFAF","C24C","C24A","C24B","C249","F1E0","C35C","F5B5","F5B4","F5B7","F5B6","C4D2","F6CB","F6CD","F6CC","C566","F7C8","F876","F877","C5F0","F964","F97D","C675","DCB0","ECB6","EFB0","F3F5","E0EF","EFB1","F1E2","F1E1","F878","C652","F965","F97E","B9A9","E8F2","E8F3","ECB7","B9AA","C35D","F1E3","F6CF","C567","F6D0","F6CE","F879","F8E9","B9AB","EFB4","EFB3","EFB2","F1E4","F1E8","F1E7","F1E6","F1E5","C35E","F3F6","F5B9","C4D3","F5B8","F6D1","F7CB","F7CA","C5C4","F7C9","F87C","F87B","F87A","BBF3","ECB8","C24D","F3F7","F3F8","F7CC","F87D","F8EA","F966","F9B9","F9D4","BBF4","C24E","F1E9","F3F9","F6D2","F87E","BEA6","EFB5","F1EA","F3FA","F3FB","F3FC","F5BE","F5BA","C568","F5BD","F5BC","C4D4","F5BB","C4D6","C4D5","F6D4","F6D3","C569","C56A","C5C6","F7CD","C5C5","F8A3","F8A4","F8A2","F8A1","C654","F8EB","F8EC","F8ED","C653","F967","F96A","F969","F968","F9D3","C073","C365","F5BF","F6D5","C5C7","F7CE","F9D5","C074","EFB6","F7CF","F9A1","C8A1","C8A3","C94A","DDFC","A14A","A157","A159","A15B","A15F","A160","A163","A164","A167","A168","A16B","A16C","A16F","A170","A173","A174","A177","A178","A17B","A17C","A1C6","A1C7","A1CA","A1CB","A1C8","A1C9","A15C","A14D","A14E","A14F","A151","A152","A153","A154","A17D","A17E","A1A1","A1A2","A1A3","A1A4","A1CC","A1CD","A1CE","A1DE","A1DF","A1E0","A1E1","A1E2","A242","A24C","A24D","A24E","A149","C8D0","A1AD","A243","A248","A1AE","C8CF","A15D","A15E","A1AF","A1CF","A141","A1D0","A144","A1FE","A2AF","A2B0","A2B1","A2B2","A2B3","A2B4","A2B5","A2B6","A2B7","A2B8","A147","A146","A1D5","A1D7","A1D6","A148","A249","A2CF","A2D0","A2D1","A2D2","A2D3","A2D4","A2D5","A2D6","A2D7","A2D8","A2D9","A2DA","A2DB","A2DC","A2DD","A2DE","A2DF","A2E0","A2E1","A2E2","A2E3","A2E4","A2E5","A2E6","A2E7","A2E8","C6E4","A240","C6E5","A1C4","A2E9","A2EA","A2EB","A2EC","A2ED","A2EE","A2EF","A2F0","A2F1","A2F2","A2F3","A2F4","A2F5","A2F6","A2F7","A2F8","A2F9","A2FA","A2FB","A2FC","A2FD","A2FE","A340","A341","A342","A343","A161","A155","A162","A1E3","A246","A247","C8CD","A1C3","C8CE","A244","F9FE","C87C","C87A","FCAD","FC47","FBCB","FC5F","FA54","FA74","FA63","FA60","FA62","FD6A","FD54","FA65","FA6E","FA40","FA76","FBD0","FA7B","FE6D","FA42","FC7B","FD63","FAC2","FAE9","FAC6","FA72","FE5C","FAEE","FBF6","FDA4","FAFE","FBA7","FAF6","FCDE","FAFC","FB4E","FB73","FB5F","FA56","FC4B","FEDE","FCB7","FB68","FB6A","FDC8","FAD6","FAD7","FBA1","FBA4","FBC2","FEDB","FBB1","FE61","FACE","FAFB","FAC8","FCAF","FBFB","FBAD","FBAA","FBAC","FCB8","FBB0","FCD2","FC63","FEA6","FB49","FBC0","FD6B","FA78","FEA8","FA4C","FBCE","FEAD","FBD5","FBD9","FBE1","FED0","FBE2","FCE3","FBE6","FC5B","FBEC","FA73","FD59","FBED","FC4E","FC50","FD73","FDA7","FA58","FA5E","FA75","FBBE","FEEF","FC5D","FB41","FC68","FD67","FCA1","FE70","FCA8","FCE9","FCA7","FAEC","FCB1","FCBB","FE4E","FDC2","FCCC","FCDB","FCDC","FCE6","FCD8","FCD9","FDA6","FCE4","FA77","FB51","FED7","FA70","FCFC","FCD6","FD40","FD46","FA5C","FD4B","FD58","FD5B","FAAE","FCF0","FD69","FD6E","FA7C","FA4E","FDA1","FD79","FD7B","FDBC","FDC6","FD70","FAC1","FDBE","FDD9","FC7E","FA43","FAEB","FAC3","FDD8","FDD7","FB4A","FDC9","FDB2","FCBD","FDD5","FDDA","FDDE","FDD3","FDD6","FDDC","FDDD","FEA1","FD7C","FB61","FCF4","FE60","FDEC","FB78","FDF6","FB6B","FDF3","FCEC","FB4C","FDF8","FBF5","FE54","FD74","FD7E","FBEB","FD7D","FA52","FCCE","FA49","FE49","FDB1","FC6E","FE66","FC6F","FA6D","FBB7","FE65","FCB3","FE67","FEDF","FCED","FDFA","FCC8","FE62","FE6B","FDF9","FCC7","FEA2","FE7E","FAA7","FA5D","FB6D","FEB7","FEB5","FECE","FECF","FCE0","FED5","FC5A","FCE7","FEE6","FE46","FCBE","FADF","FEF4","FBCF","FEEE","FDDF","FEC2","FC4D","FEC7","FEC6","FE5D","FABC","FBDA","FD75","FAF5","FA55","FB66","FBA8","FDAA","FA5A","FBA9","FCFD","FDAB","FBAE","FDAC","FBB3","FC53","FDB3","FC78","FDB5","FECC","FBBD","FC5C","FEFA","C8A4","FEEA","FBC8","FC70","FA5B","FEF6","FBFE","FCE5","FAD3","FE57","FE55","FED9","FEDA","FA53","FB40","FA4A","FC60","FC56","FD4F","FADA","FABE","FD43","FDBA","FEED","FB5D","FEA3","FCA3","FDFC","FEA5","FEAF","FE72","FEAC","FCF5","FEFD","FDD0","FD60","FCA4","FB77"], [167,168,175,176,177,183,215,247,248,331,339,592,596,603,618,629,643,650,710,711,713,714,715,717,729,913,914,915,916,917,918,919,920,921,922,923,924,925,926,927,928,929,931,932,933,934,935,936,937,945,946,947,948,949,950,951,952,953,954,955,956,957,958,959,960,961,963,964,965,966,967,968,969,1025,1040,1041,1042,1043,1044,1045,1046,1047,1048,1049,1050,1051,1052,1053,1054,1055,1056,1057,1058,1059,1060,1061,1062,1063,1064,1065,1066,1067,1068,1069,1070,1071,1072,1073,1074,1075,1076,1077,1078,1079,1080,1081,1082,1083,1084,1085,1086,1087,1088,1089,1090,1091,1092,1093,1094,1095,1096,1097,1098,1099,1100,1101,1102,1103,1105,8211,8212,8216,8217,8220,8221,8229,8230,8231,8242,8245,8251,8364,8451,8453,8457,8470,8481,8544,8545,8546,8547,8548,8549,8550,8551,8552,8553,8560,8561,8562,8563,8564,8565,8566,8567,8568,8569,8592,8593,8594,8595,8598,8599,8600,8601,8632,8633,8679,8725,8730,8734,8735,8736,8739,8741,8745,8746,8747,8750,8756,8757,8786,8800,8801,8806,8807,8853,8857,8869,8895,9216,9217,9218,9219,9220,9221,9222,9223,9224,9225,9226,9227,9228,9229,9230,9231,9232,9233,9234,9235,9236,9237,9238,9239,9240,9241,9242,9243,9244,9245,9246,9247,9249,9312,9313,9314,9315,9316,9317,9318,9319,9320,9321,9332,9333,9334,9335,9336,9337,9338,9339,9340,9341,9472,9474,9484,9488,9492,9496,9500,9508,9516,9524,9532,9552,9553,9554,9555,9556,9557,9558,9559,9560,9561,9562,9563,9564,9565,9566,9567,9568,9569,9570,9571,9572,9573,9574,9575,9576,9577,9578,9579,9580,9581,9582,9583,9584,9585,9586,9587,9588,9601,9602,9603,9604,9605,9606,9607,9608,9609,9610,9611,9612,9613,9614,9615,9620,9621,9632,9633,9650,9651,9660,9661,9670,9671,9675,9678,9679,9698,9699,9700,9701,9733,9734,9792,9794,10045,11904,11908,11910,11911,11912,11914,11916,11917,11925,11932,11933,11941,11943,11946,11948,11950,11958,11964,11966,11974,11978,11980,11981,11983,11990,11991,11998,12003,12083,12288,12289,12290,12291,12293,12294,12295,12296,12297,12298,12299,12300,12301,12302,12303,12304,12305,12306,12308,12309,12317,12318,12321,12322,12323,12324,12325,12326,12327,12328,12329,12353,12354,12355,12356,12357,12358,12359,12360,12361,12362,12363,12364,12365,12366,12367,12368,12369,12370,12371,12372,12373,12374,12375,12376,12377,12378,12379,12380,12381,12382,12383,12384,12385,12386,12387,12388,12389,12390,12391,12392,12393,12394,12395,12396,12397,12398,12399,12400,12401,12402,12403,12404,12405,12406,12407,12408,12409,12410,12411,12412,12413,12414,12415,12416,12417,12418,12419,12420,12421,12422,12423,12424,12425,12426,12427,12428,12429,12430,12431,12432,12433,12434,12435,12443,12444,12445,12446,12449,12450,12451,12452,12453,12454,12455,12456,12457,12458,12459,12460,12461,12462,12463,12464,12465,12466,12467,12468,12469,12470,12471,12472,12473,12474,12475,12476,12477,12478,12479,12480,12481,12482,12483,12484,12485,12486,12487,12488,12489,12490,12491,12492,12493,12494,12495,12496,12497,12498,12499,12500,12501,12502,12503,12504,12505,12506,12507,12508,12509,12510,12511,12512,12513,12514,12515,12516,12517,12518,12519,12520,12521,12522,12523,12524,12525,12526,12527,12528,12529,12530,12531,12532,12533,12534,12540,12541,12542,12549,12550,12551,12552,12553,12554,12555,12556,12557,12558,12559,12560,12561,12562,12563,12564,12565,12566,12567,12568,12569,12570,12571,12572,12573,12574,12575,12576,12577,12578,12579,12580,12581,12582,12583,12584,12585,12751,12849,12963,13198,13199,13212,13213,13214,13217,13252,13262,13265,13266,13269,13434,13665,13729,13859,13881,13897,14001,14019,14021,14053,14054,14088,14093,14128,14130,14144,14147,14328,14381,14390,14496,14540,14642,14691,14849,14871,14935,14950,15019,15070,15147,15232,15265,15298,15347,15373,15580,15595,15666,15675,15694,15711,15749,15796,15815,15848,15936,15957,16040,16042,16087,16132,16217,16388,16600,16644,17077,17148,17553,17570,17605,17641,17746,17749,18342,18525,18606,18653,18669,18771,19255,19693,19728,19968,19969,19971,19975,19976,19977,19978,19979,19980,19981,19982,19983,19984,19985,19988,19989,19990,19992,19993,19998,19999,20006,20011,20013,20014,20016,20017,20018,20019,20022,20024,20025,20027,20028,20031,20034,20035,20037,20039,20040,20041,20043,20045,20046,20047,20050,20051,20054,20056,20057,20058,20060,20061,20062,20063,20073,20083,20094,20095,20098,20099,20100,20101,20102,20104,20107,20108,20109,20110,20113,20114,20115,20116,20117,20121,20123,20126,20127,20128,20129,20130,20132,20133,20134,20136,20139,20140,20141,20142,20147,20150,20151,20153,20154,20160,20161,20162,20163,20164,20166,20167,20168,20169,20170,20171,20173,20180,20181,20182,20183,20184,20185,20186,20188,20189,20190,20191,20193,20195,20196,20197,20200,20201,20208,20209,20210,20211,20212,20213,20214,20215,20219,20221,20223,20224,20225,20226,20228,20229,20232,20233,20234,20235,20237,20238,20239,20240,20241,20242,20243,20244,20245,20248,20249,20253,20258,20268,20269,20271,20272,20275,20276,20278,20280,20282,20283,20284,20285,20286,20287,20289,20291,20294,20295,20296,20297,20300,20301,20302,20303,20304,20305,20306,20307,20308,20309,20310,20311,20312,20313,20314,20315,20316,20317,20318,20319,20320,20321,20322,20323,20324,20327,20329,20330,20331,20332,20334,20335,20336,20339,20340,20341,20342,20343,20344,20345,20346,20347,20348,20349,20350,20351,20352,20353,20354,20355,20356,20357,20358,20359,20360,20361,20363,20365,20367,20368,20369,20370,20372,20373,20374,20375,20376,20378,20379,20380,20381,20382,20395,20398,20399,20402,20403,20405,20406,20407,20409,20410,20411,20415,20416,20417,20418,20419,20420,20421,20423,20425,20426,20427,20429,20430,20431,20432,20433,20435,20436,20438,20439,20440,20441,20442,20443,20444,20445,20446,20447,20448,20449,20460,20462,20463,20465,20467,20468,20469,20470,20471,20472,20474,20478,20480,20485,20486,20487,20488,20489,20491,20492,20493,20494,20495,20497,20498,20499,20500,20501,20502,20503,20504,20505,20506,20507,20508,20510,20511,20512,20513,20514,20515,20517,20518,20519,20520,20521,20522,20523,20524,20525,20527,20528,20529,20531,20533,20535,20540,20544,20545,20547,20549,20550,20551,20552,20553,20554,20555,20556,20557,20558,20559,20561,20563,20565,20567,20570,20571,20572,20573,20574,20575,20576,20577,20578,20579,20580,20581,20584,20585,20586,20587,20589,20590,20591,20592,20594,20595,20596,20597,20598,20599,20602,20605,20608,20610,20611,20613,20615,20619,20620,20621,20622,20624,20625,20626,20628,20629,20630,20632,20633,20634,20635,20636,20637,20638,20642,20643,20652,20653,20654,20655,20656,20657,20658,20659,20660,20661,20662,20663,20664,20666,20667,20669,20670,20671,20673,20674,20676,20677,20678,20679,20680,20681,20682,20683,20686,20687,20689,20691,20692,20693,20694,20695,20698,20699,20701,20704,20705,20707,20708,20709,20710,20711,20712,20713,20714,20716,20717,20718,20719,20720,20721,20723,20725,20726,20728,20729,20731,20733,20734,20735,20736,20737,20738,20739,20740,20741,20742,20743,20744,20745,20746,20747,20748,20750,20752,20753,20754,20755,20756,20757,20759,20760,20762,20764,20767,20768,20769,20770,20772,20773,20774,20777,20778,20781,20782,20784,20785,20786,20787,20788,20789,20791,20792,20793,20794,20795,20796,20797,20799,20800,20801,20803,20804,20805,20806,20807,20808,20809,20811,20812,20813,20818,20820,20821,20823,20825,20826,20827,20828,20829,20830,20831,20832,20833,20834,20835,20837,20839,20840,20841,20842,20843,20844,20845,20846,20849,20853,20854,20855,20856,20860,20864,20866,20871,20872,20873,20874,20877,20879,20881,20882,20883,20884,20885,20886,20887,20888,20894,20896,20898,20900,20901,20906,20907,20908,20912,20913,20914,20915,20917,20918,20919,20920,20921,20924,20925,20926,20932,20933,20934,20935,20936,20937,20938,20939,20940,20941,20942,20943,20944,20945,20947,20948,20951,20952,20956,20957,20958,20959,20960,20961,20962,20973,20976,20977,20979,20980,20981,20982,20984,20985,20986,20989,20992,20993,20994,20995,20998,20999,21000,21001,21002,21003,21004,21006,21008,21009,21010,21011,21014,21015,21020,21021,21022,21025,21028,21029,21030,21032,21033,21034,21038,21040,21041,21042,21043,21045,21046,21047,21048,21050,21051,21052,21057,21059,21060,21062,21063,21065,21066,21067,21068,21069,21070,21074,21076,21077,21078,21079,21082,21083,21084,21085,21086,21087,21089,21090,21097,21098,21099,21100,21101,21102,21103,21106,21108,21109,21111,21112,21114,21115,21116,21117,21119,21120,21121,21122,21123,21124,21127,21128,21129,21130,21131,21132,21133,21137,21139,21140,21142,21143,21144,21145,21147,21151,21152,21155,21158,21161,21162,21163,21164,21165,21166,21173,21177,21179,21180,21182,21184,21185,21186,21187,21189,21191,21193,21197,21200,21202,21203,21205,21206,21207,21208,21209,21211,21213,21214,21215,21216,21218,21219,21220,21222,21225,21227,21231,21232,21233,21235,21236,21237,21239,21240,21241,21242,21243,21244,21246,21247,21253,21254,21256,21257,21258,21259,21261,21262,21263,21264,21265,21266,21269,21270,21271,21273,21274,21276,21277,21279,21280,21281,21282,21283,21290,21293,21295,21296,21297,21299,21300,21303,21304,21305,21308,21309,21310,21311,21312,21313,21314,21315,21316,21317,21319,21320,21321,21322,21324,21325,21329,21330,21331,21332,21335,21338,21340,21342,21344,21345,21347,21350,21353,21356,21357,21358,21359,21360,21361,21362,21363,21364,21365,21367,21368,21369,21371,21372,21374,21375,21378,21380,21386,21390,21391,21394,21395,21396,21398,21399,21400,21401,21402,21404,21405,21406,21407,21408,21412,21413,21414,21415,21416,21419,21420,21421,21422,21426,21428,21430,21433,21435,21441,21443,21445,21448,21449,21450,21451,21453,21460,21462,21463,21465,21467,21471,21472,21473,21474,21475,21476,21477,21478,21480,21481,21482,21483,21484,21485,21486,21487,21488,21489,21490,21491,21493,21494,21496,21499,21500,21502,21505,21507,21508,21510,21511,21512,21513,21514,21515,21516,21517,21518,21519,21520,21521,21522,21523,21528,21529,21531,21532,21533,21534,21535,21536,21537,21540,21541,21542,21543,21544,21545,21546,21547,21548,21549,21550,21551,21552,21553,21555,21557,21558,21559,21560,21561,21563,21564,21565,21566,21568,21569,21570,21571,21573,21574,21575,21576,21578,21582,21583,21588,21600,21601,21602,21603,21604,21605,21606,21607,21608,21611,21612,21613,21615,21616,21617,21618,21619,21620,21621,21622,21623,21624,21626,21627,21628,21629,21630,21631,21632,21633,21634,21636,21638,21639,21640,21643,21644,21645,21646,21647,21648,21649,21650,21653,21654,21656,21658,21664,21665,21666,21668,21669,21670,21671,21672,21673,21674,21675,21676,21677,21678,21679,21680,21681,21683,21686,21687,21688,21690,21691,21692,21693,21694,21695,21696,21697,21698,21699,21700,21701,21702,21703,21704,21705,21710,21711,21718,21726,21728,21729,21730,21732,21733,21734,21735,21736,21737,21738,21739,21741,21742,21745,21746,21747,21751,21752,21754,21755,21756,21757,21759,21761,21763,21764,21765,21766,21767,21768,21769,21770,21771,21772,21774,21775,21776,21777,21778,21780,21783,21786,21798,21799,21802,21803,21804,21805,21806,21807,21808,21809,21810,21811,21812,21813,21814,21815,21816,21817,21819,21820,21822,21824,21825,21827,21828,21829,21830,21831,21832,21833,21834,21835,21837,21838,21839,21840,21841,21842,21843,21845,21846,21847,21852,21853,21854,21855,21857,21858,21859,21860,21861,21862,21866,21877,21878,21879,21883,21884,21885,21886,21887,21888,21889,21890,21891,21892,21894,21895,21896,21897,21898,21899,21900,21901,21902,21903,21905,21906,21907,21908,21909,21912,21913,21914,21916,21917,21919,21921,21922,21923,21924,21925,21926,21927,21928,21929,21930,21931,21932,21933,21934,21937,21938,21939,21941,21947,21951,21952,21954,21955,21956,21957,21958,21959,21960,21961,21962,21963,21964,21965,21966,21967,21968,21969,21970,21971,21972,21973,21974,21977,21978,21979,21980,21981,21983,21985,21986,21987,21988,21989,21990,21991,21992,21993,21999,22002,22006,22007,22009,22010,22012,22013,22014,22015,22016,22017,22018,22020,22021,22022,22024,22025,22028,22029,22030,22031,22032,22034,22035,22036,22037,22038,22039,22043,22044,22045,22047,22055,22057,22058,22060,22062,22063,22064,22066,22067,22068,22069,22070,22071,22072,22073,22074,22075,22077,22078,22079,22080,22081,22082,22085,22086,22088,22089,22090,22092,22093,22094,22095,22099,22103,22104,22105,22106,22110,22112,22114,22115,22116,22117,22118,22120,22121,22122,22123,22124,22125,22126,22127,22128,22129,22130,22131,22132,22134,22135,22136,22137,22142,22143,22144,22145,22146,22147,22148,22149,22150,22151,22156,22157,22158,22159,22160,22163,22165,22167,22168,22169,22170,22172,22173,22181,22182,22183,22184,22186,22187,22188,22189,22190,22194,22195,22196,22197,22198,22199,22204,22205,22206,22208,22209,22210,22211,22213,22214,22216,22217,22218,22219,22220,22221,22225,22227,22228,22231,22234,22235,22237,22238,22239,22240,22241,22244,22245,22247,22250,22251,22253,22254,22255,22256,22257,22263,22265,22266,22271,22272,22273,22274,22275,22276,22279,22280,22281,22282,22283,22284,22285,22290,22291,22292,22294,22296,22298,22299,22300,22302,22303,22304,22306,22307,22312,22313,22314,22316,22317,22318,22319,22320,22323,22324,22331,22334,22336,22337,22338,22339,22341,22342,22343,22345,22346,22347,22348,22349,22350,22351,22352,22353,22354,22369,22370,22372,22374,22376,22377,22378,22379,22381,22383,22384,22385,22386,22387,22388,22389,22390,22391,22395,22396,22397,22400,22402,22403,22411,22412,22415,22416,22419,22420,22421,22423,22424,22425,22426,22427,22429,22430,22431,22432,22434,22435,22436,22437,22446,22453,22454,22456,22457,22458,22460,22461,22463,22465,22466,22467,22470,22471,22475,22476,22478,22479,22480,22482,22484,22485,22492,22495,22496,22497,22498,22499,22500,22501,22502,22503,22505,22508,22509,22510,22512,22513,22514,22515,22516,22517,22518,22519,22520,22521,22522,22523,22524,22525,22526,22528,22529,22530,22531,22532,22533,22534,22535,22536,22537,22538,22539,22540,22541,22542,22544,22548,22553,22555,22556,22557,22558,22560,22561,22563,22564,22565,22566,22567,22568,22569,22570,22572,22573,22574,22575,22576,22577,22578,22579,22580,22581,22582,22583,22584,22585,22587,22589,22591,22600,22601,22602,22603,22604,22605,22606,22607,22609,22610,22611,22612,22613,22615,22616,22617,22618,22619,22620,22621,22622,22626,22627,22628,22629,22632,22635,22637,22639,22641,22644,22645,22646,22649,22650,22651,22652,22653,22654,22655,22656,22657,22658,22659,22661,22662,22663,22664,22665,22666,22667,22670,22671,22672,22673,22675,22676,22680,22684,22685,22686,22687,22688,22689,22691,22693,22694,22696,22697,22698,22699,22700,22702,22703,22705,22707,22714,22715,22716,22717,22718,22719,22721,22722,22725,22726,22727,22728,22729,22734,22735,22737,22738,22739,22740,22741,22742,22744,22745,22746,22747,22748,22749,22750,22751,22752,22754,22755,22756,22759,22760,22761,22763,22764,22767,22772,22777,22778,22779,22780,22781,22782,22783,22787,22790,22794,22796,22797,22798,22799,22801,22802,22804,22805,22806,22807,22809,22810,22812,22815,22816,22818,22820,22821,22823,22825,22826,22827,22828,22829,22830,22831,22833,22839,22840,22844,22846,22848,22852,22853,22855,22856,22857,22858,22862,22863,22864,22865,22867,22868,22869,22871,22872,22874,22876,22880,22881,22882,22887,22889,22890,22891,22893,22894,22896,22897,22898,22899,22900,22902,22903,22904,22905,22907,22908,22909,22910,22911,22912,22913,22914,22915,22916,22917,22922,22925,22926,22927,22928,22930,22931,22934,22935,22936,22937,22941,22942,22944,22945,22946,22947,22948,22949,22950,22951,22952,22958,22959,22961,22962,22963,22964,22965,22966,22969,22970,22971,22972,22973,22974,22976,22977,22979,22981,22982,22983,22984,22985,22986,22987,22988,22989,22990,22991,22992,22993,22994,22995,22996,22998,23000,23001,23002,23003,23004,23005,23006,23008,23009,23011,23012,23013,23014,23016,23017,23018,23020,23021,23022,23025,23026,23027,23028,23029,23030,23031,23033,23034,23035,23036,23037,23038,23039,23040,23041,23043,23049,23050,23052,23055,23057,23059,23061,23062,23063,23064,23065,23067,23068,23070,23071,23072,23075,23076,23077,23081,23085,23086,23091,23093,23094,23095,23096,23097,23100,23102,23104,23105,23106,23107,23108,23110,23111,23112,23113,23114,23116,23117,23120,23121,23122,23123,23125,23126,23127,23128,23130,23131,23132,23133,23134,23135,23136,23138,23139,23140,23141,23142,23143,23145,23146,23148,23149,23152,23159,23160,23162,23163,23164,23165,23167,23171,23172,23176,23178,23179,23180,23182,23183,23184,23185,23186,23187,23188,23189,23190,23191,23194,23195,23196,23197,23198,23199,23200,23202,23205,23206,23207,23209,23211,23212,23214,23215,23216,23217,23218,23219,23220,23221,23222,23223,23224,23225,23226,23227,23228,23229,23230,23231,23232,23233,23234,23236,23238,23239,23240,23241,23242,23243,23244,23245,23251,23253,23254,23255,23256,23257,23258,23259,23260,23261,23262,23263,23264,23265,23266,23267,23269,23270,23272,23273,23274,23275,23276,23277,23278,23283,23284,23285,23286,23287,23288,23289,23290,23291,23293,23295,23297,23298,23299,23301,23303,23304,23305,23307,23308,23311,23312,23315,23316,23318,23319,23321,23322,23323,23325,23326,23328,23329,23331,23332,23333,23334,23335,23336,23338,23340,23341,23342,23343,23344,23346,23348,23352,23356,23357,23358,23359,23360,23363,23365,23366,23367,23368,23371,23372,23373,23374,23376,23377,23379,23380,23381,23382,23383,23384,23386,23387,23388,23389,23391,23394,23395,23396,23397,23401,23403,23404,23406,23408,23409,23410,23411,23413,23414,23415,23416,23418,23419,23420,23421,23423,23424,23425,23426,23427,23428,23429,23431,23432,23433,23435,23436,23438,23439,23442,23443,23445,23447,23448,23449,23450,23451,23452,23453,23458,23459,23460,23461,23462,23463,23464,23466,23468,23469,23470,23472,23475,23476,23477,23478,23480,23481,23487,23488,23489,23490,23492,23493,23494,23495,23498,23499,23500,23501,23502,23504,23505,23506,23507,23508,23509,23510,23511,23512,23513,23518,23519,23520,23521,23522,23523,23524,23525,23526,23527,23528,23529,23530,23531,23532,23534,23535,23536,23537,23538,23539,23541,23542,23544,23546,23551,23553,23555,23556,23557,23559,23560,23561,23562,23563,23564,23565,23566,23567,23568,23569,23570,23571,23572,23573,23574,23578,23583,23586,23588,23589,23592,23594,23596,23600,23601,23603,23607,23608,23609,23610,23611,23612,23614,23615,23616,23617,23620,23621,23622,23623,23624,23625,23627,23628,23629,23630,23631,23632,23633,23636,23637,23638,23640,23641,23644,23645,23648,23650,23651,23652,23653,23655,23656,23657,23658,23660,23661,23662,23663,23665,23667,23668,23673,23674,23675,23676,23678,23685,23686,23688,23689,23690,23691,23692,23693,23695,23696,23697,23698,23699,23700,23701,23708,23709,23711,23712,23713,23714,23715,23716,23717,23718,23719,23720,23721,23722,23723,23724,23725,23726,23727,23728,23729,23731,23733,23734,23735,23736,23738,23750,23751,23752,23753,23754,23755,23756,23758,23759,23760,23762,23763,23764,23766,23767,23768,23769,23770,23771,23774,23775,23784,23785,23786,23788,23789,23790,23791,23792,23793,23796,23798,23799,23800,23801,23803,23805,23807,23808,23809,23814,23815,23819,23820,23821,23822,23823,23824,23825,23826,23828,23830,23831,23832,23833,23834,23835,23837,23838,23839,23840,23842,23843,23844,23845,23846,23847,23848,23849,23854,23856,23857,23858,23859,23860,23861,23862,23863,23864,23865,23866,23868,23869,23871,23872,23873,23874,23875,23877,23878,23879,23881,23882,23883,23884,23886,23888,23889,23890,23893,23897,23902,23906,23907,23909,23911,23912,23913,23915,23916,23919,23921,23922,23927,23929,23930,23932,23933,23934,23935,23936,23937,23938,23940,23942,23943,23944,23945,23946,23947,23949,23954,23955,23956,23957,23959,23961,23962,23964,23965,23966,23967,23968,23969,23970,23975,23976,23977,23978,23980,23981,23982,23983,23984,23985,23986,23988,23989,23990,23991,23992,23994,23996,23997,24000,24002,24003,24006,24007,24009,24011,24013,24015,24017,24018,24020,24021,24022,24023,24024,24027,24029,24030,24031,24032,24033,24034,24037,24038,24039,24040,24043,24046,24048,24049,24050,24051,24052,24055,24057,24061,24062,24063,24066,24067,24068,24070,24073,24074,24076,24078,24081,24084,24085,24086,24087,24088,24089,24090,24091,24093,24095,24096,24097,24098,24099,24100,24101,24104,24105,24107,24109,24110,24115,24116,24118,24119,24120,24125,24126,24128,24129,24131,24132,24133,24138,24139,24140,24141,24142,24143,24147,24148,24149,24151,24152,24153,24155,24156,24157,24159,24160,24161,24162,24163,24166,24167,24168,24169,24170,24171,24172,24173,24174,24175,24176,24178,24179,24180,24181,24182,24184,24185,24186,24187,24188,24189,24190,24191,24192,24194,24195,24196,24199,24200,24201,24202,24203,24204,24205,24207,24213,24214,24215,24218,24219,24220,24224,24226,24227,24228,24229,24230,24231,24232,24234,24235,24236,24237,24238,24240,24241,24242,24243,24244,24245,24246,24247,24248,24249,24254,24257,24258,24260,24261,24262,24263,24264,24265,24266,24267,24268,24270,24272,24273,24274,24275,24276,24277,24278,24279,24280,24281,24282,24283,24284,24285,24286,24287,24288,24289,24290,24291,24293,24294,24295,24296,24297,24300,24302,24303,24305,24306,24307,24308,24310,24311,24313,24314,24315,24316,24318,24319,24321,24322,24324,24325,24327,24328,24330,24331,24333,24335,24338,24339,24340,24341,24343,24344,24346,24347,24349,24351,24354,24355,24356,24358,24359,24360,24361,24365,24366,24368,24369,24371,24373,24374,24375,24376,24378,24380,24384,24387,24388,24390,24392,24393,24394,24395,24396,24398,24399,24400,24404,24406,24407,24408,24409,24413,24417,24418,24419,24420,24421,24423,24425,24426,24427,24428,24429,24431,24432,24433,24434,24435,24436,24438,24439,24440,24441,24443,24444,24445,24446,24447,24448,24449,24450,24453,24454,24455,24456,24457,24458,24459,24460,24464,24465,24466,24470,24471,24472,24473,24475,24476,24478,24479,24480,24481,24484,24485,24486,24487,24488,24489,24490,24491,24492,24493,24494,24495,24498,24501,24502,24503,24505,24507,24508,24509,24510,24511,24512,24513,24515,24517,24521,24524,24525,24527,24528,24529,24530,24532,24533,24534,24535,24536,24537,24541,24542,24543,24544,24545,24547,24548,24549,24552,24554,24555,24557,24558,24559,24561,24563,24564,24565,24567,24568,24570,24571,24573,24575,24576,24585,24586,24587,24588,24589,24590,24591,24592,24593,24594,24595,24596,24597,24598,24599,24601,24602,24603,24604,24605,24606,24608,24609,24610,24611,24612,24613,24614,24615,24616,24617,24618,24619,24620,24621,24622,24623,24625,24626,24627,24628,24629,24631,24633,24640,24641,24642,24643,24644,24645,24646,24647,24649,24650,24652,24653,24656,24658,24659,24660,24661,24664,24665,24666,24667,24669,24670,24671,24674,24675,24676,24677,24678,24679,24680,24681,24682,24683,24684,24685,24686,24687,24688,24690,24703,24704,24705,24707,24708,24709,24710,24711,24712,24713,24714,24716,24717,24718,24720,24722,24724,24725,24726,24727,24730,24731,24732,24733,24734,24735,24736,24738,24739,24740,24744,24752,24753,24754,24755,24756,24757,24758,24759,24760,24761,24762,24763,24764,24765,24766,24767,24768,24769,24771,24772,24773,24774,24775,24776,24777,24778,24779,24780,24781,24782,24783,24785,24787,24788,24789,24792,24793,24794,24795,24796,24797,24798,24799,24800,24801,24802,24804,24806,24816,24817,24818,24819,24820,24821,24822,24823,24824,24825,24826,24827,24828,24830,24831,24832,24833,24835,24836,24837,24838,24840,24841,24842,24843,24845,24846,24847,24848,24850,24851,24852,24853,24854,24856,24858,24859,24860,24861,24863,24867,24871,24872,24873,24875,24876,24878,24879,24882,24884,24886,24887,24891,24894,24895,24896,24897,24898,24900,24901,24902,24903,24904,24905,24906,24907,24908,24909,24910,24911,24914,24915,24916,24917,24918,24920,24922,24923,24924,24925,24926,24927,24928,24929,24930,24931,24932,24933,24934,24935,24936,24938,24939,24940,24942,24944,24945,24946,24947,24948,24949,24950,24951,24953,24954,24956,24958,24960,24961,24962,24963,24967,24969,24970,24971,24972,24973,24974,24976,24977,24978,24979,24980,24982,24986,24987,24989,24991,24993,24994,24996,24999,25000,25001,25002,25003,25004,25005,25006,25007,25008,25009,25010,25011,25012,25013,25014,25016,25018,25020,25022,25023,25025,25026,25027,25029,25030,25031,25032,25033,25034,25035,25036,25037,25040,25043,25046,25048,25054,25055,25056,25059,25060,25061,25062,25063,25064,25065,25066,25067,25069,25070,25072,25073,25074,25077,25078,25079,25080,25081,25082,25083,25084,25085,25086,25087,25088,25089,25091,25092,25095,25096,25097,25098,25100,25101,25102,25104,25105,25106,25108,25109,25110,25113,25114,25115,25119,25120,25121,25122,25123,25124,25125,25127,25129,25130,25131,25133,25134,25136,25138,25139,25140,25142,25146,25149,25150,25151,25152,25153,25154,25155,25158,25159,25160,25161,25162,25163,25165,25166,25168,25169,25170,25171,25172,25176,25177,25178,25179,25180,25182,25184,25185,25186,25187,25188,25189,25190,25197,25198,25199,25200,25201,25202,25203,25204,25206,25207,25209,25210,25211,25212,25213,25214,25215,25216,25217,25219,25220,25222,25223,25224,25225,25226,25228,25230,25231,25233,25234,25235,25236,25237,25238,25239,25240,25256,25257,25258,25259,25260,25261,25262,25263,25264,25265,25267,25268,25269,25270,25272,25273,25275,25276,25277,25278,25279,25282,25284,25285,25286,25287,25288,25289,25290,25291,25292,25293,25294,25295,25296,25297,25298,25299,25300,25301,25302,25303,25304,25305,25306,25307,25308,25311,25323,25324,25325,25326,25327,25328,25329,25330,25331,25332,25333,25334,25335,25336,25337,25338,25339,25340,25341,25342,25343,25344,25345,25346,25347,25351,25352,25353,25355,25356,25357,25358,25359,25360,25361,25363,25364,25365,25366,25384,25385,25386,25387,25388,25389,25391,25394,25395,25396,25398,25400,25401,25402,25403,25404,25405,25406,25408,25409,25410,25411,25412,25413,25414,25415,25416,25417,25418,25419,25420,25421,25422,25423,25424,25425,25428,25429,25430,25431,25432,25433,25434,25445,25447,25448,25449,25451,25452,25453,25454,25455,25456,25457,25458,25461,25462,25463,25464,25466,25467,25468,25469,25472,25473,25474,25475,25476,25477,25479,25480,25481,25482,25483,25484,25485,25486,25487,25488,25489,25490,25492,25494,25495,25496,25497,25499,25500,25501,25502,25503,25504,25505,25506,25507,25508,25509,25511,25512,25513,25514,25515,25516,25517,25518,25519,25520,25521,25533,25534,25536,25538,25539,25540,25541,25542,25543,25544,25545,25546,25547,25548,25549,25550,25551,25552,25554,25555,25557,25558,25559,25560,25561,25562,25563,25564,25565,25567,25568,25569,25571,25572,25573,25575,25576,25577,25578,25579,25581,25582,25583,25584,25585,25586,25587,25588,25589,25590,25592,25593,25606,25609,25610,25611,25612,25613,25614,25615,25616,25618,25619,25620,25621,25622,25623,25624,25626,25627,25628,25630,25631,25632,25633,25634,25635,25636,25637,25638,25639,25640,25642,25643,25644,25645,25646,25647,25648,25651,25652,25653,25654,25655,25656,25657,25658,25661,25662,25663,25664,25665,25667,25675,25677,25678,25680,25681,25682,25683,25684,25688,25689,25691,25692,25693,25694,25695,25696,25697,25701,25702,25703,25704,25705,25707,25708,25709,25710,25711,25712,25714,25715,25716,25717,25718,25719,25720,25721,25722,25723,25725,25727,25730,25733,25735,25736,25737,25738,25739,25740,25743,25744,25746,25747,25749,25750,25751,25752,25753,25754,25756,25757,25758,25759,25760,25762,25763,25764,25765,25766,25769,25771,25772,25773,25774,25775,25776,25777,25778,25779,25787,25788,25789,25790,25791,25793,25794,25795,25796,25797,25799,25801,25802,25803,25805,25806,25807,25808,25810,25812,25814,25815,25816,25817,25818,25819,25824,25825,25826,25827,25828,25829,25830,25832,25833,25835,25836,25837,25839,25840,25841,25842,25843,25844,25847,25848,25850,25851,25852,25853,25854,25855,25856,25857,25859,25860,25862,25863,25865,25868,25869,25870,25871,25872,25875,25876,25877,25878,25879,25880,25881,25883,25884,25885,25886,25888,25889,25890,25891,25892,25893,25894,25897,25898,25899,25900,25901,25902,25903,25906,25907,25908,25910,25911,25912,25913,25915,25917,25918,25919,25921,25923,25925,25926,25928,25929,25930,25933,25935,25937,25939,25940,25941,25942,25943,25944,25945,25948,25949,25950,25951,25954,25955,25956,25957,25958,25959,25960,25962,25964,25965,25967,25970,25971,25972,25973,25974,25975,25976,25977,25978,25979,25980,25983,25984,25985,25986,25987,25988,25989,25991,25992,25996,26000,26001,26002,26004,26005,26006,26007,26009,26011,26012,26013,26014,26015,26016,26017,26018,26020,26021,26023,26024,26026,26027,26028,26030,26031,26032,26034,26035,26037,26038,26039,26040,26041,26043,26044,26045,26047,26049,26050,26051,26052,26053,26054,26059,26060,26061,26062,26063,26064,26065,26066,26067,26070,26071,26074,26075,26077,26078,26079,26080,26081,26082,26083,26085,26086,26088,26089,26092,26093,26094,26095,26096,26097,26098,26099,26100,26101,26106,26107,26108,26109,26112,26114,26115,26116,26117,26118,26119,26120,26121,26122,26123,26124,26125,26126,26127,26128,26129,26130,26131,26132,26133,26136,26140,26141,26143,26144,26145,26146,26147,26148,26149,26150,26151,26152,26155,26157,26158,26159,26161,26162,26163,26164,26165,26166,26169,26170,26177,26178,26179,26180,26181,26183,26185,26186,26187,26188,26191,26193,26194,26199,26201,26202,26203,26204,26205,26206,26207,26209,26210,26212,26213,26214,26215,26216,26218,26220,26222,26223,26224,26225,26226,26227,26228,26230,26231,26232,26233,26234,26235,26236,26238,26240,26244,26245,26246,26247,26248,26249,26250,26251,26252,26253,26256,26257,26260,26261,26262,26263,26264,26265,26269,26271,26272,26273,26274,26276,26280,26281,26282,26283,26285,26286,26287,26288,26289,26290,26291,26292,26293,26295,26296,26297,26298,26299,26301,26302,26304,26308,26310,26311,26312,26313,26314,26315,26316,26317,26319,26322,26326,26328,26329,26330,26331,26332,26333,26334,26336,26339,26340,26342,26344,26345,26347,26348,26349,26350,26352,26354,26355,26356,26358,26359,26360,26361,26364,26366,26367,26368,26369,26370,26371,26372,26373,26376,26377,26378,26379,26381,26383,26384,26386,26387,26388,26389,26391,26392,26395,26397,26399,26400,26401,26402,26403,26406,26407,26408,26410,26411,26412,26413,26414,26417,26419,26420,26421,26424,26425,26426,26427,26428,26429,26430,26431,26436,26437,26438,26439,26440,26441,26443,26444,26445,26446,26447,26448,26449,26451,26453,26454,26455,26457,26458,26460,26461,26462,26463,26464,26466,26471,26474,26476,26477,26479,26480,26481,26482,26483,26484,26485,26486,26487,26488,26489,26490,26491,26492,26493,26494,26495,26497,26499,26500,26501,26502,26503,26505,26507,26508,26509,26510,26512,26513,26514,26515,26516,26517,26519,26520,26521,26522,26524,26525,26527,26540,26542,26543,26544,26545,26546,26547,26548,26549,26550,26551,26552,26553,26554,26555,26560,26561,26562,26563,26564,26565,26566,26568,26569,26570,26571,26572,26573,26574,26575,26576,26577,26578,26579,26580,26583,26584,26585,26586,26587,26588,26589,26590,26591,26594,26595,26596,26597,26598,26599,26601,26602,26603,26604,26605,26606,26607,26608,26609,26610,26611,26612,26613,26614,26615,26616,26618,26620,26623,26625,26637,26642,26643,26644,26646,26647,26648,26650,26651,26652,26653,26655,26656,26657,26658,26661,26662,26664,26665,26666,26667,26669,26670,26671,26673,26674,26675,26676,26677,26679,26680,26681,26682,26683,26684,26685,26686,26688,26689,26690,26691,26692,26693,26694,26696,26697,26699,26700,26701,26702,26703,26704,26705,26707,26708,26717,26731,26733,26734,26735,26737,26738,26740,26741,26742,26743,26744,26745,26747,26748,26749,26750,26751,26752,26753,26754,26755,26757,26758,26759,26761,26762,26763,26764,26767,26768,26769,26770,26771,26772,26774,26775,26779,26780,26781,26783,26784,26785,26786,26787,26788,26790,26791,26792,26793,26794,26795,26796,26797,26798,26799,26800,26801,26802,26803,26804,26805,26813,26819,26820,26822,26823,26824,26825,26826,26827,26828,26829,26830,26832,26833,26834,26835,26836,26837,26838,26839,26840,26842,26844,26845,26846,26847,26848,26849,26851,26852,26854,26855,26856,26857,26858,26859,26860,26862,26863,26864,26865,26866,26867,26868,26869,26870,26871,26872,26873,26874,26875,26876,26877,26880,26881,26882,26883,26884,26885,26886,26887,26888,26890,26891,26892,26893,26894,26895,26896,26897,26898,26899,26900,26901,26903,26917,26922,26927,26928,26930,26931,26932,26933,26934,26935,26936,26937,26939,26940,26941,26942,26943,26944,26945,26946,26948,26949,26952,26953,26954,26955,26956,26958,26959,26961,26962,26963,26964,26966,26967,26968,26969,26970,26971,26972,26973,26974,26975,26976,26977,26978,26979,26981,26982,26984,26985,26986,26987,26988,26989,26990,26991,26992,26993,26995,26996,26997,26998,26999,27000,27001,27002,27003,27010,27011,27014,27021,27022,27024,27025,27027,27028,27029,27030,27031,27032,27033,27034,27035,27036,27038,27040,27041,27043,27044,27045,27046,27047,27048,27049,27050,27051,27052,27053,27054,27055,27056,27057,27059,27060,27061,27062,27063,27065,27067,27068,27069,27070,27071,27073,27074,27075,27076,27078,27081,27082,27083,27084,27085,27086,27087,27088,27091,27092,27097,27105,27106,27108,27109,27110,27111,27112,27115,27116,27117,27118,27121,27122,27123,27124,27126,27127,27128,27129,27130,27131,27132,27133,27134,27135,27136,27137,27138,27140,27141,27142,27143,27144,27145,27146,27147,27149,27151,27153,27155,27156,27157,27158,27159,27160,27161,27163,27165,27166,27167,27168,27169,27171,27173,27174,27175,27176,27180,27186,27188,27189,27192,27193,27194,27195,27196,27197,27198,27199,27200,27201,27204,27205,27206,27207,27208,27209,27211,27213,27214,27215,27216,27217,27220,27221,27222,27224,27225,27226,27227,27229,27230,27231,27232,27233,27234,27236,27238,27239,27240,27241,27242,27243,27245,27247,27254,27262,27263,27264,27265,27267,27268,27269,27271,27273,27274,27276,27277,27278,27280,27281,27282,27283,27284,27285,27286,27287,27290,27291,27292,27293,27294,27295,27296,27297,27298,27299,27300,27301,27302,27304,27308,27309,27310,27311,27315,27316,27318,27319,27320,27321,27322,27323,27325,27330,27331,27333,27334,27335,27336,27339,27340,27341,27343,27344,27345,27347,27353,27354,27355,27356,27357,27358,27359,27360,27361,27365,27367,27368,27370,27371,27372,27374,27375,27376,27377,27379,27384,27385,27386,27387,27388,27392,27394,27395,27396,27400,27401,27402,27403,27407,27408,27409,27410,27411,27414,27415,27416,27417,27418,27421,27422,27424,27425,27427,27429,27432,27436,27437,27439,27441,27442,27443,27444,27445,27446,27447,27448,27449,27450,27451,27452,27453,27454,27455,27457,27458,27459,27461,27462,27463,27464,27465,27466,27467,27468,27469,27470,27472,27473,27476,27477,27478,27481,27483,27484,27486,27487,27488,27489,27490,27491,27492,27493,27494,27495,27498,27501,27506,27508,27510,27511,27512,27513,27515,27518,27519,27520,27522,27523,27524,27526,27528,27529,27530,27532,27533,27534,27535,27537,27540,27541,27542,27543,27544,27545,27547,27550,27551,27552,27554,27555,27556,27557,27558,27559,27562,27563,27565,27566,27567,27568,27570,27571,27573,27574,27575,27578,27580,27581,27583,27584,27587,27588,27589,27590,27591,27592,27593,27594,27595,27596,27597,27599,27600,27602,27603,27604,27606,27607,27608,27610,27611,27612,27614,27616,27617,27618,27619,27620,27622,27623,27624,27627,27628,27631,27632,27634,27635,27639,27640,27641,27643,27644,27645,27646,27647,27648,27649,27650,27651,27652,27653,27654,27656,27657,27659,27660,27661,27663,27664,27665,27667,27668,27669,27670,27672,27673,27674,27675,27677,27679,27680,27681,27683,27684,27685,27686,27687,27688,27690,27691,27692,27694,27695,27696,27699,27700,27702,27703,27704,27707,27710,27711,27712,27713,27714,27715,27718,27721,27722,27723,27724,27725,27726,27727,27728,27730,27732,27733,27735,27737,27738,27739,27740,27741,27742,27743,27744,27745,27749,27750,27751,27752,27753,27754,27755,27757,27758,27759,27760,27761,27762,27763,27764,27766,27768,27769,27770,27771,27773,27774,27776,27777,27778,27779,27780,27781,27782,27783,27784,27785,27786,27787,27788,27789,27790,27791,27792,27794,27795,27796,27797,27798,27800,27801,27802,27803,27804,27805,27819,27820,27821,27822,27824,27825,27827,27828,27830,27831,27832,27833,27834,27835,27836,27837,27838,27839,27840,27841,27842,27843,27844,27845,27846,27847,27849,27850,27851,27852,27853,27855,27856,27857,27858,27859,27860,27861,27862,27863,27865,27866,27867,27868,27869,27870,27872,27873,27874,27875,27877,27879,27880,27881,27883,27884,27885,27886,27887,27888,27889,27890,27891,27893,27897,27904,27905,27907,27908,27911,27912,27913,27914,27915,27916,27917,27918,27919,27920,27921,27922,27926,27927,27928,27929,27930,27931,27933,27934,27935,27936,27938,27941,27943,27944,27945,27946,27947,27948,27949,27950,27951,27952,27953,27954,27955,27956,27957,27958,27959,27960,27961,27962,27963,27964,27965,27966,27967,27968,27969,27970,27992,27993,27994,27995,27996,27998,27999,28000,28001,28002,28003,28004,28005,28006,28007,28008,28009,28010,28012,28013,28014,28015,28016,28017,28020,28021,28022,28023,28024,28025,28026,28027,28028,28029,28030,28031,28032,28034,28035,28036,28037,28038,28039,28040,28041,28042,28043,28044,28045,28046,28048,28049,28050,28051,28052,28053,28054,28055,28056,28074,28075,28076,28078,28079,28082,28083,28084,28085,28087,28088,28090,28091,28092,28093,28094,28095,28096,28098,28100,28101,28102,28103,28104,28105,28106,28107,28108,28109,28111,28112,28113,28114,28115,28116,28117,28118,28119,28120,28121,28122,28123,28124,28125,28126,28127,28128,28129,28130,28131,28132,28133,28134,28136,28137,28138,28139,28140,28141,28142,28143,28144,28145,28146,28147,28148,28149,28150,28151,28153,28154,28155,28156,28157,28160,28163,28165,28185,28186,28187,28188,28189,28191,28192,28193,28194,28195,28196,28197,28198,28199,28200,28202,28203,28204,28205,28206,28207,28208,28209,28210,28211,28212,28213,28214,28216,28217,28218,28219,28220,28221,28222,28223,28224,28225,28227,28228,28229,28230,28231,28233,28234,28235,28237,28238,28239,28240,28241,28242,28243,28244,28245,28246,28248,28249,28250,28251,28252,28253,28254,28255,28256,28257,28258,28259,28260,28261,28262,28263,28264,28265,28267,28270,28271,28273,28274,28275,28276,28279,28280,28281,28284,28296,28297,28301,28302,28303,28304,28306,28307,28308,28310,28311,28312,28313,28315,28316,28317,28318,28319,28320,28321,28322,28323,28324,28325,28326,28327,28330,28331,28334,28335,28336,28337,28338,28339,28340,28341,28342,28343,28345,28346,28348,28349,28350,28351,28352,28353,28354,28355,28356,28357,28358,28359,28360,28361,28362,28363,28364,28365,28366,28367,28368,28369,28370,28371,28372,28373,28374,28376,28377,28379,28380,28395,28396,28397,28398,28399,28401,28402,28404,28405,28406,28407,28408,28409,28410,28411,28412,28413,28414,28415,28416,28417,28418,28419,28420,28421,28422,28423,28424,28425,28426,28429,28430,28431,28434,28435,28436,28437,28438,28440,28441,28442,28444,28446,28447,28448,28449,28450,28451,28453,28454,28455,28457,28458,28459,28460,28461,28462,28463,28464,28465,28466,28467,28469,28470,28471,28472,28473,28474,28475,28476,28477,28478,28479,28480,28481,28483,28484,28494,28495,28496,28497,28498,28499,28500,28501,28503,28504,28506,28507,28508,28509,28510,28511,28512,28513,28514,28515,28516,28518,28519,28521,28522,28523,28524,28525,28526,28527,28528,28530,28531,28532,28534,28535,28536,28538,28539,28540,28541,28542,28543,28544,28546,28548,28549,28550,28551,28552,28553,28554,28555,28556,28557,28558,28560,28562,28563,28564,28565,28566,28567,28573,28574,28575,28576,28577,28578,28579,28580,28581,28582,28583,28584,28585,28586,28587,28588,28589,28590,28591,28592,28593,28594,28595,28596,28598,28600,28601,28602,28604,28605,28607,28608,28609,28610,28611,28612,28614,28615,28616,28617,28618,28619,28620,28621,28622,28623,28627,28628,28629,28632,28635,28636,28637,28638,28639,28640,28641,28642,28643,28644,28646,28647,28648,28649,28651,28652,28653,28654,28655,28656,28657,28658,28660,28663,28664,28666,28667,28668,28670,28671,28672,28673,28676,28677,28678,28679,28681,28682,28683,28684,28685,28686,28687,28689,28692,28693,28694,28695,28696,28697,28698,28699,28700,28701,28703,28704,28705,28706,28707,28708,28710,28711,28712,28713,28714,28715,28719,28720,28721,28722,28723,28724,28725,28727,28728,28729,28730,28731,28732,28734,28735,28736,28737,28738,28739,28740,28741,28742,28744,28745,28746,28747,28748,28753,28754,28757,28758,28759,28760,28762,28763,28765,28766,28767,28768,28769,28770,28771,28772,28773,28774,28776,28777,28778,28779,28784,28785,28788,28790,28792,28793,28794,28796,28797,28798,28801,28802,28803,28804,28805,28806,28810,28811,28814,28817,28818,28819,28820,28821,28822,28824,28825,28826,28831,28833,28836,28839,28841,28843,28844,28845,28846,28847,28848,28849,28851,28852,28853,28855,28856,28858,28861,28862,28869,28870,28871,28872,28874,28875,28876,28877,28878,28879,28881,28882,28883,28884,28885,28886,28887,28888,28889,28890,28892,28893,28894,28895,28896,28897,28898,28900,28911,28912,28915,28916,28918,28919,28920,28921,28922,28923,28924,28925,28927,28928,28930,28932,28934,28937,28938,28939,28940,28941,28942,28944,28947,28951,28953,28954,28955,28956,28958,28959,28960,28961,28962,28963,28965,28966,28968,28971,28972,28974,28975,28976,28977,28978,28982,28986,28993,28994,28995,28996,28997,28998,28999,29001,29002,29003,29004,29005,29006,29007,29008,29010,29011,29012,29014,29016,29017,29018,29020,29021,29022,29023,29024,29025,29026,29027,29028,29029,29030,29031,29032,29033,29034,29036,29038,29040,29042,29048,29051,29053,29056,29057,29058,29060,29061,29062,29063,29065,29066,29071,29072,29074,29076,29078,29079,29081,29082,29083,29084,29085,29086,29087,29088,29089,29092,29093,29095,29096,29097,29098,29100,29103,29104,29105,29106,29107,29108,29109,29112,29113,29114,29116,29117,29118,29119,29120,29121,29122,29123,29124,29125,29126,29127,29128,29129,29130,29131,29134,29135,29136,29138,29140,29141,29142,29144,29145,29146,29147,29148,29149,29151,29152,29153,29154,29156,29157,29158,29159,29160,29164,29165,29166,29168,29169,29170,29172,29176,29177,29179,29180,29181,29182,29183,29185,29186,29187,29189,29190,29191,29194,29196,29197,29198,29200,29203,29204,29205,29209,29210,29211,29213,29214,29215,29218,29219,29220,29222,29223,29224,29225,29226,29228,29229,29230,29232,29237,29238,29240,29241,29242,29243,29245,29246,29247,29248,29249,29250,29252,29254,29255,29256,29257,29258,29259,29260,29263,29266,29267,29269,29270,29271,29272,29273,29274,29275,29277,29278,29279,29280,29281,29282,29283,29287,29289,29290,29292,29294,29295,29296,29298,29299,29300,29302,29303,29304,29305,29307,29308,29309,29310,29311,29312,29313,29314,29316,29317,29318,29320,29321,29323,29324,29325,29326,29328,29329,29330,29331,29333,29334,29335,29336,29338,29339,29341,29342,29345,29346,29347,29348,29349,29350,29351,29352,29353,29354,29356,29358,29359,29360,29364,29365,29370,29373,29375,29376,29377,29378,29379,29380,29381,29382,29385,29386,29387,29388,29390,29392,29393,29394,29396,29398,29399,29400,29401,29402,29404,29407,29408,29409,29411,29412,29414,29416,29417,29418,29419,29427,29428,29430,29431,29432,29433,29434,29435,29436,29437,29438,29439,29440,29441,29447,29448,29450,29451,29452,29455,29457,29458,29459,29462,29463,29464,29465,29467,29468,29469,29470,29474,29475,29477,29478,29479,29480,29481,29482,29483,29485,29486,29488,29489,29490,29491,29492,29493,29494,29495,29498,29499,29500,29502,29503,29504,29506,29507,29508,29509,29513,29514,29516,29517,29518,29520,29521,29522,29527,29528,29529,29530,29531,29533,29534,29535,29536,29537,29538,29541,29542,29543,29544,29545,29546,29547,29548,29550,29551,29552,29554,29555,29557,29558,29559,29560,29562,29563,29564,29565,29566,29567,29568,29569,29570,29571,29572,29573,29574,29575,29576,29577,29578,29579,29582,29586,29587,29588,29589,29590,29591,29596,29597,29599,29600,29601,29602,29604,29605,29606,29607,29608,29609,29610,29611,29612,29613,29618,29619,29620,29621,29622,29623,29624,29625,29627,29628,29630,29631,29632,29634,29635,29637,29638,29639,29640,29641,29642,29643,29644,29645,29646,29647,29650,29651,29652,29653,29654,29655,29656,29657,29658,29659,29660,29661,29662,29664,29667,29668,29669,29671,29672,29673,29674,29675,29677,29678,29679,29684,29685,29686,29687,29688,29690,29692,29693,29694,29695,29696,29697,29699,29700,29701,29702,29703,29704,29705,29706,29707,29708,29709,29717,29718,29719,29722,29723,29725,29728,29729,29730,29731,29732,29733,29734,29736,29737,29738,29739,29740,29741,29742,29743,29744,29745,29746,29747,29748,29749,29750,29753,29754,29759,29760,29761,29762,29764,29766,29767,29769,29770,29771,29773,29774,29775,29776,29777,29778,29779,29780,29781,29783,29785,29786,29787,29788,29789,29790,29791,29792,29794,29795,29796,29797,29799,29801,29802,29804,29805,29806,29807,29808,29809,29810,29811,29812,29813,29814,29817,29820,29821,29822,29823,29824,29825,29826,29827,29829,29830,29831,29832,29833,29834,29835,29840,29842,29844,29845,29847,29848,29850,29852,29854,29855,29856,29857,29859,29860,29861,29862,29863,29864,29865,29866,29867,29869,29871,29872,29873,29874,29877,29878,29879,29880,29882,29883,29885,29886,29887,29888,29889,29890,29891,29893,29896,29898,29899,29903,29908,29909,29910,29911,29912,29913,29914,29915,29916,29917,29918,29919,29920,29921,29922,29923,29924,29925,29926,29928,29929,29932,29934,29940,29941,29942,29943,29947,29949,29950,29951,29952,29954,29955,29956,29959,29960,29963,29964,29965,29966,29967,29968,29969,29970,29971,29972,29973,29974,29975,29976,29978,29980,29981,29982,29983,29985,29986,29989,29990,29992,29993,29994,29995,29996,29997,29998,29999,30000,30001,30002,30003,30007,30008,30009,30010,30013,30014,30015,30016,30023,30024,30026,30027,30028,30029,30030,30031,30036,30041,30042,30043,30044,30045,30047,30050,30051,30052,30053,30054,30055,30058,30059,30060,30062,30063,30064,30070,30071,30072,30073,30077,30078,30079,30080,30084,30086,30087,30090,30091,30092,30094,30095,30096,30097,30098,30100,30101,30104,30105,30106,30109,30110,30114,30115,30116,30117,30119,30122,30123,30128,30130,30131,30132,30133,30134,30136,30137,30138,30139,30140,30141,30142,30143,30144,30145,30146,30148,30149,30151,30154,30155,30156,30157,30158,30159,30160,30161,30162,30164,30165,30167,30168,30169,30170,30171,30173,30174,30175,30176,30177,30178,30179,30180,30182,30183,30189,30191,30192,30193,30194,30195,30196,30197,30198,30199,30200,30201,30202,30203,30204,30205,30206,30207,30208,30209,30210,30211,30216,30217,30218,30219,30220,30221,30223,30224,30225,30227,30228,30229,30230,30233,30234,30235,30236,30237,30238,30239,30240,30241,30242,30243,30244,30245,30246,30247,30248,30249,30252,30253,30255,30256,30257,30258,30259,30260,30261,30264,30266,30268,30269,30274,30275,30278,30279,30280,30281,30284,30287,30288,30289,30290,30291,30294,30295,30296,30297,30298,30300,30303,30304,30305,30306,30308,30309,30313,30314,30316,30317,30318,30319,30320,30321,30322,30325,30326,30328,30329,30331,30332,30333,30334,30335,30337,30338,30340,30342,30343,30344,30345,30346,30347,30350,30351,30352,30354,30355,30357,30358,30361,30362,30363,30364,30365,30366,30369,30372,30373,30374,30378,30379,30381,30382,30383,30384,30388,30389,30391,30392,30394,30395,30397,30398,30399,30402,30403,30404,30405,30406,30408,30409,30410,30412,30413,30414,30418,30419,30420,30426,30427,30428,30429,30430,30431,30433,30435,30436,30437,30438,30439,30441,30442,30444,30445,30446,30447,30448,30449,30450,30451,30452,30453,30455,30456,30457,30458,30459,30460,30462,30465,30467,30468,30469,30471,30472,30473,30474,30475,30480,30481,30482,30483,30485,30489,30490,30491,30493,30494,30495,30496,30498,30499,30501,30502,30503,30504,30505,30509,30511,30513,30514,30515,30516,30517,30518,30519,30520,30521,30522,30523,30524,30525,30526,30528,30532,30533,30534,30535,30538,30539,30540,30541,30542,30543,30546,30548,30549,30550,30552,30553,30554,30555,30556,30558,30559,30560,30561,30562,30563,30565,30566,30567,30568,30569,30570,30571,30572,30573,30574,30575,30585,30588,30589,30590,30591,30592,30593,30594,30595,30596,30597,30599,30600,30601,30603,30604,30605,30606,30607,30609,30613,30615,30617,30618,30619,30620,30621,30622,30623,30624,30625,30626,30627,30629,30631,30632,30634,30635,30636,30637,30639,30640,30641,30642,30643,30644,30645,30646,30647,30650,30651,30652,30653,30655,30658,30660,30661,30663,30665,30666,30668,30669,30670,30671,30672,30675,30676,30677,30679,30680,30681,30682,30683,30684,30686,30688,30690,30691,30693,30694,30695,30696,30697,30700,30701,30702,30703,30704,30705,30706,30707,30708,30711,30712,30713,30714,30715,30716,30717,30722,30723,30725,30726,30729,30732,30733,30734,30735,30736,30737,30738,30739,30740,30749,30751,30752,30753,30754,30755,30757,30758,30759,30760,30761,30762,30763,30764,30765,30766,30767,30768,30769,30770,30771,30772,30773,30775,30776,30780,30787,30789,30792,30793,30794,30796,30797,30798,30800,30802,30812,30813,30814,30816,30818,30820,30821,30824,30825,30826,30827,30828,30829,30830,30831,30832,30833,30841,30843,30844,30846,30847,30848,30849,30851,30852,30853,30854,30855,30857,30860,30862,30863,30865,30867,30868,30869,30870,30871,30872,30873,30874,30878,30879,30880,30881,30882,30883,30884,30885,30887,30888,30889,30890,30891,30892,30893,30895,30896,30897,30898,30899,30900,30906,30907,30908,30910,30913,30915,30916,30917,30919,30920,30921,30922,30923,30924,30925,30926,30927,30928,30929,30931,30932,30933,30935,30938,30939,30941,30942,30943,30944,30945,30946,30947,30949,30951,30952,30953,30954,30956,30957,30959,30961,30962,30963,30964,30967,30969,30970,30971,30972,30973,30974,30975,30977,30978,30980,30981,30985,30988,30990,30992,30993,30994,30995,30996,30999,31001,31003,31004,31005,31006,31009,31011,31012,31013,31014,31015,31016,31017,31018,31019,31020,31021,31022,31023,31025,31028,31029,31032,31033,31034,31036,31037,31038,31039,31040,31041,31042,31044,31045,31046,31047,31048,31049,31050,31051,31052,31055,31056,31057,31058,31059,31060,31061,31062,31063,31066,31067,31068,31069,31070,31071,31072,31073,31075,31076,31077,31079,31080,31081,31082,31083,31085,31088,31089,31090,31091,31092,31097,31098,31100,31101,31103,31104,31105,31106,31110,31112,31114,31115,31117,31118,31119,31120,31122,31123,31124,31125,31126,31127,31128,31130,31131,31132,31133,31136,31137,31138,31140,31142,31143,31144,31145,31146,31147,31148,31149,31150,31152,31153,31154,31155,31156,31158,31159,31160,31161,31162,31163,31165,31166,31167,31168,31169,31173,31174,31176,31177,31179,31180,31181,31182,31183,31185,31186,31188,31189,31190,31192,31196,31197,31198,31199,31200,31203,31204,31206,31207,31209,31210,31211,31212,31213,31214,31222,31223,31224,31226,31227,31232,31234,31235,31236,31237,31240,31242,31243,31244,31245,31248,31249,31250,31251,31252,31253,31255,31256,31257,31258,31259,31260,31262,31263,31264,31266,31270,31272,31275,31278,31279,31280,31281,31287,31289,31291,31292,31293,31295,31296,31300,31302,31303,31304,31306,31307,31308,31309,31310,31316,31318,31319,31320,31322,31323,31324,31327,31328,31329,31330,31335,31336,31337,31339,31340,31341,31342,31344,31345,31348,31349,31350,31352,31353,31354,31355,31358,31359,31360,31361,31364,31365,31366,31367,31368,31369,31370,31371,31372,31375,31376,31378,31380,31381,31382,31383,31384,31385,31390,31391,31392,31394,31395,31400,31401,31402,31403,31404,31406,31407,31409,31410,31411,31412,31413,31414,31415,31416,31418,31422,31423,31424,31425,31428,31429,31431,31434,31435,31441,31448,31449,31455,31456,31459,31460,31461,31462,31467,31469,31470,31471,31478,31479,31481,31482,31483,31485,31487,31488,31489,31492,31493,31494,31496,31497,31498,31502,31503,31504,31505,31506,31507,31512,31513,31514,31515,31517,31518,31520,31522,31523,31524,31525,31526,31528,31530,31531,31532,31533,31534,31535,31536,31537,31538,31539,31540,31541,31544,31547,31552,31556,31557,31558,31559,31560,31561,31562,31563,31564,31565,31566,31567,31568,31569,31570,31572,31574,31576,31584,31585,31587,31588,31589,31590,31591,31593,31597,31598,31600,31601,31602,31603,31604,31605,31606,31607,31608,31618,31620,31621,31623,31624,31626,31627,31628,31629,31630,31631,31632,31633,31636,31637,31638,31639,31640,31641,31643,31644,31645,31648,31649,31652,31660,31661,31663,31665,31668,31669,31671,31672,31673,31678,31680,31681,31684,31686,31687,31689,31690,31691,31692,31694,31700,31701,31704,31705,31706,31707,31708,31709,31710,31711,31712,31713,31714,31715,31716,31717,31718,31719,31720,31721,31722,31723,31728,31729,31730,31731,31732,31735,31736,31737,31739,31741,31742,31743,31744,31745,31746,31747,31749,31750,31751,31753,31754,31755,31756,31757,31758,31759,31760,31761,31765,31769,31772,31773,31774,31775,31776,31777,31778,31779,31781,31782,31783,31784,31785,31786,31787,31788,31789,31792,31795,31799,31800,31801,31803,31804,31805,31806,31807,31808,31811,31813,31815,31816,31817,31818,31820,31821,31824,31827,31828,31831,31833,31834,31835,31836,31839,31840,31843,31844,31845,31846,31847,31849,31850,31851,31852,31854,31855,31858,31859,31861,31864,31865,31866,31869,31871,31872,31873,31876,31877,31880,31881,31882,31884,31885,31889,31890,31892,31893,31894,31895,31896,31902,31903,31905,31906,31907,31909,31911,31912,31919,31921,31922,31923,31924,31925,31929,31930,31931,31932,31933,31934,31935,31941,31944,31946,31947,31948,31950,31952,31953,31954,31956,31957,31958,31959,31961,31964,31965,31966,31967,31968,31970,31975,31976,31978,31980,31982,31983,31984,31985,31986,31988,31990,31991,31992,31995,31997,31998,32000,32001,32002,32003,32004,32005,32006,32007,32008,32009,32010,32011,32012,32013,32014,32015,32016,32017,32018,32019,32020,32021,32022,32023,32024,32025,32026,32027,32028,32029,32030,32031,32032,32033,32034,32040,32041,32043,32044,32046,32047,32048,32049,32050,32051,32053,32054,32056,32057,32058,32059,32060,32061,32062,32063,32064,32065,32066,32067,32068,32069,32070,32071,32074,32078,32079,32080,32081,32082,32083,32084,32085,32086,32088,32091,32092,32094,32095,32097,32098,32099,32102,32103,32104,32105,32106,32107,32109,32110,32111,32112,32113,32114,32115,32121,32122,32123,32124,32125,32127,32128,32129,32131,32132,32133,32134,32136,32140,32141,32142,32143,32145,32146,32147,32148,32150,32156,32157,32158,32159,32160,32161,32162,32163,32166,32167,32169,32170,32172,32173,32174,32175,32176,32177,32178,32180,32181,32183,32184,32185,32186,32187,32188,32189,32190,32191,32192,32193,32194,32196,32197,32198,32199,32201,32202,32203,32204,32206,32210,32215,32216,32217,32218,32219,32221,32222,32223,32224,32225,32227,32230,32231,32232,32233,32234,32236,32238,32239,32240,32241,32242,32243,32244,32246,32247,32249,32250,32251,32259,32264,32265,32266,32267,32268,32269,32270,32271,32272,32273,32274,32275,32276,32277,32278,32279,32282,32283,32284,32285,32286,32287,32288,32289,32290,32291,32292,32293,32297,32298,32299,32301,32302,32303,32304,32305,32306,32307,32308,32309,32310,32311,32312,32313,32314,32315,32316,32317,32318,32319,32320,32321,32322,32323,32324,32325,32326,32327,32328,32329,32332,32336,32337,32338,32339,32340,32341,32342,32343,32344,32345,32346,32348,32350,32351,32352,32353,32354,32355,32360,32361,32362,32363,32365,32367,32368,32370,32371,32372,32373,32374,32375,32376,32377,32378,32379,32380,32381,32382,32384,32385,32386,32390,32391,32392,32394,32395,32396,32397,32399,32401,32403,32404,32405,32406,32407,32408,32409,32410,32411,32412,32566,32568,32569,32570,32573,32574,32575,32579,32580,32581,32584,32586,32587,32588,32589,32591,32592,32593,32596,32597,32600,32603,32604,32605,32606,32607,32608,32609,32611,32613,32614,32615,32616,32617,32618,32619,32620,32621,32622,32624,32626,32627,32629,32630,32631,32633,32634,32635,32636,32637,32638,32639,32643,32645,32646,32647,32648,32649,32650,32651,32652,32653,32654,32655,32657,32658,32660,32661,32662,32666,32667,32668,32669,32670,32672,32673,32674,32676,32677,32678,32679,32680,32681,32684,32685,32687,32688,32689,32690,32691,32693,32694,32695,32696,32697,32698,32699,32700,32701,32702,32703,32704,32705,32706,32707,32709,32711,32713,32714,32715,32716,32717,32718,32719,32720,32721,32722,32724,32725,32727,32731,32732,32734,32735,32736,32737,32738,32739,32741,32742,32744,32745,32746,32747,32748,32749,32750,32751,32752,32753,32754,32755,32756,32757,32759,32760,32761,32763,32764,32765,32766,32767,32768,32769,32771,32772,32773,32774,32775,32779,32780,32781,32782,32783,32784,32785,32786,32788,32789,32790,32791,32792,32793,32795,32796,32798,32799,32801,32804,32806,32808,32809,32810,32812,32816,32819,32820,32821,32822,32823,32825,32829,32830,32831,32835,32838,32839,32840,32842,32847,32848,32849,32850,32854,32856,32858,32860,32861,32862,32868,32871,32876,32879,32880,32881,32882,32883,32885,32886,32887,32888,32889,32893,32894,32895,32898,32900,32901,32902,32903,32905,32906,32907,32908,32911,32912,32914,32915,32917,32918,32920,32921,32922,32923,32924,32925,32929,32930,32931,32933,32937,32938,32939,32941,32942,32943,32945,32946,32948,32949,32952,32954,32962,32963,32964,32965,32967,32968,32969,32970,32972,32973,32974,32975,32976,32977,32980,32981,32982,32983,32984,32985,32986,32987,32988,32989,32990,32992,32993,32995,32996,32997,32998,33005,33007,33008,33009,33010,33011,33012,33013,33016,33017,33018,33019,33020,33021,33022,33024,33025,33026,33029,33030,33032,33034,33045,33046,33048,33049,33051,33053,33054,33055,33057,33058,33059,33060,33061,33063,33065,33067,33068,33069,33071,33072,33081,33082,33085,33086,33091,33092,33094,33095,33098,33099,33100,33101,33102,33103,33104,33105,33106,33107,33108,33109,33115,33116,33118,33120,33121,33122,33124,33125,33126,33127,33129,33131,33134,33135,33136,33137,33138,33139,33140,33142,33143,33144,33145,33146,33151,33152,33154,33155,33158,33159,33160,33161,33162,33163,33164,33165,33167,33173,33175,33176,33177,33178,33179,33180,33181,33182,33183,33184,33186,33187,33190,33191,33192,33193,33195,33196,33198,33200,33201,33202,33203,33204,33205,33207,33209,33210,33211,33212,33213,33214,33215,33216,33218,33219,33220,33221,33222,33223,33225,33226,33228,33229,33231,33232,33233,33234,33237,33239,33240,33241,33242,33243,33245,33246,33247,33248,33249,33250,33251,33253,33254,33255,33256,33257,33258,33260,33261,33262,33263,33266,33267,33268,33271,33272,33273,33274,33275,33276,33278,33279,33280,33281,33282,33284,33285,33287,33288,33289,33290,33291,33292,33293,33296,33297,33298,33300,33301,33302,33307,33308,33309,33310,33311,33312,33313,33314,33317,33318,33320,33322,33323,33324,33327,33330,33331,33332,33333,33334,33335,33336,33337,33338,33340,33341,33343,33344,33346,33348,33349,33351,33353,33355,33358,33359,33360,33361,33362,33363,33365,33366,33367,33368,33369,33370,33371,33372,33374,33375,33377,33379,33380,33382,33384,33385,33387,33388,33389,33390,33391,33393,33394,33396,33397,33399,33400,33404,33405,33406,33407,33408,33411,33412,33413,33415,33418,33419,33421,33422,33423,33424,33425,33426,33427,33428,33432,33433,33434,33435,33437,33438,33439,33440,33441,33442,33443,33444,33445,33447,33448,33449,33451,33452,33453,33454,33455,33456,33457,33459,33460,33461,33462,33463,33464,33465,33466,33467,33468,33469,33470,33472,33474,33475,33489,33490,33491,33492,33493,33494,33495,33497,33499,33500,33502,33503,33504,33505,33507,33508,33509,33510,33511,33512,33514,33515,33516,33517,33518,33519,33520,33521,33522,33523,33524,33525,33526,33529,33530,33531,33532,33534,33536,33537,33538,33539,33540,33541,33542,33543,33544,33545,33548,33549,33558,33559,33561,33563,33564,33566,33568,33570,33572,33573,33574,33575,33576,33577,33578,33579,33580,33581,33583,33585,33586,33587,33588,33589,33590,33591,33592,33593,33594,33595,33596,33599,33600,33601,33602,33603,33604,33605,33607,33608,33609,33610,33611,33612,33613,33614,33615,33616,33617,33618,33619,33620,33622,33635,33651,33652,33653,33654,33655,33656,33658,33659,33660,33661,33662,33663,33665,33667,33670,33671,33672,33673,33674,33675,33676,33677,33678,33679,33680,33682,33683,33684,33685,33686,33687,33688,33689,33690,33691,33693,33694,33696,33698,33699,33700,33701,33702,33703,33704,33705,33706,33707,33710,33711,33712,33721,33725,33727,33728,33729,33730,33731,33732,33733,33734,33735,33736,33737,33738,33739,33740,33741,33742,33743,33745,33748,33749,33750,33751,33752,33753,33755,33756,33757,33758,33759,33760,33761,33762,33763,33764,33765,33767,33768,33769,33770,33771,33772,33774,33775,33776,33777,33778,33779,33780,33781,33782,33784,33785,33786,33787,33788,33789,33790,33791,33793,33795,33796,33798,33799,33801,33802,33803,33804,33805,33806,33807,33808,33809,33810,33811,33814,33819,33827,33828,33833,33835,33836,33837,33838,33839,33840,33841,33842,33843,33844,33845,33846,33847,33848,33849,33850,33851,33852,33853,33855,33856,33858,33859,33860,33861,33862,33863,33865,33867,33868,33869,33870,33872,33873,33874,33876,33878,33879,33881,33882,33883,33885,33886,33887,33888,33889,33890,33891,33893,33894,33895,33896,33897,33899,33900,33901,33902,33903,33904,33907,33908,33909,33910,33911,33912,33913,33914,33917,33918,33922,33926,33933,33934,33935,33936,33937,33940,33943,33944,33945,33946,33947,33948,33949,33950,33951,33952,33953,33954,33956,33959,33960,33961,33962,33963,33964,33966,33967,33968,33969,33970,33972,33974,33976,33977,33978,33979,33980,33983,33984,33985,33986,33988,33989,33990,33991,33993,33994,33995,33996,33997,33998,33999,34000,34001,34002,34003,34004,34006,34007,34011,34023,34024,34025,34026,34027,34028,34030,34031,34032,34033,34034,34035,34036,34038,34039,34041,34042,34043,34044,34045,34046,34047,34048,34050,34053,34054,34055,34056,34057,34058,34059,34060,34061,34062,34063,34065,34066,34067,34068,34069,34070,34071,34072,34073,34074,34076,34077,34078,34079,34080,34081,34083,34084,34085,34086,34087,34088,34089,34090,34091,34092,34093,34094,34095,34096,34097,34099,34107,34109,34110,34112,34113,34114,34115,34116,34117,34118,34119,34120,34121,34122,34124,34125,34126,34129,34131,34132,34133,34134,34135,34136,34137,34139,34141,34142,34144,34145,34146,34147,34148,34149,34150,34151,34152,34153,34154,34155,34156,34157,34158,34161,34165,34166,34167,34168,34169,34170,34171,34172,34174,34176,34177,34178,34179,34180,34181,34182,34183,34184,34185,34186,34187,34188,34189,34190,34191,34192,34193,34196,34197,34198,34200,34201,34202,34203,34204,34205,34206,34207,34208,34209,34210,34211,34212,34214,34215,34216,34217,34218,34223,34224,34225,34227,34228,34229,34230,34231,34232,34233,34234,34237,34238,34239,34240,34242,34243,34244,34245,34246,34247,34248,34249,34251,34253,34254,34255,34256,34257,34258,34261,34263,34264,34265,34266,34268,34269,34270,34271,34273,34274,34275,34276,34277,34278,34280,34281,34282,34283,34284,34285,34287,34288,34289,34290,34294,34295,34296,34297,34298,34299,34301,34302,34303,34304,34305,34308,34309,34310,34311,34313,34314,34315,34316,34317,34321,34327,34328,34329,34330,34331,34332,34334,34335,34336,34337,34338,34339,34340,34341,34342,34343,34345,34346,34348,34349,34350,34353,34354,34355,34356,34357,34358,34360,34361,34362,34363,34364,34366,34367,34368,34371,34374,34375,34376,34379,34380,34381,34382,34384,34386,34387,34388,34389,34390,34393,34395,34396,34398,34399,34401,34402,34403,34404,34405,34407,34408,34409,34410,34411,34413,34414,34415,34416,34417,34419,34420,34423,34425,34426,34427,34428,34437,34438,34439,34442,34443,34444,34445,34446,34448,34449,34451,34452,34453,34454,34455,34456,34457,34458,34460,34461,34462,34465,34466,34467,34468,34469,34471,34472,34473,34474,34479,34480,34481,34483,34484,34485,34486,34487,34488,34489,34490,34491,34492,34493,34494,34495,34496,34497,34498,34499,34500,34501,34502,34503,34504,34505,34507,34508,34512,34513,34515,34516,34518,34519,34520,34521,34522,34523,34524,34525,34526,34527,34530,34531,34532,34534,34536,34537,34538,34539,34540,34541,34549,34550,34551,34552,34553,34554,34555,34558,34560,34561,34562,34563,34564,34565,34566,34567,34568,34569,34570,34571,34572,34573,34574,34577,34578,34579,34584,34585,34586,34587,34588,34590,34592,34593,34594,34595,34596,34597,34598,34599,34600,34601,34602,34604,34605,34606,34608,34609,34610,34611,34612,34613,34615,34616,34618,34619,34620,34622,34623,34624,34625,34626,34627,34630,34636,34637,34638,34639,34640,34641,34642,34643,34644,34645,34646,34647,34648,34649,34650,34651,34652,34653,34654,34655,34656,34657,34658,34659,34660,34661,34662,34663,34664,34665,34666,34667,34668,34669,34670,34671,34675,34676,34677,34678,34679,34680,34681,34682,34683,34685,34689,34690,34691,34692,34693,34695,34696,34697,34701,34703,34704,34705,34706,34707,34708,34710,34711,34712,34714,34715,34716,34717,34718,34719,34722,34723,34724,34730,34731,34732,34733,34734,34735,34736,34738,34739,34740,34741,34742,34743,34744,34745,34746,34747,34748,34749,34750,34751,34752,34754,34755,34756,34757,34758,34760,34761,34762,34763,34764,34769,34770,34771,34772,34775,34776,34777,34779,34780,34781,34782,34783,34784,34785,34786,34787,34788,34789,34790,34791,34792,34794,34795,34796,34797,34799,34802,34803,34804,34806,34807,34809,34810,34811,34812,34814,34815,34816,34817,34818,34819,34821,34822,34824,34825,34826,34827,34828,34829,34832,34833,34835,34836,34837,34838,34839,34841,34843,34844,34845,34847,34848,34849,34850,34851,34852,34853,34854,34856,34857,34858,34859,34860,34862,34863,34864,34865,34866,34867,34869,34870,34871,34872,34873,34875,34876,34877,34878,34879,34880,34881,34883,34884,34888,34890,34891,34892,34893,34894,34898,34899,34901,34902,34903,34905,34906,34907,34909,34913,34914,34915,34919,34920,34921,34922,34923,34925,34927,34928,34929,34930,34932,34933,34934,34935,34937,34940,34941,34942,34943,34944,34945,34946,34947,34948,34952,34953,34955,34956,34957,34958,34961,34962,34963,34965,34966,34967,34968,34969,34970,34971,34974,34975,34977,34978,34980,34983,34984,34986,34987,34988,34993,34994,34998,34999,35000,35001,35002,35004,35005,35006,35008,35009,35010,35017,35018,35019,35020,35021,35022,35023,35024,35026,35028,35029,35030,35031,35032,35033,35034,35035,35036,35037,35038,35039,35041,35047,35048,35051,35052,35054,35055,35056,35057,35058,35059,35060,35062,35063,35064,35065,35066,35067,35068,35069,35070,35072,35073,35074,35077,35078,35079,35081,35082,35083,35084,35086,35088,35089,35090,35091,35092,35093,35094,35095,35096,35097,35098,35102,35103,35105,35106,35107,35109,35110,35111,35113,35114,35115,35116,35117,35118,35119,35120,35121,35122,35123,35125,35126,35127,35128,35131,35132,35133,35134,35137,35138,35140,35142,35145,35147,35148,35151,35152,35153,35154,35155,35158,35159,35160,35161,35162,35163,35164,35165,35166,35167,35168,35169,35170,35171,35172,35174,35177,35178,35179,35180,35181,35182,35183,35185,35186,35187,35188,35190,35193,35194,35195,35196,35198,35199,35201,35202,35203,35205,35206,35208,35211,35215,35219,35221,35222,35223,35224,35227,35228,35229,35230,35231,35233,35234,35235,35236,35238,35242,35244,35245,35246,35247,35250,35254,35255,35257,35258,35261,35262,35263,35264,35282,35283,35284,35285,35286,35289,35290,35291,35292,35293,35295,35296,35297,35298,35299,35300,35301,35302,35304,35305,35307,35308,35309,35312,35313,35314,35315,35316,35318,35319,35320,35322,35323,35324,35326,35327,35328,35330,35331,35332,35335,35336,35338,35340,35342,35343,35344,35345,35346,35347,35349,35350,35351,35352,35355,35357,35358,35359,35362,35363,35365,35367,35370,35372,35373,35376,35377,35380,35382,35385,35386,35387,35388,35390,35391,35392,35393,35396,35397,35398,35400,35402,35404,35405,35406,35407,35408,35409,35410,35412,35413,35414,35415,35416,35417,35419,35422,35424,35425,35426,35427,35430,35432,35433,35435,35436,35437,35438,35440,35441,35442,35443,35444,35445,35446,35447,35449,35450,35451,35452,35455,35457,35458,35459,35460,35461,35462,35463,35467,35468,35469,35471,35473,35474,35475,35477,35478,35480,35481,35482,35486,35488,35489,35491,35492,35493,35494,35495,35496,35498,35499,35504,35506,35510,35512,35513,35514,35515,35516,35517,35518,35519,35520,35522,35523,35524,35525,35526,35527,35528,35529,35531,35533,35535,35537,35538,35539,35540,35541,35542,35543,35544,35545,35547,35548,35549,35550,35551,35552,35553,35554,35556,35558,35559,35560,35563,35565,35566,35567,35568,35569,35570,35571,35572,35573,35574,35575,35576,35578,35579,35580,35582,35583,35584,35585,35586,35588,35589,35590,35591,35592,35594,35595,35597,35598,35599,35600,35601,35602,35603,35604,35605,35606,35607,35608,35609,35610,35611,35612,35613,35614,35616,35618,35619,35620,35621,35622,35623,35624,35626,35627,35628,35630,35631,35632,35633,35635,35637,35638,35639,35641,35642,35643,35644,35645,35646,35648,35649,35650,35653,35654,35655,35656,35657,35658,35659,35662,35663,35664,35665,35666,35667,35668,35669,35670,35671,35672,35673,35674,35676,35677,35679,35680,35683,35685,35686,35687,35688,35690,35691,35692,35693,35695,35696,35700,35703,35704,35705,35706,35707,35709,35710,35711,35712,35714,35716,35717,35718,35720,35722,35723,35724,35726,35730,35731,35732,35733,35734,35736,35737,35738,35740,35742,35743,35895,35897,35899,35900,35901,35902,35903,35905,35906,35907,35909,35910,35911,35912,35913,35914,35915,35916,35917,35918,35919,35920,35924,35925,35926,35927,35930,35932,35933,35935,35937,35938,35940,35941,35942,35944,35945,35946,35947,35948,35949,35951,35952,35953,35954,35955,35957,35958,35959,35960,35961,35962,35963,35965,35968,35969,35970,35972,35973,35974,35977,35978,35980,35981,35983,35984,35985,35986,35987,35988,35989,35991,35992,35993,35994,35996,35997,35998,36000,36001,36002,36003,36004,36005,36007,36008,36009,36010,36011,36012,36015,36016,36018,36019,36020,36021,36022,36023,36024,36025,36026,36027,36028,36029,36030,36031,36032,36033,36034,36035,36036,36037,36039,36040,36042,36044,36047,36049,36050,36051,36053,36055,36057,36058,36060,36061,36062,36063,36064,36065,36066,36067,36068,36069,36070,36071,36072,36074,36076,36077,36078,36080,36081,36083,36084,36085,36088,36089,36090,36091,36092,36093,36094,36096,36098,36100,36101,36102,36103,36104,36105,36106,36109,36111,36112,36115,36116,36117,36118,36119,36121,36123,36196,36198,36199,36200,36201,36203,36204,36205,36206,36207,36208,36210,36211,36212,36214,36215,36216,36217,36219,36221,36224,36225,36228,36229,36233,36234,36236,36237,36238,36239,36240,36241,36242,36243,36244,36245,36246,36249,36251,36252,36255,36256,36257,36259,36261,36263,36264,36266,36267,36268,36269,36270,36271,36274,36275,36276,36277,36278,36279,36281,36282,36284,36286,36287,36289,36290,36293,36294,36295,36296,36299,36300,36301,36302,36303,36304,36305,36307,36309,36310,36311,36312,36313,36314,36315,36316,36317,36319,36320,36321,36322,36323,36324,36326,36327,36328,36329,36330,36331,36332,36334,36335,36336,36337,36338,36339,36340,36346,36348,36349,36350,36351,36352,36354,36355,36356,36357,36358,36359,36361,36362,36365,36367,36368,36369,36370,36371,36372,36373,36374,36375,36376,36377,36378,36379,36380,36381,36382,36383,36384,36385,36386,36387,36388,36389,36390,36391,36393,36395,36398,36400,36401,36403,36404,36405,36406,36408,36409,36412,36413,36414,36415,36416,36417,36418,36420,36421,36423,36424,36425,36426,36427,36428,36429,36430,36432,36435,36436,36437,36438,36439,36441,36442,36443,36444,36445,36446,36447,36448,36449,36450,36451,36452,36453,36454,36455,36457,36458,36460,36461,36463,36466,36467,36468,36470,36472,36474,36475,36476,36481,36482,36484,36485,36486,36487,36488,36489,36490,36491,36492,36493,36494,36496,36497,36498,36499,36500,36501,36502,36503,36504,36506,36509,36510,36511,36512,36513,36515,36516,36517,36518,36520,36521,36522,36523,36524,36530,36538,36541,36544,36546,36553,36554,36555,36556,36557,36559,36561,36562,36563,36564,36567,36568,36571,36572,36573,36574,36575,36576,36577,36581,36582,36583,36584,36585,36587,36588,36590,36591,36593,36596,36597,36598,36599,36600,36601,36602,36603,36604,36606,36607,36608,36609,36610,36611,36613,36614,36615,36616,36617,36618,36619,36621,36622,36624,36625,36626,36627,36628,36629,36630,36631,36632,36634,36635,36636,36637,36638,36639,36640,36643,36644,36645,36646,36649,36650,36652,36654,36655,36658,36659,36660,36661,36662,36663,36664,36665,36667,36670,36671,36672,36674,36675,36676,36677,36678,36679,36680,36681,36683,36685,36686,36687,36688,36689,36690,36691,36692,36693,36694,36695,36696,36697,36698,36699,36701,36702,36703,36704,36705,36706,36707,36708,36763,36764,36767,36771,36774,36776,36781,36782,36783,36784,36785,36786,36788,36789,36792,36799,36802,36804,36805,36806,36809,36811,36813,36814,36817,36818,36819,36820,36821,36822,36823,36832,36833,36834,36835,36836,36837,36838,36840,36842,36843,36845,36846,36848,36852,36853,36854,36855,36856,36858,36859,36860,36861,36862,36863,36864,36865,36866,36867,36868,36869,36870,36875,36876,36877,36879,36880,36881,36884,36885,36886,36887,36889,36890,36891,36892,36893,36894,36895,36896,36897,36898,36899,36900,36909,36910,36911,36913,36914,36916,36917,36918,36920,36924,36925,36926,36927,36929,36930,36932,36935,36937,36938,36939,36941,36942,36943,36944,36945,36946,36947,36948,36949,36952,36953,36955,36956,36957,36958,36960,36962,36963,36967,36968,36969,36971,36973,36974,36975,36976,36978,36979,36980,36981,36982,36983,36984,36985,36986,36987,36988,36989,36990,36991,36992,36993,36994,36995,36996,36997,36998,36999,37000,37002,37003,37005,37007,37008,37009,37012,37013,37015,37016,37017,37019,37022,37023,37024,37025,37026,37027,37029,37030,37031,37034,37039,37040,37041,37042,37043,37044,37045,37046,37048,37053,37054,37055,37057,37059,37061,37063,37064,37066,37067,37070,37076,37077,37078,37079,37080,37081,37082,37083,37084,37085,37087,37088,37089,37090,37091,37092,37093,37096,37097,37098,37099,37100,37101,37103,37104,37105,37106,37107,37108,37109,37113,37114,37115,37116,37117,37118,37119,37120,37121,37122,37123,37124,37125,37126,37127,37128,37129,37131,37133,37134,37135,37136,37137,37138,37140,37142,37143,37144,37145,37146,37147,37148,37149,37150,37151,37152,37153,37154,37155,37156,37158,37159,37160,37161,37162,37163,37164,37165,37166,37167,37168,37169,37170,37171,37172,37173,37174,37176,37177,37178,37179,37182,37183,37184,37185,37187,37188,37189,37190,37191,37192,37193,37194,37195,37196,37197,37198,37199,37200,37202,37203,37205,37206,37207,37208,37210,37215,37216,37217,37218,37219,37220,37221,37224,37225,37226,37228,37230,37231,37234,37235,37236,37237,37239,37240,37241,37242,37248,37249,37250,37251,37252,37253,37254,37255,37257,37258,37259,37261,37263,37264,37265,37266,37267,37273,37274,37275,37276,37277,37278,37279,37280,37281,37282,37283,37285,37287,37288,37290,37291,37292,37293,37294,37295,37296,37297,37298,37299,37300,37301,37303,37305,37306,37308,37309,37310,37312,37313,37314,37315,37317,37318,37319,37321,37323,37324,37325,37326,37327,37328,37329,37331,37332,37333,37334,37335,37336,37337,37338,37340,37341,37346,37347,37348,37350,37351,37352,37353,37354,37355,37356,37357,37358,37361,37363,37364,37365,37367,37368,37369,37373,37375,37376,37377,37378,37379,37380,37381,37382,37383,37385,37386,37388,37389,37391,37392,37393,37394,37396,37397,37398,37399,37401,37402,37404,37406,37411,37412,37413,37414,37415,37421,37422,37424,37425,37426,37427,37428,37430,37431,37432,37433,37434,37436,37437,37438,37439,37440,37445,37446,37448,37449,37450,37451,37452,37453,37454,37455,37456,37457,37458,37459,37460,37462,37463,37466,37467,37470,37472,37473,37475,37476,37477,37478,37479,37484,37485,37487,37488,37490,37494,37496,37497,37498,37499,37500,37501,37502,37503,37504,37506,37507,37509,37510,37511,37512,37514,37515,37516,37517,37518,37521,37523,37524,37525,37526,37527,37528,37529,37530,37531,37532,37533,37536,37537,37538,37539,37540,37541,37542,37543,37544,37545,37546,37547,37548,37554,37555,37556,37557,37558,37559,37561,37563,37564,37568,37569,37570,37571,37572,37573,37574,37575,37576,37577,37578,37579,37580,37581,37582,37583,37584,37585,37586,37587,37589,37591,37592,37593,37595,37597,37598,37599,37600,37601,37604,37606,37607,37608,37609,37610,37614,37615,37616,37617,37623,37624,37625,37626,37627,37628,37630,37631,37632,37633,37634,37636,37638,37640,37641,37643,37644,37645,37646,37647,37648,37650,37651,37652,37653,37654,37656,37657,37658,37659,37661,37662,37663,37664,37665,37666,37667,37668,37669,37670,37671,37672,37673,37674,37675,37677,37678,37679,37683,37684,37685,37686,37688,37689,37692,37696,37697,37700,37702,37703,37705,37706,37707,37708,37709,37710,37711,37712,37713,37714,37716,37717,37718,37719,37720,37721,37722,37723,37724,37726,37728,37729,37730,37731,37732,37733,37735,37738,37740,37741,37744,37745,37749,37750,37751,37753,37754,37755,37756,37758,37760,37762,37763,37765,37768,37769,37770,37772,37773,37774,37775,37777,37778,37780,37781,37782,37783,37784,37785,37786,37787,37789,37790,37791,37793,37794,37795,37796,37797,37798,37799,37800,37801,37802,37804,37806,37807,37808,37809,37810,37811,37812,37813,37815,37824,37826,37827,37828,37831,37832,37834,37836,37837,37838,37839,37840,37841,37842,37843,37844,37845,37846,37847,37848,37849,37850,37852,37853,37854,37855,37857,37858,37859,37860,37862,37863,37864,37868,37870,37872,37873,37877,37878,37879,37880,37881,37882,37883,37884,37885,37886,37887,37888,37891,37894,37895,37897,37898,37899,37900,37901,37902,37903,37904,37905,37906,37907,37908,37909,37910,37912,37913,37915,37920,37928,37929,37930,37931,37932,37934,37936,37937,37938,37939,37941,37942,37943,37944,37945,37946,37947,37948,37949,37951,37952,37954,37955,37956,37957,37958,37959,37960,37961,37962,37963,37964,37967,37968,37969,37970,37973,37975,37981,37982,37984,37986,37987,37988,37992,37993,37994,37995,37997,37998,37999,38000,38001,38002,38003,38004,38005,38006,38007,38008,38011,38012,38013,38014,38015,38016,38017,38018,38019,38263,38266,38267,38268,38269,38272,38274,38275,38278,38280,38281,38283,38284,38285,38286,38287,38288,38289,38290,38291,38292,38296,38299,38300,38302,38303,38305,38307,38308,38309,38312,38313,38315,38316,38317,38318,38320,38321,38325,38326,38327,38329,38330,38331,38332,38333,38334,38335,38336,38339,38341,38342,38343,38344,38345,38346,38347,38348,38349,38352,38353,38354,38355,38356,38357,38358,38362,38363,38364,38366,38367,38368,38369,38370,38371,38372,38373,38428,38430,38432,38433,38434,38435,38436,38440,38442,38444,38445,38446,38447,38448,38449,38450,38457,38458,38459,38460,38461,38463,38464,38466,38467,38468,38474,38475,38476,38477,38478,38479,38480,38481,38483,38484,38488,38491,38492,38493,38494,38495,38497,38498,38499,38500,38506,38507,38508,38509,38511,38512,38513,38514,38515,38516,38517,38518,38519,38520,38524,38525,38526,38528,38531,38532,38533,38534,38535,38536,38537,38538,38539,38541,38542,38545,38546,38547,38548,38549,38551,38552,38553,38555,38556,38558,38561,38562,38564,38565,38567,38568,38569,38570,38572,38574,38576,38577,38579,38580,38582,38584,38585,38587,38588,38591,38592,38593,38594,38595,38596,38597,38598,38599,38600,38601,38602,38603,38604,38605,38606,38610,38611,38612,38613,38614,38615,38616,38617,38618,38619,38620,38621,38622,38623,38625,38626,38627,38629,38632,38633,38634,38639,38640,38641,38642,38645,38646,38647,38648,38649,38650,38651,38653,38655,38656,38658,38660,38661,38662,38663,38664,38665,38667,38669,38670,38671,38672,38673,38674,38675,38678,38680,38681,38684,38685,38686,38687,38688,38690,38691,38692,38693,38694,38695,38696,38697,38698,38699,38700,38702,38703,38704,38706,38709,38712,38713,38714,38717,38718,38719,38722,38723,38724,38726,38727,38728,38729,38731,38738,38742,38744,38746,38747,38748,38750,38752,38753,38754,38758,38760,38761,38762,38764,38766,38768,38770,38771,38772,38774,38775,38776,38778,38779,38780,38781,38782,38783,38784,38785,38786,38787,38788,38789,38791,38792,38794,38795,38797,38798,38799,38804,38807,38808,38809,38810,38812,38813,38814,38816,38817,38818,38819,38820,38821,38822,38824,38826,38827,38828,38829,38830,38835,38838,38839,38841,38843,38845,38847,38849,38851,38852,38853,38854,38855,38857,38859,38860,38861,38862,38863,38864,38867,38868,38869,38870,38871,38872,38873,38876,38877,38878,38879,38881,38883,38885,38893,38896,38897,38899,38902,38904,38905,38906,38907,38909,38910,38911,38912,38913,38914,38915,38916,38917,38918,38919,38920,38922,38924,38925,38926,38927,38928,38929,38930,38931,38932,38934,38935,38936,38939,38940,38941,38942,38944,38945,38948,38950,38951,38952,38953,38955,38957,38959,38960,38962,38965,38967,38968,38969,38971,38977,38979,38980,38981,38982,38984,38985,38986,38988,38989,38990,38991,38992,38993,38994,38995,38999,39000,39001,39003,39004,39005,39006,39007,39008,39010,39011,39012,39013,39015,39017,39018,39019,39020,39023,39024,39025,39026,39027,39028,39080,39081,39084,39085,39086,39087,39089,39090,39091,39094,39096,39098,39099,39100,39101,39102,39103,39104,39105,39106,39108,39110,39113,39115,39116,39131,39135,39138,39139,39141,39143,39145,39146,39147,39149,39151,39154,39156,39158,39161,39162,39164,39165,39166,39168,39170,39171,39173,39175,39176,39177,39178,39180,39184,39185,39186,39187,39188,39189,39190,39191,39192,39194,39195,39198,39199,39201,39204,39205,39207,39208,39209,39210,39211,39212,39213,39214,39215,39216,39217,39218,39219,39221,39226,39228,39229,39230,39231,39233,39235,39237,39239,39240,39241,39243,39244,39246,39248,39249,39250,39251,39252,39253,39254,39255,39256,39257,39259,39260,39262,39263,39265,39318,39319,39320,39321,39324,39325,39326,39329,39331,39332,39333,39334,39335,39336,39339,39340,39341,39342,39343,39344,39345,39346,39347,39348,39349,39353,39354,39355,39357,39361,39362,39363,39367,39369,39371,39372,39373,39374,39375,39376,39377,39378,39379,39380,39381,39382,39383,39384,39385,39387,39388,39389,39391,39394,39395,39396,39397,39399,39401,39402,39404,39405,39406,39408,39409,39412,39414,39415,39416,39417,39418,39419,39420,39421,39422,39423,39425,39426,39427,39428,39429,39430,39431,39433,39434,39435,39437,39438,39439,39441,39444,39445,39446,39449,39450,39451,39452,39453,39454,39456,39458,39459,39460,39461,39463,39465,39466,39467,39468,39469,39470,39472,39473,39474,39476,39477,39478,39479,39480,39481,39482,39485,39486,39487,39488,39489,39490,39491,39492,39493,39494,39496,39497,39498,39500,39501,39502,39503,39504,39506,39507,39508,39509,39510,39511,39513,39514,39515,39518,39519,39520,39522,39524,39525,39526,39527,39528,39529,39530,39531,39592,39595,39597,39599,39600,39601,39603,39604,39607,39608,39609,39611,39612,39614,39615,39616,39617,39618,39622,39623,39626,39629,39631,39632,39633,39634,39635,39636,39637,39638,39640,39644,39647,39649,39651,39654,39655,39659,39660,39661,39662,39663,39665,39666,39667,39670,39671,39673,39674,39675,39676,39677,39678,39681,39683,39684,39685,39686,39688,39690,39691,39692,39693,39694,39696,39697,39698,39701,39702,39703,39704,39705,39706,39710,39711,39712,39714,39715,39716,39717,39719,39720,39721,39723,39726,39727,39729,39730,39731,39733,39735,39738,39739,39740,39742,39743,39745,39746,39747,39748,39749,39750,39752,39754,39755,39756,39757,39758,39759,39761,39762,39764,39765,39766,39768,39769,39770,39771,39775,39776,39777,39780,39782,39783,39784,39788,39791,39792,39793,39796,39797,39798,39799,39802,39803,39804,39805,39806,39808,39810,39811,39813,39814,39815,39816,39824,39825,39826,39827,39829,39834,39835,39838,39840,39841,39842,39844,39845,39846,39848,39850,39851,39853,39854,39855,39861,39862,39864,39865,39869,39871,39872,39873,39875,39876,39878,39879,39880,39881,39882,39891,39892,39893,39894,39895,39897,39898,39899,39900,39902,39904,39905,39906,39908,39909,39910,39911,39912,39914,39915,39916,39920,39927,39928,39933,39941,39942,39943,39944,39945,39947,39949,39950,39954,39955,39956,39959,39964,39965,39969,39971,39972,39973,39976,39977,39979,39980,39981,39985,39986,39987,39988,39990,39991,39993,39995,39996,39997,39998,39999,40000,40001,40004,40006,40008,40009,40010,40011,40012,40013,40014,40016,40018,40020,40021,40022,40023,40024,40025,40030,40031,40032,40034,40035,40038,40039,40040,40045,40046,40049,40051,40052,40053,40055,40056,40057,40058,40165,40166,40167,40169,40170,40173,40177,40178,40179,40180,40181,40182,40183,40185,40186,40187,40188,40189,40191,40192,40195,40196,40197,40198,40199,40200,40201,40208,40210,40212,40213,40215,40216,40217,40219,40221,40222,40223,40224,40226,40227,40229,40230,40232,40233,40237,40238,40239,40240,40241,40243,40244,40246,40247,40248,40251,40253,40254,40255,40256,40257,40258,40259,40261,40266,40267,40268,40271,40273,40274,40275,40276,40278,40279,40280,40281,40282,40283,40284,40285,40287,40288,40289,40295,40296,40297,40298,40299,40300,40303,40304,40305,40306,40307,40308,40309,40311,40312,40313,40315,40317,40319,40320,40321,40322,40324,40325,40326,40327,40328,40329,40330,40331,40332,40336,40338,40340,40342,40343,40344,40345,40346,40347,40348,40349,40350,40351,40352,40353,40354,40355,40356,40358,40359,40360,40361,40362,40364,40365,40367,40369,40370,40371,40372,40373,40374,40375,40376,40377,40378,40379,40380,40382,40383,40385,40386,40387,40389,40391,40392,40394,40395,40396,40397,40398,40399,40400,40401,40402,40403,40405,40406,40407,40408,40409,40410,40411,40412,40413,40414,40415,40417,40418,40419,40420,40421,40422,40424,40425,40427,40428,40429,40430,40431,40432,40434,40435,40436,40437,40438,40439,40440,40441,40442,40443,40445,40446,40447,40448,40449,40450,40451,40452,40453,40454,40455,40457,40459,40461,40463,40464,40465,40466,40467,40468,40469,40471,40473,40474,40475,40477,40478,40565,40569,40570,40572,40573,40575,40576,40578,40579,40582,40583,40584,40585,40586,40587,40588,40589,40590,40593,40594,40595,40596,40599,40601,40602,40603,40604,40605,40607,40608,40609,40612,40613,40615,40617,40621,40622,40624,40628,40629,40630,40631,40635,40636,40638,40640,40642,40643,40648,40652,40653,40654,40655,40656,40657,40659,40660,40661,40662,40664,40666,40667,40668,40669,40670,40671,40672,40676,40677,40678,40679,40680,40683,40685,40686,40687,40688,40690,40691,40692,40693,40694,40695,40697,40698,40699,40700,40701,40703,40704,40705,40710,40711,40713,40714,40718,40719,40720,40722,40723,40725,40726,40728,40729,40730,40731,40732,40734,40736,40738,40739,40740,40741,40744,40745,40746,40747,40748,40749,40750,40751,40752,40753,40754,40755,40756,40757,40758,40759,40760,40763,40765,40766,40768,40769,40770,40771,40774,40775,40776,40777,40778,40779,40780,40781,40782,40783,40786,40788,40789,40790,40791,40792,40793,40795,40796,40797,40798,40799,40800,40801,40803,40804,40805,40806,40807,40810,40811,40812,40814,40815,40816,40817,40818,40820,40821,40822,40823,40824,40825,40826,40827,40830,40845,40848,40849,40850,40852,40853,40856,40860,40864,40866,40868,40880,40881,64012,64013,65072,65073,65075,65076,65077,65078,65079,65080,65081,65082,65083,65084,65085,65086,65087,65088,65089,65090,65091,65092,65097,65098,65099,65100,65101,65102,65103,65104,65105,65106,65108,65109,65110,65111,65113,65114,65115,65116,65117,65118,65119,65120,65121,65122,65123,65124,65125,65126,65128,65129,65130,65131,65281,65282,65283,65284,65285,65286,65287,65288,65289,65290,65291,65292,65293,65294,65295,65296,65297,65298,65299,65300,65301,65302,65303,65304,65305,65306,65307,65308,65309,65310,65311,65312,65313,65314,65315,65316,65317,65318,65319,65320,65321,65322,65323,65324,65325,65326,65327,65328,65329,65330,65331,65332,65333,65334,65335,65336,65337,65338,65339,65340,65341,65343,65345,65346,65347,65348,65349,65350,65351,65352,65353,65354,65355,65356,65357,65358,65359,65360,65361,65362,65363,65364,65365,65366,65367,65368,65369,65370,65371,65372,65373,65374,65504,65505,65506,65507,65508,65509,65517,131210,131276,131352,131521,131588,131742,131744,131813,131911,131966,131968,132021,132041,132127,132231,132423,132494,132517,132531,132547,132575,132694,133411,133460,134031,134114,134143,134202,134263,134355,134450,134477,134478,134516,134773,134877,135348,135740,135759,135941,135994,136054,136078,136214,136276,136343,136598,136723,136766,136850,136888,136897,136917,136935,136936,136973,136998,137072,137073,137138,137180,137211,137212,137256,137257,137335,137374,137431,137489,137560,137603,137608,137691,137780,137803,138052,138405,138745,138780,138807,138813,138916,139114,139337,139516,139562,139635,140240,140401,140427,140433,140525,140563,140592,140628,140828,140904,140952,141236,141261,141407,141485,141505,141559,141696,142755,142861,142902,143428,143435,143486,143780,143781,143816,143863,143864,143865,143921,143923,144010,144044,144308,144358,144384,144495,144845,144896,144956,144985,145155,145197,145346,146139,146266,146585,146632,146633,146876,146877,146991,146992,147196,147328,147330,147380,147383,147397,147435,147436,147440,147513,147514,147517,147589,147596,147601,147775,147776,147780,147797,147798,147799,147807,147876,147877,147917,147938,148087,148088,148161,148323,148325,148380,148533,148570,148668,149539,149745,149760,149761,149772,149811,149824,149826,149876,149887,149890,149901,149903,149924,149929,149943,149983,150011,150034,150049,150050,150051,150052,150054,150055,150056,150137,150156,150182,150193,150517,150803,151426,151709,151933,152038,152096,152144,152339,153056,153093,153373,153458,153687,154060,155779,156125,156661,157832,157966,157969,158254,158483,158504,158545,158547,158643,158846,158849,159013,159017,159018,159136,159137,159142,159211,159301,159342,159346,159442,159604,159647,159649,159992,160038,161287,161292,162318,162387,162436,162617,164994,165364,165547,165892,165931,166244,166248,166314,166332,166372,166441,166454,166474,166555,166622,166623,166689,166699,166758,166799,166813,166888,166915,167596,167602,167641,168173,168205,168304,168348,170333,172368]],
            'extended characters' => [["8859","8857","885D","885B","8866","8861","885F","886A","8868","886F","886D","88A7","8873","8871","8877","8875","887B","8879","88A2","8856","8867","885A","886C","885C","886E","8870","885E","8874","8878","8858","8869","8872","8860","8876","887A","887C","887D","887E","88A1","886B","88A8","8863","88A4","8865","88A6","88A9","88AA","8840","8841","8842","8843","8844","8846","8849","884A","884D","884F","8850","8851","8852","8854","8855","9277","96DF","8CF4","89D5","93CD","9BDF","89DA","8F59","89DB","8F5D","89DC","96F7","8ADA","8BDC","97DB","9E53","9DAA","87BE","9BEA","8A6E","8BC8","89E8","89EA","8C4B","89ED","94DD","89EE","9EB4","8AD3","92DB","94DB","89F9","89FB","9EFC","89FC","89BF","89FE","89E6","9D46","9DEE","A07E","A068","98E9","8B68","8DFD","8BBE","9FD9","8AEB","9FD7","8B6A","9C5C","8BB1","8770","9DF3","A0D0","92E9","9AEC","8FAB","8E45","9C6F","8D5C","9EDE","89EF","96E9","9EBB","94DE","9EB8","97BA","95D6","9CBB","97DA","8F45","9158","9856","9B4D","935B","95C7","97E7","9359","91F5","97B8","92FA","9357","8BA6","97B0","9CA1","91F2","91F9","8FF1","9745","9853","9251","9DAD","9BC2","9A7B","8B60","934B","9ABD","91B7","8D4B","95B4","9EF0","8D64","9269","8D67","8D68","93EB","877A","9166","93DD","8D52","8BCC","8D6D","8D6E","96A8","8D6F","8D70","8CF3","9060","8D74","97C3","8AD0","9274","9BBE","9CC8","9CBA","8D78","9EB9","955A","91B4","8A48","8D7D","8A7D","8AC2","8DA1","8AD1","8B47","93A4","9EDA","8A51","8DA6","9EC5","A078","94B5","8A6B","8DAB","8DAD","93C1","906F","8DB0","87A2","947E","90FA","9479","8DB2","997B","8DB4","8DB7","91B3","8DBB","8DBA","8DBC","9044","874B","93E4","93E0","8DC3","9BB8","93E9","93F6","8DC5","8DCA","8DCC","93B5","9CF8","9252","A0E8","9CA5","8C56","8DD6","97C0","A0DE","97D2","8DDB","8CEA","8EAF","91B5","8DEB","97C6","90FC","96D6","97C5","8DEF","97D7","8DF0","96A6","8CDF","8DF3","9449","8DF5","9872","8E6B","8F50","9DCC","8C44","996E","94A1","8F63","A0DA","9253","9DB5","9879","876A","9D5D","8D63","9669","9F70","8AC7","89D7","9EDD","98BC","95B0","9464","936F","94B9","95EC","91EE","98C3","95F6","8FFD","98C5","9766","97DD","8CAA","92D2","9761","98CB","95F0","975D","91E3","877E","98CC","9469","98CD","98CE","95FC","94A3","9662","9463","8D47","98D0","98D1","9475","9472","98D6","8AF0","98D9","8D5A","98DB","98DD","98A8","8A6D","8AFB","8AAE","8C5D","98E4","98E6","98E8","8A4D","9257","95DF","A0AC","98EB","98EC","8CC3","98F4","87D9","8AB8","9EE7","94BC","9CC6","8D4A","9E7E","8D44","98FE","9940","94C9","87C6","94D3","9946","90C0","94D1","8D4E","9573","87CE","93C2","9948","994B","8E55","994E","8EFE","8D5F","8E59","94EC","94EF","8C60","8F74","9955","9544","8CCB","9956","9959","995B","8CC4","90B7","9743","95CD","97C9","87AA","8EB9","95C6","9967","8CE3","8AB9","8DFC","8A76","9D51","9973","8740","9D4F","997A","9564","99A1","99A5","99A7","8EED","99AD","946E","8F70","99B3","A053","8D5E","965C","8CE0","97FE","92BD","8D5D","97FD","87DB","8F64","9562","97CD","9E64","924C","8EC9","99BC","9DA5","8F54","8F7C","8D55","8EA2","8F7A","97AE","96C8","8CE4","99C3","90D6","9CBE","8F76","9470","8CEF","8EC7","8D54","A0F9","8FA9","8D51","99C7","8744","90D7","8743","8747","8758","9EDF","8D59","8742","99CE","8FBA","8FEB","99CF","8FC2","92C9","97DC","875D","87CC","8D45","95B3","9C79","95B2","8D4C","8FDB","9BE3","874C","874D","9E7A","8757","9BEE","99DE","8A52","99E1","8A67","8BB5","8AAC","99E9","97DE","95D1","99F5","9BA9","9EA4","9D49","95DB","89C5","99F8","9664","9055","96D4","87C4","87AE","977C","964D","97E1","9A48","9A49","90AA","9A50","9347","8ED8","90C9","9A55","90BC","9A58","8BB8","90D5","9641","9A5A","9A5C","97C2","875C","8ABB","9BAA","90F5","9A60","9145","8C58","9A63","8C49","8BB6","966B","9A6E","914F","9746","A0E6","92D7","9675","93D4","91BB","9679","9A70","9678","91CD","9C4A","A06F","A06A","915F","8741","9FA5","89BA","874F","874E","8755","9ECD","9A79","8CF2","8D57","9DCE","8CD2","8759","9D73","96B9","96BC","9CD1","89B7","9EEE","8749","875B","9EC9","91AE","8D58","8746","8D56","9D78","9D7B","9EB3","9EB2","9DD6","994F","89CE","8BC0","9FC4","8BD4","8C72","8BF9","8946","8BC6","9C57","9AFB","89D0","89CF","89D1","89E2","927E","9DBA","8C6F","8BC7","926B","89D2","9FCF","9DA9","89D3","99E2","9267","92A4","8C73","894E","894F","9278","91B6","89D4","9FD2","92A7","95A2","926E","96EA","926F","92A3","8950","9866","8CF8","9C53","89D6","98B2","92AB","96DE","92AC","8C70","9F6E","8EF2","9F6C","89D8","92A8","9163","8C40","9F73","92AD","9BE9","92A9","92AA","89D9","9FA8","8C71","92A1","90E3","A0A6","94AB","9FCB","97C4","92AE","92A2","9268","8951","92AF","92B0","92B1","92B2","8952","945A","89DD","9E52","8953","9E55","92BA","8C5B","9A68","92BB","9BB4","89DF","8954","89E0","9F4F","89E1","9FCD","A0E7","89A6","9EFA","87BC","92C4","9F6F","8BB0","9FAC","89E3","9BD3","89E4","9FD5","8955","92C5","8956","9EDC","9F71","92C7","9A4C","8C68","89E5","9F7D","A0A9","8957","8958","8BE3","8B61","9AF1","9EB7","9EBA","9CE0","89E7","A07A","89E9","89EB","90C8","92DA","8959","9CF5","895A","9FA2","8FAD","96EF","9DEC","9DCA","89EC","9DE2","8C75","9EC0","87C5","9E56","9F79","9AC7","98A1","89F0","9E47","9DF7","9FD3","9ACA","89F1","8E5A","89F2","89F3","925D","8B51","92E0","89F4","9FD4","8A79","89F5","97A7","93BA","9E58","89F6","9E57","89F7","8A41","89F8","89FA","9E4E","94DC","95DA","9DF8","9F6A","8AB7","8A46","9148","92DE","8B53","9DF6","9BDA","9D7E","89FD","99E4","9E43","9DE9","8F52","9DF5","9DF0","99E7","8BBD","9DEF","9FB7","9DD0","9FEB","8DA9","9DCF","98E1","9DE5","9DC8","9DEB","9AA2","8AD6","9A5F","9EF5","8FB7","9AD2","9E6A","9EE8","8BBF","91C2","9D62","9260","925E","91C1","8AC5","97A3","8B6C","8D7E","9C54","9DBD","9CC5","895B","8765","98C7","9CEE","92E2","94A7","8CCC","9BD4","99E5","9AC2","91FB","A073","9F72","9FCC","98A5","92E8","9BBC","96F3","92E7","8B7D","9BF4","9EF7","9EC1","87C3","996F","96F1","8E41","954A","97E6","96F5","92E6","9F42","99A9","97E5","87C8","967D","99A2","9ABB","9A65","944E","99DF","98E3","9254","967B","8AAF","8C77","87B0","8BAF","9EBD","9EE6","8EE1","9B7D","9C7E","92EA","8C78","895C","98F0","96F2","8BC1","895D","89DE","895E","8768","895F","8960","9BCD","9DD3","984C","9752","95C3","9BB6","9AB9","97B3","9F74","92F1","8CFA","97DF","9877","9854","95C5","9D55","957E","9742","94E6","92F5","8CC5","92FD","9C51","94E9","985C","92F0","944C","916B","8B78","94E2","984F","9CD0","9271","9365","985B","9850","97BC","92F3","9340","984D","9572","92EB","97B7","876F","90A7","9741","92F4","8772","9577","9EE2","8F78","9672","9EB5","964B","8CAC","A0FA","96FC","9575","90DA","9367","90DF","9354","8961","8BB4","9DC0","8E48","9E67","8CD9","8962","8963","8773","9F6B","876D","9CBC","8B5D","934C","9AE2","8BC9","9FC9","9F44","98ED","8CE9","8DF2","8964","934D","A0F2","9868","9F58","8CE6","8D73","8C48","8774","8DA8","9C75","9878","8D60","8D61","8D62","A0A1","9C40","98AD","9EEA","8CEC","8CD4","9CEB","9F51","8D65","9CF1","8D66","9654","9FCE","9AE4","9F75","8D69","934F","934E","8965","8C7A","8C7B","8D6A","9353","9DFB","9059","9361","9362","8D6B","8CFE","95B8","8BCA","987A","8D6C","9B70","A051","8C7C","8BCB","936E","8966","9EA9","937A","A0E0","936B","A0DC","9468","8D71","9BEC","99BA","9AD0","9A61","A0E5","A05B","96AC","9740","9EF1","8C4D","9F7E","8D72","96A9","A06E","A074","A071","9C50","9379","9378","A0DD","8D75","8D76","9374","8D77","90C3","A079","8D79","8BFC","A076","8BCD","9F5A","9FF4","9FBA","8D7A","9E45","93B0","A075","87DD","9B46","A077","9DC4","8D7B","8D7C","9ED6","93AC","9F5B","93A9","A07C","8AC1","9FB4","9E4C","8FC5","93AD","9DC3","8DA2","9D4A","8DA3","9E4B","9E4D","8DA4","8AFD","93B2","8DA5","93A1","8AC6","8A5B","894D","8A78","93AB","8DA7","9F45","8A56","8EE6","8AA4","8943","93F3","9EA2","9DC7","8BCE","93B3","8DAC","8967","8C7E","9CF3","95BB","8DAE","93DB","93D5","9B71","8764","8DAF","87B5","93D8","93D3","8E76","93D1","8DB1","9859","9CBF","9B72","93BE","8CDB","9DF1","A0BB","9B7E","8DB3","8C52","9AE8","8EDC","9CF9","98E7","8CCA","8775","87BA","93E5","9A59","8DB5","8F7D","9547","9250","8968","8DB6","A07D","98FC","8969","9256","93E8","9CE3","9640","8DB8","9B4A","8FB9","896A","8DB9","917E","93F4","93E7","97EF","96A5","8DBD","9BA1","8CA2","9AB7","8EFC","9FA1","8DBE","89A4","9AD9","8DC0","97F0","93B4","9FA7","8DC2","99B6","8DC1","8E46","A0D1","9FCA","92CF","9CF4","8DC4","9B4C","9CDE","986C","97F9","9558","87B6","985E","94CD","93EE","8CA3","93F5","93EF","8EEA","8F5B","8C5E","8DC6","8DC8","8DC7","93F7","8DC9","9670","8DCB","8F65","8DCD","9DA8","94F9","8DCE","93EA","93F0","9FB6","8DCF","9763","8DD0","93F1","9FDB","93F8","8BF7","8BCF","8DD1","8DD2","8DD3","9FE7","90BD","9FD0","8BD0","9CAE","8BD1","8ADB","9EFD","95CE","8DD4","8EE3","9076","98C6","8DD5","97D1","9EB6","A042","9873","9FFC","8CA5","8DD7","92FB","8DD8","944F","8DD9","896B","97CE","9447","92B7","8DDA","9C5A","8DDC","9444","8DDD","A0D6","8C41","97D5","944A","944D","97CB","8DDE","8DDF","8DE0","8CDD","92B3","8DE1","95D3","89C1","9CB7","8DE3","8DE5","8947","8DE4","8DE7","8DE8","9445","97D6","9844","8DE9","8DEA","9DA7","95D2","8DED","9CDC","9BF6","95CF","8DEE","96EC","96EB","90B6","98AB","96ED","8DF4","8C67","8DF6","8DF7","8FFA","97D0","8BD2","87DE","8DF8","90D9","8C47","8DF9","8DFA","90A6","9970","91EB","9770","986F","98F2","9AFC","896C","995E","95BD","91E6","9454","99B8","97E9","9346","9863","95BC","9870","96F6","8EA9","9451","8E43","8B5A","9BF5","8CEE","A0DF","977E","9BD5","9AC3","97C8","A0DB","91D0","9FE4","8FDD","91E9","98E0","92CA","9857","8C51","9B49","9D76","9EAF","9CCC","8DF1","8E53","9C62","96F9","98BF","9E49","8CA7","9B76","9BCA","92DC","91CC","91E2","875F","8BD3","9455","8DBF","9E78","9456","9D61","9457","9966","8BD5","A069","98B4","A049","A04C","9E65","98B5","9975","A065","98B7","98B8","98BA","98BB","9FBC","A04A","9EC7","8CA9","98AE","92D6","91D4","8C53","87BF","94C5","98C1","975C","9773","9764","964E","9765","8C5A","89A1","95FA","92D4","98C8","90EF","98C9","98CA","946D","94B7","946B","92FC","95EB","976E","87B8","92D5","8778","947A","95FB","92D1","945D","9344","8EA6","92D3","94B8","8779","975E","8CAD","87C1","946A","93E3","98CF","A0D9","A0BF","A04D","A0B8","A0CE","A0B7","89C3","9DF4","896D","9C7B","98D2","9FA9","97D9","A0C4","9476","9978","98D3","98D4","9FB9","9471","98D5","9E5C","A044","98D7","98D8","9EEF","9FFE","9DDD","9EE1","98DA","9DDF","9EEB","9E59","A05C","9477","98DC","98DE","9FC2","8C6B","98C4","94B0","94B1","A0C1","A0CD","98E5","91E4","8FC7","94AE","8A4F","94B2","8FD4","98EA","9DE0","98EE","95C4","98EF","9B78","8C6E","A0AE","9D4C","98F1","98F3","94C1","98F5","96E2","9450","96A2","98F6","96E5","98F7","A046","96E3","98F8","9EE4","94C3","94C2","96E4","89AC","96DB","94C4","9FFB","8C59","93C9","94E8","90C5","A0A8","98FD","98FB","8EBF","8BD8","8F68","94C6","9DEA","9CDA","9C72","89C9","9941","9942","94CA","91D7","94CC","97A8","8CDE","87B3","96D1","9CBD","94D5","94D0","9944","8C63","87BB","A0B3","87B4","94CF","9FFA","91E5","9C6A","8E49","8E4C","87C9","8E4D","9A73","9947","8CB1","8E50","8E4F","9949","8E51","8E52","9AB2","89A5","994C","9FF8","8E56","994D","91CA","8E57","94E1","9047","8FD8","8E58","87A3","94EB","8E5C","9553","9FE5","9F56","954F","8E5E","996A","9C64","9CD9","9D5A","8E5D","9950","9951","8E62","9952","8E68","8E61","9F59","87D0","8BB3","8E69","87B9","9F5D","8E66","8CB2","8E6E","9F64","8E6F","9953","9954","8E70","9F61","8E72","A06B","9F40","94ED","94EE","9FBD","8E7B","9957","94F7","9F5F","8E73","9F62","94F6","9958","8E75","9072","94F8","995A","A0B0","8CB3","8E79","8E78","94F3","8E7E","98AF","A0B2","8E7A","995C","8E7C","8E7D","8BD9","89A2","9ED7","A0B6","9E42","8EA4","8EA7","8C76","8767","9542","987D","9755","8EA8","8EAA","89A3","9960","9962","8EAB","94FC","9961","94FA","8EAE","8EB2","8EB0","9963","97AA","94FB","8EB4","8EBB","8CDC","9876","8EA1","8CB4","8EB7","9DA6","9B7B","9EB0","8EB8","9D70","896E","896F","8970","8971","8972","8973","8974","8975","8EBC","8EBD","8EBE","9DD1","94FD","8BD7","8BDA","A0E2","9FE9","8AE7","8EC2","8EC4","9964","9965","954E","98B3","8ECB","8BDF","8ECD","8ECE","8ECF","9968","9969","8ED0","996B","8ED1","996C","8ED4","8ED5","996D","A0BE","8ED6","A0BC","A0B5","A0B4","8BE0","89B5","8EDD","9E5D","9971","89AE","9DE8","9565","9972","8B5C","89B1","A0C0","8EDF","9566","9974","9976","9977","9979","9DDA","8EE0","935C","9DE6","8B5F","9563","9567","9DE3","997C","997D","997E","8CFB","8B5B","99A3","99A4","99A6","99A8","8ABE","9E61","99AA","A0C8","99AB","98C2","8EE8","A0BA","8EEE","9EBF","89C2","99AC","956B","956C","99AF","994A","8976","8F48","99AE","8EFB","8CD0","8B52","99B0","8977","8F41","99B1","8F49","9DE4","8CB5","9B54","99B2","9E68","8F4A","8F42","8F51","9846","99B4","8EF5","9CCD","8978","8F53","8F6F","8E63","8F56","9FC6","8F57","9C77","8F58","9848","99B7","8F6E","9665","9DE7","9E62","96CC","8E67","987E","97FC","98F9","8F66","956E","9245","8F60","9ED1","99B9","8F62","974C","91C7","955F","87AB","99BB","8E6D","8F71","94CB","95B1","8F69","9AF2","96C3","99BD","A0CF","8F6D","99BE","8EF4","8F72","95E4","99BF","9242","87D7","99C0","8F77","99C1","8F40","99C2","8F5C","8CBD","99C4","99C5","8F7B","8776","8CB6","8FA3","8CCE","99C6","96CD","96C7","8FA5","8C61","9570","87AF","9368","8F7E","8FAA","A050","90D3","9556","8FB8","99C8","8FAF","99C9","9579","9F49","99CA","99CB","9DD5","8FB0","9E5F","99CD","A0C9","9ADB","A0C6","8FB4","A0D7","A0C7","A043","8FB5","8FB2","A061","9E5E","8FB6","9FE8","9CB2","957C","9FC7","8FBB","8FBC","8FEC","8FC0","936A","8BE4","9C7C","95A1","95A3","8C45","8CB8","8FC1","87B7","A052","99D0","8FC3","8FC4","95A4","8FC6","9E60","95A5","9CB3","99D1","99D2","9CC2","99D3","95A7","95A9","95A6","9C5D","98E2","8FC9","A0C2","8FCA","99D4","A0B9","9B58","8FCB","8FCD","8FCC","8FCE","8BE5","8979","8FD0","95B6","99D6","95E5","99D7","95B5","A0CA","9FFD","A058","8FD6","99D8","8FD3","8FE5","8FE9","99D9","8CF7","927C","9C45","8CE8","8FDE","8FDF","A04B","8FE2","A0CC","8FE3","8FE4","9BC4","9BFC","964C","9AF6","8CAE","87CB","8FE7","8FE8","8BE7","897A","897B","99DA","8FED","95C0","A0CB","9E48","99DB","8FF3","8FF9","95C1","A04E","99DC","A064","8FF7","89B0","A048","8FFB","8FF6","9DDC","99DD","8BE8","8FFE","92C1","9FD6","A0D2","9040","8AC4","99E0","9FF0","9FF3","9DBF","9FF6","95C8","9E5A","99E3","8A4A","9FF1","8AA7","99E6","9FF7","9FED","8A5C","9DAE","95C9","9048","99E8","9049","8CBA","90B1","904A","99EA","9BD1","99EB","99EC","99ED","99EE","9D57","904C","904D","95CB","9C42","97E2","95CC","9F78","897C","897D","897E","995D","9B5A","9050","8C4F","9054","9AA8","99EF","9DA3","9DA1","9943","9945","9D7D","99F0","99F1","99F2","8CBC","9D60","A0A3","905B","9EDB","9D79","99F3","9062","87BD","9F55","9BF9","9065","96E0","98BE","95D9","9068","906C","95D8","906A","906D","9C68","9FB2","9FAE","9FB0","89AD","906E","9E71","9E4A","9FDC","89AB","9FB8","9070","8B63","95DC","9071","9BDE","8949","965B","8C50","94A6","8FD5","9E73","9075","99F7","87B2","8CBF","99F9","9663","95B9","94D4","9077","90AB","9D4D","99FA","92E3","97BB","9078","99FB","97E0","96DC","9CA8","9772","9440","92F2","99FD","99FC","907A","964A","96D8","99FE","904B","9A40","975B","9A41","91DD","93FC","9A42","9A43","9659","9A44","9051","94BF","90A2","9CAB","9776","94A8","9A45","9DE1","96D9","9774","92E5","9645","91DA","90A3","92C8","90AF","97BF","914C","967A","91DE","9A46","9779","946C","9858","9266","93FB","9A47","9749","9748","934A","9CE2","9264","91DF","96D7","9343","91DB","8C6A","97AF","95DD","9348","9A4B","9A4D","91BC","90E2","90B4","95E1","9A4E","87AD","9A4F","96DD","9A51","96A7","90B0","9C4E","9443","8EBA","9A52","8BE9","9CAF","8BFD","9ABC","9AB8","9AAE","9AA7","9A53","9D74","8BEA","8BEB","90B2","95E9","95E8","95E6","90B5","9A54","90B3","95E7","8B50","8BEC","9A56","8BFB","9A57","A0AA","9FA6","99CC","9C59","99B5","90BE","9FAF","95F2","90BF","90C1","90C4","90C7","92E4","9F52","90DB","A066","90D2","876B","90D4","9A5B","95FD","87B1","8BC4","8C66","90DE","90DC","9644","90E1","9E46","9651","90E6","9650","90E7","90E8","9A5D","9F7A","9B5C","9F7C","90E9","90EA","9A5E","9F76","90EB","90EC","8BEE","90EE","91C6","90F2","90F1","8A74","9657","9CEF","9FDF","90F7","90F6","9B5E","90F8","8CFC","90F9","8CC9","8BEF","9FE0","9142","9A62","9569","9144","9143","9141","8BF0","9660","8BF1","99F6","9149","914A","914B","9A64","8ABF","9A66","9A67","9A69","9A6A","9652","914D","9666","9F7B","9A6B","A06C","9667","9A6C","9A6D","8BF2","966A","966C","91C4","9677","99F4","9A6F","9BC6","9FAB","8CBE","8EC1","9555","9152","9153","9155","955D","9671","9C6D","9673","9154","9A71","9156","966D","9557","89C6","89C7","8A6A","8B57","9FE1","9B5F","A05D","915B","915C","915E","9F5C","9F57","9F65","9A72","9160","9F5E","9161","9F60","9164","9F41","9169","9168","9A74","96B2","9A75","9EE9","8BBA","916D","A060","9FDE","9FC3","96B5","A067","96B3","9A76","95D5","9ECA","9A77","9A78","9170","916F","9FA3","9171","96B1","9F63","9F67","8BB9","9A7A","8B56","9ADA","96B0","9A7E","9DDE","96AD","96AE","9EA1","9E50","96AF","8BF4","9FA4","96BD","96F4","96B8","91A7","A05E","9A7D","8948","9EB1","9DDB","95BF","8A73","9EFE","917A","917B","9AA3","96C2","9F77","9AA4","9AA5","91A1","89B8","9173","9C6B","9AA6","89BD","89B9","917D","96BB","9FF2","8BF5","9AA9","9F54","9FE3","9EED","91AA","91AB","A070","9F6D","91AC","91AD","A0FD","9FE2","91AF","9E41","9AAA","91B0","9AAB","9AAC","9A4A","91B2","8BF6","9AAD","89B6","9AAF","9AB0","9AB1","9AA1","91B9","91BA","91BF","91BE","A041","8BB7","91C0","9AB3","91C3","A0FC","9FEE","9F69","91C8","91C9","8DE6","91CB","87AC","87A4","89C8","8DAA","9FDD","8C43","8C6D","8C74","8CB7","8CB9","8CBB","8CC0","8CD7","8CD8","8CDA","8CED","8D48","87C2","87D2","87D6","87DA","87DF","8BF8"], [192,193,200,201,202,210,211,224,225,232,233,234,236,237,242,243,249,250,252,256,257,274,275,282,283,299,332,333,363,461,462,464,465,466,468,470,472,474,476,593,609,7870,7871,7872,7873,9178,9179,12736,12737,12738,12739,12740,12741,12742,12743,12744,12745,12746,12747,12748,12749,12750,13365,13376,13386,13388,13412,13427,13437,13438,13459,13462,13477,13487,13500,13505,13512,13535,13540,13542,13563,13574,13630,13649,13651,13657,13677,13680,13682,13687,13688,13700,13719,13720,13733,13741,13759,13761,13765,13767,13770,13774,13778,13782,13787,13789,13809,13810,13811,13819,13822,13833,13848,13850,13861,13869,13877,13886,13895,13896,13902,13919,13921,13946,13953,13978,13989,13994,13996,14000,14005,14009,14012,14017,14020,14023,14024,14035,14036,14038,14045,14049,14050,14069,14081,14083,14090,14108,14114,14115,14117,14124,14125,14131,14138,14178,14191,14231,14240,14265,14270,14294,14322,14331,14351,14361,14368,14392,14435,14453,14531,14545,14548,14586,14600,14612,14631,14655,14669,14712,14720,14729,14730,14738,14745,14747,14753,14756,14776,14812,14818,14821,14828,14840,14843,14846,14851,14854,14872,14889,14890,14900,14923,14930,14940,14942,14951,14999,15037,15072,15088,15090,15093,15099,15118,15129,15138,15161,15170,15192,15200,15217,15218,15227,15228,15253,15254,15257,15292,15294,15300,15319,15325,15340,15346,15348,15377,15381,15384,15444,15499,15563,15565,15569,15574,15599,15634,15635,15645,15686,15692,15697,15714,15721,15722,15727,15733,15741,15752,15754,15759,15761,15781,15789,15807,15814,15817,15820,15821,15827,15835,15847,15851,15859,15860,15863,15868,15869,15878,15939,15944,15988,16041,16045,16049,16056,16063,16066,16071,16074,16076,16080,16081,16086,16090,16091,16094,16097,16098,16103,16105,16107,16108,16112,16115,16116,16122,16124,16127,16128,16134,16135,16142,16211,16216,16227,16252,16275,16320,16328,16343,16348,16357,16365,16377,16378,16393,16413,16441,16453,16467,16471,16482,16485,16490,16495,16497,16552,16564,16571,16575,16584,16607,16632,16634,16642,16643,16649,16654,16689,16690,16743,16748,16750,16764,16767,16769,16784,16818,16836,16842,16847,16859,16877,16879,16889,16913,16931,16960,16992,17002,17014,17018,17036,17044,17058,17081,17084,17140,17147,17162,17195,17262,17303,17306,17338,17345,17369,17375,17389,17392,17394,17409,17410,17427,17445,17453,17530,17551,17567,17568,17584,17591,17597,17600,17603,17614,17629,17630,17631,17633,17636,17642,17643,17644,17652,17667,17668,17673,17675,17686,17691,17693,17703,17710,17715,17718,17723,17725,17727,17731,17745,17752,17756,17761,17762,17770,17773,17783,17784,17797,17830,17843,17882,17897,17898,17923,17926,17935,17941,17943,18011,18042,18048,18081,18094,18107,18127,18128,18165,18167,18195,18200,18230,18244,18254,18255,18300,18328,18389,18413,18420,18432,18443,18487,18545,18587,18605,18640,18675,18682,18694,18705,18718,18725,18730,18733,18735,18736,18741,18748,18750,18757,18769,18789,18794,18802,18825,18849,18855,18911,18917,18919,18959,18973,18980,18997,19094,19108,19124,19128,19153,19172,19199,19216,19225,19232,19244,19311,19312,19314,19323,19326,19342,19344,19347,19350,19351,19357,19389,19390,19392,19460,19463,19470,19506,19515,19518,19520,19527,19543,19547,19565,19575,19579,19581,19585,19589,19620,19630,19632,19639,19661,19681,19682,19719,19721,19764,19830,19831,19849,19857,19868,19972,19994,19996,20001,20004,20008,20010,20012,20015,20023,20029,20059,20074,20088,20096,20097,20103,20105,20120,20122,20131,20155,20156,20159,20174,20202,20203,20206,20216,20227,20247,20250,20264,20265,20274,20279,20281,20290,20293,20299,20338,20362,20386,20392,20400,20404,20411,20413,20424,20428,20452,20453,20464,20466,20473,20477,20483,20526,20532,20539,20566,20568,20582,20588,20609,20616,20646,20668,20685,20688,20697,20703,20717,20724,20732,20749,20779,20822,20825,20827,20829,20830,20852,20857,20870,20890,20892,20903,20904,20910,20916,20931,20946,20955,20964,20988,20990,20997,21023,21024,21031,21044,21071,21081,21088,21096,21107,21113,21135,21136,21146,21153,21156,21160,21196,21201,21217,21249,21284,21287,21292,21298,21307,21326,21341,21343,21348,21351,21373,21410,21417,21418,21424,21442,21452,21456,21457,21458,21464,21466,21495,21524,21526,21530,21539,21554,21556,21579,21580,21581,21609,21610,21637,21651,21652,21655,21660,21662,21667,21682,21684,21689,21707,21708,21709,21712,21722,21731,21743,21762,21773,21779,21784,21790,21795,21797,21800,21823,21865,21867,21873,21874,21875,21881,21904,21936,21940,21945,21946,21948,21953,21975,21976,21982,21994,21996,22000,22001,22005,22011,22021,22033,22046,22048,22049,22050,22051,22053,22061,22083,22096,22098,22100,22109,22113,22138,22139,22140,22153,22154,22155,22162,22174,22175,22177,22180,22191,22193,22201,22207,22230,22242,22267,22269,22293,22301,22322,22333,22335,22356,22359,22363,22367,22375,22394,22398,22399,22408,22410,22413,22428,22433,22439,22442,22451,22452,22459,22462,22468,22472,22487,22493,22494,22511,22546,22552,22562,22586,22592,22596,22599,22623,22633,22636,22642,22643,22674,22678,22681,22682,22695,22704,22709,22710,22731,22736,22768,22770,22771,22786,22788,22789,22791,22813,22834,22836,22885,22901,22921,22932,22938,22943,22956,22960,22967,22968,22975,22980,23019,23023,23024,23032,23042,23051,23053,23058,23066,23073,23079,23082,23083,23084,23101,23109,23124,23129,23137,23143,23144,23147,23150,23153,23161,23166,23169,23170,23174,23193,23201,23204,23235,23246,23247,23249,23268,23280,23282,23294,23309,23313,23317,23327,23336,23339,23361,23364,23370,23375,23398,23400,23405,23412,23440,23446,23454,23455,23479,23491,23515,23568,23580,23582,23584,23587,23626,23635,23646,23672,23705,23706,23710,23745,23746,23761,23765,23781,23797,23804,23829,23852,23855,23870,23880,23894,23895,23899,23920,23924,23928,23931,23941,23950,23972,23979,23993,24001,24053,24075,24082,24130,24136,24158,24198,24210,24217,24253,24269,24272,24312,24332,24334,24357,24379,24397,24401,24412,24451,24497,24506,24514,24516,24539,24635,24674,24693,24695,24702,24740,24743,24791,24798,24803,24807,24808,24809,24829,24839,24844,24857,24866,24880,24893,24912,24921,24943,24957,24981,24984,24985,24988,25015,25017,25024,25039,25050,25058,25132,25143,25145,25164,25192,25218,25221,25232,25245,25252,25254,25281,25283,25294,25296,25317,25368,25390,25393,25397,25399,25444,25465,25471,25529,25537,25553,25566,25570,25574,25595,25596,25598,25607,25650,25659,25690,25713,25724,25741,25745,25780,25782,25792,25811,25821,25831,25834,25866,25873,25887,25904,25909,25963,25990,25993,25995,26046,26068,26111,26142,26160,26184,26189,26195,26208,26211,26219,26237,26254,26258,26266,26294,26303,26318,26343,26353,26380,26382,26390,26393,26398,26405,26422,26465,26475,26478,26498,26511,26528,26532,26559,26582,26617,26622,26624,26626,26627,26628,26640,26654,26678,26695,26698,26709,26710,26725,26756,26760,26776,26806,26809,26821,26889,26904,26905,26906,26907,26924,26947,26950,26965,26980,26983,26994,27008,27013,27018,27039,27042,27058,27072,27089,27093,27094,27113,27139,27148,27162,27164,27177,27179,27181,27187,27203,27212,27218,27219,27223,27235,27237,27249,27252,27258,27266,27279,27289,27303,27307,27313,27314,27317,27326,27337,27338,27348,27352,27382,27397,27474,27479,27503,27514,27521,27585,27626,27642,27676,27697,27701,27705,27706,27709,27722,27736,27765,27775,27807,27810,27818,27823,27826,27854,27871,27882,27903,27906,27909,27910,27940,27942,27982,27991,28007,28018,28033,28047,28068,28069,28081,28089,28158,28162,28164,28170,28175,28181,28184,28201,28247,28278,28294,28299,28314,28344,28347,28378,28381,28386,28392,28393,28427,28428,28439,28452,28468,28502,28537,28545,28597,28598,28603,28606,28633,28634,28662,28675,28702,28716,28749,28752,28756,28764,28775,28780,28782,28783,28789,28791,28799,28809,28815,28832,28835,28837,28838,28857,28859,28860,28864,28868,28880,28913,28917,28926,28933,28957,28969,28979,28980,28981,28987,28990,28992,29009,29015,29026,29035,29041,29043,29045,29046,29047,29050,29052,29054,29064,29068,29070,29073,29080,29090,29091,29101,29111,29137,29151,29163,29173,29174,29184,29185,29193,29199,29206,29207,29221,29227,29264,29276,29286,29319,29327,29332,29343,29357,29362,29389,29410,29442,29444,29456,29482,29484,29496,29497,29505,29512,29519,29553,29556,29580,29583,29592,29598,29626,29636,29648,29665,29666,29670,29683,29689,29691,29698,29713,29714,29716,29721,29724,29726,29727,29751,29752,29756,29763,29765,29768,29772,29782,29793,29800,29803,29817,29818,29836,29837,29849,29851,29853,29876,29881,29894,29900,29904,29907,29927,29936,29937,29938,29944,29957,29977,30004,30005,30011,30018,30022,30033,30035,30037,30048,30061,30066,30074,30083,30093,30129,30147,30152,30172,30201,30215,30267,30272,30285,30286,30292,30310,30311,30323,30324,30330,30422,30425,30476,30478,30479,30500,30507,30531,30578,30583,30584,30586,30587,30611,30616,30649,30654,30659,30667,30679,30685,30718,30728,30744,30748,30750,30777,30781,30786,30788,30791,30795,30801,30803,30804,30822,30842,30856,30861,30902,30904,30905,30930,30936,30948,30958,30960,30965,30982,31026,31027,31030,31035,31064,31065,31074,31102,31107,31111,31121,31129,31131,31135,31141,31172,31178,31184,31202,31203,31217,31220,31238,31274,31276,31277,31282,31290,31294,31299,31301,31305,31333,31346,31357,31363,31377,31379,31408,31419,31420,31426,31427,31432,31433,31439,31443,31450,31451,31452,31453,31458,31463,31465,31466,31484,31486,31499,31500,31508,31519,31527,31529,31545,31554,31555,31569,31573,31586,31596,31599,31611,31634,31650,31651,31662,31666,31672,31685,31695,31696,31724,31738,31740,31750,31762,31771,31797,31810,31812,31825,31830,31837,31853,31856,31860,31867,31868,31870,31875,31878,31886,31900,31910,31916,31918,31928,31938,31939,31943,31945,31949,31950,31955,31962,31974,31981,31987,31989,31993,31996,32037,32077,32090,32093,32137,32139,32149,32151,32164,32168,32171,32179,32205,32207,32208,32210,32211,32212,32220,32228,32229,32245,32253,32254,32263,32285,32295,32347,32357,32359,32364,32366,32383,32391,32398,32402,32415,32420,32428,32442,32455,32463,32479,32518,32567,32576,32577,32583,32585,32590,32594,32595,32625,32632,32659,32663,32675,32686,32692,32733,32743,32762,32770,32773,32776,32797,32800,32805,32808,32814,32815,32817,32827,32828,32852,32859,32865,32866,32867,32870,32896,32927,32935,32950,32951,32956,32957,32966,33001,33004,33014,33027,33031,33033,33036,33038,33042,33044,33047,33050,33066,33074,33076,33079,33090,33096,33110,33113,33114,33132,33133,33148,33156,33171,33189,33194,33206,33217,33224,33252,33270,33304,33306,33321,33325,33342,33364,33378,33381,33398,33401,33402,33403,33446,33450,33471,33476,33482,33487,33488,33496,33506,33527,33528,33533,33535,33547,33560,33562,33565,33597,33623,33634,33638,33647,33669,33681,33692,33708,33726,33729,33743,33747,33773,33797,33807,33812,33816,33820,33824,33825,33830,33854,33864,33866,33875,33877,33880,33884,33892,33905,33906,33919,33920,33924,33928,33938,33939,33942,33955,33960,33965,33981,33982,34010,34014,34017,34018,34020,34021,34040,34051,34052,34064,34100,34104,34123,34130,34138,34143,34159,34160,34163,34195,34199,34241,34262,34272,34286,34300,34306,34319,34320,34323,34324,34326,34344,34351,34370,34373,34412,34418,34430,34450,34464,34477,34482,34543,34672,34673,34694,34699,34700,34725,34729,34737,34753,34766,34774,34778,34798,34805,34820,34831,34840,34855,34861,34882,34885,34886,34895,34910,34912,34916,34917,34926,34951,34959,34960,34972,34973,34976,34990,34996,34997,35007,35013,35015,35046,35061,35071,35108,35139,35143,35149,35156,35173,35191,35200,35207,35209,35210,35217,35220,35237,35238,35239,35241,35260,35265,35270,35303,35356,35369,35371,35384,35389,35401,35431,35454,35472,35476,35484,35497,35503,35508,35532,35546,35562,35577,35596,35615,35629,35647,35651,35660,35661,35678,35682,35689,35713,35715,35727,35728,35739,35744,35774,35810,35921,35995,35999,36013,36045,36052,36054,36059,36073,36075,36082,36087,36099,36107,36108,36113,36114,36120,36124,36125,36215,36218,36226,36262,36265,36288,36291,36308,36353,36366,36392,36394,36397,36410,36422,36431,36456,36465,36469,36471,36478,36480,36519,36525,36528,36534,36537,36540,36547,36558,36570,36578,36580,36589,36594,36619,36633,36653,36656,36662,36673,36682,36700,36710,36711,36718,36755,36768,36773,36775,36787,36790,36791,36793,36794,36795,36796,36798,36801,36810,36812,36815,36816,36826,36857,36872,36882,36915,36919,36934,36940,36950,36961,36964,36972,37015,37032,37038,37051,37060,37117,37124,37201,37209,37212,37214,37223,37232,37238,37244,37260,37262,37289,37302,37307,37316,37332,37343,37349,37366,37370,37374,37384,37390,37395,37400,37409,37416,37418,37419,37429,37441,37444,37461,37464,37469,37471,37474,37483,37486,37495,37505,37508,37513,37519,37550,37553,37561,37562,37566,37567,37588,37603,37605,37611,37612,37618,37619,37620,37622,37629,37635,37639,37676,37680,37681,37698,37699,37701,37704,37727,37734,37736,37737,37739,37742,37747,37748,37752,37757,37761,37764,37766,37767,37776,37788,37792,37805,37816,37819,37821,37823,37830,37835,37851,37856,37875,37876,37889,37892,37896,37911,37917,37924,37925,37926,37927,37933,37935,37950,37965,37972,37976,37979,37989,37991,37996,38009,38021,38047,38050,38081,38083,38108,38134,38189,38215,38264,38271,38277,38294,38295,38297,38304,38306,38310,38311,38314,38376,38388,38429,38451,38456,38465,38469,38486,38505,38523,38529,38543,38550,38563,38582,38589,38644,38659,38683,38689,38705,38708,38710,38720,38721,38730,38737,38741,38743,38748,38749,38751,38765,38769,38793,38811,38815,38833,38834,38836,38840,38842,38846,38848,38850,38856,38866,38880,38886,38894,38898,38901,38911,38933,38943,38947,38958,38963,38964,38983,38987,38998,39014,39016,39029,39092,39095,39097,39107,39111,39112,39114,39118,39132,39134,39136,39137,39142,39148,39153,39157,39182,39193,39196,39223,39224,39225,39227,39232,39234,39242,39245,39261,39266,39267,39323,39338,39352,39356,39364,39365,39382,39386,39392,39393,39398,39413,39436,39440,39455,39457,39462,39471,39483,39484,39512,39516,39523,39532,39567,39602,39606,39610,39613,39639,39648,39650,39668,39679,39682,39689,39695,39700,39722,39725,39726,39732,39737,39744,39760,39785,39807,39809,39819,39821,39822,39823,39831,39837,39839,39856,39887,39901,39913,39917,39921,39924,39935,39938,39946,39948,39952,39957,39963,39967,39968,39974,39983,39989,39994,40005,40015,40019,40029,40050,40059,40060,40194,40204,40214,40225,40249,40260,40265,40270,40272,40286,40301,40302,40316,40318,40323,40339,40357,40363,40381,40384,40388,40393,40404,40432,40444,40458,40460,40462,40472,40476,40479,40571,40577,40580,40581,40592,40597,40598,40600,40606,40610,40614,40616,40618,40619,40620,40623,40625,40637,40639,40641,40644,40646,40647,40674,40689,40696,40702,40706,40712,40726,40727,40742,40743,40761,40772,40773,40784,40787,40794,40802,40809,40831,40846,40854,40855,40857,40863,40869,40870,40871,40872,40873,40874,40875,40876,40877,40878,40879,40882,40883,40903,40904,40905,40906,40907,63751]],
        ];
        foreach ($series as $test) {
            foreach ($test[0] as $a => $input) {
                $class = $this->testedClass;
                $char = hex2bin($input);
                $exp = $test[1][$a];
                $s = new $class($char);
                $this->assertSame($exp, $s->nextCode(), "Sequence $input did not decode to $exp.");
                $this->assertFalse($s->nextCode(), "Sequence $input did not end after one character");
            }
        }
    }
}
