<?php
/** @license MIT
 * Copyright 2018 J. King et al.
 * See LICENSE and AUTHORS files for details */

declare(strict_types=1);
namespace MensBeam\Intl\TestCase\Encoding;

use MensBeam\Intl\Encoding\EUCJP;
use MensBeam\Intl\Encoding\Encoding;
use MensBeam\Intl\Encoding\EncoderException;

class TestEUCJP extends \MensBeam\Intl\Test\CoderDecoderTest {
    protected $testedClass = EUCJP::class;
    /*
        Char 0 U+007A   (1 byte)  Offset 0
        Char 1 U+FF96   (2 bytes) Offset 1
        Char 2 U+3088   (2 bytes) Offset 3
        Char 3 U+FF0D   (2 bytes) Offset 5
        Char 4 U+005C   (1 byte)  Offset 7
        Char 5 U+FF9B   (2 bytes) Offset 8
        Char 6 U+4F58   (3 bytes) Offset 10
        End of string at char 7, offset 13
    */
    protected $seekString = "7A 8ED6 A4E8 A1DD 5C 8EDB 8FB0EF";
    protected $seekCodes = [0x7A, 0xFF96, 0x3088, 0xFF0D, 0x5C, 0xFF9B, 0x4F58];
    protected $seekOffsets = [0, 1, 3, 5, 7, 8, 10, 13];
    /* This string contains an invalid character sequence sandwiched between two null characters */
    protected $brokenChar = "00 FF 00";

    public function provideCodePoints() {
        return [
            'U+0064 (HTML)'  => [false, 0x64, "64"],
            'U+0064 (fatal)' => [true,  0x64, "64"],
            'U+00A5 (HTML)'  => [false, 0xA5, "5C"],
            'U+00A5 (fatal)' => [true,  0xA5, "5C"],
            'U+203E (HTML)'  => [false, 0x203E, "7E"],
            'U+203E (fatal)' => [true,  0x203E, "7E"],
            'U+3088 (HTML)'  => [false, 0x3088, "A4 E8"],
            'U+3088 (fatal)' => [true,  0x3088, "A4 E8"],
            'U+FF96 (HTML)'  => [false, 0xFF96, "8E D6"],
            'U+FF96 (fatal)' => [true,  0xFF96, "8E D6"],
            'U+2212 (HTML)'  => [false, 0x2212, "A1 DD"],
            'U+2212 (fatal)' => [true,  0x2212, "A1 DD"],
            'U+00E6 (HTML)'  => [false, 0xE6, bin2hex("&#230;")],
            'U+00E6 (fatal)' => [true,  0xE6, new EncoderException("", Encoding::E_UNAVAILABLE_CODE_POINT)],
            'U+FFE2 (HTML)'  => [false, 0xFFE2, "A2 CC"],
            'U+FFE2 (fatal)' => [true,  0xFFE2, "A2 CC"],
            'U+2116 (HTML)'  => [false, 0x2116, "AD E2"],
            'U+2116 (fatal)' => [true,  0x2116, "AD E2"],
            '-1 (HTML)'  => [false, -1, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
            '-1 (fatal)' => [true,  -1, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
            '0x110000 (HTML)'  => [false, 0x110000, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
            '0x110000 (fatal)' => [true,  0x110000, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
        ];
    }

    public function provideStrings() {
        return [
            'empty string' => ["", []],
            'sanity check' => ["40", [64]],
            'former ASCII deviations' => ["5C 7E", [92, 126]],
            'changed multibyte index' => ["A1DD", [65293]],
            'JIS X 0201 range' => ["8EA1 8EDF", [65377, 65439]],
            'JIS X 0201 bogus range' => ["8EA0 8EE0", [65533, 65533]],
            'JIS X 0201 truncated character 1' => ["8E", [65533]],
            'JIS X 0201 truncated character 2' => ["8E 20", [65533, 32]],
            'JIS X 0201 truncated character 3' => ["8E FF", [65533]],
            'JIS X 0212 assigned range' => ["8FA2AF 8FEDE3", [728, 40869]],
            'JIS X 0212 total range' => ["8FA1A1 8FFEFE", [65533, 65533]],
            'JIS X 0212 bogus range 1' => ["8FA0A1 8FFFFE", [65533, 65533, 65533, 65533]],
            'JIS X 0212 bogus range 2' => ["8FA1A0 8FFEFF", [65533, 65533]],
            'JIS X 0212 truncated character 1' => ["8FA2", [65533]],
            'JIS X 0212 truncated character 2' => ["8FA2 20", [65533, 32]],
            'JIS X 0212 truncated character 3' => ["8FA2 FF", [65533]],
            'JIS X 0208 assigned range' => ["A1A1 FCFE", [12288, 65282]],
            'JIS X 0208 total range' => ["A1A1 FEFE", [12288, 65533]],
            'JIS X 0208 bogus range' => ["A1A0 A0FE", [65533, 65533, 65533]],
            'JIS X 0208 truncated character 1' => ["A1", [65533]],
            'JIS X 0208 truncated character 2' => ["A1 20", [65533, 32]],
            'JIS X 0208 truncated character 3' => ["A1 FF", [65533]],
        ];
    }

    /**
     * @dataProvider provideCodePoints
     * @covers MensBeam\Intl\Encoding\Encoder
     * @covers MensBeam\Intl\Encoding\EUCJP::encode
     * @covers MensBeam\Intl\Encoding\EUCJP::errEnc
     */
    public function testEncodeCodePoints(bool $fatal, $input, $exp) {
        return parent::testEncodeCodePoints($fatal, $input, $exp);
    }

    /**
     * @dataProvider provideCodePoints
     * @covers MensBeam\Intl\Encoding\EUCJP::encode
     * @covers MensBeam\Intl\Encoding\EUCJP::errEnc
     */
    public function testEncodeCodePointsStatically(bool $fatal, $input, $exp) {
        return parent::testEncodeCodePointsStatically($fatal, $input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\EUCJP::__construct
     * @covers MensBeam\Intl\Encoding\EUCJP::nextCode
     */
    public function testDecodeMultipleCharactersAsCodePoints(string $input, array $exp) {
        return parent::testDecodeMultipleCharactersAsCodePoints($input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\EUCJP::__construct
     * @covers MensBeam\Intl\Encoding\EUCJP::nextChar
     */
    public function testDecodeMultipleCharactersAsStrings(string $input, array $exp) {
        return parent::testDecodeMultipleCharactersAsStrings($input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\EUCJP::seekBack
     */
    public function testSTepBackThroughAString(string $input, array $exp) {
        return parent::testSTepBackThroughAString($input, $exp);
    }

    /**
     * @covers MensBeam\Intl\Encoding\EUCJP::seek
     * @covers MensBeam\Intl\Encoding\EUCJP::posChar
     * @covers MensBeam\Intl\Encoding\EUCJP::posByte
     * @covers MensBeam\Intl\Encoding\EUCJP::rewind
     */
    public function testSeekThroughAString() {
        return parent::testSeekThroughAString();
    }

    /**
     * @covers MensBeam\Intl\Encoding\EUCJP::posChar
     * @covers MensBeam\Intl\Encoding\EUCJP::posByte
     */
    public function testTraversePastTheEndOfAString() {
        return parent::testTraversePastTheEndOfAString();
    }

    /**
     * @covers MensBeam\Intl\Encoding\EUCJP::peekChar
     * @covers MensBeam\Intl\Encoding\EUCJP::stateSave
     * @covers MensBeam\Intl\Encoding\EUCJP::stateApply
     */
    public function testPeekAtCharacters() {
        return parent::testPeekAtCharacters();
    }

    /**
     * @covers MensBeam\Intl\Encoding\EUCJP::peekCode
     * @covers MensBeam\Intl\Encoding\EUCJP::stateSave
     * @covers MensBeam\Intl\Encoding\EUCJP::stateApply
     */
    public function testPeekAtCodePoints() {
        return parent::testPeekAtCodePoints();
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\EUCJP::lenChar
     * @covers MensBeam\Intl\Encoding\EUCJP::lenByte
     * @covers MensBeam\Intl\Encoding\EUCJP::stateSave
     * @covers MensBeam\Intl\Encoding\EUCJP::stateApply
     */
    public function testGetStringLength(string $input, array $points) {
        return parent::testGetStringLength($input, $points);
    }

    /**
     * @covers MensBeam\Intl\Encoding\EUCJP::errDec
     */
    public function testReplacementModes() {
        return parent::testReplacementModes();
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\EUCJP::rewind
     * @covers MensBeam\Intl\Encoding\EUCJP::chars
     * @covers MensBeam\Intl\Encoding\EUCJP::codes
     */
    public function testIterateThroughAString(string $input, array $exp) {
        return parent::testIterateThroughAString($input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\EUCJP::nextCode
     */
    public function testIterateThroughAStringAllowingSurrogates(string $input, array $strictExp, array $relaxedExp = null) {
        return parent::testIterateThroughAStringAllowingSurrogates($input, $strictExp, $relaxedExp);
    }

    /**
     * @covers MensBeam\Intl\Encoding\EUCJP::seekBack
     */
    public function testSeekBackOverRandomData() {
        return parent::testSeekBackOverRandomData();
    }

    /**
     * @group optional
     */
    public function testPedanticallyDecodeSingleCharactersAsCodePoint() {
        $series = [
            'characters' => [["5C","A1F8","A1AF","A1EB","A1DE","A1AD","A2F9","A1DF","A1E0","A6A1","A6A2","A6A3","A6A4","A6A5","A6A6","A6A7","A6A8","A6A9","A6AA","A6AB","A6AC","A6AD","A6AE","A6AF","A6B0","A6B1","A6B2","A6B3","A6B4","A6B5","A6B6","A6B7","A6B8","A6C1","A6C2","A6C3","A6C4","A6C5","A6C6","A6C7","A6C8","A6C9","A6CA","A6CB","A6CC","A6CD","A6CE","A6CF","A6D0","A6D1","A6D2","A6D3","A6D4","A6D5","A6D6","A6D7","A6D8","A7A7","A7A1","A7A2","A7A3","A7A4","A7A5","A7A6","A7A8","A7A9","A7AA","A7AB","A7AC","A7AD","A7AE","A7AF","A7B0","A7B1","A7B2","A7B3","A7B4","A7B5","A7B6","A7B7","A7B8","A7B9","A7BA","A7BB","A7BC","A7BD","A7BE","A7BF","A7C0","A7C1","A7D1","A7D2","A7D3","A7D4","A7D5","A7D6","A7D8","A7D9","A7DA","A7DB","A7DC","A7DD","A7DE","A7DF","A7E0","A7E1","A7E2","A7E3","A7E4","A7E5","A7E6","A7E7","A7E8","A7E9","A7EA","A7EB","A7EC","A7ED","A7EE","A7EF","A7F0","A7F1","A7D7","A1BE","A1BD","A1C6","A1C7","A1C8","A1C9","A2F7","A2F8","A1C5","A1C4","A2F3","A1EC","A1ED","A2A8","7E","A1EE","ADE2","ADE4","A2F2","ADB5","ADB6","ADB7","ADB8","ADB9","ADBA","ADBB","ADBC","ADBD","ADBE","FCF1","FCF2","FCF3","FCF4","FCF5","FCF6","FCF7","FCF8","FCF9","FCFA","A2AB","A2AC","A2AA","A2AD","A2CD","A2CE","A2CF","A2DF","A2D0","A2E0","A2BA","A2BB","ADF4","A1DD","A2E5","A2E7","A1E7","ADF8","A2DC","A1C2","A2CA","A2CB","A2C1","A2C0","A2E9","A2EA","ADF3","A1E8","A2E8","A2E6","A2E2","A1E2","A2E1","A1E5","A1E6","A2E3","A2E4","A2BE","A2BF","A2BC","A2BD","A2DD","ADF9","A2DE","ADA1","ADA2","ADA3","ADA4","ADA5","ADA6","ADA7","ADA8","ADA9","ADAA","ADAB","ADAC","ADAD","ADAE","ADAF","ADB0","ADB1","ADB2","ADB3","ADB4","A8A1","A8AC","A8A2","A8AD","A8A3","A8AE","A8A4","A8AF","A8A6","A8B1","A8A5","A8B0","A8A7","A8BC","A8B7","A8B2","A8A9","A8BE","A8B9","A8B4","A8A8","A8B8","A8BD","A8B3","A8AA","A8BA","A8BF","A8B5","A8AB","A8BB","A8C0","A8B6","A2A3","A2A2","A2A5","A2A4","A2A7","A2A6","A2A1","A1FE","A1FB","A1FD","A1FC","A2FE","A1FA","A1F9","A1EA","A1E9","A2F6","A2F5","A2F4","A1A1","A1A2","A1A3","A1B7","A1B9","A1BA","A1BB","A1D2","A1D3","A1D4","A1D5","A1D6","A1D7","A1D8","A1D9","A1DA","A1DB","A2A9","A2AE","A1CC","A1CD","ADE0","ADE1","A4A1","A4A2","A4A3","A4A4","A4A5","A4A6","A4A7","A4A8","A4A9","A4AA","A4AB","A4AC","A4AD","A4AE","A4AF","A4B0","A4B1","A4B2","A4B3","A4B4","A4B5","A4B6","A4B7","A4B8","A4B9","A4BA","A4BB","A4BC","A4BD","A4BE","A4BF","A4C0","A4C1","A4C2","A4C3","A4C4","A4C5","A4C6","A4C7","A4C8","A4C9","A4CA","A4CB","A4CC","A4CD","A4CE","A4CF","A4D0","A4D1","A4D2","A4D3","A4D4","A4D5","A4D6","A4D7","A4D8","A4D9","A4DA","A4DB","A4DC","A4DD","A4DE","A4DF","A4E0","A4E1","A4E2","A4E3","A4E4","A4E5","A4E6","A4E7","A4E8","A4E9","A4EA","A4EB","A4EC","A4ED","A4EE","A4EF","A4F0","A4F1","A4F2","A4F3","A1AB","A1AC","A1B5","A1B6","A5A1","A5A2","A5A3","A5A4","A5A5","A5A6","A5A7","A5A8","A5A9","A5AA","A5AB","A5AC","A5AD","A5AE","A5AF","A5B0","A5B1","A5B2","A5B3","A5B4","A5B5","A5B6","A5B7","A5B8","A5B9","A5BA","A5BB","A5BC","A5BD","A5BE","A5BF","A5C0","A5C1","A5C2","A5C3","A5C4","A5C5","A5C6","A5C7","A5C8","A5C9","A5CA","A5CB","A5CC","A5CD","A5CE","A5CF","A5D0","A5D1","A5D2","A5D3","A5D4","A5D5","A5D6","A5D7","A5D8","A5D9","A5DA","A5DB","A5DC","A5DD","A5DE","A5DF","A5E0","A5E1","A5E2","A5E3","A5E4","A5E5","A5E6","A5E7","A5E8","A5E9","A5EA","A5EB","A5EC","A5ED","A5EE","A5EF","A5F0","A5F1","A5F2","A5F3","A5F4","A5F5","A5F6","A1A6","A1BC","A1B3","A1B4","ADEA","ADEB","ADEC","ADE5","ADE6","ADE7","ADE8","ADE9","ADC6","ADCA","ADC1","ADC4","ADC2","ADCC","ADCB","ADC5","ADCD","ADC7","ADCF","ADC0","ADCE","ADC3","ADC8","ADC9","ADDF","ADEF","ADEE","ADED","ADD3","ADD4","ADD0","ADD1","ADD2","ADD6","ADD5","ADE3","B0EC","C3FA","BCB7","CBFC","BEE6","BBB0","BEE5","B2BC","C9D4","CDBF","D0A2","B1AF","B3EE","D0A3","C0A4","D2C2","B5D6","CABA","BEE7","CEBE","CAC2","F9AD","D0A4","C3E6","D0A5","B6FA","D0A6","B4DD","C3B0","BCE7","D0A7","D0A8","D0A9","C7B5","B5D7","C7B7","C6E3","B8C3","CBB3","E9C9","D0AA","BEE8","D0AB","B2B5","B6E5","B8F0","CCE9","D6A6","CDF0","C6FD","B4A5","B5B5","D0AC","D0AD","CEBB","CDBD","C1E8","D0AF","BBF6","C6F3","D0B2","B1BE","B8DF","B8DE","B0E6","CFCB","CFCA","BAB3","B0A1","D0B3","D0B4","D0B5","CBB4","D0B6","B8F2","B0E7","CBF2","B5FC","B5FD","B5FE","C4E2","CEBC","D0B7","D0B8","D0B9","BFCD","BDBA","BFCE","D0BE","D0BC","D0BD","B5D8","BAA3","B2F0","D0BB","D0BA","CAA9","BBC6","BBC5","C2BE","D0BF","C9D5","C0E7","A1B8","D0C0","D0C2","F9AE","C2E5","CEE1","B0CA","D0C1","B2BE","B6C4","C3E7","B7EF","D0C3","C7A4","F9AF","F9B0","B4EB","F9B1","D0C4","B0CB","B8E0","B4EC","C9FA","C8B2","B5D9","B2F1","D0E7","C5C1","C7EC","D0C6","C8BC","CEE2","BFAD","F9B2","BBC7","BBF7","B2C0","C4D1","C3A2","D0CA","B0CC","C4E3","BDBB","BAB4","CDA4","C2CE","B2BF","F9B3","D0C9","CDBE","D0C5","D0C7","BAEE","D0C8","D5A4","D0D0","D0D3","D0D1","B2C2","CABB","D0CB","D0CF","B8F3","BBC8","B4A6","D0D4","D0CC","F9B5","CEE3","BBF8","D0CD","D0D2","F9B4","F9B7","D0D5","D0CE","F9B6","B6A1","B0CD","B6A2","B2C1","D5A5","CBF9","C9EE","B8F4","BFAF","CEB7","CAD8","B7B8","C2A5","B2E4","F9A6","BDD3","F9B8","D0D9","D0DE","D0DC","D0D7","C2AF","D0DA","D0DD","D0DB","CADD","D0D8","BFAE","CBF3","D0DF","D0E0","BDA4","D0ED","C7D0","C9B6","D0E8","CAF0","B2B6","D0EC","F9BB","D0E6","D0EF","C1D2","B8C4","C7DC","E0C7","D0EE","C5DD","D0E3","B8F6","B8F5","D0E1","F9BC","BCDA","D0E9","F9BA","CAEF","C3CD","D0E5","B7F1","D0E2","D0EA","D0E4","CED1","D0EB","CFC1","B6E6","B7F0","F9B9","F9BF","D0F0","F9BD","D0F1","D0F5","B0CE","CAD0","D0F4","D0F3","D0F7","D0F6","C4E4","B7F2","D0F8","F9BE","BCC5","C2A6","C4E5","B6F6","D0F9","B5B6","D0FA","D0FC","CBB5","B7E6","F9C0","BBB1","C8F7","D0FB","BAC5","CDC3","D0FE","D1A3","D0FD","BAC4","BDFD","B7B9","D1A4","B6CF","D1A1","D1A2","C6AF","C1FC","B6A3","CBCD","D1A5","F9C2","CEBD","D1A6","D1A9","D1A7","C1CE","D1A8","D1AA","F9C1","D1AC","D1AB","CAC8","B5B7","D1AE","D1AF","B2AF","D1AD","BCF4","D1B2","D1B1","D1B0","D0D6","D1B3","BDFE","D1B4","CDA5","CCD9","D1B6","D1B5","D1B8","D1B7","D1B9","D1BA","B0F4","B8B5","B7BB","BDBC","C3FB","B6A4","C0E8","B8F7","F9C3","B9EE","D1BC","CCC8","C5C6","BBF9","D1BB","D1BD","C5DE","B3F5","D1BE","F9C4","C6FE","C1B4","D1C0","D1C1","C8AC","B8F8","CFBB","D1C2","B6A6","CABC","C2B6","B6F1","C5B5","B7F3","D1C3","D1C4","C6E2","B1DF","D1C7","BAFD","D1C6","BAC6","D1C8","E6EE","D1C9","CBC1","D1CA","D1CB","D1CC","BEE9","BCCC","F9C5","B4A7","D1CF","D1CD","CCBD","D1CE","C9DA","D1D0","D1D1","D1D2","C5DF","D1D6","D1D4","D1D5","D1D3","BAE3","D1D7","CCEA","CEE4","D1D8","F9C6","C0A8","D1D9","BDDA","D1DA","C3FC","CEBF","C5E0","D2C5","D1DB","F4A5","B6C5","D1DC","CBDE","BDE8","C2FC","D1DE","C6E4","F9C7","D1DF","D1E0","B3AE","D1E1","B6A7","C6CC","B1FA","BDD0","C8A1","D1E2","C5E1","BFCF","D1E3","CAAC","C0DA","B4A2","B4A9","D1E4","D1E6","B7BA","D1E5","F9C8","CEF3","BDE9","C8BD","CACC","D1E7","CDF8","D1E8","D1E9","C5FE","D1EA","C0A9","BAFE","B7F4","D1EB","BBC9","B9EF","C4E6","D1ED","C2A7","BAEF","D1EE","D1EF","C1B0","D1EC","D1F1","CBB6","B9E4","D1F0","B7F5","BADE","C7ED","D1F4","D1F2","C9FB","BEEA","D1FB","B3E4","D1F5","D1F3","C1CF","D1F7","D1F6","B3C4","B7E0","D1FC","CEAD","D1F8","D1FD","D1FA","D1F9","CECF","F9C9","B8F9","B2C3","CEF4","F9CA","BDF5","C5D8","B9E5","D2A2","D2A3","FBBC","CEE5","CFAB","D2A5","B8FA","D2A4","B3AF","F9CB","D2A6","CBD6","C4BC","CDA6","CAD9","D2A7","F0D5","C6B0","D2A8","B4AA","CCB3","F9CC","BEA1","D2A9","CAE7","D2AD","C0AA","D2AA","B6D0","D2AB","B4AB","B7AE","D2AE","D2AF","D2B0","D2B1","BCDB","B8FB","CCDE","F9CD","CCE8","C6F7","CAF1","D2B2","F9CE","D2B3","D2B5","D2B7","D2B6","D2B8","B2BD","CBCC","BAFC","D2B9","C1D9","BEA2","B6A9","D2BA","F9CF","C8DB","D2BB","D2BC","D2BD","D2BE","C9A4","B6E8","B0E5","C6BF","D2BF","BDBD","C0E9","D2C1","D2C0","BEA3","B8E1","D2C3","C8BE","D2C4","C8DC","C2B4","C2EE","B6A8","C6EE","C3B1","C7EE","CBCE","D2C6","C0EA","B7B5","D2C7","D2C8","B1AC","B0F5","B4ED","F9D0","C2A8","B5D1","CDF1","D2CB","B2B7","D2CA","B6AA","D2CC","CCF1","F9D1","D2CD","CED2","B8FC","B8B6","D2CE","D2D0","D2CF","BFDF","B1B9","B1DE","D2D1","D2D2","F9D2","B8B7","D2D3","B5EE","BBB2","D2D4","CBF4","BAB5","B5DA","CDA7","C1D0","C8BF","BCFD","BDC7","BCE8","BCF5","BDF6","C8C0","F9D3","D2D7","B1C3","C1D1","B8FD","B8C5","B6E7","D2DB","C3A1","C2FE","B6AB","BEA4","D2DC","D2DA","B2C4","C2E6","BCB8","BBCB","B1A6","B3F0","B9E6","BBCA","D2DD","D2DE","B5C9","B3C6","B9E7","B5C8","C4DF","B1A5","C6B1","CCBE","B9A1","CDF9","C5C7","B8FE","B7AF","D2E7","B6E3","CBCA","C8DD","D2E6","B4DE","D2E1","D2E2","D2E4","D2E5","B5DB","BFE1","CAAD","D2E3","D2DF","B8E3","D2E0","CFA4","CAF2","C4E8","B8E2","B9F0","D2E8","C6DD","D2EC","BCFE","BCF6","D2EF","D2ED","CCA3","D2EA","D2F3","D2EE","D2F1","B8C6","CCBF","D2F2","D2F4","D2F6","F9D6","BAF0","CFC2","D2EB","D2E9","D2F5","D2F0","F9D5","D2F8","D3A3","D2FA","D2FE","F9D7","D3A1","D2FB","D3BE","BAE9","B3B1","D2F9","D3A5","B0F6","D3A4","B0A5","C9CA","D3A2","D2FC","D2F7","D2FD","BAC8","D3A6","B0F7","D3AF","D3A7","D3A8","BEA5","CBE9","D3AD","D3AC","C5AF","D3AE","D3AB","F9D8","B1B4","BAB6","BFB0","D3A9","C5E2","D3AA","B0A2","D3B4","CDA3","BEA7","D3BA","D3B9","D3B0","C2C3","D3B1","C2EF","D3B6","BEA6","D3B3","CCE4","B7BC","D3B7","D3B8","D3B5","D3BB","D3B2","D3C1","D3C6","D3C2","D3BD","D3C7","C1B1","F9D9","D3C9","B9A2","D3BF","C3FD","D3C3","D3BC","B4AD","B4EE","B3E5","D3C4","D3C0","B7F6","D3CA","D3C8","C1D3","B5CA","B6AC","D3C5","B6F4","B1C4","D3CE","D3CC","D4A7","D3D1","D3CB","D3CF","D3CD","BBCC","D3D0","D3D3","D3D8","D3D6","D3D5","C3B2","B2C5","D3D2","D3D4","BEA8","B1B3","D3D7","B2DE","D3E2","BEFC","D3DE","D3DC","D3DD","D3DF","B1BD","C1B9","D3D9","D3DA","B3FA","D3E1","B4EF","D3E4","D3E0","D3E3","CAAE","C6D5","C8B8","D3E6","D3E5","B3C5","D3E7","D3EA","D3E9","D3E8","C7B9","D3EB","D3EC","D3EE","D3ED","D3F0","D3F3","D3F1","D3EF","D3F2","D3F4","D3F5","D3F6","D3F7","D3F8","D1C5","BCFC","BBCD","B2F3","B0F8","C3C4","D3F9","BAA4","B0CF","BFDE","D3FA","B8C7","B9F1","D3FC","D3FB","CAE0","D3FD","D4A1","D3FE","D4A2","D4A3","B7F7","B1E0","D4A4","D4A6","D4A5","D4A8","C5DA","D4A9","B0B5","BADF","B7BD","C3CF","D4AA","D4AB","D4AD","D4AE","BAE4","B6D1","CBB7","D4AC","D4AF","BAC1","B9A3","F9DA","D4B3","BAA5","F9DB","C3B3","D4B0","C4DA","D4B4","BFE2","D4B2","D4B5","B7BF","D4B6","D4B7","B9A4","B3C0","D4B9","D4BA","F9DC","D4BB","D4B8","D4B1","D4BC","D4BD","F9DE","F9DD","CBE4","BEEB","D4BF","D4C0","D4BE","D4C2","C7B8","B0E8","C9D6","D4C3","BEFD","BCB9","C7DD","B4F0","BAEB","CBD9","C6B2","B7F8","C2CF","D4C1","D4C4","C2C4","D4C5","D4C6","D4C8","C4E9","B4AE","F4A1","B1E1","CAF3","BEEC","C5C8","BAE6","D4CE","CABD","CEDD","B2F4","D4CA","C1BA","D4CD","C5E3","C5C9","C5E4","C8B9","C4CD","BAC9","D4C9","B1F6","C5B6","D4CB","D4C7","BFD0","D4CF","BDCE","B6AD","D4D0","CAE8","C1FD","C4C6","F9E1","D4D2","CBCF","D4D3","D4D8","F9E2","CAAF","D4D7","D4D1","D4D4","D4D6","BAA6","CAC9","D4D9","C3C5","B2F5","BEED","D4DB","D4DA","B9E8","D4DC","D4DE","D4DD","D4E0","D4D5","D4E2","D4E1","D4DF","BBCE","BFD1","C1D4","D4E3","C0BC","B0ED","C7E4","C4DB","D4E5","D4E4","D4E6","D4E7","D4E8","D4E9","CAD1","D4EA","F9E3","B2C6","D4EB","CDBC","B3B0","D2C9","BDC8","C2BF","D4EC","CCEB","CCB4","D4EE","C2E7","C5B7","C2C0","C9D7","D4EF","D4F0","B1FB","BCBA","D4F1","B0D0","D4F2","D4F3","B1E2","B4F1","C6E0","CAF4","D4F7","C1D5","D4F6","B7C0","F9E4","CBDB","D4F5","C5E5","D4F9","D4F8","F9E5","F9E6","D4FB","D4FA","F9E7","B1FC","D4FC","BEA9","D4FE","C3A5","D4FD","CAB3","BDF7","C5DB","D5A1","B9A5","D5A2","C7A1","C8DE","CCD1","C7A5","D5AB","B5B8","CDC5","CCAF","D6AC","D5A3","D5A6","F9E8","C2C5","CBB8","C5CA","D5A7","CBE5","F9E9","BACA","BEAA","D5A8","BBD0","BBCF","B0B9","B8C8","C0AB","B0D1","D5AC","D5AD","D5AA","B1B8","B4AF","D5A9","CCC5","C9B1","B0A8","B0F9","BBD1","B0D2","B0A3","D5B2","D5B0","CCBC","D5B3","D5B1","D5AF","BFB1","D5AE","CADA","B8E4","D5B7","D5B8","BEAB","D5B4","CFAC","C7CC","D5B6","BAA7","D5B9","C9D8","D5BA","D5B5","CCBB","C7DE","D5BB","C9B2","D5BC","D5C0","D5BD","B2C7","D5BF","BCBB","D5BE","B7F9","D5CC","D5C5","D5C2","C3E4","D5C1","D5C3","D5C4","D5C6","D5C7","B4F2","D5C9","D5C8","D5CA","BEEE","D5CD","C4DC","B1C5","D5CB","D5CE","D5CF","D5D2","D5D0","D5D1","BBD2","D5D3","B9A6","D5D4","F9EA","BBFA","C2B8","D5D5","D5D6","BBDA","B9A7","CCD2","B5A8","B8C9","D5D7","B3D8","D5D8","C2B9","D5D9","D6A3","D5DA","D5DB","D5DC","D5DE","D5DF","D5E0","C2F0","B1A7","BCE9","B0C2","C1D7","B4B0","BCB5","B9A8","C5E6","BDA1","B4B1","C3E8","C4EA","B0B8","B5B9","CAF5","BCC2","B5D2","C0EB","BCBC","CDA8","D5E1","B5DC","BACB","B3B2","B1E3","BEAC","B2C8","D5E2","CDC6","BDC9","F9EB","BCE4","D5E3","B4F3","C6D2","CCA9","D5E4","D5E5","C9D9","D5E7","B4A8","B6F7","D5E6","F9ED","B4B2","BFB2","D5EB","BBA1","B2C9","D5EA","D5E8","D5EC","D5E9","C7AB","DCCD","BFB3","D5ED","F9EE","CEC0","D5EE","D5F0","C3FE","D5EF","C0A3","BBFB","C2D0","BCF7","C9F5","C0EC","BCCD","D5F1","BEAD","D5F2","D5F3","B0D3","C2BA","BFD2","D5F4","C6B3","BEAE","BEAF","D5F5","C0ED","BEB0","F9EF","D5F6","D5F7","CCE0","D5F8","B6C6","BDA2","D5F9","D5FA","BCDC","BFAC","C6F4","BFD4","C8F8","C7A2","B6C9","D5FB","B5EF","D5FC","B6FE","C6CF","B2B0","BBD3","D5FD","D6A2","D6A1","B6FD","D5FE","C5B8","C2B0","C5CB","BCC8","C1D8","CDFA","D6A4","D6A5","C6D6","BBB3","D6A7","D6A8","D6A9","B4F4","D6AA","D6AB","B2AC","F9F0","C1BB","B4E4","D6AD","CCA8","C2D2","B3D9","D6AF","D6B1","B4DF","F9F1","D6AE","D6B0","D6B3","D6B2","D6B4","D6B5","C6BD","B6AE","B2E5","D6B6","D6BB","D6B9","CAF7","CAF6","F9F2","C5E7","D6B8","BDD4","D6B7","BFF2","D6BC","BAEA","D6C2","D6C3","D6BD","B3B3","D6BE","D6C7","D6C6","D6C5","D6C1","D6C0","D6C4","F9F3","CAF8","F9F6","D6CB","D6C8","D6CA","CDF2","D6C9","F9F4","D6BF","BFF3","D6CC","F9F7","BAB7","D6CD","D6CE","D6D1","D6D0","D6CF","C5E8","D6BA","D6D7","D6D3","D6D2","D6D4","D6D5","D6D8","F9F8","F9F9","CEE6","D6D9","D6D6","D6DA","B4E0","D6DB","F9FA","D6DD","D6DC","D6DE","D6DF","C0EE","BDA3","BDE4","C1E3","B9A9","BAB8","B9AA","B5F0","D6E0","BAB9","B8CA","D6E1","CCA6","C7C3","D6E2","B9AB","B4AC","C3A7","B6D2","BBD4","C9DB","C8C1","D6E3","B4F5","D6E6","C4A1","D6E5","D6E4","D6E7","C4EB","BFE3","BBD5","C0CA","C2D3","B5A2","C4A2","D6E8","D6E9","BEEF","CBB9","D6EC","D6EB","D6EA","C9FD","D6F3","CBDA","D6ED","D6EF","CBEB","D6EE","D6F0","C8A8","D6F1","CABE","D6F2","B4B3","CABF","C7AF","D6F4","D6F5","B9AC","B4B4","D6F6","B8B8","CDC4","CDA9","B4F6","D6F8","C4A3","B9AD","BEB1","C8DF","BEB2","BDF8","C4EC","CAF9","C5B9","B9AE","C9DC","D6F9","C5D9","BAC2","B8CB","C4ED","B0C3","BDEE","B9AF","CDC7","D6FA","D6FB","C7D1","D6FC","CEF7","CFAD","D6FE","D6FD","B3C7","D7A1","D7A4","D7A5","D7A3","C9C0","BEB3","D7A7","D7A6","D7A2","D7A8","D7A9","D7AA","D7AD","D7AB","D7AC","D7AE","B1E4","C4EE","D7AF","B7FA","B2F6","C7B6","D7B0","C6FB","CADB","D7B1","CFAE","D7B2","CAC0","D7B5","D0A1","D0B1","BCB0","C6F5","D7B6","B5DD","C4A4","B0FA","D7B7","CAA6","B9B0","C3D0","C4EF","F9FB","CCEF","B8B9","B8CC","D7B8","D7B9","D7BF","BCE5","F9FC","C4A5","B6AF","D7BA","C9AB","C3C6","D7BB","F9AC","D7BC","B6B0","D7BD","D7BE","D7C0","C5F6","D7C1","D7C2","D7C3","D7B4","D7B3","D7C4","B7C1","C9A7","F9FD","BACC","C9B7","C4A6","C9CB","D7C5","BEB4","B1C6","D7C6","D7C7","CCF2","C8E0","D7CA","B1FD","C0AC","D7C9","D7C8","B7C2","C2D4","D7CE","D7CC","D7CB","CEA7","B8E5","BDF9","D7CD","C5CC","BDBE","C6C0","D7D1","D7D0","D7CF","D7D2","B8E6","D7D3","C9FC","BDDB","D7D4","C8F9","C6C1","C4A7","F9FE","C5B0","D7D5","B5AB","BFB4","C9AC","B4F7","C7A6","D7D6","BBD6","CBBA","CBBB","B1FE","D7DB","FAA1","C3E9","D7D8","B2F7","D8AD","D7DA","C7B0","D7D9","D7D7","B9FA","D7DD","D7E3","D7E9","D7E1","C5DC","D7E6","C9DD","D7E0","D7E5","CEE7","BBD7","C2D5","D7DE","B5DE","D7E8","C0AD","B1E5","D7E2","B2F8","D7E7","B6B1","D7E4","D7EA","D7EC","D7F6","D7F4","D7F1","D7F0","CEF8","D7F2","B6B2","B9B1","BDFA","D7F9","D7EB","FAA2","D7EF","D7DF","B2FA","D7F3","D7F5","C3D1","BAA8","B2B8","D7ED","D7F8","D7F7","B6B3","C2A9","B3E6","B7C3","D7EE","D7FA","D7FD","D8A1","FAA3","BCBD","FAA4","D8A7","C4F0","D7FB","D8A5","B2F9","D8A3","D8A4","D7FE","D8A2","B8E7","CDAA","B4B5","B1D9","D8A6","C7BA","B0AD","C8E1","D7DC","D8AC","D8B0","CCE5","D8A9","C5E9","D8AE","BEF0","D8AF","C6D7","CFC7","D8AB","FAA6","D8B1","B9FB","C0CB","FAA5","B0D4","D8AA","D8A8","C1DA","D7FC","BBB4","C2C6","D8BD","FAA8","C1DB","D8B8","D8B5","D8B6","BCE6","D8B9","D8BC","D8B7","BDA5","D8BA","D8B4","CCFC","CCFB","D8BE","D8BF","B0D5","FAA9","D8B3","B6F2","B0A6","B4B6","FAA7","D8BB","D8C3","D8C2","D8C7","FAAB","D8C8","FAAA","D8C6","D8C9","D8C1","D8C5","D8CA","D8CB","D8C0","BBFC","D8C4","C2D6","B9B2","D8B2","BFB5","D8D8","CAE9","D8CE","D8CF","D8D0","D8D7","D8D6","CBFD","B4B7","D8D4","B7C5","B3B4","D8D1","CEB8","D8D3","B0D6","D8D5","D8CC","D8D2","D8D9","B7C4","D8CD","CDDD","CDAB","D8DC","D8E0","C1FE","CEF9","D8E1","D8DE","D8DB","FAAC","D8DA","D8DF","CAB0","C6B4","B7C6","D8E2","D8DD","D8E3","B7FB","B2B1","D8EB","B4B8","D8E9","D8EA","BAA9","D8E8","D8E6","D8E5","D8EC","D8E4","D8EE","B2FB","D8F0","D8EF","C4A8","D8F3","D8F1","D8E7","B7FC","D8F2","D8F6","D8F5","D8F7","D8F4","D8F8","D8F9","D8FA","CAEA","D8FC","D8FB","BDBF","C0AE","B2E6","B2FC","FAAD","D8FD","B0BF","C0CC","D8FE","ECC3","D9A1","B7E1","D9A2","C0EF","D9A3","D9A4","B5BA","D9A5","D9A6","D9A7","C2D7","B8CD","CCE1","CBBC","BDEA","D9A8","C0F0","EEBD","C8E2","BCEA","BACD","D9A9","C2C7","CAA7","C2F1","D9AC","D9AA","D9AD","D9AB","D9AE","CAB1","B0B7","C9DE","C8E3","D9AF","D9B2","BEB5","B5BB","D9B0","D9B7","BEB6","D9B1","C7C4","CDDE","D9B3","D9B4","D9B8","C5EA","D9B5","B9B3","C0DE","D9C6","C8B4","C2F2","FAAE","C8E4","DAAD","CAFA","C4F1","CBF5","D9BB","B2A1","C3EA","D9C4","C3B4","D9BE","D9C5","D9C0","D9C7","D9C3","D9C2","C7EF","D9BC","B2FD","D9BA","B5F1","C2F3","D9B6","D9B9","B9B4","C0DB","BEB7","D9C1","C7D2","B5F2","B3C8","B3E7","BFA1","D9C9","D9CE","D9CA","B7FD","D9CF","BBA2","B9E9","BDA6","D9BD","BBFD","D9CC","BBD8","D9CD","B0C4","D9C8","C4A9","B5F3","B6B4","D9CB","B0A7","BAC3","BFB6","C4F2","C8D4","D9D1","C1DE","C2AA","BBAB","D9D2","D9D4","D9D0","CAE1","C4BD","C1DC","CAFB","BCCE","D9E0","D9DF","BFF8","B7FE","D9D9","BEB9","C6E8","C7B1","D9D7","C1DD","BCF8","D9DC","BEB8","D9D6","D9DB","C7D3","D9D5","B7A1","B3DD","D9DD","CEAB","BACE","C3B5","D9DA","C0DC","B9B5","BFE4","B1E6","C1BC","D9D8","B5C5","B7C7","C4CF","D9DE","C1DF","D9E1","D9E3","C2B7","D9E9","D9E4","D9E6","C9C1","C4F3","D9E7","CDAC","CDC8","B4B9","B0AE","D9E5","D9E2","B4F8","B1E7","FAAF","D9E8","CDC9","D9EC","C2BB","D9F3","D9ED","D9EA","D9F1","D9D3","D9EE","D9F2","C8C2","C5EB","D9EB","D9EF","B7C8","BAF1","C0DD","D9F7","C5A6","FAB0","D9F4","CBE0","D9F5","D9F6","CCCE","C0A2","B7E2","D9FD","BBB5","D9FA","D9F9","C7B2","FAB1","C6B5","C5B1","D9FB","D9FC","C9EF","C7C5","BBA3","C0F1","CBD0","B3C9","DAA5","D9FE","CDCA","DAA7","DAA3","DAA4","C1E0","FAB2","DAA2","D9BF","DAA6","DAA1","DAAB","DAAC","C5A7","DAAE","BBA4","DAA9","B5BC","DAAF","DAA8","DAB3","DAB2","DAB1","DAB4","DAB6","BEF1","DAB5","DAB9","DAB7","DAB8","D9F0","DABB","DABA","D9F8","DABC","DAB0","BBD9","DABD","DABE","DAC0","DABF","DAC1","B2FE","B9B6","CAFC","C0AF","B8CE","DAC3","DAC6","FAB3","C9D2","B5DF","DAC5","DAC4","C7D4","DAC7","B6B5","DAC9","DAC8","B4BA","BBB6","C6D8","B7C9","BFF4","DACA","C0B0","C5A8","C9DF","DACB","DACC","DACD","CAB8","D5DD","C0C6","C9CC","BAD8","C8E5","C8C3","C5CD","CEC1","DACF","BCD0","DAD0","B0B6","B6D4","C0CD","C9E0","DAD1","BBC2","C3C7","BBDB","BFB7","DAD2","CAFD","B1F7","BBDC","DAD5","DAD3","DAD6","CEB9","DAD4","C0FB","DAD7","C2B2","DAD8","B4FA","DADA","DAD9","DADB","DADC","B4FB","C6FC","C3B6","B5EC","BBDD","C1E1","BDDC","B0B0","DADD","B2A2","DAE1","FAB4","B9B7","DAE0","BAAB","BEBA","FAB7","DADF","BEBB","CCC0","BAAA","B0D7","C0CE","FAB5","DAE6","FAB9","C0B1","B1C7","FABA","BDD5","CBE6","BAF2","BEBC","FAB8","C0A7","F9A8","DAE5","DAE3","DAE4","FAB6","C3EB","DBA6","DAEA","BBFE","B9B8","DAE8","DAE9","BFB8","DAE7","BBAF","FABC","FABD","DAEC","DAEB","DAF0","DAF1","DAED","FABB","B3A2","DAEE","DAEF","C8D5","C9E1","B7CA","DAF2","FABF","C0B2","BEBD","C3D2","B6C7","DAF3","DAF7","B2CB","DAF4","DAF6","DAF5","BDEB","C3C8","B0C5","DAF8","FAC0","DAF9","FAC1","C4AA","CEF1","BBC3","CAEB","FAC2","CBBD","DBA2","DAFB","DAFE","DAFD","FAC3","DAFA","DBA1","C6DE","DAFC","DBA3","BDEC","DBA4","CDCB","C7F8","DBA5","DBA7","DBA8","DBA9","B6CA","B1C8","B9B9","DBAA","DBAB","BDF1","C1E2","FAC4","F9AB","D2D8","C1BE","C1BD","C2D8","BAC7","D0F2","B7EE","CDAD","CAFE","C9FE","FAC5","DBAC","BAF3","C4BF","DBAD","CFAF","CBBE","C4AB","DBAE","B4FC","DBAF","DBB0","CCDA","CCA4","CBF6","CBDC","BBA5","DBB2","BCEB","CBD1","DBB4","DBB7","DBB6","B4F9","B5E0","DBB3","DBB5","DBB8","BFF9","CDFB","B0C9","BAE0","C2BC","BCDD","BEF3","DBBB","C5CE","DBB9","C2AB","DBBA","BEF2","CCDD","DBBC","DBBD","CDE8","FAC7","DBC2","B9BA","C7D5","DBBF","C5EC","DADE","DAE2","B5CF","C7C7","DBC1","BEBE","C8C4","DBC7","C8FA","DBBE","DBC4","DBC3","C0CF","CBED","CED3","CBE7","B2CC","BBDE","CFC8","DBC6","BFF5","DBC5","DBC0","B8CF","DBCC","DBCA","B2CD","DBC8","DBCE","DBD4","FAC8","FACA","C2C8","CAC1","DBD6","C9A2","DBD5","C7F0","CBBF","B4BB","C0F7","BDC0","C4D3","CDAE","DBD1","DBD0","DBD2","DBCF","DBD7","DBCD","DBCB","DBD3","DBC9","C3EC","CCF8","BCC6","BAF4","BABA","CBEF","B3C1","FACB","C4CE","C6CA","B1C9","C0F2","C0B4","B7AA","DBD9","B9BB","B3FC","DBDB","B3F4","DBE1","DBDE","C0F3","B3CB","BAAC","B3CA","BACF","DBDC","B7E5","B7CB","C5ED","FACC","DBDA","B0C6","DBDD","DBDF","B6CD","B7AC","FAC9","B4BC","B5CB","DBE2","BAF9","CBF1","BBB7","DBE3","C9B0","DBEF","B2B3","DBE4","DBF5","DBE5","CEC2","DBEC","C7DF","DBF4","DBE7","B0B4","DBE9","B9BC","DBEB","DBEA","DBE6","DBF1","BEBF","D4ED","B8E8","CDFC","DBE8","C4F4","B3A3","BAAD","DBE0","DBF0","B3E1","DBEE","DBF2","C5EE","B4FE","DCB2","F9A9","CCC9","DBF7","B4FD","DBFE","FACD","CBC0","DCA1","DCA3","DCA7","DBF9","C3AA","C5EF","DCAB","DBFC","DCA8","DCA2","BFB9","DCAC","C0B3","DCAA","B4BD","CFD0","DBF6","DCA6","B0D8","DBF8","CCBA","DBFD","BFA2","C4C7","DBF3","DCA5","BFFA","DCAF","B3F1","B8A1","DCB1","DBFA","DCB0","DCA9","DBFB","DCAD","DCAE","DCBF","C6CE","DCA4","DCBB","DCBD","C4D8","CDCC","C9F6","DCB8","C2CA","DCBE","C1BF","DCB5","DCC2","DCC1","C6EF","DCC0","C6EA","FACF","DCC4","DCB7","B6C8","DCBA","BDDD","C7E0","DCBC","B6CB","DCB4","DCB6","DCB3","CFB0","B3DA","DCB9","DCC3","B3B5","BAE7","B1DD","DCD4","CFB1","DCD7","FAD1","BFBA","DCD6","DCD5","DCD2","DCC6","DCE3","DCC5","DCD8","DCD0","DCCB","DCC8","DCC9","DCD1","F4A2","DCCE","B9BD","C4C8","C1E4","DCCC","DCC7","DCCA","CDCD","CBEA","DCCF","DCD9","FAD2","DCE1","DCDA","DCE7","DCE5","DCE0","DCDF","C4D0","C1E5","DCDD","DCDB","DCE2","DCE8","C8F5","DCEE","DCE9","DCEC","DCE6","C3F4","C9B8","DCDC","DCE4","BEC0","CCCF","DCF8","DCEB","B8A2","B2A3","B3DF","DCD3","FAD3","BEC1","DCF0","DCF7","BCF9","B3F2","C3AE","DCED","FAD5","DCF2","DCF6","B6B6","B5CC","DCF4","B5A1","C6CB","DCF3","DCF5","FAD4","DCEF","FAD6","DCF1","FAD7","B3E0","C3C9","DCFC","DCFA","B8E9","DCF9","DDA1","DBD8","DCFB","DCFD","DCFE","DDAC","DDA8","DBED","DDA7","DDA6","DDA3","DCEA","DDA5","DDA4","DDAA","CFA6","DDAD","B6FB","DDA9","DDAB","FAD8","FAD9","C8A7","DDAE","DDB2","DDAF","CDF3","DDB0","DCDE","DDB3","DDB4","B1B5","DDB6","B7E7","BCA1","B6D5","B2A4","CDDF","DDB8","DDB7","DDBA","B5BD","B6D6","B4BE","DDBD","DDBC","DDBE","B2CE","C3B7","DDBF","B4BF","DDC1","DDC0","DDC2","DDC3","DDC4","BBDF","C0B5","BAA1","C9F0","CAE2","CFC4","BBF5","BAD0","CEF2","DDC5","DDC6","BBE0","DDC7","DDC8","DDCA","DDC9","CBD8","BDDE","BCEC","BBC4","DDCB","DDCD","BFA3","DDCC","DDCE","DDCF","DDD0","DDD1","DDD2","DDD4","DDD3","DDD5","B2A5","C3CA","DDD6","BBA6","B3CC","DDD7","C5C2","D4CC","B5A3","DDD8","DDD9","CAEC","CBE8","C6C7","DDDA","C8E6","FADA","C8FB","CCD3","DDDB","DDDD","DDDC","DDDF","DDDE","DDE1","BBE1","CCB1","DDE2","DDE3","B5A4","DDE4","DDE6","DDE5","BFE5","C9B9","B1CA","C8C5","FADB","C4F5","BDC1","B5E1","C8C6","BCAE","DDE8","B4C0","B1F8","FADC","C6F2","DDE7","B9BE","C3D3","DDE9","DDF1","DDEA","FADE","C2C1","B5E2","DDF2","B7E8","B5A5","DDF0","DDEE","DDEB","CDE0","FADD","C4C0","C6D9","DDEC","DDF4","DDF3","B7A3","B2AD","BABB","DDED","DDEF","CBD7","C2F4","CBF7","DDFC","DDFD","B2CF","CAA8","CCFD","DEA1","BCA3","BEC2","DDF8","DDFE","B1E8","B6B7","DDF5","DDFA","C0F4","C7F1","C8E7","DDF7","CBA1","DDF9","DEA4","FADF","DEA2","DDFB","CBA2","C7C8","B5E3","C5A5","C3ED","DEA5","DEA3","C2D9","DDF6","B1CB","FAE0","CDCE","DEB0","DEAF","C0F6","DEAC","CDEC","C6B6","DEA6","C4C5","B1CC","B9BF","DEA9","BDA7","DEAE","DEAD","DEA8","DEAB","B3E8","DEAA","C7C9","CEAE","BEF4","C0F5","DEB6","DEB4","C9CD","DEB1","DEB3","B1BA","B9C0","CFB2","B3BD","C9E2","FAE2","CDE1","B3A4","BFBB","DEB5","DEBA","FAE1","BEC3","CDB0","DEB7","DEB2","DEB8","FAE3","CEDE","C5F3","C6C2","FAE4","B3B6","B1D5","DEBE","DEC1","CEC3","CDE4","DEC8","DEC2","DEBF","CED4","DEC5","FAE5","BDCA","DEC7","DECC","C5F1","DECA","DEC4","C3B8","DECB","DEC0","DEC6","DECD","B0FC","DEC3","DECE","BFBC","FAE7","BDDF","CAA5","BAAE","FAE6","DEBB","DEC9","C5BA","FAE8","C0B6","B3E9","BAD1","BEC4","DEBD","BDC2","B7CC","DEBC","DED2","BDED","B8BA","DEE1","DEDB","B5F4","C5CF","DED6","DEDF","B0AF","B1B2","FAEB","B2B9","DED8","C2AC","DECF","DED1","B9C1","DEE2","FAE9","DEDD","FAEC","DED5","DEDC","CCAB","DEDA","DEDE","B8D0","BEC5","C3B9","FAEA","DED4","CDAF","DED7","DED0","C5F2","DED3","DED9","CFD1","BCBE","CBFE","DEE3","C8AE","DEEF","B8BB","BDE0","DEE5","CEAF","B9C2","DEF2","B0EE","DEF0","DEE4","DEEA","DEEC","CDCF","DEE7","C5AE","DEE9","FAED","DEF1","DEEB","CCC7","DEE6","BCA2","DEFE","B3EA","DEE8","DEED","DEEE","C2EC","C2DA","DEF6","DEFC","DEFA","C5A9","DFA3","DEF7","DEF8","DEE0","B5F9","C9BA","BCBF","B9F7","CFB3","DEF4","DFA2","B1E9","C1E6","C7F9","B4C1","CEFA","CCA1","C4D2","DEFB","DEFD","C1B2","DFA1","DEF9","DEF3","B4C3","B7E9","DFAF","DFAA","C0F8","B3E3","BDE1","DFB3","DFAC","C4AC","DFA9","C4D9","DFCC","DFA6","DFA5","DFAE","DFA8","DFA7","DFAD","C0A1","DFA4","FAEE","DFB0","DFB1","B4C2","DFB6","DFB5","DFB7","DFBA","C5C3","DFB4","FAEF","DFB8","B7E3","C2F9","DFB2","C7BB","DFB9","DFBE","DFBC","DFBF","DFC2","DFBB","B9EA","C7A8","DEB9","CDF4","DFBD","DFC1","C2F5","DFC0","DFAB","FAF0","EFE9","DFC5","DFC9","DFC7","FAF1","FAF2","DFC3","DFC4","DFC8","DFC6","C9CE","DFCE","DFCB","DFCA","DFCD","C6D4","DFCF","C3F5","C2ED","FAF3","C0A5","DFD0","DFD2","DFD1","DEF5","DFD3","C6E7","DFD4","B2D0","C5F4","B3A5","B5E4","BCDE","BAD2","FAF4","CFA7","BFE6","B1EA","DFD6","DFD5","FAF5","DFD9","C3BA","DFDC","DFD7","DFDB","DFDA","C5C0","B0D9","F9A7","CEF5","DFDE","B1A8","DFE0","DFDF","DFDD","DFD8","CBA3","DFE2","FAF7","DFE1","FAF6","B1EB","DFE4","CAB2","DFE3","CCB5","BEC7","C1B3","BEC6","FAF9","FAFA","CEFB","DFEA","C0F9","DFE6","DFEB","B1EC","FAF8","DFE9","C7E1","DFE5","DFE8","BEC8","C8D1","DFEC","BCD1","C0FA","DFEF","DFE7","B7A7","DFED","CDD0","DFF0","F4A6","BDCF","DFF1","DFF2","C7AE","DFF4","DFF5","FAFC","C7B3","C5F5","DFF7","DFF9","CED5","DFF6","DFF8","B1ED","DFF3","D3DB","DFFA","C1E7","BBB8","DFFC","DFFB","BFA4","D2D9","DFFD","E0A1","DFEE","DFFE","FAFD","E0A2","C7FA","E0A3","E0A4","E0A5","E0A6","C4DE","E0A8","E0A7","E0A9","E0AA","BCDF","C9E3","CCEC","E0AB","E0AC","C1D6","BCA4","E0AD","E0AE","E0AF","CAD2","C8C7","E0B0","C7D7","C4AD","E0B1","B2E7","B5ED","CCC6","CCB6","B2B4","CFB4","CBD2","CAAA","C0B7","E0B2","C6C3","B8A3","E0B3","BAD4","E0B5","E0B4","E0B6","E0B7","E0B8","B5BE","E0B9","E0BA","B8A4","C8C8","FAFE","E0BC","BEF5","E0BB","FBA1","B6B8","E0BD","E0BF","E0BE","E0C0","B8D1","E0C1","B6E9","C1C0","B9FD","E0C3","E0C4","E0C2","BCED","C6C8","B6B9","E0C6","C3AC","E0C5","CFB5","C7E2","E0C9","E0CB","E0C8","CCD4","E0CA","E0CC","CEC4","FBA2","E0D0","E0CF","C3F6","C7AD","B8A5","E0CE","E0CD","CDB1","CDB2","E0D1","B1EE","B9F6","BBE2","E0D2","E0D3","E0D5","BDC3","E0D7","E0D6","E0D8","B3CD","E0DA","FBA4","E0D9","E0DC","E0DB","B8BC","CEA8","B6CC","B2A6","B6EA","B4E1","CEE8","E0DE","E0E0","FBA5","E0E1","B2D1","E0DD","FBA6","BBB9","C4C1","E0DF","FBA9","FBA7","E0E4","BCEE","FBA8","E0E2","B7BE","C8C9","E0E3","E0FE","FBAB","E0E9","B8BD","B5E5","E0E6","CDFD","FBAA","CEB0","C2F6","E0E8","FBAC","FBAE","FBAD","FBAF","E0EA","CED6","B6D7","C8FC","C7CA","E0EB","E0ED","E0F0","E0EC","E0EF","B8EA","B1CD","E0F1","BFF0","E0EE","CEDC","FBB0","E0F4","F4A4","E0F2","E0F5","E0E7","E0F3","BABC","E0F6","E0F7","CDFE","FBB1","E0F8","E0F9","FBB2","E0E5","E0FA","B4C4","BCA5","E0FB","E0FC","E0FD","B1BB","E1A1","C9BB","E1A2","B4A4","E1A3","E1A4","E1A5","E1A7","E1A8","E1A6","C9D3","E1AA","E1A9","FBB3","E1AC","E1AB","E1AD","E1AE","E1B0","E1AF","B9F9","E1B2","E1B1","B4C5","BFD3","C5BC","E1B3","C0B8","BBBA","B1F9","E1B4","CDD1","CAE3","E1B5","F9EC","C5C4","CDB3","B9C3","BFBD","C3CB","D2B4","C4AE","B2E8","E1B6","E1B7","E1BC","E1BA","E1B9","DAC2","B3A6","E1B8","B0DA","C8AA","C8CA","CEB1","E1BD","E1BB","C3DC","C0A6","C8AB","C9AD","E1BF","CEAC","B7CD","E1C0","E1BE","C8D6","E1C1","E1C2","FBB4","B0DB","BEF6","E1C7","E1C4","C6ED","E1C3","B5A6","E1CA","E1C5","E1C6","E1C9","E1C8","C9A5","C1C2","C1C1","B5BF","E1CB","E1CC","E1CD","E1CF","E1CE","B1D6","E1D7","C8E8","E1D1","E1D3","E1D5","BFBE","E1D6","E1D4","BCC0","E1D0","E1D2","C9C2","BEC9","E1D9","E1D8","E1DA","BCA6","BAAF","C5F7","E1DB","C4CB","E1DD","CEA1","E1DC","C1E9","E1E2","E1E4","E1E5","C3D4","E1E3","E1E0","E1DE","E1DF","E1E1","E1E8","E1E6","E1E7","E1E9","E1EB","E1EC","E1ED","E1EE","E1EA","E1F0","E1EF","E1F1","CEC5","E1F4","E1F2","E1F3","B4E2","CCFE","CACA","E1F6","E1F5","E1F7","E1F8","E1FC","E1F9","E1FA","E1FB","E1FD","E1FE","E2A1","E2A2","E2A3","C8AF","C5D0","E2A4","C7F2","C9B4","E2A5","FBB5","E2A6","C5AA","B3A7","B9C4","E2A7","E2A8","E2A9","BBA9","E2AB","E2AA","E2AC","E2AD","FBB8","FBB6","FBB7","FBB9","C8E9","E2AE","E2AF","F3E9","E2B0","E2B1","E2B2","BBAE","E2B3","C7D6","CBDF","B1CE","B1D7","E2B4","E2B6","E2B5","C5F0","C0B9","DDB9","E2B7","CCC1","E2B8","B4C6","C8D7","E2B9","E2BA","E2BB","CCDC","CCD5","C4BE","C1EA","E2BD","BDE2","BECA","E2C0","E2BF","E2BE","C8FD","B4C7","B8A9","E2C6","E2C3","BFBF","CCB2","E2C2","E2C4","E2C5","E2C1","E2C7","E2C8","C4AF","B4E3","C3E5","FBBB","E2C9","E2CA","E2CD","BFE7","C6C4","E2CE","CBD3","E2CB","E2CC","E2D1","E2D0","E2CF","E2D3","E2D2","E2D4","E2D6","E2D5","CACD","BDD6","CEC6","E2D7","C6B7","E2D8","E2D9","E2DD","E2DB","E2DC","E2DA","E2DE","E2DF","E2E0","E2E1","CCB7","E2E2","CCF0","E2E3","C3CE","C7EA","B6EB","C3BB","E2E4","B6BA","C0D0","E2E5","BABD","E2E6","E2E7","B8A6","BAD5","E2E9","FBBD","C5D6","BAD6","B5CE","CBA4","C7CB","C5D7","B9DC","E2EB","FBBE","BECB","FBBF","CEB2","B9C5","B8A7","C8A3","E2ED","FBC0","E2EF","B8EB","E2EE","C4F6","E2F1","B3B7","E2EC","C8EA","B1B0","BAEC","CFD2","E2F0","E2F2","CACB","C0D9","E2F4","E2F5","E2F3","B3CE","E2FB","E2FA","BCA7","E2FC","E2F7","E2FD","E2F8","C8D8","E2F6","E2F9","E3A2","E3A1","CBE1","E2FE","B0EB","E3A4","E3A3","BECC","E3A5","C1C3","E3A7","E3A6","E3A8","E2E8","E2EA","E3AA","E3A9","FBC1","BCA8","CEE9","BCD2","E3AB","B7B7","B5C0","B5A7","BBE3","CDB4","E3B1","E3B0","C1C4","E3AD","E3AF","BDCB","BFC0","E3AE","E3AC","C7AA","BECD","C9BC","BAD7","C5F8","E3B2","E3B3","E3C9","B6D8","CFBD","C1B5","E3B4","B2D2","C4F7","CAA1","FBC5","FBC7","E3B5","B5FA","E3B6","E3B8","E3B9","C7A9","E3BA","E3BB","E3BC","B6D9","B2D3","C6C5","BDA8","BBE4","E3BD","BDA9","B2CA","C9C3","E3BE","C8EB","C1C5","E3C1","E3C2","C7E9","BFC1","E3BF","C3E1","E3C0","BECE","B0DC","B5A9","E3C3","C4F8","E3C4","C0C7","CCAD","C9A3","E3C5","E3C6","C3D5","CEC7","E3C8","E3C7","BCEF","E3CA","B0F0","E3CD","E3CB","B2D4","B7CE","E3CC","B9C6","B9F2","CAE6","E3CE","CBD4","E3D0","C0D1","B1CF","B2BA","B0AC","E3CF","E3D1","E3D2","BEF7","E3D3","B3CF","E3D5","B7EA","B5E6","E3D6","B6F5","E3D7","C0FC","C6CD","C0E0","BAF5","E3D8","C3E2","C1EB","E3DA","E3DC","E3D9","E3DB","B7A2","E3DD","B7A6","B5E7","CDD2","E3DF","E3E0","B1AE","E3E3","B3F6","E3E2","E3E1","E3E5","E3DE","E3E6","CEA9","E3E7","E3E8","FBC8","D4F4","E3EA","E3E9","E3EB","E3EC","CEB5","E3ED","F0EF","BECF","E3EE","E3EF","BDD7","C6B8","E3F0","FBC9","C3A8","FBCB","E3F1","C3BC","E3F2","B6A5","D1BF","C3DD","BCB3","B4C8","E3F3","E4A2","E3F6","B5E8","E3F5","E4A4","E3F4","BED0","E3F8","E3F9","C5AB","E3FA","B3DE","BFDA","C9E4","E3FC","C2E8","E3F7","E3FB","E3FD","BAFB","E4A6","C9AE","C8A6","C5F9","B6DA","E4A5","E4A3","C8B5","E3FE","C3DE","C5FB","C5FA","BAF6","E4B8","E4A8","E4AA","E4AD","E4AE","E4AB","E4AC","E4A9","E4A7","E4A1","CACF","B2D5","E4B5","E4B2","E4B7","E4B6","C7F3","CCA7","BBBB","E4B0","E4B9","E4B4","E4B3","E4AF","FBCC","E4B1","B4C9","C3BD","C0FD","C8A2","E4BE","C8A4","C0E1","E4BB","C8CF","E4BF","CAD3","C3DB","E4BA","E4BC","E4BD","E4C0","BCC4","C6C6","E4C5","E4C4","E4C1","CFB6","E4CA","E4CE","E4CB","E4C7","E4C8","E4CD","E4C2","D2D5","E4C9","E4C3","E4CC","E4D2","B4CA","E4CF","E4D0","E4D1","E4D4","E4D3","C8F6","E4D5","CEFC","CAED","E4DA","E4D7","E4D6","C0D2","E4D9","E4DB","E4D8","E4DF","E4DC","E4DD","E4C6","E4DE","E4E0","E4E1","CAC6","E4E2","CCE2","B6CE","B7A9","E4E3","CAB4","BFE8","CCB0","E4E4","CEB3","C7F4","C1C6","C7B4","BDCD","B0C0","E4E9","E4E7","E4E5","B4A1","BED1","E4EA","E4E8","E4E6","E4EE","E4ED","E4EC","E4EB","E4EF","E4F0","C0BA","E4F1","E4F3","E4F2","B8D2","C1B8","E4F5","C5FC","E4F4","E4F6","CAB5","C1EC","B9C7","E4F7","CEC8","E4F9","E4FA","E4FB","E4FC","BBE5","E4FD","B7CF","B5EA","B5AA","E5A1","CCF3","B9C8","E4FE","E5A4","CCE6","C7BC","C9B3","BDE3","E5A3","BCD3","B9C9","BBE6","B5E9","CAB6","E5A2","C1C7","CBC2","BAF7","BBE7","C4DD","E5A7","CEDF","BAD9","E5A8","BFC2","E5AA","BED2","BAB0","E5A9","BDAA","B8BE","C1C8","E5A5","E5AB","FBCE","E5A6","B7D0","E5AE","E5B2","B7EB","E5AD","E5B6","FBCF","B9CA","CDED","B0BC","E5B3","B5EB","E5B0","E5B1","C5FD","E5AF","E5AC","B3A8","C0E4","B8A8","E5B8","E5B5","E5B7","E5B4","B7D1","C2B3","E5B9","C1EE","E5C6","FBD1","E5C2","E5BC","E5C0","BCFA","B0DD","E5BB","E5C3","E5C7","B9CB","CCD6","C4D6","E5BD","FBD0","E5C5","E5BA","C3BE","E5BF","B0BD","CCCA","E5BE","B6DB","C8EC","C1ED","CED0","BDEF","E5EE","FBD2","E5C8","C0FE","E5C4","E5C9","E5CB","C4F9","E5CE","E5CA","CAD4","B4CB","CCCB","B0DE","E5CD","CEFD","E5CC","B1EF","C6EC","E5CF","E5D6","E5D0","E5D7","E5D3","C7FB","BCCA","E5D5","E5D2","E5D8","E5D1","BDC4","CBA5","BDCC","E5D4","E5E0","E5DC","E5DF","E5DD","E5E1","E5DB","E5C1","C0D3","C8CB","E5DE","E5D9","C1A1","B7D2","BDAB","FBD3","BFA5","C1B6","E5E4","E5E6","E5E7","E5E3","E5E5","E5DA","E5E2","E5EA","E5E9","CBFA","B7AB","E5E8","E5EC","E5EB","E5EF","E5F1","BBBC","E5ED","E5F2","E5F3","F9A1","E5F4","E5FA","C5BB","E5F6","E5F5","E5F7","E5F8","E5F9","E5FB","E5FC","B4CC","E5FD","E5FE","E6A1","FBD4","E6A2","E6A3","E6A4","E6A5","E6A6","E6A8","E6A7","E6A9","E6AA","E6AB","E6AE","E6AC","E6AD","BAE1","B7D3","C3D6","C8B3","BDF0","C7CD","C8ED","E6AF","D8ED","E6B0","E6B2","CDE5","E6B1","E6B4","E6B3","CDD3","E6B5","C8FE","E6B6","E6B9","E6B8","E6B7","FBD5","E6BA","B7B2","C1A2","B5C1","E6BE","E6BB","E6BC","E6BF","E6C0","E6BD","B1A9","B2A7","E6C2","E6C3","E6C4","CDE2","BDAC","E6C6","E6C5","BFE9","E6C7","E6C8","E6C9","B4E5","B4CD","E6CA","E6CB","CBDD","CDE3","CDD4","CFB7","B9CD","E6CE","BCD4","E6CD","E6CF","BCA9","C2D1","E6D0","B9CC","CCD7","E6D1","E6D2","E6D3","E6D4","E6D5","BCAA","CCED","E6D7","C3BF","E6D6","E6D9","E6D8","E6DA","C0BB","E6DB","E6DC","CAB9","E6DD","C1EF","E6DE","E6DF","CEFE","E6E2","E6E1","E6E0","C4B0","E6E3","BFA6","E6E4","E6E5","CFB8","E6E6","E6E7","E6E9","E6E8","C8A5","C6F9","CFBE","C8A9","E6EB","BED3","C9AA","E6EC","E6EA","B4CE","B8D4","BBE8","C8EE","B8AA","CBC3","E6EF","E6ED","B9CE","B9CF","B0E9","BAE8","C7D9","B0DF","E6F4","C3C0","C7D8","C2DB","E6F6","E6F2","E6F5","E6F0","E6F3","CBA6","B8D5","B0FD","E6F1","E6F8","E6F9","C6B9","B6BB","E7A6","C7BD","BBE9","B6BC","C0C8","CFC6","CCAE","E6F7","C0D4","B5D3","E6FA","E6FC","E6FB","E6FD","C3A6","C7BE","C4B1","E7A3","E7A2","E6FE","BFD5","C9E5","E7A5","E7A4","B9D0","CFD3","E7B5","E7A9","E7AA","BCF0","E7A8","B9F8","E7A7","E7AB","C4B2","CAA2","C1A3","C2DC","E7AF","E7B0","E7AC","E7AD","E7AE","B9D1","E7B6","E7B2","C9E6","CBEC","C9A8","E7B1","E7B4","E7B3","CBC4","E7B7","E7B8","C1B7","E7B9","E7BB","E7BF","E7BC","E7BA","C7BF","E7BD","E7BE","B2B2","E7C5","E7C0","E7C1","E7C2","C2A1","E7C4","E7C3","E7C6","E7C7","E7C8","BFC3","B2E9","E7C9","CED7","BCAB","BDAD","BBEA","C3D7","E7CA","E7CB","B1B1","E7CC","E7CD","E7CE","E7CF","E7D0","B6BD","DAAA","E7D1","C0E5","E7D2","BCCB","E7D3","D0B0","E7D4","CADE","B4DC","C1A4","BDD8","C9F1","BDAE","E7D5","B9D2","E7D6","C8CC","E7E4","E7D8","C2C9","C7F5","B8BF","E7D7","C1A5","E7D9","C4FA","E7DB","E7DA","E7DD","E7DC","E7DE","E7E0","E7DF","B4CF","E7E1","E7E2","E7E3","BAB1","CEC9","E7E5","BFA7","B1F0","E7E6","E7E7","E7E8","B0F2","E7E9","E7EA","C9E7","BCC7","E7EC","B3A9","B0B2","E7EB","E7EE","C7CE","BFC4","B2D6","CBA7","B7DD","B6DC","E7ED","B2EA","B4A3","B1F1","E7F2","CEEA","C2DD","C9C4","E7FE","B2D7","E7FC","E7FA","E7F1","E7EF","E7F0","BCE3","B6EC","C3F7","C6D1","B1D1","E7F4","E7F3","E7F9","E7F5","E7F8","FBD7","CCD0","E7F7","B2D8","B3FD","E7FB","E7FD","B7D4","E8A3","E8AC","E8AD","B0AB","E8B4","B0F1","E8AB","E8AA","E8A5","E8A4","E8A2","E8A1","C3E3","C2FB","E8A7","E8A6","E8A9","C1F0","B7D5","B1C1","E8A8","B9D3","C1F1","FBD8","E8BA","E8BB","B2D9","B2AE","E8B8","FBD9","E8AE","E8B6","E8BD","E8B7","E8B5","E7F6","E8B3","E8AF","B4D0","E8B1","E8BC","E8B2","E8BE","E8B0","C7FC","CDE9","E8B9","E8CF","E8C7","BFFB","FBDA","B5C6","B6DD","E8C2","B2DB","BED4","E8C5","BADA","C5D1","E8CA","CAEE","E8C1","B2DA","B8D6","C9A9","E8CB","E8BF","FBDB","E8C8","E8D2","E8C3","E8C4","C6BA","E8C9","E8C6","CBA8","E8CC","B0E0","E8C0","E8CE","E8CD","C7EB","E8D4","E8DF","B3FE","E8E2","E8D0","E8D5","CDEE","E8DE","FBDC","CDD5","CEAA","C3F8","B3EB","C9F2","E8E4","C6A1","B0B1","E8DD","E8D9","C1F2","E8D3","E8DB","E8E0","C7AC","B0AA","E8D8","E8E1","C9F8","E8DC","E8D7","BED5","BDAF","BCAC","CCD8","C9C7","E8E7","E8F0","E8DA","B3F7","FBDD","BEF8","E8E5","E8EA","C1F3","E8E6","E8ED","C3DF","E8EE","CDD6","E8E3","B3B8","E8E9","E8EC","CCAC","E8EF","E8E8","E8EB","F9A5","CBA9","CFA1","E8F3","E8FA","E8F2","BCC3","E8D1","CACE","CCA2","E8F9","E8F8","E8F4","E8F5","B1B6","E8F7","E8F1","C4D5","E8F6","B0FE","C2A2","CAC3","E8FB","E9A1","C8D9","E8FE","BED6","BCC9","E9A3","B6BE","FBDE","E9A4","C9F9","E8FD","FBDF","E8D6","E8FC","CFCF","C6A2","C9F3","FBE0","E9AB","E9B1","E9B2","E9A5","C7F6","E9AF","E9A7","E9A9","E9B3","E9A8","E9AC","B1F2","C6E5","E9AD","E9B0","E9A6","C1A6","E9AA","BBA7","BFC5","B7B0","CCF4","CCF9","BDF2","FBE2","E9B7","E9B5","CFCE","E9B4","CDF5","E9B6","E9B8","E9B9","E9BC","E9BA","C6A3","E9BB","C8CD","E9AE","BDF3","E9BD","E9C2","C1F4","E9C1","E9A2","E9C3","C1C9","E9BE","E9C0","E9BF","DDB1","DDA2","E9C5","E9C4","CDF6","E2BC","E9C6","E9C7","E9C8","B8D7","B5D4","E9CA","D1DD","B5F5","CEBA","B6F3","E9CB","E9CC","C3EE","E9CD","C6FA","B0BA","B2E3","E9D2","E9D3","E9CE","BBBD","E9CF","C7C2","E9D0","E9D1","E9DB","E9D5","E9D8","E9D4","E9D6","E9D7","BCD8","E9D9","C3C1","B7D6","B3C2","E9DC","B3BF","E9E1","E9DD","E9E0","C8BA","E9DE","E9DF","C9C8","C8DA","E9E2","C2FD","E9EC","E9E8","B2EB","E9E6","CBAA","E9E7","E9E4","E9E5","E9EA","E9ED","E9EB","E9E9","E9E3","C3D8","E9F4","CCAA","E9F2","E9F3","E9EE","E9F0","E9F1","E9EF","C0E6","CFB9","E9F8","E9F9","EAA1","BFAA","E9FB","E9FE","E9F6","E9F5","EAA2","B2DC","E9FC","EAA3","E9FD","E9FA","C4B3","E9F7","C7E8","EAA7","CDBB","EAA6","EAA5","EAAE","EAA8","EAB0","CDE6","EAB3","EAAA","EAAB","EAAF","EAB2","EAB1","EAA9","EAAC","EABD","EAB6","EAB4","EAB5","EABA","EABB","B3AA","B5C2","EAB9","EAA4","FBE5","EAB8","EABC","EAB7","EABE","EAC0","EABF","EAC2","EAC1","E9DA","EAC6","EAC3","EAC4","EAC5","EAC7","B7EC","EAC9","EAC8","BDB0","B9D4","DEA7","EACA","BDD1","B3B9","EACB","B1D2","BED7","EACC","B9D5","EACD","B0E1","C9BD","EACE","BFEA","EAD5","EAD2","C3EF","EAD3","EAD0","B6DE","EACF","EAD6","B7B6","C2DE","EADC","EAD8","C2B5","EAD7","EADA","EAD1","EADB","EADD","C8EF","EAD9","EADE","EAE0","B8D3","EAD4","B0C1","EADF","BADB","CEF6","EAE1","EAE2","C1F5","CEA2","EAE3","CDB5","EAE4","EAE5","CAE4","EAE6","BAC0","CEA3","EAEB","EAEC","BED8","EAEA","FBE6","CDE7","EAE7","EAE9","C0BD","BFFE","EAE8","EAED","CAA3","EAEF","EAEE","B3EC","CBAB","EAF0","F9A2","EAFC","EAF2","EAF3","EAF4","EAF5","EAF9","EAFA","EAF8","EAF6","EAF1","EAF7","EAFB","F0B7","B2A8","EAFE","B6DF","EAFD","EBA2","EBA1","EBA4","EBA3","EBA5","BDB1","EBA6","EBA7","EBA8","C0BE","CDD7","EBA9","CAA4","C7C6","EBAA","EBAB","B8AB","B5AC","EBAC","BBEB","C7C1","EBAD","B3D0","EBAE","EBB0","CDF7","EBAF","BFC6","EBB1","EBB2","EBB3","B4D1","EBB4","EBB5","EBB6","EBB7","B3D1","EBB8","EBB9","EBBA","B2F2","BFA8","EBBB","EBBC","EBBD","B8C0","C4FB","EBBE","B7D7","BFD6","EBC1","C6A4","EBC0","FBE7","B7B1","EBBF","C2F7","B5AD","EBC2","EBC3","BED9","B7ED","EBC4","CBAC","C0DF","B5F6","CCF5","C1CA","EBC5","FBE8","BFC7","C3F0","BEDA","EBC6","EBC9","EBCA","BABE","C2C2","EBC8","BEDB","C9BE","EBC7","BBEC","B1D3","EBCE","B7D8","BBEE","BBED","CFCD","EBCD","EBCC","C1A7","B5CD","CFC3","B3BA","BEDC","FBE9","EBCB","EBD0","EBD1","EBCF","B8D8","CDC0","BBEF","C7A7","EBD4","C0C0","C3C2","CDB6","EBD7","B8EC","C0BF","EBD3","EBD8","B8ED","EBD5","EBD6","FBEA","EBD2","C0E2","C6C9","C3AF","B2DD","C8F0","B5C3","FBEB","C4B4","EBDB","EBD9","C3CC","C0C1","B4D2","EBDA","BFDB","CECA","CFC0","EBDC","EBE7","C4B5","EBE6","FBEC","EBE3","EBEB","EBE4","EBE0","C4FC","EBDF","EBDD","CDA1","BBF0","EBE1","EBDE","FBEE","EBE5","BDF4","B8C1","C2FA","CBC5","B1DA","B0E2","C6A5","EBE9","EBE8","C6E6","EBED","EBE2","EBEC","EBEE","B8AC","EBEA","B9D6","BCD5","EBEF","CDD8","EBF2","EBF5","EBF3","C9B5","EBF0","B6E0","EBF4","EBF6","EBFA","EBF7","EBF9","EBF8","FBEF","EBFB","BCB1","EBFD","EBFC","C9E8","ECA1","B7D9","EBFE","ECA2","ECA3","B5C4","E6C1","BEF9","ECA4","B8EE","ECA5","FBF0","ECA6","BBBE","DACE","ECA7","ECA8","BDB2","ECA9","ECAA","ECAB","ECAC","ECAD","C3AB","ECAE","ECB0","ECAF","C6A6","ECB1","CBAD","ECB2","ECB3","ECB4","ECB5","C6DA","BEDD","ECB6","B9EB","D0AE","ECB7","ECB8","C9BF","ECB9","ECC1","ECBA","ECBC","ECBB","ECBD","CBC6","ECBE","ECBF","ECC0","ECC2","B3AD","C4E7","C9E9","BAE2","B9D7","C9CF","B2DF","C8CE","ECC5","B4D3","C0D5","ECC4","ECC9","C3F9","CCE3","ECC7","ECC8","B5AE","ECCA","C7E3","C2DF","C8F1","C5BD","ECC6","CBC7","B2EC","ECCC","CFA8","C4C2","CFC5","BBF1","ECCB","C2B1","ECDC","C1A8","C6F8","C9D0","ECCF","BBBF","BBF2","BEDE","C7E5","B8AD","ECCE","ECCD","C9EA","BCC1","C5D2","FBF1","FBF2","ECD1","ECD2","B9D8","ECD0","ECD3","ECD4","ECD6","C2A3","ECD5","B4E6","ECD8","ECD7","ECD9","FBF3","ECDB","ECDD","ECDE","C0D6","BCCF","ECDF","B3D2","ECE0","C1F6","ECE1","ECE2","C9EB","FBF4","B5AF","ECE3","C4B6","B1DB","ECE4","BCF1","BFF6","C2AD","ECE7","ECE6","ECE5","ECED","ECEB","ECE8","ECEA","ECE9","ECEC","B5F7","ECF0","C0D7","ECF1","B8D9","ECEE","ECEF","CFA9","C4B7","C1A9","ECF2","ECF5","ECF3","ECF4","CDD9","C6A7","ECF8","ECF6","ECF7","ECF9","EDA9","ECFC","ECFD","ECFB","ECFA","C4FD","EDA1","EDA5","EDA2","ECFE","EDA3","EDA4","EDAB","EDA6","C0D8","EDA8","EDAA","EDA7","EDAD","BDB3","EDAC","EDAE","EDAF","EDB2","EDB1","EDB0","EDB4","EDB3","CCF6","EDB6","EDB5","EDB7","EDB8","EDBA","EDB9","BFC8","EDBB","B6ED","EDBC","EDBE","EDBF","EDC0","EDBD","EDC1","BCD6","EDC2","B5B0","B7B3","FBF6","B8AE","EDC3","C6F0","C5BE","EDC4","EDC7","BCB4","EDC6","EDC5","B7DA","EDC8","B3D3","EDCA","BADC","EDC9","EDD2","EDCC","EDCE","CAE5","EDCB","EDCD","EDD1","EDCF","B5B1","EDD0","EDD3","C7DA","CED8","BDB4","EDD4","CDA2","EDD6","EDD5","EDD9","CDC1","EDD8","B3ED","EDD7","EDDC","EDDB","EDDA","C5B2","EDDD","EDDE","EDDF","B9EC","B7A5","EDE0","EDE1","EDE2","BFC9","EDE3","BCAD","EDE4","EDE5","D2A1","D1FE","EDE6","E5F0","EDE7","C3A4","BFAB","C7C0","EDE8","CAD5","C4D4","B9FE","C3A9","B1AA","CBF8","BFD7","B7DE","B6E1","CAD6","EDE9","EDEB","EDEA","B2E0","C6F6","EDEC","C7F7","C5B3","EDED","BDD2","EDEF","CCC2","EDFE","EDF1","EDF2","C4C9","C2E0","C1F7","C6A8","EDF0","B5D5","EDF9","EDF6","EEA5","C6A9","C3E0","EDF3","C4FE","C5D3","EDF4","EDF8","BFE0","C7E7","C4CC","C0C2","EDF7","C2AE","C2A4","EDF5","B0A9","CFA2","EDFA","C2E1","BDB5","BFCA","EDFC","EDFB","B0EF","EDFD","C9AF","EEA7","C6DB","BFEB","C3D9","B6F8","EEA6","CDB7","B1BF","CAD7","B2E1","EEA1","EEA2","EEA3","EEA4","C6BB","C3A3","B0E3","EEA8","EEA9","F4A3","C2BD","EEAA","B1F3","C1CC","B8AF","CDDA","FBF9","EEAB","C5AC","C1F8","BCD7","EEAC","EEAF","BDE5","EEAD","C1AB","C1AA","B0E4","CECB","EEB1","C8F2","EEB3","EEB2","EEB0","E3E4","B4D4","EDEE","EEB5","EEB4","EEB6","CDB8","C6E1","CBAE","EEB7","BCD9","EEB8","EEB9","EEBA","C5A1","B0EA","B9D9","CFBA","EEBE","FBFA","B7B4","EEBB","EEBC","C9F4","B3D4","CDB9","B6BF","C5D4","EEBF","EEC0","FBFC","EEC1","FBFD","C5A2","EEC3","EEC2","C6D3","EEC4","BDB6","BCE0","C7DB","C3F1","BCF2","BFEC","EEC5","EEC6","BFDD","EEC7","EEC8","EEC9","CDEF","BDB7","EECB","EECA","B9DA","B9F3","BBC0","EECE","BDE6","EECD","EECC","C2E9","B8EF","C0C3","C8B0","BDB9","EECF","BEDF","EED2","EED0","EED1","EED4","EED3","BEFA","EED5","EED6","EED7","C8D0","BAD3","BCE1","EED8","EED9","CEA4","BDC5","CCEE","CECC","EEDA","B6E2","EEDB","FCA1","C5A3","FBFE","EEDE","B3F8","BFCB","FCA2","EEDC","EEDD","C4E0","FCA5","FCA6","CBD5","B6FC","FCA3","FCA4","EEE0","EEE1","EEDF","EEE3","FCA7","FCA9","C6DF","B3C3","FCA8","EEE7","EEE4","EEE6","EEE2","EFCF","EEE5","CEEB","B8DA","FCB0","FCAA","FCAC","EEEF","FCAB","C5B4","EEEA","EEED","EEEB","EEF0","FCAD","EEF1","FCAF","EEE9","FCAE","EEF6","B1F4","EEE8","C8AD","EEEC","BEE0","FCB1","B9DB","FCB3","FCB4","CBC8","B6E4","BDC6","C6BC","F9A4","C1AD","EEF4","EEEE","EEF3","CCC3","C4B8","EEF5","EEF2","FCB2","C1AC","EEF9","EEF8","EEF7","FCB8","CBAF","FCBC","FCBA","FCB6","FCB7","FCBB","BDFB","FCB5","EEFA","CADF","B1D4","C9C6","C3F2","B5F8","F9AA","EEFC","FCBF","B9DD","FCC2","FCC4","BBAC","EEFB","BFED","BFEE","EFA1","EFA3","FCC3","FCC1","BEFB","FCBE","EFA2","EFA4","FCBD","B6D3","C9C5","BCE2","CFA3","EEFE","BAF8","CFBF","EFA6","EFA5","EFA7","EEFD","F9A3","C6E9","C5D5","C4D7","EFAC","FCC6","C3C3","EFA8","EFA9","B7AD","EFAB","FCC5","B8B0","EFAA","BEE1","B3F9","EFB0","BABF","C1F9","C4CA","FCC7","B3BB","EFAE","EFAF","C4C3","EFAD","EFB1","EFB7","FCC8","EFBA","EFB9","C5AD","EFB2","EFB3","EFB6","EFB8","FCC9","B6C0","EFBB","EFB5","EFB4","FCCA","EFBF","EFC0","EFC1","EFBE","EFBD","BEE2","C6AA","EFBC","EFC5","EFC3","FCCB","EFC4","EFC2","C2F8","EFC6","EFC7","EFC9","FCCC","FCCD","B4D5","EFC8","CCFA","EFD4","EFCA","EFCD","EFCB","EFCC","EFCE","EFD0","EFD1","EFD2","EFD5","EFD3","EFD6","EFD8","EFD7","C4B9","CCE7","EFD9","C1AE","EFDA","CAC4","EFDB","B3AB","B1BC","B4D7","FCCE","B4D6","EFDC","EFDD","EFDE","EFDF","EFE0","B4D8","B3D5","B9DE","C8B6","EFE2","EFE1","EFE3","B1DC","EFE6","EFE5","EFE4","EFE7","EFEA","B0C7","EFE8","EFEC","EFEB","EFEE","EFED","EFEF","C6AE","EFF0","EFF1","EFF3","EFF2","C9EC","EFF4","EFF5","BAE5","EFF6","EFF7","CBC9","C1CB","B0A4","C2CB","EFF8","C9ED","EFFB","EFF9","B9DF","EFFA","B8C2","CAC5","EFFD","F0A1","EFFE","F0A2","B1A1","BFD8","BDFC","B4D9","F0A3","C7E6","F0A5","B1A2","F0A4","C4C4","CECD","C6AB","EFFC","CEA6","B8B1","CDDB","B6F9","CEB4","B7A8","C2E2","E7A1","F0A6","B3AC","BFEF","B3D6","F0A8","F0A9","F0A7","B7E4","BADD","BEE3","FCD1","B1A3","CED9","F0AB","EEAE","F0AA","FCD2","F0AE","F0AC","F0AD","F0AF","F0B0","CEEC","F0B1","F0B2","C0C9","C8BB","BFFD","B4E7","CDBA","B2ED","BDB8","B8DB","F0B5","F0B4","BBF3","F0B6","F0B3","BBA8","F0BA","EAAD","D2D6","BFF7","F0B8","CEA5","C6F1","B1AB","C0E3","BCB6","CAB7","B1C0","CEED","CDEB","F0BB","C5C5","BCFB","F0BC","F0BD","BFCC","F0BE","CEEE","F0B9","F0C0","F0C2","F0C1","F0BF","F0C3","F0C4","C1FA","B2E2","F0C5","CCB8","F0C6","F0C7","CFAA","FCD3","DBB1","F0C8","FCD4","F0C9","F0CA","F0CE","FCD5","F0CB","F0CC","F0CD","F0CF","FCD6","FCD7","FCD8","C0C4","FCD9","CCF7","C0C5","F0D0","C8F3","F0D1","F3D3","CCCC","F0D2","F0D3","F0D4","B3D7","F0D6","BFD9","F0D7","B7A4","F0D8","F0DC","F0DA","F0DB","B3F3","F0D9","F0DD","F0DE","B0C8","F0DF","F0E0","BEE4","F0E1","B5C7","F0E4","F0E3","F0E2","EBF1","CADC","F0E5","F0E6","F0E7","F0E8","F0E9","F0EA","B4DA","F0EB","F0EC","C7A3","F0EE","B2BB","F0F1","F0F0","B1A4","B6C1","CAC7","C4BA","BAA2","B9E0","BDE7","BFDC","F0F3","F0F2","CDC2","B4E8","C8D2","C6DC","BFFC","CECE","B7DB","F0F6","F0F5","CBCB","C6AC","B1D0","F0F7","F0F4","C9D1","CDEA","F0F8","F0F9","F0FB","C2EA","B3DB","B3DC","F0FA","B4E9","B8B2","FCDA","B4EA","C5BF","CEE0","FCDB","B8DC","F0FC","F0FD","F0FE","F1A1","F1A3","F1A2","C9F7","F1A4","F1A5","F1A6","F1A7","F1A9","F1A8","F1AA","C8F4","E6CC","BFA9","B5B2","F1AB","F1AC","D2AC","DDBB","C8D3","B0FB","B0BB","BBF4","CBB0","BEFE","F1AD","CCDF","F1AE","CDDC","B1C2","BBC1","F1AF","B2EE","F1B0","F1B1","F1B3","F1B4","F1B6","F1B2","F1B5","FCDE","B4DB","F1B7","F1B8","F1B9","F1BA","F1BB","F1BD","F1BC","F1BF","F1C2","F1BE","F1C0","F1C1","F1C3","B6C2","BCF3","F1C4","F1C5","B9E1","FCE0","F1C6","B3BE","C7CF","F1C7","F1C8","C3DA","C6EB","F1C9","C7FD","C2CC","B1D8","B6EE","B6EF","C3F3","F1CE","B6F0","B2EF","F1CD","F1CB","F1CC","F1CA","F1D8","F1CF","F1D0","F1D1","F1D2","F1D4","F1D3","BDD9","F1D5","F1D7","B5B3","F1D6","C1FB","B8B3","F1D9","C2CD","F1DA","C6AD","F1DB","F1E0","F1DE","F1DD","F1DF","F1DC","F1E2","FCE1","F1E1","F1E4","B6C3","F1E3","F1E5","F1E6","F1E8","F1E7","F1E9","F1EB","F1EA","B9FC","F1EC","F1ED","B3BC","F1EE","F1EF","BFF1","F1F0","F1F1","F1F2","F1F3","B9E2","FCE2","FCE3","F1F4","F1F5","F1F6","F1F7","F1F8","C8B1","F1FA","C9A6","F1FB","F1F9","F1FD","F1FC","F1FE","F2A1","F2A2","F2A3","F2A4","F2A5","F2A6","F2A7","F2A8","F2A9","F2AA","F2AB","F2AC","F2AD","F2AE","DDB5","F2AF","E4F8","B5B4","B3A1","BAB2","F2B1","F2B0","CCA5","F2B3","F2B4","F2B2","F2B5","CBE2","F2B6","B5FB","CFA5","FCE5","F2B7","FCE4","F2B9","B0BE","FCE6","F2BA","CAAB","F2B8","F2BB","F2BC","F2BD","F2BE","F2BF","CBEE","BBAD","BAFA","C1AF","FCE7","F2C0","F2C3","FCE8","F2C1","F2C4","B8F1","F2C2","F2C5","F2C6","F2C7","F2CB","BBAA","C2E4","F2CC","F2C9","F2C8","F2CA","B7DF","F2D0","F2CF","F2CE","B0B3","FCE9","F2DA","F2D6","F2D7","F2D3","F2D9","F2D5","B3E2","CFCC","F2D8","F2D4","F2D2","F2D1","F2DC","F2DF","F2DE","F2DD","C9C9","F2DB","B0F3","F2E0","F2E2","B3EF","F2CD","B1B7","F2E4","F2E3","F2E1","C3AD","CBF0","CEDA","F2E5","F2E6","F2E7","F2E8","F2E9","C4BB","F2EA","C8B7","F2EF","F2EB","F2EC","CBB1","CCC4","C6D0","F2F0","F2F1","C6BE","F2EE","F2ED","B2AA","F2F9","F2F8","B1F5","F2F6","F2F5","F2F3","B3FB","F2F2","BCB2","B2A9","B9E3","F2FC","F2FB","F2FA","F2F7","F2FD","F2FE","F3A5","F3A4","F3A6","B1AD","F3A1","F3A2","B9F4","CCB9","F3A3","FCEB","CBB2","F3AB","FCEA","F3A7","F3AC","F3A9","F3A8","B7DC","F3AD","F3AE","F3AF","F3AA","F2F4","F3B0","C4E1","F3B4","F3B5","F3B3","F3B2","F3B8","F3B1","F3B6","F3B7","F3BA","F3B9","F3BC","F3BD","F3BE","CFC9","F3BB","C2EB","BAED","F3BF","FCED","F3C0","F3C1","F3C2","F3C3","B8B4","F3C4","F3C5","BCAF","F3C6","F3C7","F3C8","F3C9","F3CC","F3CA","CFBC","F3CB","CEEF","F3CD","CEDB","F3CE","C7FE","F3CF","F3D1","F3D2","F3D0","B9ED","CCCD","CBE3","D6F7","DDE0","CBFB","B2AB","F3D4","B5D0","F3D5","F3D6","F3D7","FCEE","B9F5","F3D8","E0D4","CCDB","C2E3","F3D9","F3DB","F3DA","F3DC","F3DD","F3DE","F3DF","F3E0","F3E1","F3E2","F3E3","F3E4","F3E5","F3E6","F3E7","F3E8","C5A4","B8DD","F3EA","C1CD","F3EB","F3EC","C9A1","F3ED","F3EE","E3B7","ECDA","F0ED","F3EF","F3F0","F3F2","F3F3","F3F4","CEF0","F3F1","F3F5","F3F6","F3F8","F3F7","F3FA","F3FB","F3F9","CEB6","F3FC","F3FD","E3D4","F3FE","FAC6","FCCF","F9D4","F9DF","F9E0","F9F5","FABE","FACE","FAD0","FAFB","FBA3","FBBA","FBC2","FBC3","FBC4","FBC6","FBCA","FBCD","FBD6","FBE1","FBE3","FBE4","FBED","FBF5","FBF7","FBF8","FBFB","FCB9","FCC0","FCD0","FCDC","FCDD","FCDF","FCEC","A1AA","FCFE","A1F4","A1F0","A1F3","A1F5","FCFD","A1CA","A1CB","A1F6","A1DC","A1A4","A1DD","A1A5","A1BF","A3B0","A3B1","A3B2","A3B3","A3B4","A3B5","A3B6","A3B7","A3B8","A3B9","A1A7","A1A8","A1E3","A1E1","A1E4","A1A9","A1F7","A3C1","A3C2","A3C3","A3C4","A3C5","A3C6","A3C7","A3C8","A3C9","A3CA","A3CB","A3CC","A3CD","A3CE","A3CF","A3D0","A3D1","A3D2","A3D3","A3D4","A3D5","A3D6","A3D7","A3D8","A3D9","A3DA","A1CE","A1C0","A1CF","A1B0","A1B2","A1AE","A3E1","A3E2","A3E3","A3E4","A3E5","A3E6","A3E7","A3E8","A3E9","A3EA","A3EB","A3EC","A3ED","A3EE","A3EF","A3F0","A3F1","A3F2","A3F3","A3F4","A3F5","A3F6","A3F7","A3F8","A3F9","A3FA","A1D0","A1C3","A1D1","A1C1","8EA1","8EA2","8EA3","8EA4","8EA5","8EA6","8EA7","8EA8","8EA9","8EAA","8EAB","8EAC","8EAD","8EAE","8EAF","8EB0","8EB1","8EB2","8EB3","8EB4","8EB5","8EB6","8EB7","8EB8","8EB9","8EBA","8EBB","8EBC","8EBD","8EBE","8EBF","8EC0","8EC1","8EC2","8EC3","8EC4","8EC5","8EC6","8EC7","8EC8","8EC9","8ECA","8ECB","8ECC","8ECD","8ECE","8ECF","8ED0","8ED1","8ED2","8ED3","8ED4","8ED5","8ED6","8ED7","8ED8","8ED9","8EDA","8EDB","8EDC","8EDD","8EDE","8EDF","A1F1","A1F2","A2CC","A1B1","FCFC","A1EF","8FA2AF","8FA2B0","8FA2B1","8FA2B2","8FA2B3","8FA2B4","8FA2B5","8FA2B6","8FA2B7","8FA2B8","8FA2B9","8FA2C2","8FA2C3","8FA2C4","8FA2EB","8FA2EC","8FA2ED","8FA2EE","8FA2EF","8FA2F0","8FA2F1","8FA6E1","8FA6E2","8FA6E3","8FA6E4","8FA6E5","8FA6E7","8FA6E9","8FA6EA","8FA6EC","8FA6F1","8FA6F2","8FA6F3","8FA6F4","8FA6F5","8FA6F6","8FA6F7","8FA6F8","8FA6F9","8FA6FA","8FA6FB","8FA6FC","8FA7C2","8FA7C3","8FA7C4","8FA7C5","8FA7C6","8FA7C7","8FA7C8","8FA7C9","8FA7CA","8FA7CB","8FA7CC","8FA7CD","8FA7CE","8FA7F2","8FA7F3","8FA7F4","8FA7F5","8FA7F6","8FA7F7","8FA7F8","8FA7F9","8FA7FA","8FA7FB","8FA7FC","8FA7FD","8FA7FE","8FA9A1","8FA9A2","8FA9A4","8FA9A6","8FA9A8","8FA9A9","8FA9AB","8FA9AC","8FA9AD","8FA9AF","8FA9B0","8FA9C1","8FA9C2","8FA9C3","8FA9C4","8FA9C5","8FA9C6","8FA9C7","8FA9C8","8FA9C9","8FA9CA","8FA9CB","8FA9CC","8FA9CD","8FA9CE","8FA9CF","8FA9D0","8FAAA1","8FAAA2","8FAAA3","8FAAA4","8FAAA5","8FAAA6","8FAAA7","8FAAA8","8FAAA9","8FAAAA","8FAAAB","8FAAAC","8FAAAD","8FAAAE","8FAAAF","8FAAB0","8FAAB1","8FAAB2","8FAAB3","8FAAB4","8FAAB5","8FAAB6","8FAAB7","8FAAB8","8FAABA","8FAABB","8FAABC","8FAABD","8FAABE","8FAABF","8FAAC0","8FAAC1","8FAAC2","8FAAC3","8FAAC4","8FAAC5","8FAAC6","8FAAC7","8FAAC8","8FAAC9","8FAACA","8FAACB","8FAACC","8FAACD","8FAACE","8FAACF","8FAAD0","8FAAD1","8FAAD2","8FAAD3","8FAAD4","8FAAD5","8FAAD6","8FAAD7","8FAAD8","8FAAD9","8FAADA","8FAADB","8FAADC","8FAADD","8FAADE","8FAADF","8FAAE0","8FAAE1","8FAAE2","8FAAE3","8FAAE4","8FAAE5","8FAAE6","8FAAE7","8FAAE8","8FAAE9","8FAAEA","8FAAEB","8FAAEC","8FAAED","8FAAEE","8FAAEF","8FAAF0","8FAAF1","8FAAF2","8FAAF3","8FAAF4","8FAAF5","8FAAF6","8FAAF7","8FABA1","8FABA2","8FABA3","8FABA4","8FABA5","8FABA6","8FABA7","8FABA8","8FABA9","8FABAA","8FABAB","8FABAC","8FABAD","8FABAE","8FABAF","8FABB0","8FABB1","8FABB2","8FABB3","8FABB4","8FABB5","8FABB6","8FABB7","8FABB8","8FABB9","8FABBA","8FABBB","8FABBD","8FABBE","8FABBF","8FABC0","8FABC1","8FABC2","8FABC3","8FABC5","8FABC6","8FABC7","8FABC8","8FABC9","8FABCA","8FABCB","8FABCC","8FABCD","8FABCE","8FABCF","8FABD0","8FABD1","8FABD2","8FABD3","8FABD4","8FABD5","8FABD6","8FABD7","8FABD8","8FABD9","8FABDA","8FABDB","8FABDC","8FABDD","8FABDE","8FABDF","8FABE0","8FABE1","8FABE2","8FABE3","8FABE4","8FABE5","8FABE6","8FABE7","8FABE8","8FABE9","8FABEA","8FABEB","8FABEC","8FABED","8FABEE","8FABEF","8FABF0","8FABF1","8FABF2","8FABF3","8FABF4","8FABF5","8FABF6","8FABF7","8FB0A1","8FB0A2","8FB0A3","8FB0A4","8FB0A5","8FB0A6","8FB0A7","8FB0A8","8FB0A9","8FB0AA","8FB0AB","8FB0AC","8FB0AD","8FB0AE","8FB0AF","8FB0B0","8FB0B1","8FB0B2","8FB0B3","8FB0B4","8FB0B5","8FB0B6","8FB0B7","8FB0B8","8FB0B9","8FB0BA","8FB0BB","8FB0BC","8FB0BD","8FB0BE","8FB0BF","8FB0C0","8FB0C1","8FB0C2","8FB0C3","8FB0C4","8FB0C5","8FB0C6","8FB0C7","8FB0C8","8FB0C9","8FB0CA","8FB0CB","8FB0CC","8FB0CD","8FB0CE","8FB0CF","8FB0D0","8FB0D1","8FB0D2","8FB0D3","8FB0D4","8FB0D5","8FB0D6","8FB0D7","8FB0D8","8FB0D9","8FB0DA","8FB0DB","8FB0DC","8FB0DD","8FB0DE","8FB0DF","8FB0E0","8FB0E1","8FB0E2","8FB0E3","8FB0E4","8FB0E5","8FB0E6","8FB0E7","8FB0E8","8FB0E9","8FB0EA","8FB0EB","8FB0EC","8FB0ED","8FB0EE","8FB0EF","8FB0F0","8FB0F1","8FB0F2","8FB0F3","8FB0F4","8FB0F5","8FB0F6","8FB0F7","8FB0F8","8FB0F9","8FB0FA","8FB0FB","8FB0FC","8FB0FD","8FB0FE","8FB1A1","8FB1A2","8FB1A3","8FB1A4","8FB1A5","8FB1A6","8FB1A7","8FB1A8","8FB1A9","8FB1AA","8FB1AB","8FB1AC","8FB1AD","8FB1AE","8FB1AF","8FB1B0","8FB1B1","8FB1B2","8FB1B3","8FB1B4","8FB1B5","8FB1B6","8FB1B7","8FB1B8","8FB1B9","8FB1BA","8FB1BB","8FB1BC","8FB1BD","8FB1BE","8FB1BF","8FB1C0","8FB1C1","8FB1C2","8FB1C3","8FB1C4","8FB1C5","8FB1C6","8FB1C7","8FB1C8","8FB1C9","8FB1CA","8FB1CB","8FB1CC","8FB1CD","8FB1CE","8FB1CF","8FB1D0","8FB1D1","8FB1D2","8FB1D3","8FB1D4","8FB1D5","8FB1D6","8FB1D7","8FB1D8","8FB1D9","8FB1DA","8FB1DB","8FB1DC","8FB1DD","8FB1DE","8FB1DF","8FB1E0","8FB1E1","8FB1E2","8FB1E3","8FB1E4","8FB1E5","8FB1E6","8FB1E7","8FB1E8","8FB1E9","8FB1EA","8FB1EB","8FB1EC","8FB1ED","8FB1EE","8FB1EF","8FB1F0","8FB1F1","8FB1F2","8FB1F3","8FB1F4","8FB1F5","8FB1F6","8FB1F7","8FB1F8","8FB1F9","8FB1FA","8FB1FB","8FB1FC","8FB1FD","8FB1FE","8FB2A1","8FB2A2","8FB2A3","8FB2A4","8FB2A5","8FB2A6","8FB2A7","8FB2A8","8FB2A9","8FB2AA","8FB2AB","8FB2AC","8FB2AD","8FB2AE","8FB2AF","8FB2B0","8FB2B1","8FB2B2","8FB2B3","8FB2B4","8FB2B5","8FB2B6","8FB2B7","8FB2B8","8FB2B9","8FB2BA","8FB2BB","8FB2BC","8FB2BD","8FB2BE","8FB2BF","8FB2C0","8FB2C1","8FB2C2","8FB2C3","8FB2C4","8FB2C5","8FB2C6","8FB2C7","8FB2C8","8FB2C9","8FB2CA","8FB2CB","8FB2CC","8FB2CD","8FB2CE","8FB2CF","8FB2D0","8FB2D1","8FB2D2","8FB2D3","8FB2D4","8FB2D5","8FB2D6","8FB2D7","8FB2D8","8FB2D9","8FB2DA","8FB2DB","8FB2DC","8FB2DD","8FB2DE","8FB2DF","8FB2E0","8FB2E1","8FB2E2","8FB2E3","8FB2E4","8FB2E5","8FB2E6","8FB2E7","8FB2E8","8FB2E9","8FB2EA","8FB2EB","8FB2EC","8FB2ED","8FB2EE","8FB2EF","8FB2F0","8FB2F1","8FB2F2","8FB2F3","8FB2F4","8FB2F5","8FB2F6","8FB2F7","8FB2F8","8FB2F9","8FB2FA","8FB2FB","8FB2FC","8FB2FD","8FB2FE","8FB3A1","8FB3A2","8FB3A3","8FB3A4","8FB3A5","8FB3A6","8FB3A7","8FB3A8","8FB3A9","8FB3AA","8FB3AB","8FB3AC","8FB3AD","8FB3AE","8FB3AF","8FB3B0","8FB3B1","8FB3B2","8FB3B3","8FB3B4","8FB3B5","8FB3B6","8FB3B7","8FB3B8","8FB3B9","8FB3BA","8FB3BB","8FB3BC","8FB3BD","8FB3BE","8FB3BF","8FB3C0","8FB3C1","8FB3C2","8FB3C3","8FB3C4","8FB3C5","8FB3C6","8FB3C7","8FB3C8","8FB3C9","8FB3CA","8FB3CB","8FB3CC","8FB3CD","8FB3CE","8FB3CF","8FB3D0","8FB3D1","8FB3D2","8FB3D3","8FB3D4","8FB3D5","8FB3D6","8FB3D7","8FB3D8","8FB3D9","8FB3DA","8FB3DB","8FB3DC","8FB3DD","8FB3DE","8FB3DF","8FB3E0","8FB3E1","8FB3E2","8FB3E3","8FB3E4","8FB3E5","8FB3E6","8FB3E7","8FB3E8","8FB3E9","8FB3EA","8FB3EB","8FB3EC","8FB3ED","8FB3EE","8FB3EF","8FB3F0","8FB3F1","8FB3F2","8FB3F3","8FB3F4","8FB3F5","8FB3F6","8FB3F7","8FB3F8","8FB3F9","8FB3FA","8FB3FB","8FB3FC","8FB3FD","8FB3FE","8FB4A1","8FB4A2","8FB4A3","8FB4A4","8FB4A5","8FB4A6","8FB4A7","8FB4A8","8FB4A9","8FB4AA","8FB4AB","8FB4AC","8FB4AD","8FB4AE","8FB4AF","8FB4B0","8FB4B1","8FB4B2","8FB4B3","8FB4B4","8FB4B5","8FB4B6","8FB4B7","8FB4B8","8FB4B9","8FB4BA","8FB4BB","8FB4BC","8FB4BD","8FB4BE","8FB4BF","8FB4C0","8FB4C1","8FB4C2","8FB4C3","8FB4C4","8FB4C5","8FB4C6","8FB4C7","8FB4C8","8FB4C9","8FB4CA","8FB4CB","8FB4CC","8FB4CD","8FB4CE","8FB4CF","8FB4D0","8FB4D1","8FB4D2","8FB4D3","8FB4D4","8FB4D5","8FB4D6","8FB4D7","8FB4D8","8FB4D9","8FB4DA","8FB4DB","8FB4DC","8FB4DD","8FB4DE","8FB4DF","8FB4E0","8FB4E1","8FB4E2","8FB4E3","8FB4E4","8FB4E5","8FB4E6","8FB4E7","8FB4E8","8FB4E9","8FB4EA","8FB4EB","8FB4EC","8FB4ED","8FB4EE","8FB4EF","8FB4F0","8FB4F1","8FB4F2","8FB4F3","8FB4F4","8FB4F5","8FB4F6","8FB4F7","8FB4F8","8FB4F9","8FB4FA","8FB4FB","8FB4FC","8FB4FD","8FB4FE","8FB5A1","8FB5A2","8FB5A3","8FB5A4","8FB5A5","8FB5A6","8FB5A7","8FB5A8","8FB5A9","8FB5AA","8FB5AB","8FB5AC","8FB5AD","8FB5AE","8FB5AF","8FB5B0","8FB5B1","8FB5B2","8FB5B3","8FB5B4","8FB5B5","8FB5B6","8FB5B7","8FB5B8","8FB5B9","8FB5BA","8FB5BB","8FB5BC","8FB5BD","8FB5BE","8FB5BF","8FB5C0","8FB5C1","8FB5C2","8FB5C3","8FB5C4","8FB5C5","8FB5C6","8FB5C7","8FB5C8","8FB5C9","8FB5CA","8FB5CB","8FB5CC","8FB5CD","8FB5CE","8FB5CF","8FB5D0","8FB5D1","8FB5D2","8FB5D3","8FB5D4","8FB5D5","8FB5D6","8FB5D7","8FB5D8","8FB5D9","8FB5DA","8FB5DB","8FB5DC","8FB5DD","8FB5DE","8FB5DF","8FB5E0","8FB5E1","8FB5E2","8FB5E3","8FB5E4","8FB5E5","8FB5E6","8FB5E7","8FB5E8","8FB5E9","8FB5EA","8FB5EB","8FB5EC","8FB5ED","8FB5EE","8FB5EF","8FB5F0","8FB5F1","8FB5F2","8FB5F3","8FB5F4","8FB5F5","8FB5F6","8FB5F7","8FB5F8","8FB5F9","8FB5FA","8FB5FB","8FB5FC","8FB5FD","8FB5FE","8FB6A1","8FB6A2","8FB6A3","8FB6A4","8FB6A5","8FB6A6","8FB6A7","8FB6A8","8FB6A9","8FB6AA","8FB6AB","8FB6AC","8FB6AD","8FB6AE","8FB6AF","8FB6B0","8FB6B1","8FB6B2","8FB6B3","8FB6B4","8FB6B5","8FB6B6","8FB6B7","8FB6B8","8FB6B9","8FB6BA","8FB6BB","8FB6BC","8FB6BD","8FB6BE","8FB6BF","8FB6C0","8FB6C1","8FB6C2","8FB6C3","8FB6C4","8FB6C5","8FB6C6","8FB6C7","8FB6C8","8FB6C9","8FB6CA","8FB6CB","8FB6CC","8FB6CD","8FB6CE","8FB6CF","8FB6D0","8FB6D1","8FB6D2","8FB6D3","8FB6D4","8FB6D5","8FB6D6","8FB6D7","8FB6D8","8FB6D9","8FB6DA","8FB6DB","8FB6DC","8FB6DD","8FB6DE","8FB6DF","8FB6E0","8FB6E1","8FB6E2","8FB6E3","8FB6E4","8FB6E5","8FB6E6","8FB6E7","8FB6E8","8FB6E9","8FB6EA","8FB6EB","8FB6EC","8FB6ED","8FB6EE","8FB6EF","8FB6F0","8FB6F1","8FB6F2","8FB6F3","8FB6F4","8FB6F5","8FB6F6","8FB6F7","8FB6F8","8FB6F9","8FB6FA","8FB6FB","8FB6FC","8FB6FD","8FB6FE","8FB7A1","8FB7A2","8FB7A3","8FB7A4","8FB7A5","8FB7A6","8FB7A7","8FB7A8","8FB7A9","8FB7AA","8FB7AB","8FB7AC","8FB7AD","8FB7AE","8FB7AF","8FB7B0","8FB7B1","8FB7B2","8FB7B3","8FB7B4","8FB7B5","8FB7B6","8FB7B7","8FB7B8","8FB7B9","8FB7BA","8FB7BB","8FB7BC","8FB7BD","8FB7BE","8FB7BF","8FB7C0","8FB7C1","8FB7C2","8FB7C3","8FB7C4","8FB7C5","8FB7C6","8FB7C7","8FB7C8","8FB7C9","8FB7CA","8FB7CB","8FB7CC","8FB7CD","8FB7CE","8FB7CF","8FB7D0","8FB7D1","8FB7D2","8FB7D3","8FB7D4","8FB7D5","8FB7D6","8FB7D7","8FB7D8","8FB7D9","8FB7DA","8FB7DB","8FB7DC","8FB7DD","8FB7DE","8FB7DF","8FB7E0","8FB7E1","8FB7E2","8FB7E3","8FB7E4","8FB7E5","8FB7E6","8FB7E7","8FB7E8","8FB7E9","8FB7EA","8FB7EB","8FB7EC","8FB7ED","8FB7EE","8FB7EF","8FB7F0","8FB7F1","8FB7F2","8FB7F3","8FB7F4","8FB7F5","8FB7F6","8FB7F7","8FB7F8","8FB7F9","8FB7FA","8FB7FB","8FB7FC","8FB7FD","8FB7FE","8FB8A1","8FB8A2","8FB8A3","8FB8A4","8FB8A5","8FB8A6","8FB8A7","8FB8A8","8FB8A9","8FB8AA","8FB8AB","8FB8AC","8FB8AD","8FB8AE","8FB8AF","8FB8B0","8FB8B1","8FB8B2","8FB8B3","8FB8B4","8FB8B5","8FB8B6","8FB8B7","8FB8B8","8FB8B9","8FB8BA","8FB8BB","8FB8BC","8FB8BD","8FB8BE","8FB8BF","8FB8C0","8FB8C1","8FB8C2","8FB8C3","8FB8C4","8FB8C5","8FB8C6","8FB8C7","8FB8C8","8FB8C9","8FB8CA","8FB8CB","8FB8CC","8FB8CD","8FB8CE","8FB8CF","8FB8D0","8FB8D1","8FB8D2","8FB8D3","8FB8D4","8FB8D5","8FB8D6","8FB8D7","8FB8D8","8FB8D9","8FB8DA","8FB8DB","8FB8DC","8FB8DD","8FB8DE","8FB8DF","8FB8E0","8FB8E1","8FB8E2","8FB8E3","8FB8E4","8FB8E5","8FB8E6","8FB8E7","8FB8E8","8FB8E9","8FB8EA","8FB8EB","8FB8EC","8FB8ED","8FB8EE","8FB8EF","8FB8F0","8FB8F1","8FB8F2","8FB8F3","8FB8F4","8FB8F5","8FB8F6","8FB8F7","8FB8F8","8FB8F9","8FB8FA","8FB8FB","8FB8FC","8FB8FD","8FB8FE","8FB9A1","8FB9A2","8FB9A3","8FB9A4","8FB9A5","8FB9A6","8FB9A7","8FB9A8","8FB9A9","8FB9AA","8FB9AB","8FB9AC","8FB9AD","8FB9AE","8FB9AF","8FB9B0","8FB9B1","8FB9B2","8FB9B3","8FB9B4","8FB9B5","8FB9B6","8FB9B7","8FB9B8","8FB9B9","8FB9BA","8FB9BB","8FB9BC","8FB9BD","8FB9BE","8FB9BF","8FB9C0","8FB9C1","8FB9C2","8FB9C3","8FB9C4","8FB9C5","8FB9C6","8FB9C7","8FB9C8","8FB9C9","8FB9CA","8FB9CB","8FB9CC","8FB9CD","8FB9CE","8FB9CF","8FB9D0","8FB9D1","8FB9D2","8FB9D3","8FB9D4","8FB9D5","8FB9D6","8FB9D7","8FB9D8","8FB9D9","8FB9DA","8FB9DB","8FB9DC","8FB9DD","8FB9DE","8FB9DF","8FB9E0","8FB9E1","8FB9E2","8FB9E3","8FB9E4","8FB9E5","8FB9E6","8FB9E7","8FB9E8","8FB9E9","8FB9EA","8FB9EB","8FB9EC","8FB9ED","8FB9EE","8FB9EF","8FB9F0","8FB9F1","8FB9F2","8FB9F3","8FB9F4","8FB9F5","8FB9F6","8FB9F7","8FB9F8","8FB9F9","8FB9FA","8FB9FB","8FB9FC","8FB9FD","8FB9FE","8FBAA1","8FBAA2","8FBAA3","8FBAA4","8FBAA5","8FBAA6","8FBAA7","8FBAA8","8FBAA9","8FBAAA","8FBAAB","8FBAAC","8FBAAD","8FBAAE","8FBAAF","8FBAB0","8FBAB1","8FBAB2","8FBAB3","8FBAB4","8FBAB5","8FBAB6","8FBAB7","8FBAB8","8FBAB9","8FBABA","8FBABB","8FBABC","8FBABD","8FBABE","8FBABF","8FBAC0","8FBAC1","8FBAC2","8FBAC3","8FBAC4","8FBAC5","8FBAC6","8FBAC7","8FBAC8","8FBAC9","8FBACA","8FBACB","8FBACC","8FBACD","8FBACE","8FBACF","8FBAD0","8FBAD1","8FBAD2","8FBAD3","8FBAD4","8FBAD5","8FBAD6","8FBAD7","8FBAD8","8FBAD9","8FBADA","8FBADB","8FBADC","8FBADD","8FBADE","8FBADF","8FBAE0","8FBAE1","8FBAE2","8FBAE3","8FBAE4","8FBAE5","8FBAE6","8FBAE7","8FBAE8","8FBAE9","8FBAEA","8FBAEB","8FBAEC","8FBAED","8FBAEE","8FBAEF","8FBAF0","8FBAF1","8FBAF2","8FBAF3","8FBAF4","8FBAF5","8FBAF6","8FBAF7","8FBAF8","8FBAF9","8FBAFA","8FBAFB","8FBAFC","8FBAFD","8FBAFE","8FBBA1","8FBBA2","8FBBA3","8FBBA4","8FBBA5","8FBBA6","8FBBA7","8FBBA8","8FBBA9","8FBBAA","8FBBAB","8FBBAC","8FBBAD","8FBBAE","8FBBAF","8FBBB0","8FBBB1","8FBBB2","8FBBB3","8FBBB4","8FBBB5","8FBBB6","8FBBB7","8FBBB8","8FBBB9","8FBBBA","8FBBBB","8FBBBC","8FBBBD","8FBBBE","8FBBBF","8FBBC0","8FBBC1","8FBBC2","8FBBC3","8FBBC4","8FBBC5","8FBBC6","8FBBC7","8FBBC8","8FBBC9","8FBBCA","8FBBCB","8FBBCC","8FBBCD","8FBBCE","8FBBCF","8FBBD0","8FBBD1","8FBBD2","8FBBD3","8FBBD4","8FBBD5","8FBBD6","8FBBD7","8FBBD8","8FBBD9","8FBBDA","8FBBDB","8FBBDC","8FBBDD","8FBBDE","8FBBDF","8FBBE0","8FBBE1","8FBBE2","8FBBE3","8FBBE4","8FBBE5","8FBBE6","8FBBE7","8FBBE8","8FBBE9","8FBBEA","8FBBEB","8FBBEC","8FBBED","8FBBEE","8FBBEF","8FBBF0","8FBBF1","8FBBF2","8FBBF3","8FBBF4","8FBBF5","8FBBF6","8FBBF7","8FBBF8","8FBBF9","8FBBFA","8FBBFB","8FBBFC","8FBBFD","8FBBFE","8FBCA1","8FBCA2","8FBCA3","8FBCA4","8FBCA5","8FBCA6","8FBCA7","8FBCA8","8FBCA9","8FBCAA","8FBCAB","8FBCAC","8FBCAD","8FBCAE","8FBCAF","8FBCB0","8FBCB1","8FBCB2","8FBCB3","8FBCB4","8FBCB5","8FBCB6","8FBCB7","8FBCB8","8FBCB9","8FBCBA","8FBCBB","8FBCBC","8FBCBD","8FBCBE","8FBCBF","8FBCC0","8FBCC1","8FBCC2","8FBCC3","8FBCC4","8FBCC5","8FBCC6","8FBCC7","8FBCC8","8FBCC9","8FBCCA","8FBCCB","8FBCCC","8FBCCD","8FBCCE","8FBCCF","8FBCD0","8FBCD1","8FBCD2","8FBCD3","8FBCD4","8FBCD5","8FBCD6","8FBCD7","8FBCD8","8FBCD9","8FBCDA","8FBCDB","8FBCDC","8FBCDD","8FBCDE","8FBCDF","8FBCE0","8FBCE1","8FBCE2","8FBCE3","8FBCE4","8FBCE5","8FBCE6","8FBCE7","8FBCE8","8FBCE9","8FBCEA","8FBCEB","8FBCEC","8FBCED","8FBCEE","8FBCEF","8FBCF0","8FBCF1","8FBCF2","8FBCF3","8FBCF4","8FBCF5","8FBCF6","8FBCF7","8FBCF8","8FBCF9","8FBCFA","8FBCFB","8FBCFC","8FBCFD","8FBCFE","8FBDA1","8FBDA2","8FBDA3","8FBDA4","8FBDA5","8FBDA6","8FBDA7","8FBDA8","8FBDA9","8FBDAA","8FBDAB","8FBDAC","8FBDAD","8FBDAE","8FBDAF","8FBDB0","8FBDB1","8FBDB2","8FBDB3","8FBDB4","8FBDB5","8FBDB6","8FBDB7","8FBDB8","8FBDB9","8FBDBA","8FBDBB","8FBDBC","8FBDBD","8FBDBE","8FBDBF","8FBDC0","8FBDC1","8FBDC2","8FBDC3","8FBDC4","8FBDC5","8FBDC6","8FBDC7","8FBDC8","8FBDC9","8FBDCA","8FBDCB","8FBDCC","8FBDCD","8FBDCE","8FBDCF","8FBDD0","8FBDD1","8FBDD2","8FBDD3","8FBDD4","8FBDD5","8FBDD6","8FBDD7","8FBDD8","8FBDD9","8FBDDA","8FBDDB","8FBDDC","8FBDDD","8FBDDE","8FBDDF","8FBDE0","8FBDE1","8FBDE2","8FBDE3","8FBDE4","8FBDE5","8FBDE6","8FBDE7","8FBDE8","8FBDE9","8FBDEA","8FBDEB","8FBDEC","8FBDED","8FBDEE","8FBDEF","8FBDF0","8FBDF1","8FBDF2","8FBDF3","8FBDF4","8FBDF5","8FBDF6","8FBDF7","8FBDF8","8FBDF9","8FBDFA","8FBDFB","8FBDFC","8FBDFD","8FBDFE","8FBEA1","8FBEA2","8FBEA3","8FBEA4","8FBEA5","8FBEA6","8FBEA7","8FBEA8","8FBEA9","8FBEAA","8FBEAB","8FBEAC","8FBEAD","8FBEAE","8FBEAF","8FBEB0","8FBEB1","8FBEB2","8FBEB3","8FBEB4","8FBEB5","8FBEB6","8FBEB7","8FBEB8","8FBEB9","8FBEBA","8FBEBB","8FBEBC","8FBEBD","8FBEBE","8FBEBF","8FBEC0","8FBEC1","8FBEC2","8FBEC3","8FBEC4","8FBEC5","8FBEC6","8FBEC7","8FBEC8","8FBEC9","8FBECA","8FBECB","8FBECC","8FBECD","8FBECE","8FBECF","8FBED0","8FBED1","8FBED2","8FBED3","8FBED4","8FBED5","8FBED6","8FBED7","8FBED8","8FBED9","8FBEDA","8FBEDB","8FBEDC","8FBEDD","8FBEDE","8FBEDF","8FBEE0","8FBEE1","8FBEE2","8FBEE3","8FBEE4","8FBEE5","8FBEE6","8FBEE7","8FBEE8","8FBEE9","8FBEEA","8FBEEB","8FBEEC","8FBEED","8FBEEE","8FBEEF","8FBEF0","8FBEF1","8FBEF2","8FBEF3","8FBEF4","8FBEF5","8FBEF6","8FBEF7","8FBEF8","8FBEF9","8FBEFA","8FBEFB","8FBEFC","8FBEFD","8FBEFE","8FBFA1","8FBFA2","8FBFA3","8FBFA4","8FBFA5","8FBFA6","8FBFA7","8FBFA8","8FBFA9","8FBFAA","8FBFAB","8FBFAC","8FBFAD","8FBFAE","8FBFAF","8FBFB0","8FBFB1","8FBFB2","8FBFB3","8FBFB4","8FBFB5","8FBFB6","8FBFB7","8FBFB8","8FBFB9","8FBFBA","8FBFBB","8FBFBC","8FBFBD","8FBFBE","8FBFBF","8FBFC0","8FBFC1","8FBFC2","8FBFC3","8FBFC4","8FBFC5","8FBFC6","8FBFC7","8FBFC8","8FBFC9","8FBFCA","8FBFCB","8FBFCC","8FBFCD","8FBFCE","8FBFCF","8FBFD0","8FBFD1","8FBFD2","8FBFD3","8FBFD4","8FBFD5","8FBFD6","8FBFD7","8FBFD8","8FBFD9","8FBFDA","8FBFDB","8FBFDC","8FBFDD","8FBFDE","8FBFDF","8FBFE0","8FBFE1","8FBFE2","8FBFE3","8FBFE4","8FBFE5","8FBFE6","8FBFE7","8FBFE8","8FBFE9","8FBFEA","8FBFEB","8FBFEC","8FBFED","8FBFEE","8FBFEF","8FBFF0","8FBFF1","8FBFF2","8FBFF3","8FBFF4","8FBFF5","8FBFF6","8FBFF7","8FBFF8","8FBFF9","8FBFFA","8FBFFB","8FBFFC","8FBFFD","8FBFFE","8FC0A1","8FC0A2","8FC0A3","8FC0A4","8FC0A5","8FC0A6","8FC0A7","8FC0A8","8FC0A9","8FC0AA","8FC0AB","8FC0AC","8FC0AD","8FC0AE","8FC0AF","8FC0B0","8FC0B1","8FC0B2","8FC0B3","8FC0B4","8FC0B5","8FC0B6","8FC0B7","8FC0B8","8FC0B9","8FC0BA","8FC0BB","8FC0BC","8FC0BD","8FC0BE","8FC0BF","8FC0C0","8FC0C1","8FC0C2","8FC0C3","8FC0C4","8FC0C5","8FC0C6","8FC0C7","8FC0C8","8FC0C9","8FC0CA","8FC0CB","8FC0CC","8FC0CD","8FC0CE","8FC0CF","8FC0D0","8FC0D1","8FC0D2","8FC0D3","8FC0D4","8FC0D5","8FC0D6","8FC0D7","8FC0D8","8FC0D9","8FC0DA","8FC0DB","8FC0DC","8FC0DD","8FC0DE","8FC0DF","8FC0E0","8FC0E1","8FC0E2","8FC0E3","8FC0E4","8FC0E5","8FC0E6","8FC0E7","8FC0E8","8FC0E9","8FC0EA","8FC0EB","8FC0EC","8FC0ED","8FC0EE","8FC0EF","8FC0F0","8FC0F1","8FC0F2","8FC0F3","8FC0F4","8FC0F5","8FC0F6","8FC0F7","8FC0F8","8FC0F9","8FC0FA","8FC0FB","8FC0FC","8FC0FD","8FC0FE","8FC1A1","8FC1A2","8FC1A3","8FC1A4","8FC1A5","8FC1A6","8FC1A7","8FC1A8","8FC1A9","8FC1AA","8FC1AB","8FC1AC","8FC1AD","8FC1AE","8FC1AF","8FC1B0","8FC1B1","8FC1B2","8FC1B3","8FC1B4","8FC1B5","8FC1B6","8FC1B7","8FC1B8","8FC1B9","8FC1BA","8FC1BB","8FC1BC","8FC1BD","8FC1BE","8FC1BF","8FC1C0","8FC1C1","8FC1C2","8FC1C3","8FC1C4","8FC1C5","8FC1C6","8FC1C7","8FC1C8","8FC1C9","8FC1CA","8FC1CB","8FC1CC","8FC1CD","8FC1CE","8FC1CF","8FC1D0","8FC1D1","8FC1D2","8FC1D3","8FC1D4","8FC1D5","8FC1D6","8FC1D7","8FC1D8","8FC1D9","8FC1DA","8FC1DB","8FC1DC","8FC1DD","8FC1DE","8FC1DF","8FC1E0","8FC1E1","8FC1E2","8FC1E3","8FC1E4","8FC1E5","8FC1E6","8FC1E7","8FC1E8","8FC1E9","8FC1EA","8FC1EB","8FC1EC","8FC1ED","8FC1EE","8FC1EF","8FC1F0","8FC1F1","8FC1F2","8FC1F3","8FC1F4","8FC1F5","8FC1F6","8FC1F7","8FC1F8","8FC1F9","8FC1FA","8FC1FB","8FC1FC","8FC1FD","8FC1FE","8FC2A1","8FC2A2","8FC2A3","8FC2A4","8FC2A5","8FC2A6","8FC2A7","8FC2A8","8FC2A9","8FC2AA","8FC2AB","8FC2AC","8FC2AD","8FC2AE","8FC2AF","8FC2B0","8FC2B1","8FC2B2","8FC2B3","8FC2B4","8FC2B5","8FC2B6","8FC2B7","8FC2B8","8FC2B9","8FC2BA","8FC2BB","8FC2BC","8FC2BD","8FC2BE","8FC2BF","8FC2C0","8FC2C1","8FC2C2","8FC2C3","8FC2C4","8FC2C5","8FC2C6","8FC2C7","8FC2C8","8FC2C9","8FC2CA","8FC2CB","8FC2CC","8FC2CD","8FC2CE","8FC2CF","8FC2D0","8FC2D1","8FC2D2","8FC2D3","8FC2D4","8FC2D5","8FC2D6","8FC2D7","8FC2D8","8FC2D9","8FC2DA","8FC2DB","8FC2DC","8FC2DD","8FC2DE","8FC2DF","8FC2E0","8FC2E1","8FC2E2","8FC2E3","8FC2E4","8FC2E5","8FC2E6","8FC2E7","8FC2E8","8FC2E9","8FC2EA","8FC2EB","8FC2EC","8FC2ED","8FC2EE","8FC2EF","8FC2F0","8FC2F1","8FC2F2","8FC2F3","8FC2F4","8FC2F5","8FC2F6","8FC2F7","8FC2F8","8FC2F9","8FC2FA","8FC2FB","8FC2FC","8FC2FD","8FC2FE","8FC3A1","8FC3A2","8FC3A3","8FC3A4","8FC3A5","8FC3A6","8FC3A7","8FC3A8","8FC3A9","8FC3AA","8FC3AB","8FC3AC","8FC3AD","8FC3AE","8FC3AF","8FC3B0","8FC3B1","8FC3B2","8FC3B3","8FC3B4","8FC3B5","8FC3B6","8FC3B7","8FC3B8","8FC3B9","8FC3BA","8FC3BB","8FC3BC","8FC3BD","8FC3BE","8FC3BF","8FC3C0","8FC3C1","8FC3C2","8FC3C3","8FC3C4","8FC3C5","8FC3C6","8FC3C7","8FC3C8","8FC3C9","8FC3CA","8FC3CB","8FC3CC","8FC3CD","8FC3CE","8FC3CF","8FC3D0","8FC3D1","8FC3D2","8FC3D3","8FC3D4","8FC3D5","8FC3D6","8FC3D7","8FC3D8","8FC3D9","8FC3DA","8FC3DB","8FC3DC","8FC3DD","8FC3DE","8FC3DF","8FC3E0","8FC3E1","8FC3E2","8FC3E3","8FC3E4","8FC3E5","8FC3E6","8FC3E7","8FC3E8","8FC3E9","8FC3EA","8FC3EB","8FC3EC","8FC3ED","8FC3EE","8FC3EF","8FC3F0","8FC3F1","8FC3F2","8FC3F3","8FC3F4","8FC3F5","8FC3F6","8FC3F7","8FC3F8","8FC3F9","8FC3FA","8FC3FB","8FC3FC","8FC3FD","8FC3FE","8FC4A1","8FC4A2","8FC4A3","8FC4A4","8FC4A5","8FC4A6","8FC4A7","8FC4A8","8FC4A9","8FC4AA","8FC4AB","8FC4AC","8FC4AD","8FC4AE","8FC4AF","8FC4B0","8FC4B1","8FC4B2","8FC4B3","8FC4B4","8FC4B5","8FC4B6","8FC4B7","8FC4B8","8FC4B9","8FC4BA","8FC4BB","8FC4BC","8FC4BD","8FC4BE","8FC4BF","8FC4C0","8FC4C1","8FC4C2","8FC4C3","8FC4C4","8FC4C5","8FC4C6","8FC4C7","8FC4C8","8FC4C9","8FC4CA","8FC4CB","8FC4CC","8FC4CD","8FC4CE","8FC4CF","8FC4D0","8FC4D1","8FC4D2","8FC4D3","8FC4D4","8FC4D5","8FC4D6","8FC4D7","8FC4D8","8FC4D9","8FC4DA","8FC4DB","8FC4DC","8FC4DD","8FC4DE","8FC4DF","8FC4E0","8FC4E1","8FC4E2","8FC4E3","8FC4E4","8FC4E5","8FC4E6","8FC4E7","8FC4E8","8FC4E9","8FC4EA","8FC4EB","8FC4EC","8FC4ED","8FC4EE","8FC4EF","8FC4F0","8FC4F1","8FC4F2","8FC4F3","8FC4F4","8FC4F5","8FC4F6","8FC4F7","8FC4F8","8FC4F9","8FC4FA","8FC4FB","8FC4FC","8FC4FD","8FC4FE","8FC5A1","8FC5A2","8FC5A3","8FC5A4","8FC5A5","8FC5A6","8FC5A7","8FC5A8","8FC5A9","8FC5AA","8FC5AB","8FC5AC","8FC5AD","8FC5AE","8FC5AF","8FC5B0","8FC5B1","8FC5B2","8FC5B3","8FC5B4","8FC5B5","8FC5B6","8FC5B7","8FC5B8","8FC5B9","8FC5BA","8FC5BB","8FC5BC","8FC5BD","8FC5BE","8FC5BF","8FC5C0","8FC5C1","8FC5C2","8FC5C3","8FC5C4","8FC5C5","8FC5C6","8FC5C7","8FC5C8","8FC5C9","8FC5CA","8FC5CB","8FC5CC","8FC5CD","8FC5CE","8FC5CF","8FC5D0","8FC5D1","8FC5D2","8FC5D3","8FC5D4","8FC5D5","8FC5D6","8FC5D7","8FC5D8","8FC5D9","8FC5DA","8FC5DB","8FC5DC","8FC5DD","8FC5DE","8FC5DF","8FC5E0","8FC5E1","8FC5E2","8FC5E3","8FC5E4","8FC5E5","8FC5E6","8FC5E7","8FC5E8","8FC5E9","8FC5EA","8FC5EB","8FC5EC","8FC5ED","8FC5EE","8FC5EF","8FC5F0","8FC5F1","8FC5F2","8FC5F3","8FC5F4","8FC5F5","8FC5F6","8FC5F7","8FC5F8","8FC5F9","8FC5FA","8FC5FB","8FC5FC","8FC5FD","8FC5FE","8FC6A1","8FC6A2","8FC6A3","8FC6A4","8FC6A5","8FC6A6","8FC6A7","8FC6A8","8FC6A9","8FC6AA","8FC6AB","8FC6AC","8FC6AD","8FC6AE","8FC6AF","8FC6B0","8FC6B1","8FC6B2","8FC6B3","8FC6B4","8FC6B5","8FC6B6","8FC6B7","8FC6B8","8FC6B9","8FC6BA","8FC6BB","8FC6BC","8FC6BD","8FC6BE","8FC6BF","8FC6C0","8FC6C1","8FC6C2","8FC6C3","8FC6C4","8FC6C5","8FC6C6","8FC6C7","8FC6C8","8FC6C9","8FC6CA","8FC6CB","8FC6CC","8FC6CD","8FC6CE","8FC6CF","8FC6D0","8FC6D1","8FC6D2","8FC6D3","8FC6D4","8FC6D5","8FC6D6","8FC6D7","8FC6D8","8FC6D9","8FC6DA","8FC6DB","8FC6DC","8FC6DD","8FC6DE","8FC6DF","8FC6E0","8FC6E1","8FC6E2","8FC6E3","8FC6E4","8FC6E5","8FC6E6","8FC6E7","8FC6E8","8FC6E9","8FC6EA","8FC6EB","8FC6EC","8FC6ED","8FC6EE","8FC6EF","8FC6F0","8FC6F1","8FC6F2","8FC6F3","8FC6F4","8FC6F5","8FC6F6","8FC6F7","8FC6F8","8FC6F9","8FC6FA","8FC6FB","8FC6FC","8FC6FD","8FC6FE","8FC7A1","8FC7A2","8FC7A3","8FC7A4","8FC7A5","8FC7A6","8FC7A7","8FC7A8","8FC7A9","8FC7AA","8FC7AB","8FC7AC","8FC7AD","8FC7AE","8FC7AF","8FC7B0","8FC7B1","8FC7B2","8FC7B3","8FC7B4","8FC7B5","8FC7B6","8FC7B7","8FC7B8","8FC7B9","8FC7BA","8FC7BB","8FC7BC","8FC7BD","8FC7BE","8FC7BF","8FC7C0","8FC7C1","8FC7C2","8FC7C3","8FC7C4","8FC7C5","8FC7C6","8FC7C7","8FC7C8","8FC7C9","8FC7CA","8FC7CB","8FC7CC","8FC7CD","8FC7CE","8FC7CF","8FC7D0","8FC7D1","8FC7D2","8FC7D3","8FC7D4","8FC7D5","8FC7D6","8FC7D7","8FC7D8","8FC7D9","8FC7DA","8FC7DB","8FC7DC","8FC7DD","8FC7DE","8FC7DF","8FC7E0","8FC7E1","8FC7E2","8FC7E3","8FC7E4","8FC7E5","8FC7E6","8FC7E7","8FC7E8","8FC7E9","8FC7EA","8FC7EB","8FC7EC","8FC7ED","8FC7EE","8FC7EF","8FC7F0","8FC7F1","8FC7F2","8FC7F3","8FC7F4","8FC7F5","8FC7F6","8FC7F7","8FC7F8","8FC7F9","8FC7FA","8FC7FB","8FC7FC","8FC7FD","8FC7FE","8FC8A1","8FC8A2","8FC8A3","8FC8A4","8FC8A5","8FC8A6","8FC8A7","8FC8A8","8FC8A9","8FC8AA","8FC8AB","8FC8AC","8FC8AD","8FC8AE","8FC8AF","8FC8B0","8FC8B1","8FC8B2","8FC8B3","8FC8B4","8FC8B5","8FC8B6","8FC8B7","8FC8B8","8FC8B9","8FC8BA","8FC8BB","8FC8BC","8FC8BD","8FC8BE","8FC8BF","8FC8C0","8FC8C1","8FC8C2","8FC8C3","8FC8C4","8FC8C5","8FC8C6","8FC8C7","8FC8C8","8FC8C9","8FC8CA","8FC8CB","8FC8CC","8FC8CD","8FC8CE","8FC8CF","8FC8D0","8FC8D1","8FC8D2","8FC8D3","8FC8D4","8FC8D5","8FC8D6","8FC8D7","8FC8D8","8FC8D9","8FC8DA","8FC8DB","8FC8DC","8FC8DD","8FC8DE","8FC8DF","8FC8E0","8FC8E1","8FC8E2","8FC8E3","8FC8E4","8FC8E5","8FC8E6","8FC8E7","8FC8E8","8FC8E9","8FC8EA","8FC8EB","8FC8EC","8FC8ED","8FC8EE","8FC8EF","8FC8F0","8FC8F1","8FC8F2","8FC8F3","8FC8F4","8FC8F5","8FC8F6","8FC8F7","8FC8F8","8FC8F9","8FC8FA","8FC8FB","8FC8FC","8FC8FD","8FC8FE","8FC9A1","8FC9A2","8FC9A3","8FC9A4","8FC9A5","8FC9A6","8FC9A7","8FC9A8","8FC9A9","8FC9AA","8FC9AB","8FC9AC","8FC9AD","8FC9AE","8FC9AF","8FC9B0","8FC9B1","8FC9B2","8FC9B3","8FC9B4","8FC9B5","8FC9B6","8FC9B7","8FC9B8","8FC9B9","8FC9BA","8FC9BB","8FC9BC","8FC9BD","8FC9BE","8FC9BF","8FC9C0","8FC9C1","8FC9C2","8FC9C3","8FC9C4","8FC9C5","8FC9C6","8FC9C7","8FC9C8","8FC9C9","8FC9CA","8FC9CB","8FC9CC","8FC9CD","8FC9CE","8FC9CF","8FC9D0","8FC9D1","8FC9D2","8FC9D3","8FC9D4","8FC9D5","8FC9D6","8FC9D7","8FC9D8","8FC9D9","8FC9DA","8FC9DB","8FC9DC","8FC9DD","8FC9DE","8FC9DF","8FC9E0","8FC9E1","8FC9E2","8FC9E3","8FC9E4","8FC9E5","8FC9E6","8FC9E7","8FC9E8","8FC9E9","8FC9EA","8FC9EB","8FC9EC","8FC9ED","8FC9EE","8FC9EF","8FC9F0","8FC9F1","8FC9F2","8FC9F3","8FC9F4","8FC9F5","8FC9F6","8FC9F7","8FC9F8","8FC9F9","8FC9FA","8FC9FB","8FC9FC","8FC9FD","8FC9FE","8FCAA1","8FCAA2","8FCAA3","8FCAA4","8FCAA5","8FCAA6","8FCAA7","8FCAA8","8FCAA9","8FCAAA","8FCAAB","8FCAAC","8FCAAD","8FCAAE","8FCAAF","8FCAB0","8FCAB1","8FCAB2","8FCAB3","8FCAB4","8FCAB5","8FCAB6","8FCAB7","8FCAB8","8FCAB9","8FCABA","8FCABB","8FCABC","8FCABD","8FCABE","8FCABF","8FCAC0","8FCAC1","8FCAC2","8FCAC3","8FCAC4","8FCAC5","8FCAC6","8FCAC7","8FCAC8","8FCAC9","8FCACA","8FCACB","8FCACC","8FCACD","8FCACE","8FCACF","8FCAD0","8FCAD1","8FCAD2","8FCAD3","8FCAD4","8FCAD5","8FCAD6","8FCAD7","8FCAD8","8FCAD9","8FCADA","8FCADB","8FCADC","8FCADD","8FCADE","8FCADF","8FCAE0","8FCAE1","8FCAE2","8FCAE3","8FCAE4","8FCAE5","8FCAE6","8FCAE7","8FCAE8","8FCAE9","8FCAEA","8FCAEB","8FCAEC","8FCAED","8FCAEE","8FCAEF","8FCAF0","8FCAF1","8FCAF2","8FCAF3","8FCAF4","8FCAF5","8FCAF6","8FCAF7","8FCAF8","8FCAF9","8FCAFA","8FCAFB","8FCAFC","8FCAFD","8FCAFE","8FCBA1","8FCBA2","8FCBA3","8FCBA4","8FCBA5","8FCBA6","8FCBA7","8FCBA8","8FCBA9","8FCBAA","8FCBAB","8FCBAC","8FCBAD","8FCBAE","8FCBAF","8FCBB0","8FCBB1","8FCBB2","8FCBB3","8FCBB4","8FCBB5","8FCBB6","8FCBB7","8FCBB8","8FCBB9","8FCBBA","8FCBBB","8FCBBC","8FCBBD","8FCBBE","8FCBBF","8FCBC0","8FCBC1","8FCBC2","8FCBC3","8FCBC4","8FCBC5","8FCBC6","8FCBC7","8FCBC8","8FCBC9","8FCBCA","8FCBCB","8FCBCC","8FCBCD","8FCBCE","8FCBCF","8FCBD0","8FCBD1","8FCBD2","8FCBD3","8FCBD4","8FCBD5","8FCBD6","8FCBD7","8FCBD8","8FCBD9","8FCBDA","8FCBDB","8FCBDC","8FCBDD","8FCBDE","8FCBDF","8FCBE0","8FCBE1","8FCBE2","8FCBE3","8FCBE4","8FCBE5","8FCBE6","8FCBE7","8FCBE8","8FCBE9","8FCBEA","8FCBEB","8FCBEC","8FCBED","8FCBEE","8FCBEF","8FCBF0","8FCBF1","8FCBF2","8FCBF3","8FCBF4","8FCBF5","8FCBF6","8FCBF7","8FCBF8","8FCBF9","8FCBFA","8FCBFB","8FCBFC","8FCBFD","8FCBFE","8FCCA1","8FCCA2","8FCCA3","8FCCA4","8FCCA5","8FCCA6","8FCCA7","8FCCA8","8FCCA9","8FCCAA","8FCCAB","8FCCAC","8FCCAD","8FCCAE","8FCCAF","8FCCB0","8FCCB1","8FCCB2","8FCCB3","8FCCB4","8FCCB5","8FCCB6","8FCCB7","8FCCB8","8FCCB9","8FCCBA","8FCCBB","8FCCBC","8FCCBD","8FCCBE","8FCCBF","8FCCC0","8FCCC1","8FCCC2","8FCCC3","8FCCC4","8FCCC5","8FCCC6","8FCCC7","8FCCC8","8FCCC9","8FCCCA","8FCCCB","8FCCCC","8FCCCD","8FCCCE","8FCCCF","8FCCD0","8FCCD1","8FCCD2","8FCCD3","8FCCD4","8FCCD5","8FCCD6","8FCCD7","8FCCD8","8FCCD9","8FCCDA","8FCCDB","8FCCDC","8FCCDD","8FCCDE","8FCCDF","8FCCE0","8FCCE1","8FCCE2","8FCCE3","8FCCE4","8FCCE5","8FCCE6","8FCCE7","8FCCE8","8FCCE9","8FCCEA","8FCCEB","8FCCEC","8FCCED","8FCCEE","8FCCEF","8FCCF0","8FCCF1","8FCCF2","8FCCF3","8FCCF4","8FCCF5","8FCCF6","8FCCF7","8FCCF8","8FCCF9","8FCCFA","8FCCFB","8FCCFC","8FCCFD","8FCCFE","8FCDA1","8FCDA2","8FCDA3","8FCDA4","8FCDA5","8FCDA6","8FCDA7","8FCDA8","8FCDA9","8FCDAA","8FCDAB","8FCDAC","8FCDAD","8FCDAE","8FCDAF","8FCDB0","8FCDB1","8FCDB2","8FCDB3","8FCDB4","8FCDB5","8FCDB6","8FCDB7","8FCDB8","8FCDB9","8FCDBA","8FCDBB","8FCDBC","8FCDBD","8FCDBE","8FCDBF","8FCDC0","8FCDC1","8FCDC2","8FCDC3","8FCDC4","8FCDC5","8FCDC6","8FCDC7","8FCDC8","8FCDC9","8FCDCA","8FCDCB","8FCDCC","8FCDCD","8FCDCE","8FCDCF","8FCDD0","8FCDD1","8FCDD2","8FCDD3","8FCDD4","8FCDD5","8FCDD6","8FCDD7","8FCDD8","8FCDD9","8FCDDA","8FCDDB","8FCDDC","8FCDDD","8FCDDE","8FCDDF","8FCDE0","8FCDE1","8FCDE2","8FCDE3","8FCDE4","8FCDE5","8FCDE6","8FCDE7","8FCDE8","8FCDE9","8FCDEA","8FCDEB","8FCDEC","8FCDED","8FCDEE","8FCDEF","8FCDF0","8FCDF1","8FCDF2","8FCDF3","8FCDF4","8FCDF5","8FCDF6","8FCDF7","8FCDF8","8FCDF9","8FCDFA","8FCDFB","8FCDFC","8FCDFD","8FCDFE","8FCEA1","8FCEA2","8FCEA3","8FCEA4","8FCEA5","8FCEA6","8FCEA7","8FCEA8","8FCEA9","8FCEAA","8FCEAB","8FCEAC","8FCEAD","8FCEAE","8FCEAF","8FCEB0","8FCEB1","8FCEB2","8FCEB3","8FCEB4","8FCEB5","8FCEB6","8FCEB7","8FCEB8","8FCEB9","8FCEBA","8FCEBB","8FCEBC","8FCEBD","8FCEBE","8FCEBF","8FCEC0","8FCEC1","8FCEC2","8FCEC3","8FCEC4","8FCEC5","8FCEC6","8FCEC7","8FCEC8","8FCEC9","8FCECA","8FCECB","8FCECC","8FCECD","8FCECE","8FCECF","8FCED0","8FCED1","8FCED2","8FCED3","8FCED4","8FCED5","8FCED6","8FCED7","8FCED8","8FCED9","8FCEDA","8FCEDB","8FCEDC","8FCEDD","8FCEDE","8FCEDF","8FCEE0","8FCEE1","8FCEE2","8FCEE3","8FCEE4","8FCEE5","8FCEE6","8FCEE7","8FCEE8","8FCEE9","8FCEEA","8FCEEB","8FCEEC","8FCEED","8FCEEE","8FCEEF","8FCEF0","8FCEF1","8FCEF2","8FCEF3","8FCEF4","8FCEF5","8FCEF6","8FCEF7","8FCEF8","8FCEF9","8FCEFA","8FCEFB","8FCEFC","8FCEFD","8FCEFE","8FCFA1","8FCFA2","8FCFA3","8FCFA4","8FCFA5","8FCFA6","8FCFA7","8FCFA8","8FCFA9","8FCFAA","8FCFAB","8FCFAC","8FCFAD","8FCFAE","8FCFAF","8FCFB0","8FCFB1","8FCFB2","8FCFB3","8FCFB4","8FCFB5","8FCFB6","8FCFB7","8FCFB8","8FCFB9","8FCFBA","8FCFBB","8FCFBC","8FCFBD","8FCFBE","8FCFBF","8FCFC0","8FCFC1","8FCFC2","8FCFC3","8FCFC4","8FCFC5","8FCFC6","8FCFC7","8FCFC8","8FCFC9","8FCFCA","8FCFCB","8FCFCC","8FCFCD","8FCFCE","8FCFCF","8FCFD0","8FCFD1","8FCFD2","8FCFD3","8FCFD4","8FCFD5","8FCFD6","8FCFD7","8FCFD8","8FCFD9","8FCFDA","8FCFDB","8FCFDC","8FCFDD","8FCFDE","8FCFDF","8FCFE0","8FCFE1","8FCFE2","8FCFE3","8FCFE4","8FCFE5","8FCFE6","8FCFE7","8FCFE8","8FCFE9","8FCFEA","8FCFEB","8FCFEC","8FCFED","8FCFEE","8FCFEF","8FCFF0","8FCFF1","8FCFF2","8FCFF3","8FCFF4","8FCFF5","8FCFF6","8FCFF7","8FCFF8","8FCFF9","8FCFFA","8FCFFB","8FCFFC","8FCFFD","8FCFFE","8FD0A1","8FD0A2","8FD0A3","8FD0A4","8FD0A5","8FD0A6","8FD0A7","8FD0A8","8FD0A9","8FD0AA","8FD0AB","8FD0AC","8FD0AD","8FD0AE","8FD0AF","8FD0B0","8FD0B1","8FD0B2","8FD0B3","8FD0B4","8FD0B5","8FD0B6","8FD0B7","8FD0B8","8FD0B9","8FD0BA","8FD0BB","8FD0BC","8FD0BD","8FD0BE","8FD0BF","8FD0C0","8FD0C1","8FD0C2","8FD0C3","8FD0C4","8FD0C5","8FD0C6","8FD0C7","8FD0C8","8FD0C9","8FD0CA","8FD0CB","8FD0CC","8FD0CD","8FD0CE","8FD0CF","8FD0D0","8FD0D1","8FD0D2","8FD0D3","8FD0D4","8FD0D5","8FD0D6","8FD0D7","8FD0D8","8FD0D9","8FD0DA","8FD0DB","8FD0DC","8FD0DD","8FD0DE","8FD0DF","8FD0E0","8FD0E1","8FD0E2","8FD0E3","8FD0E4","8FD0E5","8FD0E6","8FD0E7","8FD0E8","8FD0E9","8FD0EA","8FD0EB","8FD0EC","8FD0ED","8FD0EE","8FD0EF","8FD0F0","8FD0F1","8FD0F2","8FD0F3","8FD0F4","8FD0F5","8FD0F6","8FD0F7","8FD0F8","8FD0F9","8FD0FA","8FD0FB","8FD0FC","8FD0FD","8FD0FE","8FD1A1","8FD1A2","8FD1A3","8FD1A4","8FD1A5","8FD1A6","8FD1A7","8FD1A8","8FD1A9","8FD1AA","8FD1AB","8FD1AC","8FD1AD","8FD1AE","8FD1AF","8FD1B0","8FD1B1","8FD1B2","8FD1B3","8FD1B4","8FD1B5","8FD1B6","8FD1B7","8FD1B8","8FD1B9","8FD1BA","8FD1BB","8FD1BC","8FD1BD","8FD1BE","8FD1BF","8FD1C0","8FD1C1","8FD1C2","8FD1C3","8FD1C4","8FD1C5","8FD1C6","8FD1C7","8FD1C8","8FD1C9","8FD1CA","8FD1CB","8FD1CC","8FD1CD","8FD1CE","8FD1CF","8FD1D0","8FD1D1","8FD1D2","8FD1D3","8FD1D4","8FD1D5","8FD1D6","8FD1D7","8FD1D8","8FD1D9","8FD1DA","8FD1DB","8FD1DC","8FD1DD","8FD1DE","8FD1DF","8FD1E0","8FD1E1","8FD1E2","8FD1E3","8FD1E4","8FD1E5","8FD1E6","8FD1E7","8FD1E8","8FD1E9","8FD1EA","8FD1EB","8FD1EC","8FD1ED","8FD1EE","8FD1EF","8FD1F0","8FD1F1","8FD1F2","8FD1F3","8FD1F4","8FD1F5","8FD1F6","8FD1F7","8FD1F8","8FD1F9","8FD1FA","8FD1FB","8FD1FC","8FD1FD","8FD1FE","8FD2A1","8FD2A2","8FD2A3","8FD2A4","8FD2A5","8FD2A6","8FD2A7","8FD2A8","8FD2A9","8FD2AA","8FD2AB","8FD2AC","8FD2AD","8FD2AE","8FD2AF","8FD2B0","8FD2B1","8FD2B2","8FD2B3","8FD2B4","8FD2B5","8FD2B6","8FD2B7","8FD2B8","8FD2B9","8FD2BA","8FD2BB","8FD2BC","8FD2BD","8FD2BE","8FD2BF","8FD2C0","8FD2C1","8FD2C2","8FD2C3","8FD2C4","8FD2C5","8FD2C6","8FD2C7","8FD2C8","8FD2C9","8FD2CA","8FD2CB","8FD2CC","8FD2CD","8FD2CE","8FD2CF","8FD2D0","8FD2D1","8FD2D2","8FD2D3","8FD2D4","8FD2D5","8FD2D6","8FD2D7","8FD2D8","8FD2D9","8FD2DA","8FD2DB","8FD2DC","8FD2DD","8FD2DE","8FD2DF","8FD2E0","8FD2E1","8FD2E2","8FD2E3","8FD2E4","8FD2E5","8FD2E6","8FD2E7","8FD2E8","8FD2E9","8FD2EA","8FD2EB","8FD2EC","8FD2ED","8FD2EE","8FD2EF","8FD2F0","8FD2F1","8FD2F2","8FD2F3","8FD2F4","8FD2F5","8FD2F6","8FD2F7","8FD2F8","8FD2F9","8FD2FA","8FD2FB","8FD2FC","8FD2FD","8FD2FE","8FD3A1","8FD3A2","8FD3A3","8FD3A4","8FD3A5","8FD3A6","8FD3A7","8FD3A8","8FD3A9","8FD3AA","8FD3AB","8FD3AC","8FD3AD","8FD3AE","8FD3AF","8FD3B0","8FD3B1","8FD3B2","8FD3B3","8FD3B4","8FD3B5","8FD3B6","8FD3B7","8FD3B8","8FD3B9","8FD3BA","8FD3BB","8FD3BC","8FD3BD","8FD3BE","8FD3BF","8FD3C0","8FD3C1","8FD3C2","8FD3C3","8FD3C4","8FD3C5","8FD3C6","8FD3C7","8FD3C8","8FD3C9","8FD3CA","8FD3CB","8FD3CC","8FD3CD","8FD3CE","8FD3CF","8FD3D0","8FD3D1","8FD3D2","8FD3D3","8FD3D4","8FD3D5","8FD3D6","8FD3D7","8FD3D8","8FD3D9","8FD3DA","8FD3DB","8FD3DC","8FD3DD","8FD3DE","8FD3DF","8FD3E0","8FD3E1","8FD3E2","8FD3E3","8FD3E4","8FD3E5","8FD3E6","8FD3E7","8FD3E8","8FD3E9","8FD3EA","8FD3EB","8FD3EC","8FD3ED","8FD3EE","8FD3EF","8FD3F0","8FD3F1","8FD3F2","8FD3F3","8FD3F4","8FD3F5","8FD3F6","8FD3F7","8FD3F8","8FD3F9","8FD3FA","8FD3FB","8FD3FC","8FD3FD","8FD3FE","8FD4A1","8FD4A2","8FD4A3","8FD4A4","8FD4A5","8FD4A6","8FD4A7","8FD4A8","8FD4A9","8FD4AA","8FD4AB","8FD4AC","8FD4AD","8FD4AE","8FD4AF","8FD4B0","8FD4B1","8FD4B2","8FD4B3","8FD4B4","8FD4B5","8FD4B6","8FD4B7","8FD4B8","8FD4B9","8FD4BA","8FD4BB","8FD4BC","8FD4BD","8FD4BE","8FD4BF","8FD4C0","8FD4C1","8FD4C2","8FD4C3","8FD4C4","8FD4C5","8FD4C6","8FD4C7","8FD4C8","8FD4C9","8FD4CA","8FD4CB","8FD4CC","8FD4CD","8FD4CE","8FD4CF","8FD4D0","8FD4D1","8FD4D2","8FD4D3","8FD4D4","8FD4D5","8FD4D6","8FD4D7","8FD4D8","8FD4D9","8FD4DA","8FD4DB","8FD4DC","8FD4DD","8FD4DE","8FD4DF","8FD4E0","8FD4E1","8FD4E2","8FD4E3","8FD4E4","8FD4E5","8FD4E6","8FD4E7","8FD4E8","8FD4E9","8FD4EA","8FD4EB","8FD4EC","8FD4ED","8FD4EE","8FD4EF","8FD4F0","8FD4F1","8FD4F2","8FD4F3","8FD4F4","8FD4F5","8FD4F6","8FD4F7","8FD4F8","8FD4F9","8FD4FA","8FD4FB","8FD4FC","8FD4FD","8FD4FE","8FD5A1","8FD5A2","8FD5A3","8FD5A4","8FD5A5","8FD5A6","8FD5A7","8FD5A8","8FD5A9","8FD5AA","8FD5AB","8FD5AC","8FD5AD","8FD5AE","8FD5AF","8FD5B0","8FD5B1","8FD5B2","8FD5B3","8FD5B4","8FD5B5","8FD5B6","8FD5B7","8FD5B8","8FD5B9","8FD5BA","8FD5BB","8FD5BC","8FD5BD","8FD5BE","8FD5BF","8FD5C0","8FD5C1","8FD5C2","8FD5C3","8FD5C4","8FD5C5","8FD5C6","8FD5C7","8FD5C8","8FD5C9","8FD5CA","8FD5CB","8FD5CC","8FD5CD","8FD5CE","8FD5CF","8FD5D0","8FD5D1","8FD5D2","8FD5D3","8FD5D4","8FD5D5","8FD5D6","8FD5D7","8FD5D8","8FD5D9","8FD5DA","8FD5DB","8FD5DC","8FD5DD","8FD5DE","8FD5DF","8FD5E0","8FD5E1","8FD5E2","8FD5E3","8FD5E4","8FD5E5","8FD5E6","8FD5E7","8FD5E8","8FD5E9","8FD5EA","8FD5EB","8FD5EC","8FD5ED","8FD5EE","8FD5EF","8FD5F0","8FD5F1","8FD5F2","8FD5F3","8FD5F4","8FD5F5","8FD5F6","8FD5F7","8FD5F8","8FD5F9","8FD5FA","8FD5FB","8FD5FC","8FD5FD","8FD5FE","8FD6A1","8FD6A2","8FD6A3","8FD6A4","8FD6A5","8FD6A6","8FD6A7","8FD6A8","8FD6A9","8FD6AA","8FD6AB","8FD6AC","8FD6AD","8FD6AE","8FD6AF","8FD6B0","8FD6B1","8FD6B2","8FD6B3","8FD6B4","8FD6B5","8FD6B6","8FD6B7","8FD6B8","8FD6B9","8FD6BA","8FD6BB","8FD6BC","8FD6BD","8FD6BE","8FD6BF","8FD6C0","8FD6C1","8FD6C2","8FD6C3","8FD6C4","8FD6C5","8FD6C6","8FD6C7","8FD6C8","8FD6C9","8FD6CA","8FD6CB","8FD6CC","8FD6CD","8FD6CE","8FD6CF","8FD6D0","8FD6D1","8FD6D2","8FD6D3","8FD6D4","8FD6D5","8FD6D6","8FD6D7","8FD6D8","8FD6D9","8FD6DA","8FD6DB","8FD6DC","8FD6DD","8FD6DE","8FD6DF","8FD6E0","8FD6E1","8FD6E2","8FD6E3","8FD6E4","8FD6E5","8FD6E6","8FD6E7","8FD6E8","8FD6E9","8FD6EA","8FD6EB","8FD6EC","8FD6ED","8FD6EE","8FD6EF","8FD6F0","8FD6F1","8FD6F2","8FD6F3","8FD6F4","8FD6F5","8FD6F6","8FD6F7","8FD6F8","8FD6F9","8FD6FA","8FD6FB","8FD6FC","8FD6FD","8FD6FE","8FD7A1","8FD7A2","8FD7A3","8FD7A4","8FD7A5","8FD7A6","8FD7A7","8FD7A8","8FD7A9","8FD7AA","8FD7AB","8FD7AC","8FD7AD","8FD7AE","8FD7AF","8FD7B0","8FD7B1","8FD7B2","8FD7B3","8FD7B4","8FD7B5","8FD7B6","8FD7B7","8FD7B8","8FD7B9","8FD7BA","8FD7BB","8FD7BC","8FD7BD","8FD7BE","8FD7BF","8FD7C0","8FD7C1","8FD7C2","8FD7C3","8FD7C4","8FD7C5","8FD7C6","8FD7C7","8FD7C8","8FD7C9","8FD7CA","8FD7CB","8FD7CC","8FD7CD","8FD7CE","8FD7CF","8FD7D0","8FD7D1","8FD7D2","8FD7D3","8FD7D4","8FD7D5","8FD7D6","8FD7D7","8FD7D8","8FD7D9","8FD7DA","8FD7DB","8FD7DC","8FD7DD","8FD7DE","8FD7DF","8FD7E0","8FD7E1","8FD7E2","8FD7E3","8FD7E4","8FD7E5","8FD7E6","8FD7E7","8FD7E8","8FD7E9","8FD7EA","8FD7EB","8FD7EC","8FD7ED","8FD7EE","8FD7EF","8FD7F0","8FD7F1","8FD7F2","8FD7F3","8FD7F4","8FD7F5","8FD7F6","8FD7F7","8FD7F8","8FD7F9","8FD7FA","8FD7FB","8FD7FC","8FD7FD","8FD7FE","8FD8A1","8FD8A2","8FD8A3","8FD8A4","8FD8A5","8FD8A6","8FD8A7","8FD8A8","8FD8A9","8FD8AA","8FD8AB","8FD8AC","8FD8AD","8FD8AE","8FD8AF","8FD8B0","8FD8B1","8FD8B2","8FD8B3","8FD8B4","8FD8B5","8FD8B6","8FD8B7","8FD8B8","8FD8B9","8FD8BA","8FD8BB","8FD8BC","8FD8BD","8FD8BE","8FD8BF","8FD8C0","8FD8C1","8FD8C2","8FD8C3","8FD8C4","8FD8C5","8FD8C6","8FD8C7","8FD8C8","8FD8C9","8FD8CA","8FD8CB","8FD8CC","8FD8CD","8FD8CE","8FD8CF","8FD8D0","8FD8D1","8FD8D2","8FD8D3","8FD8D4","8FD8D5","8FD8D6","8FD8D7","8FD8D8","8FD8D9","8FD8DA","8FD8DB","8FD8DC","8FD8DD","8FD8DE","8FD8DF","8FD8E0","8FD8E1","8FD8E2","8FD8E3","8FD8E4","8FD8E5","8FD8E6","8FD8E7","8FD8E8","8FD8E9","8FD8EA","8FD8EB","8FD8EC","8FD8ED","8FD8EE","8FD8EF","8FD8F0","8FD8F1","8FD8F2","8FD8F3","8FD8F4","8FD8F5","8FD8F6","8FD8F7","8FD8F8","8FD8F9","8FD8FA","8FD8FB","8FD8FC","8FD8FD","8FD8FE","8FD9A1","8FD9A2","8FD9A3","8FD9A4","8FD9A5","8FD9A6","8FD9A7","8FD9A8","8FD9A9","8FD9AA","8FD9AB","8FD9AC","8FD9AD","8FD9AE","8FD9AF","8FD9B0","8FD9B1","8FD9B2","8FD9B3","8FD9B4","8FD9B5","8FD9B6","8FD9B7","8FD9B8","8FD9B9","8FD9BA","8FD9BB","8FD9BC","8FD9BD","8FD9BE","8FD9BF","8FD9C0","8FD9C1","8FD9C2","8FD9C3","8FD9C4","8FD9C5","8FD9C6","8FD9C7","8FD9C8","8FD9C9","8FD9CA","8FD9CB","8FD9CC","8FD9CD","8FD9CE","8FD9CF","8FD9D0","8FD9D1","8FD9D2","8FD9D3","8FD9D4","8FD9D5","8FD9D6","8FD9D7","8FD9D8","8FD9D9","8FD9DA","8FD9DB","8FD9DC","8FD9DD","8FD9DE","8FD9DF","8FD9E0","8FD9E1","8FD9E2","8FD9E3","8FD9E4","8FD9E5","8FD9E6","8FD9E7","8FD9E8","8FD9E9","8FD9EA","8FD9EB","8FD9EC","8FD9ED","8FD9EE","8FD9EF","8FD9F0","8FD9F1","8FD9F2","8FD9F3","8FD9F4","8FD9F5","8FD9F6","8FD9F7","8FD9F8","8FD9F9","8FD9FA","8FD9FB","8FD9FC","8FD9FD","8FD9FE","8FDAA1","8FDAA2","8FDAA3","8FDAA4","8FDAA5","8FDAA6","8FDAA7","8FDAA8","8FDAA9","8FDAAA","8FDAAB","8FDAAC","8FDAAD","8FDAAE","8FDAAF","8FDAB0","8FDAB1","8FDAB2","8FDAB3","8FDAB4","8FDAB5","8FDAB6","8FDAB7","8FDAB8","8FDAB9","8FDABA","8FDABB","8FDABC","8FDABD","8FDABE","8FDABF","8FDAC0","8FDAC1","8FDAC2","8FDAC3","8FDAC4","8FDAC5","8FDAC6","8FDAC7","8FDAC8","8FDAC9","8FDACA","8FDACB","8FDACC","8FDACD","8FDACE","8FDACF","8FDAD0","8FDAD1","8FDAD2","8FDAD3","8FDAD4","8FDAD5","8FDAD6","8FDAD7","8FDAD8","8FDAD9","8FDADA","8FDADB","8FDADC","8FDADD","8FDADE","8FDADF","8FDAE0","8FDAE1","8FDAE2","8FDAE3","8FDAE4","8FDAE5","8FDAE6","8FDAE7","8FDAE8","8FDAE9","8FDAEA","8FDAEB","8FDAEC","8FDAED","8FDAEE","8FDAEF","8FDAF0","8FDAF1","8FDAF2","8FDAF3","8FDAF4","8FDAF5","8FDAF6","8FDAF7","8FDAF8","8FDAF9","8FDAFA","8FDAFB","8FDAFC","8FDAFD","8FDAFE","8FDBA1","8FDBA2","8FDBA3","8FDBA4","8FDBA5","8FDBA6","8FDBA7","8FDBA8","8FDBA9","8FDBAA","8FDBAB","8FDBAC","8FDBAD","8FDBAE","8FDBAF","8FDBB0","8FDBB1","8FDBB2","8FDBB3","8FDBB4","8FDBB5","8FDBB6","8FDBB7","8FDBB8","8FDBB9","8FDBBA","8FDBBB","8FDBBC","8FDBBD","8FDBBE","8FDBBF","8FDBC0","8FDBC1","8FDBC2","8FDBC3","8FDBC4","8FDBC5","8FDBC6","8FDBC7","8FDBC8","8FDBC9","8FDBCA","8FDBCB","8FDBCC","8FDBCD","8FDBCE","8FDBCF","8FDBD0","8FDBD1","8FDBD2","8FDBD3","8FDBD4","8FDBD5","8FDBD6","8FDBD7","8FDBD8","8FDBD9","8FDBDA","8FDBDB","8FDBDC","8FDBDD","8FDBDE","8FDBDF","8FDBE0","8FDBE1","8FDBE2","8FDBE3","8FDBE4","8FDBE5","8FDBE6","8FDBE7","8FDBE8","8FDBE9","8FDBEA","8FDBEB","8FDBEC","8FDBED","8FDBEE","8FDBEF","8FDBF0","8FDBF1","8FDBF2","8FDBF3","8FDBF4","8FDBF5","8FDBF6","8FDBF7","8FDBF8","8FDBF9","8FDBFA","8FDBFB","8FDBFC","8FDBFD","8FDBFE","8FDCA1","8FDCA2","8FDCA3","8FDCA4","8FDCA5","8FDCA6","8FDCA7","8FDCA8","8FDCA9","8FDCAA","8FDCAB","8FDCAC","8FDCAD","8FDCAE","8FDCAF","8FDCB0","8FDCB1","8FDCB2","8FDCB3","8FDCB4","8FDCB5","8FDCB6","8FDCB7","8FDCB8","8FDCB9","8FDCBA","8FDCBB","8FDCBC","8FDCBD","8FDCBE","8FDCBF","8FDCC0","8FDCC1","8FDCC2","8FDCC3","8FDCC4","8FDCC5","8FDCC6","8FDCC7","8FDCC8","8FDCC9","8FDCCA","8FDCCB","8FDCCC","8FDCCD","8FDCCE","8FDCCF","8FDCD0","8FDCD1","8FDCD2","8FDCD3","8FDCD4","8FDCD5","8FDCD6","8FDCD7","8FDCD8","8FDCD9","8FDCDA","8FDCDB","8FDCDC","8FDCDD","8FDCDE","8FDCDF","8FDCE0","8FDCE1","8FDCE2","8FDCE3","8FDCE4","8FDCE5","8FDCE6","8FDCE7","8FDCE8","8FDCE9","8FDCEA","8FDCEB","8FDCEC","8FDCED","8FDCEE","8FDCEF","8FDCF0","8FDCF1","8FDCF2","8FDCF3","8FDCF4","8FDCF5","8FDCF6","8FDCF7","8FDCF8","8FDCF9","8FDCFA","8FDCFB","8FDCFC","8FDCFD","8FDCFE","8FDDA1","8FDDA2","8FDDA3","8FDDA4","8FDDA5","8FDDA6","8FDDA7","8FDDA8","8FDDA9","8FDDAA","8FDDAB","8FDDAC","8FDDAD","8FDDAE","8FDDAF","8FDDB0","8FDDB1","8FDDB2","8FDDB3","8FDDB4","8FDDB5","8FDDB6","8FDDB7","8FDDB8","8FDDB9","8FDDBA","8FDDBB","8FDDBC","8FDDBD","8FDDBE","8FDDBF","8FDDC0","8FDDC1","8FDDC2","8FDDC3","8FDDC4","8FDDC5","8FDDC6","8FDDC7","8FDDC8","8FDDC9","8FDDCA","8FDDCB","8FDDCC","8FDDCD","8FDDCE","8FDDCF","8FDDD0","8FDDD1","8FDDD2","8FDDD3","8FDDD4","8FDDD5","8FDDD6","8FDDD7","8FDDD8","8FDDD9","8FDDDA","8FDDDB","8FDDDC","8FDDDD","8FDDDE","8FDDDF","8FDDE0","8FDDE1","8FDDE2","8FDDE3","8FDDE4","8FDDE5","8FDDE6","8FDDE7","8FDDE8","8FDDE9","8FDDEA","8FDDEB","8FDDEC","8FDDED","8FDDEE","8FDDEF","8FDDF0","8FDDF1","8FDDF2","8FDDF3","8FDDF4","8FDDF5","8FDDF6","8FDDF7","8FDDF8","8FDDF9","8FDDFA","8FDDFB","8FDDFC","8FDDFD","8FDDFE","8FDEA1","8FDEA2","8FDEA3","8FDEA4","8FDEA5","8FDEA6","8FDEA7","8FDEA8","8FDEA9","8FDEAA","8FDEAB","8FDEAC","8FDEAD","8FDEAE","8FDEAF","8FDEB0","8FDEB1","8FDEB2","8FDEB3","8FDEB4","8FDEB5","8FDEB6","8FDEB7","8FDEB8","8FDEB9","8FDEBA","8FDEBB","8FDEBC","8FDEBD","8FDEBE","8FDEBF","8FDEC0","8FDEC1","8FDEC2","8FDEC3","8FDEC4","8FDEC5","8FDEC6","8FDEC7","8FDEC8","8FDEC9","8FDECA","8FDECB","8FDECC","8FDECD","8FDECE","8FDECF","8FDED0","8FDED1","8FDED2","8FDED3","8FDED4","8FDED5","8FDED6","8FDED7","8FDED8","8FDED9","8FDEDA","8FDEDB","8FDEDC","8FDEDD","8FDEDE","8FDEDF","8FDEE0","8FDEE1","8FDEE2","8FDEE3","8FDEE4","8FDEE5","8FDEE6","8FDEE7","8FDEE8","8FDEE9","8FDEEA","8FDEEB","8FDEEC","8FDEED","8FDEEE","8FDEEF","8FDEF0","8FDEF1","8FDEF2","8FDEF3","8FDEF4","8FDEF5","8FDEF6","8FDEF7","8FDEF8","8FDEF9","8FDEFA","8FDEFB","8FDEFC","8FDEFD","8FDEFE","8FDFA1","8FDFA2","8FDFA3","8FDFA4","8FDFA5","8FDFA6","8FDFA7","8FDFA8","8FDFA9","8FDFAA","8FDFAB","8FDFAC","8FDFAD","8FDFAE","8FDFAF","8FDFB0","8FDFB1","8FDFB2","8FDFB3","8FDFB4","8FDFB5","8FDFB6","8FDFB7","8FDFB8","8FDFB9","8FDFBA","8FDFBB","8FDFBC","8FDFBD","8FDFBE","8FDFBF","8FDFC0","8FDFC1","8FDFC2","8FDFC3","8FDFC4","8FDFC5","8FDFC6","8FDFC7","8FDFC8","8FDFC9","8FDFCA","8FDFCB","8FDFCC","8FDFCD","8FDFCE","8FDFCF","8FDFD0","8FDFD1","8FDFD2","8FDFD3","8FDFD4","8FDFD5","8FDFD6","8FDFD7","8FDFD8","8FDFD9","8FDFDA","8FDFDB","8FDFDC","8FDFDD","8FDFDE","8FDFDF","8FDFE0","8FDFE1","8FDFE2","8FDFE3","8FDFE4","8FDFE5","8FDFE6","8FDFE7","8FDFE8","8FDFE9","8FDFEA","8FDFEB","8FDFEC","8FDFED","8FDFEE","8FDFEF","8FDFF0","8FDFF1","8FDFF2","8FDFF3","8FDFF4","8FDFF5","8FDFF6","8FDFF7","8FDFF8","8FDFF9","8FDFFA","8FDFFB","8FDFFC","8FDFFD","8FDFFE","8FE0A1","8FE0A2","8FE0A3","8FE0A4","8FE0A5","8FE0A6","8FE0A7","8FE0A8","8FE0A9","8FE0AA","8FE0AB","8FE0AC","8FE0AD","8FE0AE","8FE0AF","8FE0B0","8FE0B1","8FE0B2","8FE0B3","8FE0B4","8FE0B5","8FE0B6","8FE0B7","8FE0B8","8FE0B9","8FE0BA","8FE0BB","8FE0BC","8FE0BD","8FE0BE","8FE0BF","8FE0C0","8FE0C1","8FE0C2","8FE0C3","8FE0C4","8FE0C5","8FE0C6","8FE0C7","8FE0C8","8FE0C9","8FE0CA","8FE0CB","8FE0CC","8FE0CD","8FE0CE","8FE0CF","8FE0D0","8FE0D1","8FE0D2","8FE0D3","8FE0D4","8FE0D5","8FE0D6","8FE0D7","8FE0D8","8FE0D9","8FE0DA","8FE0DB","8FE0DC","8FE0DD","8FE0DE","8FE0DF","8FE0E0","8FE0E1","8FE0E2","8FE0E3","8FE0E4","8FE0E5","8FE0E6","8FE0E7","8FE0E8","8FE0E9","8FE0EA","8FE0EB","8FE0EC","8FE0ED","8FE0EE","8FE0EF","8FE0F0","8FE0F1","8FE0F2","8FE0F3","8FE0F4","8FE0F5","8FE0F6","8FE0F7","8FE0F8","8FE0F9","8FE0FA","8FE0FB","8FE0FC","8FE0FD","8FE0FE","8FE1A1","8FE1A2","8FE1A3","8FE1A4","8FE1A5","8FE1A6","8FE1A7","8FE1A8","8FE1A9","8FE1AA","8FE1AB","8FE1AC","8FE1AD","8FE1AE","8FE1AF","8FE1B0","8FE1B1","8FE1B2","8FE1B3","8FE1B4","8FE1B5","8FE1B6","8FE1B7","8FE1B8","8FE1B9","8FE1BA","8FE1BB","8FE1BC","8FE1BD","8FE1BE","8FE1BF","8FE1C0","8FE1C1","8FE1C2","8FE1C3","8FE1C4","8FE1C5","8FE1C6","8FE1C7","8FE1C8","8FE1C9","8FE1CA","8FE1CB","8FE1CC","8FE1CD","8FE1CE","8FE1CF","8FE1D0","8FE1D1","8FE1D2","8FE1D3","8FE1D4","8FE1D5","8FE1D6","8FE1D7","8FE1D8","8FE1D9","8FE1DA","8FE1DB","8FE1DC","8FE1DD","8FE1DE","8FE1DF","8FE1E0","8FE1E1","8FE1E2","8FE1E3","8FE1E4","8FE1E5","8FE1E6","8FE1E7","8FE1E8","8FE1E9","8FE1EA","8FE1EB","8FE1EC","8FE1ED","8FE1EE","8FE1EF","8FE1F0","8FE1F1","8FE1F2","8FE1F3","8FE1F4","8FE1F5","8FE1F6","8FE1F7","8FE1F8","8FE1F9","8FE1FA","8FE1FB","8FE1FC","8FE1FD","8FE1FE","8FE2A1","8FE2A2","8FE2A3","8FE2A4","8FE2A5","8FE2A6","8FE2A7","8FE2A8","8FE2A9","8FE2AA","8FE2AB","8FE2AC","8FE2AD","8FE2AE","8FE2AF","8FE2B0","8FE2B1","8FE2B2","8FE2B3","8FE2B4","8FE2B5","8FE2B6","8FE2B7","8FE2B8","8FE2B9","8FE2BA","8FE2BB","8FE2BC","8FE2BD","8FE2BE","8FE2BF","8FE2C0","8FE2C1","8FE2C2","8FE2C3","8FE2C4","8FE2C5","8FE2C6","8FE2C7","8FE2C8","8FE2C9","8FE2CA","8FE2CB","8FE2CC","8FE2CD","8FE2CE","8FE2CF","8FE2D0","8FE2D1","8FE2D2","8FE2D3","8FE2D4","8FE2D5","8FE2D6","8FE2D7","8FE2D8","8FE2D9","8FE2DA","8FE2DB","8FE2DC","8FE2DD","8FE2DE","8FE2DF","8FE2E0","8FE2E1","8FE2E2","8FE2E3","8FE2E4","8FE2E5","8FE2E6","8FE2E7","8FE2E8","8FE2E9","8FE2EA","8FE2EB","8FE2EC","8FE2ED","8FE2EE","8FE2EF","8FE2F0","8FE2F1","8FE2F2","8FE2F3","8FE2F4","8FE2F5","8FE2F6","8FE2F7","8FE2F8","8FE2F9","8FE2FA","8FE2FB","8FE2FC","8FE2FD","8FE2FE","8FE3A1","8FE3A2","8FE3A3","8FE3A4","8FE3A5","8FE3A6","8FE3A7","8FE3A8","8FE3A9","8FE3AA","8FE3AB","8FE3AC","8FE3AD","8FE3AE","8FE3AF","8FE3B0","8FE3B1","8FE3B2","8FE3B3","8FE3B4","8FE3B5","8FE3B6","8FE3B7","8FE3B8","8FE3B9","8FE3BA","8FE3BB","8FE3BC","8FE3BD","8FE3BE","8FE3BF","8FE3C0","8FE3C1","8FE3C2","8FE3C3","8FE3C4","8FE3C5","8FE3C6","8FE3C7","8FE3C8","8FE3C9","8FE3CA","8FE3CB","8FE3CC","8FE3CD","8FE3CE","8FE3CF","8FE3D0","8FE3D1","8FE3D2","8FE3D3","8FE3D4","8FE3D5","8FE3D6","8FE3D7","8FE3D8","8FE3D9","8FE3DA","8FE3DB","8FE3DC","8FE3DD","8FE3DE","8FE3DF","8FE3E0","8FE3E1","8FE3E2","8FE3E3","8FE3E4","8FE3E5","8FE3E6","8FE3E7","8FE3E8","8FE3E9","8FE3EA","8FE3EB","8FE3EC","8FE3ED","8FE3EE","8FE3EF","8FE3F0","8FE3F1","8FE3F2","8FE3F3","8FE3F4","8FE3F5","8FE3F6","8FE3F7","8FE3F8","8FE3F9","8FE3FA","8FE3FB","8FE3FC","8FE3FD","8FE3FE","8FE4A1","8FE4A2","8FE4A3","8FE4A4","8FE4A5","8FE4A6","8FE4A7","8FE4A8","8FE4A9","8FE4AA","8FE4AB","8FE4AC","8FE4AD","8FE4AE","8FE4AF","8FE4B0","8FE4B1","8FE4B2","8FE4B3","8FE4B4","8FE4B5","8FE4B6","8FE4B7","8FE4B8","8FE4B9","8FE4BA","8FE4BB","8FE4BC","8FE4BD","8FE4BE","8FE4BF","8FE4C0","8FE4C1","8FE4C2","8FE4C3","8FE4C4","8FE4C5","8FE4C6","8FE4C7","8FE4C8","8FE4C9","8FE4CA","8FE4CB","8FE4CC","8FE4CD","8FE4CE","8FE4CF","8FE4D0","8FE4D1","8FE4D2","8FE4D3","8FE4D4","8FE4D5","8FE4D6","8FE4D7","8FE4D8","8FE4D9","8FE4DA","8FE4DB","8FE4DC","8FE4DD","8FE4DE","8FE4DF","8FE4E0","8FE4E1","8FE4E2","8FE4E3","8FE4E4","8FE4E5","8FE4E6","8FE4E7","8FE4E8","8FE4E9","8FE4EA","8FE4EB","8FE4EC","8FE4ED","8FE4EE","8FE4EF","8FE4F0","8FE4F1","8FE4F2","8FE4F3","8FE4F4","8FE4F5","8FE4F6","8FE4F7","8FE4F8","8FE4F9","8FE4FA","8FE4FB","8FE4FC","8FE4FD","8FE4FE","8FE5A1","8FE5A2","8FE5A3","8FE5A4","8FE5A5","8FE5A6","8FE5A7","8FE5A8","8FE5A9","8FE5AA","8FE5AB","8FE5AC","8FE5AD","8FE5AE","8FE5AF","8FE5B0","8FE5B1","8FE5B2","8FE5B3","8FE5B4","8FE5B5","8FE5B6","8FE5B7","8FE5B8","8FE5B9","8FE5BA","8FE5BB","8FE5BC","8FE5BD","8FE5BE","8FE5BF","8FE5C0","8FE5C1","8FE5C2","8FE5C3","8FE5C4","8FE5C5","8FE5C6","8FE5C7","8FE5C8","8FE5C9","8FE5CA","8FE5CB","8FE5CC","8FE5CD","8FE5CE","8FE5CF","8FE5D0","8FE5D1","8FE5D2","8FE5D3","8FE5D4","8FE5D5","8FE5D6","8FE5D7","8FE5D8","8FE5D9","8FE5DA","8FE5DB","8FE5DC","8FE5DD","8FE5DE","8FE5DF","8FE5E0","8FE5E1","8FE5E2","8FE5E3","8FE5E4","8FE5E5","8FE5E6","8FE5E7","8FE5E8","8FE5E9","8FE5EA","8FE5EB","8FE5EC","8FE5ED","8FE5EE","8FE5EF","8FE5F0","8FE5F1","8FE5F2","8FE5F3","8FE5F4","8FE5F5","8FE5F6","8FE5F7","8FE5F8","8FE5F9","8FE5FA","8FE5FB","8FE5FC","8FE5FD","8FE5FE","8FE6A1","8FE6A2","8FE6A3","8FE6A4","8FE6A5","8FE6A6","8FE6A7","8FE6A8","8FE6A9","8FE6AA","8FE6AB","8FE6AC","8FE6AD","8FE6AE","8FE6AF","8FE6B0","8FE6B1","8FE6B2","8FE6B3","8FE6B4","8FE6B5","8FE6B6","8FE6B7","8FE6B8","8FE6B9","8FE6BA","8FE6BB","8FE6BC","8FE6BD","8FE6BE","8FE6BF","8FE6C0","8FE6C1","8FE6C2","8FE6C3","8FE6C4","8FE6C5","8FE6C6","8FE6C7","8FE6C8","8FE6C9","8FE6CA","8FE6CB","8FE6CC","8FE6CD","8FE6CE","8FE6CF","8FE6D0","8FE6D1","8FE6D2","8FE6D3","8FE6D4","8FE6D5","8FE6D6","8FE6D7","8FE6D8","8FE6D9","8FE6DA","8FE6DB","8FE6DC","8FE6DD","8FE6DE","8FE6DF","8FE6E0","8FE6E1","8FE6E2","8FE6E3","8FE6E4","8FE6E5","8FE6E6","8FE6E7","8FE6E8","8FE6E9","8FE6EA","8FE6EB","8FE6EC","8FE6ED","8FE6EE","8FE6EF","8FE6F0","8FE6F1","8FE6F2","8FE6F3","8FE6F4","8FE6F5","8FE6F6","8FE6F7","8FE6F8","8FE6F9","8FE6FA","8FE6FB","8FE6FC","8FE6FD","8FE6FE","8FE7A1","8FE7A2","8FE7A3","8FE7A4","8FE7A5","8FE7A6","8FE7A7","8FE7A8","8FE7A9","8FE7AA","8FE7AB","8FE7AC","8FE7AD","8FE7AE","8FE7AF","8FE7B0","8FE7B1","8FE7B2","8FE7B3","8FE7B4","8FE7B5","8FE7B6","8FE7B7","8FE7B8","8FE7B9","8FE7BA","8FE7BB","8FE7BC","8FE7BD","8FE7BE","8FE7BF","8FE7C0","8FE7C1","8FE7C2","8FE7C3","8FE7C4","8FE7C5","8FE7C6","8FE7C7","8FE7C8","8FE7C9","8FE7CA","8FE7CB","8FE7CC","8FE7CD","8FE7CE","8FE7CF","8FE7D0","8FE7D1","8FE7D2","8FE7D3","8FE7D4","8FE7D5","8FE7D6","8FE7D7","8FE7D8","8FE7D9","8FE7DA","8FE7DB","8FE7DC","8FE7DD","8FE7DE","8FE7DF","8FE7E0","8FE7E1","8FE7E2","8FE7E3","8FE7E4","8FE7E5","8FE7E6","8FE7E7","8FE7E8","8FE7E9","8FE7EA","8FE7EB","8FE7EC","8FE7ED","8FE7EE","8FE7EF","8FE7F0","8FE7F1","8FE7F2","8FE7F3","8FE7F4","8FE7F5","8FE7F6","8FE7F7","8FE7F8","8FE7F9","8FE7FA","8FE7FB","8FE7FC","8FE7FD","8FE7FE","8FE8A1","8FE8A2","8FE8A3","8FE8A4","8FE8A5","8FE8A6","8FE8A7","8FE8A8","8FE8A9","8FE8AA","8FE8AB","8FE8AC","8FE8AD","8FE8AE","8FE8AF","8FE8B0","8FE8B1","8FE8B2","8FE8B3","8FE8B4","8FE8B5","8FE8B6","8FE8B7","8FE8B8","8FE8B9","8FE8BA","8FE8BB","8FE8BC","8FE8BD","8FE8BE","8FE8BF","8FE8C0","8FE8C1","8FE8C2","8FE8C3","8FE8C4","8FE8C5","8FE8C6","8FE8C7","8FE8C8","8FE8C9","8FE8CA","8FE8CB","8FE8CC","8FE8CD","8FE8CE","8FE8CF","8FE8D0","8FE8D1","8FE8D2","8FE8D3","8FE8D4","8FE8D5","8FE8D6","8FE8D7","8FE8D8","8FE8D9","8FE8DA","8FE8DB","8FE8DC","8FE8DD","8FE8DE","8FE8DF","8FE8E0","8FE8E1","8FE8E2","8FE8E3","8FE8E4","8FE8E5","8FE8E6","8FE8E7","8FE8E8","8FE8E9","8FE8EA","8FE8EB","8FE8EC","8FE8ED","8FE8EE","8FE8EF","8FE8F0","8FE8F1","8FE8F2","8FE8F3","8FE8F4","8FE8F5","8FE8F6","8FE8F7","8FE8F8","8FE8F9","8FE8FA","8FE8FB","8FE8FC","8FE8FD","8FE8FE","8FE9A1","8FE9A2","8FE9A3","8FE9A4","8FE9A5","8FE9A6","8FE9A7","8FE9A8","8FE9A9","8FE9AA","8FE9AB","8FE9AC","8FE9AD","8FE9AE","8FE9AF","8FE9B0","8FE9B1","8FE9B2","8FE9B3","8FE9B4","8FE9B5","8FE9B6","8FE9B7","8FE9B8","8FE9B9","8FE9BA","8FE9BB","8FE9BC","8FE9BD","8FE9BE","8FE9BF","8FE9C0","8FE9C1","8FE9C2","8FE9C3","8FE9C4","8FE9C5","8FE9C6","8FE9C7","8FE9C8","8FE9C9","8FE9CA","8FE9CB","8FE9CC","8FE9CD","8FE9CE","8FE9CF","8FE9D0","8FE9D1","8FE9D2","8FE9D3","8FE9D4","8FE9D5","8FE9D6","8FE9D7","8FE9D8","8FE9D9","8FE9DA","8FE9DB","8FE9DC","8FE9DD","8FE9DE","8FE9DF","8FE9E0","8FE9E1","8FE9E2","8FE9E3","8FE9E4","8FE9E5","8FE9E6","8FE9E7","8FE9E8","8FE9E9","8FE9EA","8FE9EB","8FE9EC","8FE9ED","8FE9EE","8FE9EF","8FE9F0","8FE9F1","8FE9F2","8FE9F3","8FE9F4","8FE9F5","8FE9F6","8FE9F7","8FE9F8","8FE9F9","8FE9FA","8FE9FB","8FE9FC","8FE9FD","8FE9FE","8FEAA1","8FEAA2","8FEAA3","8FEAA4","8FEAA5","8FEAA6","8FEAA7","8FEAA8","8FEAA9","8FEAAA","8FEAAB","8FEAAC","8FEAAD","8FEAAE","8FEAAF","8FEAB0","8FEAB1","8FEAB2","8FEAB3","8FEAB4","8FEAB5","8FEAB6","8FEAB7","8FEAB8","8FEAB9","8FEABA","8FEABB","8FEABC","8FEABD","8FEABE","8FEABF","8FEAC0","8FEAC1","8FEAC2","8FEAC3","8FEAC4","8FEAC5","8FEAC6","8FEAC7","8FEAC8","8FEAC9","8FEACA","8FEACB","8FEACC","8FEACD","8FEACE","8FEACF","8FEAD0","8FEAD1","8FEAD2","8FEAD3","8FEAD4","8FEAD5","8FEAD6","8FEAD7","8FEAD8","8FEAD9","8FEADA","8FEADB","8FEADC","8FEADD","8FEADE","8FEADF","8FEAE0","8FEAE1","8FEAE2","8FEAE3","8FEAE4","8FEAE5","8FEAE6","8FEAE7","8FEAE8","8FEAE9","8FEAEA","8FEAEB","8FEAEC","8FEAED","8FEAEE","8FEAEF","8FEAF0","8FEAF1","8FEAF2","8FEAF3","8FEAF4","8FEAF5","8FEAF6","8FEAF7","8FEAF8","8FEAF9","8FEAFA","8FEAFB","8FEAFC","8FEAFD","8FEAFE","8FEBA1","8FEBA2","8FEBA3","8FEBA4","8FEBA5","8FEBA6","8FEBA7","8FEBA8","8FEBA9","8FEBAA","8FEBAB","8FEBAC","8FEBAD","8FEBAE","8FEBAF","8FEBB0","8FEBB1","8FEBB2","8FEBB3","8FEBB4","8FEBB5","8FEBB6","8FEBB7","8FEBB8","8FEBB9","8FEBBA","8FEBBB","8FEBBC","8FEBBD","8FEBBE","8FEBBF","8FEBC0","8FEBC1","8FEBC2","8FEBC3","8FEBC4","8FEBC5","8FEBC6","8FEBC7","8FEBC8","8FEBC9","8FEBCA","8FEBCB","8FEBCC","8FEBCD","8FEBCE","8FEBCF","8FEBD0","8FEBD1","8FEBD2","8FEBD3","8FEBD4","8FEBD5","8FEBD6","8FEBD7","8FEBD8","8FEBD9","8FEBDA","8FEBDB","8FEBDC","8FEBDD","8FEBDE","8FEBDF","8FEBE0","8FEBE1","8FEBE2","8FEBE3","8FEBE4","8FEBE5","8FEBE6","8FEBE7","8FEBE8","8FEBE9","8FEBEA","8FEBEB","8FEBEC","8FEBED","8FEBEE","8FEBEF","8FEBF0","8FEBF1","8FEBF2","8FEBF3","8FEBF4","8FEBF5","8FEBF6","8FEBF7","8FEBF8","8FEBF9","8FEBFA","8FEBFB","8FEBFC","8FEBFD","8FEBFE","8FECA1","8FECA2","8FECA3","8FECA4","8FECA5","8FECA6","8FECA7","8FECA8","8FECA9","8FECAA","8FECAB","8FECAC","8FECAD","8FECAE","8FECAF","8FECB0","8FECB1","8FECB2","8FECB3","8FECB4","8FECB5","8FECB6","8FECB7","8FECB8","8FECB9","8FECBA","8FECBB","8FECBC","8FECBD","8FECBE","8FECBF","8FECC0","8FECC1","8FECC2","8FECC3","8FECC4","8FECC5","8FECC6","8FECC7","8FECC8","8FECC9","8FECCA","8FECCB","8FECCC","8FECCD","8FECCE","8FECCF","8FECD0","8FECD1","8FECD2","8FECD3","8FECD4","8FECD5","8FECD6","8FECD7","8FECD8","8FECD9","8FECDA","8FECDB","8FECDC","8FECDD","8FECDE","8FECDF","8FECE0","8FECE1","8FECE2","8FECE3","8FECE4","8FECE5","8FECE6","8FECE7","8FECE8","8FECE9","8FECEA","8FECEB","8FECEC","8FECED","8FECEE","8FECEF","8FECF0","8FECF1","8FECF2","8FECF3","8FECF4","8FECF5","8FECF6","8FECF7","8FECF8","8FECF9","8FECFA","8FECFB","8FECFC","8FECFD","8FECFE","8FEDA1","8FEDA2","8FEDA3","8FEDA4","8FEDA5","8FEDA6","8FEDA7","8FEDA8","8FEDA9","8FEDAA","8FEDAB","8FEDAC","8FEDAD","8FEDAE","8FEDAF","8FEDB0","8FEDB1","8FEDB2","8FEDB3","8FEDB4","8FEDB5","8FEDB6","8FEDB7","8FEDB8","8FEDB9","8FEDBA","8FEDBB","8FEDBC","8FEDBD","8FEDBE","8FEDBF","8FEDC0","8FEDC1","8FEDC2","8FEDC3","8FEDC4","8FEDC5","8FEDC6","8FEDC7","8FEDC8","8FEDC9","8FEDCA","8FEDCB","8FEDCC","8FEDCD","8FEDCE","8FEDCF","8FEDD0","8FEDD1","8FEDD2","8FEDD3","8FEDD4","8FEDD5","8FEDD6","8FEDD7","8FEDD8","8FEDD9","8FEDDA","8FEDDB","8FEDDC","8FEDDD","8FEDDE","8FEDDF","8FEDE0","8FEDE1","8FEDE2","8FEDE3"], [92,167,168,176,177,180,182,215,247,913,914,915,916,917,918,919,920,921,922,923,924,925,926,927,928,929,931,932,933,934,935,936,937,945,946,947,948,949,950,951,952,953,954,955,956,957,958,959,960,961,963,964,965,966,967,968,969,1025,1040,1041,1042,1043,1044,1045,1046,1047,1048,1049,1050,1051,1052,1053,1054,1055,1056,1057,1058,1059,1060,1061,1062,1063,1064,1065,1066,1067,1068,1069,1070,1071,1072,1073,1074,1075,1076,1077,1078,1079,1080,1081,1082,1083,1084,1085,1086,1087,1088,1089,1090,1091,1092,1093,1094,1095,1096,1097,1098,1099,1100,1101,1102,1103,1105,8208,8213,8216,8217,8220,8221,8224,8225,8229,8230,8240,8242,8243,8251,126,8451,8470,8481,8491,8544,8545,8546,8547,8548,8549,8550,8551,8552,8553,8560,8561,8562,8563,8564,8565,8566,8567,8568,8569,8592,8593,8594,8595,8658,8660,8704,8706,8707,8711,8712,8715,8721,65293,8730,8733,8734,8735,8736,8741,8743,8744,8745,8746,8747,8748,8750,8756,8757,8765,8786,8800,8801,8806,8807,8810,8811,8834,8835,8838,8839,8869,8895,8978,9312,9313,9314,9315,9316,9317,9318,9319,9320,9321,9322,9323,9324,9325,9326,9327,9328,9329,9330,9331,9472,9473,9474,9475,9484,9487,9488,9491,9492,9495,9496,9499,9500,9501,9504,9507,9508,9509,9512,9515,9516,9519,9520,9523,9524,9527,9528,9531,9532,9535,9538,9547,9632,9633,9650,9651,9660,9661,9670,9671,9675,9678,9679,9711,9733,9734,9792,9794,9834,9837,9839,12288,12289,12290,12291,12293,12294,12295,12296,12297,12298,12299,12300,12301,12302,12303,12304,12305,12306,12307,12308,12309,12317,12319,12353,12354,12355,12356,12357,12358,12359,12360,12361,12362,12363,12364,12365,12366,12367,12368,12369,12370,12371,12372,12373,12374,12375,12376,12377,12378,12379,12380,12381,12382,12383,12384,12385,12386,12387,12388,12389,12390,12391,12392,12393,12394,12395,12396,12397,12398,12399,12400,12401,12402,12403,12404,12405,12406,12407,12408,12409,12410,12411,12412,12413,12414,12415,12416,12417,12418,12419,12420,12421,12422,12423,12424,12425,12426,12427,12428,12429,12430,12431,12432,12433,12434,12435,12443,12444,12445,12446,12449,12450,12451,12452,12453,12454,12455,12456,12457,12458,12459,12460,12461,12462,12463,12464,12465,12466,12467,12468,12469,12470,12471,12472,12473,12474,12475,12476,12477,12478,12479,12480,12481,12482,12483,12484,12485,12486,12487,12488,12489,12490,12491,12492,12493,12494,12495,12496,12497,12498,12499,12500,12501,12502,12503,12504,12505,12506,12507,12508,12509,12510,12511,12512,12513,12514,12515,12516,12517,12518,12519,12520,12521,12522,12523,12524,12525,12526,12527,12528,12529,12530,12531,12532,12533,12534,12539,12540,12541,12542,12849,12850,12857,12964,12965,12966,12967,12968,13059,13069,13076,13080,13090,13091,13094,13095,13099,13110,13115,13129,13130,13133,13137,13143,13179,13180,13181,13182,13198,13199,13212,13213,13214,13217,13252,13261,19968,19969,19971,19975,19976,19977,19978,19979,19981,19982,19984,19985,19988,19989,19990,19991,19992,19993,19998,20001,20006,20008,20010,20013,20017,20018,20022,20024,20025,20027,20028,20031,20034,20035,20037,20043,20045,20046,20047,20053,20054,20055,20056,20057,20061,20062,20063,20066,20081,20083,20094,20096,20098,20101,20102,20104,20105,20106,20107,20108,20110,20113,20114,20116,20117,20120,20121,20123,20124,20126,20127,20128,20129,20130,20132,20133,20134,20136,20139,20140,20141,20142,20144,20147,20150,20154,20160,20161,20162,20164,20166,20167,20170,20171,20173,20174,20175,20180,20181,20182,20183,20184,20185,20189,20190,20191,20193,20195,20196,20197,20205,20206,20208,20210,20214,20215,20219,20220,20224,20225,20227,20233,20234,20237,20238,20239,20240,20241,20250,20252,20253,20271,20272,20276,20278,20280,20281,20282,20284,20285,20291,20294,20295,20301,20302,20303,20304,20305,20307,20309,20310,20311,20313,20314,20315,20316,20317,20318,20329,20335,20336,20339,20341,20342,20347,20348,20351,20355,20358,20360,20362,20363,20365,20367,20369,20370,20372,20374,20376,20378,20379,20381,20384,20385,20395,20397,20398,20399,20405,20406,20415,20418,20419,20420,20425,20426,20429,20430,20432,20433,20436,20439,20440,20442,20443,20445,20447,20449,20451,20452,20453,20462,20463,20467,20469,20470,20472,20474,20478,20479,20485,20486,20489,20491,20493,20495,20497,20498,20500,20502,20505,20506,20510,20511,20513,20514,20515,20516,20517,20518,20520,20521,20522,20523,20524,20525,20534,20537,20544,20546,20547,20550,20551,20552,20553,20559,20560,20565,20566,20570,20572,20581,20588,20592,20594,20596,20597,20598,20600,20605,20608,20613,20621,20625,20628,20632,20633,20634,20652,20653,20658,20659,20660,20661,20663,20670,20674,20677,20681,20682,20685,20687,20689,20693,20694,20696,20698,20702,20707,20709,20711,20717,20718,20724,20725,20729,20731,20736,20737,20738,20740,20745,20754,20756,20757,20758,20760,20762,20767,20769,20778,20786,20791,20794,20795,20796,20799,20800,20801,20803,20804,20805,20806,20807,20808,20809,20810,20811,20812,20813,20814,20816,20818,20820,20826,20828,20834,20836,20837,20840,20841,20842,20843,20844,20845,20846,20849,20853,20854,20855,20856,20860,20864,20866,20869,20870,20873,20874,20876,20877,20879,20880,20881,20882,20883,20885,20886,20887,20889,20893,20896,20898,20900,20901,20902,20904,20905,20906,20907,20908,20912,20913,20914,20915,20916,20917,20918,20919,20925,20926,20932,20933,20934,20937,20939,20940,20941,20950,20955,20956,20957,20960,20961,20966,20967,20969,20970,20972,20973,20976,20977,20981,20982,20984,20985,20986,20989,20990,20992,20995,20996,20998,20999,21000,21002,21003,21006,21009,21012,21013,21015,21021,21028,21029,21031,21033,21034,21038,21040,21043,21046,21047,21048,21049,21050,21051,21059,21060,21063,21066,21067,21068,21069,21071,21076,21078,21083,21086,21091,21092,21093,21097,21098,21103,21104,21105,21106,21107,21108,21109,21117,21119,21123,21127,21128,21129,21133,21137,21138,21140,21147,21148,21151,21152,21155,21158,21161,21162,21163,21164,21165,21167,21169,21172,21173,21177,21180,21182,21184,21185,21187,21189,21191,21193,21197,21202,21205,21207,21208,21209,21211,21213,21214,21215,21216,21218,21219,21220,21222,21223,21234,21235,21237,21240,21241,21242,21246,21247,21248,21249,21250,21253,21254,21255,21256,21261,21263,21264,21269,21270,21271,21273,21274,21277,21280,21281,21283,21284,21290,21295,21297,21299,21304,21305,21306,21307,21311,21312,21313,21315,21317,21318,21319,21320,21321,21322,21325,21329,21330,21331,21332,21335,21336,21338,21340,21342,21344,21350,21353,21358,21359,21360,21361,21362,21363,21364,21365,21367,21368,21371,21375,21378,21380,21395,21398,21400,21402,21407,21408,21413,21414,21416,21417,21421,21422,21424,21426,21427,21430,21435,21442,21443,21448,21449,21450,21451,21452,21453,21454,21460,21462,21463,21465,21467,21469,21471,21473,21474,21475,21476,21477,21480,21481,21482,21483,21484,21485,21486,21487,21488,21489,21490,21491,21494,21495,21496,21498,21505,21507,21508,21512,21513,21514,21515,21516,21517,21518,21519,21520,21521,21531,21533,21535,21536,21542,21545,21547,21548,21549,21550,21558,21560,21561,21563,21564,21565,21566,21568,21570,21574,21576,21577,21578,21582,21585,21599,21608,21610,21616,21617,21619,21621,21622,21623,21627,21628,21629,21632,21636,21638,21642,21643,21644,21646,21647,21648,21650,21660,21666,21668,21669,21672,21673,21675,21676,21679,21682,21683,21688,21692,21693,21694,21696,21697,21698,21700,21703,21704,21705,21720,21729,21730,21733,21734,21736,21737,21741,21742,21746,21754,21757,21759,21764,21766,21767,21775,21776,21780,21782,21806,21807,21809,21811,21816,21817,21822,21824,21828,21829,21830,21836,21839,21843,21846,21847,21852,21853,21859,21883,21884,21886,21888,21891,21892,21894,21895,21897,21898,21899,21912,21913,21914,21916,21917,21918,21919,21927,21928,21929,21930,21931,21932,21934,21936,21942,21956,21957,21959,21972,21978,21980,21983,21987,21988,22007,22009,22013,22014,22022,22025,22036,22038,22039,22040,22043,22057,22063,22065,22066,22068,22070,22072,22082,22092,22094,22096,22107,22116,22120,22122,22123,22124,22132,22136,22138,22144,22150,22151,22154,22159,22164,22176,22178,22181,22190,22196,22198,22204,22208,22209,22210,22211,22216,22222,22225,22227,22231,22232,22234,22235,22238,22240,22243,22254,22256,22258,22259,22265,22266,22269,22271,22272,22275,22276,22280,22281,22283,22285,22287,22290,22291,22294,22296,22300,22303,22310,22311,22312,22317,22320,22327,22328,22331,22336,22338,22343,22346,22350,22351,22352,22353,22361,22369,22372,22373,22374,22377,22378,22399,22402,22408,22409,22411,22419,22432,22434,22435,22436,22442,22444,22448,22451,22464,22467,22470,22471,22472,22475,22478,22482,22483,22484,22486,22492,22495,22496,22499,22516,22519,22521,22522,22524,22528,22530,22533,22534,22538,22539,22549,22553,22557,22561,22564,22570,22575,22576,22577,22580,22581,22586,22589,22592,22593,22602,22603,22609,22610,22612,22615,22616,22617,22618,22622,22626,22633,22635,22640,22642,22645,22649,22654,22659,22661,22675,22679,22684,22686,22687,22696,22699,22702,22706,22707,22712,22713,22714,22715,22718,22721,22725,22727,22730,22732,22737,22739,22741,22743,22744,22745,22748,22750,22751,22756,22757,22763,22764,22766,22767,22768,22769,22770,22775,22777,22778,22779,22780,22781,22786,22793,22794,22795,22799,22800,22805,22806,22808,22809,22810,22811,22812,22818,22821,22823,22825,22826,22827,22828,22829,22830,22833,22834,22839,22840,22846,22852,22855,22856,22857,22862,22863,22864,22865,22867,22868,22869,22871,22872,22874,22875,22877,22880,22882,22883,22885,22887,22888,22889,22890,22892,22894,22899,22900,22904,22909,22913,22914,22915,22916,22922,22925,22931,22934,22937,22939,22941,22947,22948,22949,22952,22956,22962,22969,22970,22971,22974,22982,22985,22987,22992,22993,22995,22996,23001,23002,23004,23013,23014,23016,23018,23019,23030,23035,23039,23041,23043,23049,23057,23064,23066,23068,23071,23072,23077,23081,23087,23093,23094,23100,23104,23105,23110,23113,23130,23138,23142,23146,23148,23167,23186,23194,23195,23228,23229,23230,23233,23234,23241,23243,23244,23248,23254,23255,23265,23267,23270,23273,23290,23291,23305,23307,23308,23318,23330,23338,23340,23344,23346,23350,23358,23360,23363,23365,23376,23377,23380,23381,23382,23383,23384,23386,23387,23388,23389,23391,23395,23396,23397,23398,23401,23403,23408,23409,23411,23413,23416,23418,23424,23427,23429,23431,23432,23433,23435,23436,23437,23439,23445,23447,23448,23449,23450,23451,23452,23453,23455,23458,23459,23460,23461,23462,23470,23472,23475,23476,23477,23478,23480,23481,23487,23488,23490,23491,23492,23493,23494,23495,23497,23500,23504,23506,23507,23508,23512,23515,23517,23518,23519,23521,23522,23524,23525,23526,23527,23528,23529,23531,23532,23534,23536,23539,23541,23542,23544,23546,23550,23551,23553,23554,23556,23557,23558,23559,23560,23561,23562,23563,23565,23566,23567,23569,23571,23574,23578,23582,23584,23586,23588,23592,23597,23601,23608,23609,23610,23611,23612,23613,23614,23615,23616,23617,23621,23622,23624,23626,23627,23629,23630,23631,23632,23633,23635,23637,23646,23648,23649,23652,23653,23660,23662,23663,23665,23670,23673,23692,23696,23697,23700,23713,23718,23720,23721,23723,23724,23729,23731,23734,23735,23736,23738,23739,23740,23742,23749,23751,23769,23776,23777,23784,23785,23786,23789,23791,23792,23797,23798,23802,23803,23805,23815,23819,23822,23825,23828,23829,23830,23831,23832,23833,23834,23835,23839,23842,23847,23849,23874,23883,23884,23886,23888,23890,23891,23900,23913,23916,23917,23919,23923,23926,23938,23940,23943,23947,23948,23952,23965,23970,23980,23982,23991,23992,23993,23994,23996,23997,24009,24012,24013,24016,24018,24019,24022,24027,24029,24030,24033,24035,24037,24038,24039,24040,24043,24046,24049,24050,24051,24052,24053,24055,24059,24061,24062,24066,24067,24070,24075,24076,24081,24086,24089,24090,24091,24093,24101,24107,24109,24111,24112,24115,24118,24119,24120,24125,24128,24131,24132,24133,24135,24140,24142,24148,24149,24151,24159,24161,24162,24163,24164,24178,24179,24180,24181,24182,24184,24185,24186,24187,24188,24189,24190,24191,24193,24195,24196,24199,24202,24207,24213,24214,24215,24218,24220,24224,24230,24231,24235,24237,24245,24246,24247,24248,24257,24258,24259,24264,24265,24266,24271,24272,24275,24278,24282,24283,24285,24287,24288,24289,24290,24291,24296,24297,24300,24304,24305,24307,24308,24310,24311,24312,24314,24315,24316,24318,24319,24321,24323,24324,24329,24330,24331,24332,24333,24335,24336,24337,24339,24340,24341,24342,24343,24344,24347,24351,24353,24357,24358,24359,24361,24365,24367,24369,24372,24373,24375,24376,24380,24382,24385,24389,24392,24394,24396,24398,24401,24403,24406,24407,24409,24412,24413,24417,24418,24422,24423,24425,24426,24427,24428,24429,24432,24433,24435,24439,24441,24444,24447,24448,24449,24450,24451,24452,24453,24455,24456,24458,24459,24460,24464,24465,24466,24467,24471,24472,24473,24478,24480,24481,24488,24489,24490,24493,24494,24499,24500,24503,24505,24508,24509,24515,24517,24524,24525,24534,24535,24536,24537,24540,24541,24542,24544,24548,24555,24560,24561,24565,24568,24571,24573,24575,24590,24591,24592,24594,24597,24598,24601,24603,24604,24605,24608,24609,24613,24614,24615,24616,24617,24618,24619,24623,24625,24634,24641,24642,24643,24646,24650,24651,24653,24656,24658,24661,24665,24666,24669,24671,24672,24674,24675,24676,24677,24680,24681,24682,24683,24684,24685,24687,24688,24693,24695,24705,24707,24708,24709,24713,24714,24715,24716,24717,24722,24724,24726,24727,24730,24731,24735,24736,24739,24742,24743,24745,24746,24754,24755,24756,24757,24758,24760,24764,24765,24773,24774,24775,24785,24787,24789,24792,24794,24796,24798,24799,24800,24801,24803,24807,24808,24816,24817,24818,24819,24820,24822,24823,24825,24826,24827,24832,24833,24835,24838,24840,24841,24845,24846,24847,24849,24853,24858,24859,24863,24864,24865,24871,24872,24876,24880,24884,24887,24892,24893,24894,24895,24898,24900,24903,24904,24906,24907,24908,24909,24910,24915,24917,24920,24921,24922,24925,24927,24930,24931,24933,24935,24936,24939,24942,24943,24944,24945,24947,24948,24949,24950,24951,24958,24962,24967,24970,24974,24976,24977,24980,24982,24984,24985,24986,24996,24999,25001,25003,25004,25006,25010,25014,25018,25022,25027,25030,25031,25032,25033,25034,25035,25036,25037,25040,25059,25062,25074,25076,25078,25079,25080,25082,25084,25085,25086,25087,25088,25096,25097,25098,25100,25101,25102,25104,25105,25106,25107,25108,25110,25114,25115,25117,25118,25119,25121,25126,25130,25134,25135,25136,25138,25139,25140,25144,25147,25151,25152,25153,25159,25160,25161,25163,25165,25166,25171,25173,25176,25179,25182,25184,25187,25192,25198,25201,25206,25209,25212,25214,25215,25216,25218,25219,25220,25225,25226,25233,25234,25235,25236,25237,25238,25239,25240,25243,25244,25246,25254,25259,25260,25265,25269,25273,25275,25276,25277,25282,25285,25286,25287,25288,25289,25290,25292,25293,25295,25296,25297,25298,25299,25300,25303,25304,25305,25307,25308,25309,25312,25313,25324,25325,25326,25327,25329,25331,25333,25334,25335,25342,25343,25345,25346,25351,25352,25353,25356,25361,25369,25375,25383,25384,25387,25391,25402,25405,25406,25407,25417,25420,25421,25423,25424,25429,25431,25436,25447,25448,25449,25451,25454,25458,25462,25463,25466,25467,25472,25475,25480,25481,25484,25486,25487,25490,25494,25496,25499,25503,25504,25505,25506,25507,25509,25511,25512,25513,25514,25515,25516,25522,25524,25525,25531,25534,25536,25539,25540,25542,25545,25551,25552,25554,25558,25562,25563,25569,25571,25577,25582,25588,25589,25590,25594,25606,25613,25615,25619,25622,25623,25628,25638,25640,25644,25645,25652,25654,25658,25662,25666,25678,25688,25696,25703,25705,25711,25718,25720,25722,25731,25736,25746,25747,25749,25754,25757,25758,25764,25765,25769,25771,25773,25774,25776,25778,25785,25787,25788,25793,25794,25797,25799,25805,25806,25810,25812,25816,25818,25824,25825,25826,25827,25830,25831,25836,25839,25841,25842,25844,25846,25850,25853,25854,25856,25861,25880,25884,25885,25891,25892,25898,25899,25900,25903,25908,25909,25910,25911,25912,25913,25915,25918,25919,25925,25928,25933,25934,25935,25937,25941,25942,25943,25944,25945,25949,25950,25954,25955,25958,25964,25968,25970,25972,25973,25975,25976,25986,25987,25991,25992,25993,25996,25998,26000,26001,26007,26009,26011,26012,26015,26017,26020,26021,26023,26027,26028,26029,26031,26032,26039,26041,26044,26045,26049,26051,26052,26053,26054,26059,26060,26063,26066,26071,26073,26075,26080,26081,26082,26085,26086,26087,26088,26089,26092,26093,26097,26106,26107,26112,26114,26115,26118,26119,26121,26122,26124,26126,26127,26131,26132,26133,26140,26142,26143,26144,26148,26149,26151,26152,26157,26158,26159,26161,26164,26165,26166,26171,26172,26175,26177,26178,26179,26180,26185,26187,26191,26194,26199,26201,26205,26206,26207,26210,26212,26213,26214,26215,26216,26217,26222,26223,26224,26227,26228,26230,26234,26241,26243,26244,26247,26248,26249,26254,26257,26262,26263,26264,26265,26269,26272,26274,26278,26283,26286,26290,26292,26296,26297,26300,26302,26303,26305,26308,26311,26313,26326,26329,26330,26332,26333,26336,26342,26345,26352,26354,26355,26356,26357,26359,26360,26361,26362,26363,26364,26365,26366,26367,26368,26371,26376,26377,26379,26381,26382,26383,26388,26389,26390,26391,26395,26397,26398,26399,26406,26407,26408,26410,26411,26412,26413,26414,26417,26420,26422,26423,26424,26426,26429,26431,26433,26438,26441,26446,26447,26448,26449,26451,26454,26457,26460,26462,26463,26464,26465,26466,26467,26468,26469,26470,26474,26477,26479,26480,26481,26482,26483,26485,26487,26492,26494,26495,26501,26503,26505,26507,26508,26512,26517,26519,26522,26524,26525,26528,26529,26530,26534,26537,26543,26547,26548,26550,26551,26552,26553,26555,26560,26561,26564,26566,26570,26574,26575,26576,26577,26579,26580,26584,26586,26589,26590,26594,26596,26599,26601,26604,26606,26607,26609,26611,26612,26613,26619,26622,26623,26625,26626,26627,26628,26643,26646,26647,26654,26657,26658,26665,26666,26667,26674,26676,26680,26681,26684,26685,26688,26689,26690,26691,26692,26694,26696,26701,26702,26704,26705,26706,26707,26708,26713,26716,26717,26719,26723,26727,26740,26742,26743,26750,26751,26753,26755,26757,26765,26767,26771,26772,26775,26779,26781,26783,26784,26786,26790,26791,26792,26797,26799,26800,26801,26803,26805,26806,26809,26810,26812,26820,26822,26824,26825,26826,26827,26829,26831,26834,26836,26837,26839,26840,26842,26847,26848,26849,26851,26855,26862,26863,26866,26873,26874,26880,26881,26884,26885,26888,26891,26892,26893,26894,26895,26898,26905,26906,26907,26908,26913,26914,26915,26917,26918,26920,26922,26928,26932,26934,26937,26941,26943,26954,26963,26964,26965,26969,26970,26972,26973,26974,26976,26977,26978,26984,26986,26987,26989,26990,26991,26995,26996,26997,26999,27000,27001,27004,27005,27006,27009,27010,27018,27022,27025,27028,27029,27032,27035,27036,27040,27047,27054,27057,27058,27060,27067,27070,27071,27073,27075,27079,27082,27083,27084,27085,27086,27088,27091,27096,27097,27101,27102,27106,27111,27112,27115,27117,27122,27129,27131,27133,27135,27138,27141,27146,27147,27148,27154,27155,27156,27159,27161,27163,27166,27167,27169,27170,27171,27177,27178,27179,27182,27184,27189,27190,27192,27193,27194,27197,27204,27206,27207,27208,27211,27224,27225,27231,27233,27234,27238,27243,27250,27251,27256,27262,27263,27264,27268,27277,27278,27280,27287,27292,27296,27298,27299,27306,27308,27310,27315,27320,27323,27329,27330,27331,27345,27347,27354,27355,27358,27359,27362,27364,27368,27370,27386,27387,27396,27397,27402,27410,27414,27421,27423,27424,27425,27427,27431,27442,27447,27448,27449,27450,27453,27454,27459,27463,27465,27468,27470,27472,27475,27476,27481,27483,27487,27489,27490,27491,27492,27494,27497,27498,27503,27507,27508,27512,27513,27515,27519,27520,27523,27524,27526,27529,27530,27531,27533,27541,27542,27544,27550,27556,27562,27563,27567,27569,27570,27571,27572,27573,27575,27578,27579,27580,27583,27584,27589,27590,27595,27597,27598,27602,27603,27604,27606,27608,27611,27615,27627,27628,27631,27635,27656,27663,27665,27667,27668,27671,27675,27683,27684,27700,27703,27704,27710,27711,27712,27713,27714,27726,27728,27733,27735,27738,27740,27741,27742,27743,27744,27746,27752,27754,27759,27760,27762,27763,27770,27773,27774,27777,27778,27779,27782,27784,27788,27789,27792,27794,27795,27798,27801,27802,27803,27809,27810,27819,27822,27825,27827,27832,27833,27834,27835,27836,27837,27838,27839,27841,27844,27845,27849,27850,27852,27859,27861,27863,27865,27866,27867,27869,27873,27874,27875,27877,27880,27882,27887,27888,27889,27891,27908,27915,27916,27922,27927,27929,27931,27934,27935,27941,27945,27946,27947,27954,27955,27957,27958,27960,27963,27965,27966,27969,27972,27973,27993,27994,27996,28003,28004,28006,28009,28010,28012,28014,28015,28020,28023,28024,28025,28037,28039,28040,28044,28046,28051,28053,28054,28057,28059,28060,28076,28079,28082,28085,28088,28092,28096,28101,28102,28103,28107,28108,28111,28113,28114,28117,28120,28121,28126,28129,28132,28134,28136,28138,28139,28140,28142,28145,28146,28147,28149,28151,28152,28153,28154,28155,28156,28165,28167,28168,28169,28170,28171,28179,28181,28185,28186,28187,28189,28191,28192,28193,28195,28196,28197,28198,28199,28201,28203,28204,28205,28206,28207,28216,28217,28218,28220,28222,28227,28234,28237,28238,28246,28248,28251,28252,28255,28263,28267,28270,28271,28274,28278,28286,28287,28288,28290,28300,28303,28304,28310,28312,28316,28317,28319,28322,28325,28330,28335,28338,28342,28343,28346,28349,28351,28354,28356,28357,28361,28363,28364,28369,28371,28372,28373,28381,28382,28396,28399,28402,28404,28407,28408,28414,28415,28417,28418,28422,28425,28431,28433,28435,28436,28437,28448,28450,28451,28459,28460,28465,28466,28472,28478,28479,28481,28485,28500,28504,28507,28508,28511,28516,28518,28525,28526,28527,28528,28532,28536,28538,28540,28544,28545,28546,28548,28550,28552,28558,28561,28567,28577,28579,28580,28586,28593,28595,28597,28601,28608,28609,28610,28611,28614,28628,28629,28632,28635,28639,28640,28641,28644,28651,28652,28654,28655,28657,28659,28661,28662,28666,28670,28673,28677,28679,28681,28683,28687,28689,28693,28696,28698,28699,28701,28702,28703,28710,28711,28712,28716,28720,28722,28734,28748,28753,28760,28771,28779,28783,28784,28792,28796,28797,28805,28809,28810,28814,28818,28825,28843,28844,28845,28846,28847,28851,28856,28857,28858,28859,28872,28875,28879,28889,28893,28895,28913,28921,28925,28932,28937,28943,28948,28953,28954,28956,28961,28966,28982,28988,28998,28999,29001,29004,29006,29013,29014,29017,29020,29026,29028,29029,29030,29031,29033,29036,29038,29053,29060,29064,29066,29071,29076,29077,29081,29087,29096,29100,29105,29113,29118,29121,29123,29128,29129,29134,29136,29138,29140,29141,29143,29151,29152,29157,29158,29159,29164,29165,29166,29173,29177,29179,29180,29182,29183,29190,29197,29200,29211,29224,29226,29228,29229,29232,29234,29237,29238,29242,29243,29244,29245,29246,29247,29248,29254,29255,29256,29259,29260,29266,29272,29273,29275,29277,29279,29281,29282,29287,29289,29298,29300,29305,29309,29310,29312,29313,29314,29319,29330,29334,29344,29346,29351,29356,29359,29361,29362,29366,29369,29374,29378,29379,29380,29382,29390,29392,29394,29399,29401,29403,29408,29409,29410,29417,29420,29421,29431,29432,29433,29436,29437,29450,29462,29463,29467,29468,29469,29471,29476,29477,29481,29482,29483,29486,29487,29492,29494,29495,29502,29503,29508,29509,29518,29519,29527,29539,29544,29546,29552,29554,29557,29559,29560,29562,29563,29572,29575,29577,29579,29590,29609,29618,29619,29627,29629,29632,29634,29640,29641,29642,29645,29646,29650,29654,29662,29664,29667,29669,29674,29677,29678,29681,29685,29688,29694,29699,29701,29702,29703,29705,29730,29733,29734,29737,29738,29742,29746,29747,29748,29749,29750,29754,29759,29761,29781,29785,29786,29787,29788,29790,29791,29792,29794,29795,29796,29801,29802,29807,29808,29811,29814,29822,29827,29833,29835,29854,29855,29858,29863,29872,29885,29898,29903,29908,29916,29920,29922,29923,29926,29927,29929,29934,29936,29937,29938,29942,29943,29944,29953,29955,29956,29957,29964,29965,29966,29969,29971,29973,29976,29978,29980,29982,29983,29987,29989,29990,29992,29995,29996,29999,30000,30001,30002,30003,30007,30008,30010,30011,30012,30020,30022,30025,30026,30027,30028,30029,30031,30033,30036,30041,30042,30043,30044,30045,30048,30050,30052,30053,30054,30055,30057,30058,30059,30061,30063,30064,30067,30068,30070,30071,30072,30079,30082,30086,30087,30089,30090,30091,30094,30095,30097,30100,30106,30109,30115,30117,30123,30129,30130,30131,30133,30136,30137,30140,30141,30142,30146,30147,30149,30151,30154,30157,30162,30164,30165,30168,30169,30171,30174,30178,30179,30185,30192,30194,30195,30196,30202,30204,30206,30207,30209,30217,30219,30221,30239,30240,30241,30242,30244,30247,30256,30260,30267,30274,30278,30279,30280,30284,30290,30294,30296,30300,30305,30306,30311,30312,30313,30314,30316,30320,30322,30326,30328,30330,30331,30332,30333,30334,30336,30338,30339,30340,30342,30343,30344,30347,30350,30352,30355,30358,30361,30362,30363,30364,30366,30374,30382,30384,30388,30391,30392,30393,30394,30399,30402,30403,30406,30408,30410,30413,30418,30422,30423,30427,30428,30430,30431,30433,30435,30436,30437,30439,30442,30446,30450,30452,30456,30459,30462,30465,30468,30471,30472,30473,30475,30476,30491,30494,30495,30496,30500,30501,30502,30505,30519,30520,30522,30524,30528,30534,30535,30554,30555,30561,30563,30565,30566,30568,30571,30585,30590,30591,30603,30606,30609,30622,30624,30629,30636,30637,30640,30643,30646,30649,30651,30652,30653,30655,30663,30669,30679,30682,30683,30684,30690,30691,30693,30695,30697,30701,30702,30703,30707,30716,30722,30732,30738,30740,30741,30752,30753,30757,30758,30759,30770,30772,30778,30783,30789,30798,30813,30820,30827,30828,30831,30834,30836,30842,30844,30849,30854,30855,30860,30861,30862,30865,30867,30869,30871,30874,30883,30887,30889,30890,30895,30901,30906,30908,30910,30913,30917,30918,30922,30923,30928,30929,30932,30938,30951,30952,30956,30959,30964,30973,30977,30983,30990,30993,30994,31001,31014,31018,31019,31020,31024,31034,31036,31038,31040,31041,31047,31048,31049,31056,31059,31061,31062,31063,31066,31069,31070,31071,31072,31074,31077,31080,31085,31095,31098,31103,31104,31105,31108,31109,31114,31117,31118,31119,31124,31131,31133,31142,31143,31146,31150,31152,31155,31161,31162,31165,31166,31167,31168,31169,31177,31179,31185,31186,31189,31192,31199,31201,31203,31204,31206,31207,31209,31212,31216,31227,31232,31240,31243,31245,31246,31252,31255,31256,31257,31258,31260,31263,31264,31278,31281,31282,31287,31291,31292,31293,31294,31295,31296,31298,31299,31302,31305,31309,31310,31311,31312,31319,31329,31330,31331,31337,31339,31344,31348,31350,31353,31354,31357,31359,31361,31363,31364,31368,31378,31379,31381,31382,31383,31384,31391,31401,31402,31406,31407,31408,31414,31418,31423,31427,31428,31429,31431,31432,31434,31435,31437,31439,31441,31442,31443,31445,31449,31450,31452,31453,31455,31456,31457,31458,31459,31461,31462,31463,31466,31467,31469,31471,31472,31478,31480,31481,31482,31487,31490,31492,31494,31496,31498,31499,31503,31505,31512,31513,31515,31518,31520,31525,31526,31528,31532,31539,31541,31542,31545,31557,31558,31560,31561,31563,31564,31565,31567,31568,31569,31570,31572,31574,31581,31589,31591,31596,31598,31600,31601,31604,31605,31610,31622,31623,31627,31629,31631,31634,31636,31637,31639,31640,31641,31642,31644,31645,31646,31647,31649,31658,31661,31665,31668,31672,31680,31681,31684,31686,31687,31689,31691,31692,31695,31709,31712,31716,31717,31718,31721,31725,31731,31734,31735,31744,31751,31757,31761,31762,31763,31764,31767,31775,31777,31779,31783,31786,31787,31799,31800,31805,31806,31807,31808,31811,31820,31821,31823,31824,31828,31830,31832,31839,31840,31844,31845,31852,31859,31861,31870,31873,31874,31875,31881,31883,31885,31888,31890,31893,31895,31896,31899,31903,31905,31906,31908,31909,31911,31912,31915,31917,31918,31921,31922,31923,31929,31933,31934,31936,31938,31941,31946,31950,31954,31958,31960,31964,31966,31967,31968,31970,31975,31983,31986,31988,31990,31992,31994,31995,31998,32000,32002,32004,32005,32006,32010,32011,32013,32016,32020,32021,32023,32024,32025,32026,32027,32028,32032,32033,32034,32043,32044,32046,32047,32048,32050,32051,32053,32057,32058,32063,32066,32067,32068,32069,32070,32072,32075,32076,32078,32079,32080,32086,32091,32092,32094,32097,32098,32099,32102,32104,32110,32113,32114,32115,32117,32118,32121,32125,32137,32143,32147,32153,32154,32155,32156,32159,32160,32162,32163,32171,32172,32173,32174,32175,32176,32177,32178,32180,32181,32183,32184,32186,32187,32189,32190,32191,32199,32202,32203,32207,32209,32210,32213,32214,32216,32218,32220,32221,32222,32224,32225,32228,32232,32233,32236,32239,32242,32244,32251,32257,32260,32261,32265,32266,32267,32274,32283,32286,32287,32289,32290,32291,32294,32299,32302,32305,32306,32309,32311,32313,32314,32315,32317,32318,32321,32323,32326,32330,32331,32333,32338,32340,32341,32342,32345,32346,32349,32350,32358,32359,32361,32362,32365,32368,32377,32379,32380,32381,32383,32386,32387,32392,32393,32394,32396,32398,32399,32400,32402,32403,32404,32406,32411,32412,32566,32568,32570,32581,32583,32588,32589,32590,32592,32593,32596,32597,32600,32607,32608,32615,32616,32617,32618,32619,32622,32624,32626,32629,32631,32632,32633,32642,32643,32645,32646,32647,32648,32650,32652,32654,32660,32666,32669,32670,32673,32675,32676,32680,32681,32686,32687,32690,32694,32696,32697,32701,32705,32709,32710,32714,32716,32722,32724,32725,32736,32737,32742,32745,32747,32752,32755,32761,32763,32764,32768,32769,32771,32772,32773,32774,32779,32780,32784,32786,32789,32791,32792,32793,32796,32801,32808,32819,32822,32827,32829,32831,32838,32842,32850,32854,32856,32858,32862,32863,32865,32866,32872,32879,32880,32882,32883,32884,32886,32887,32889,32893,32894,32895,32900,32901,32902,32903,32905,32907,32908,32915,32918,32920,32922,32923,32925,32929,32930,32933,32937,32938,32940,32941,32943,32945,32946,32948,32954,32963,32964,32966,32972,32974,32982,32985,32986,32987,32989,32990,32993,32996,32997,33007,33009,33012,33016,33020,33021,33026,33029,33030,33031,33032,33033,33034,33050,33051,33059,33065,33071,33073,33075,33081,33086,33094,33099,33102,33104,33105,33107,33108,33109,33119,33125,33126,33131,33134,33136,33137,33140,33144,33145,33146,33151,33152,33154,33155,33160,33162,33167,33171,33173,33178,33180,33181,33184,33187,33188,33192,33193,33200,33203,33205,33208,33210,33213,33214,33215,33216,33218,33222,33224,33225,33229,33233,33235,33240,33241,33242,33247,33248,33251,33253,33255,33256,33258,33261,33267,33268,33274,33275,33276,33278,33281,33282,33285,33287,33288,33289,33290,33292,33293,33294,33296,33298,33302,33303,33304,33307,33308,33310,33311,33321,33322,33323,33324,33326,33331,33333,33334,33335,33336,33337,33344,33351,33368,33369,33370,33373,33375,33378,33380,33382,33384,33386,33387,33390,33391,33393,33394,33398,33399,33400,33406,33419,33421,33426,33433,33437,33439,33445,33446,33451,33452,33453,33455,33457,33459,33464,33465,33467,33469,33477,33489,33490,33491,33492,33495,33497,33499,33500,33502,33503,33505,33507,33509,33510,33511,33515,33521,33523,33524,33529,33530,33531,33537,33538,33539,33540,33541,33542,33545,33550,33558,33559,33560,33564,33571,33576,33579,33583,33585,33586,33588,33589,33590,33592,33593,33600,33605,33609,33610,33615,33616,33618,33624,33634,33651,33653,33655,33659,33660,33663,33669,33671,33673,33674,33678,33683,33686,33690,33694,33695,33696,33698,33704,33706,33707,33713,33717,33725,33729,33733,33735,33738,33740,33742,33747,33750,33752,33756,33759,33760,33769,33771,33775,33776,33777,33778,33780,33782,33783,33787,33789,33795,33796,33799,33803,33804,33805,33806,33811,33824,33826,33833,33834,33836,33841,33845,33848,33852,33853,33862,33864,33865,33870,33879,33883,33889,33890,33891,33894,33897,33899,33900,33901,33902,33903,33905,33909,33911,33913,33914,33922,33924,33931,33936,33940,33945,33948,33951,33953,33965,33970,33972,33976,33977,33979,33980,33983,33985,33988,33990,33993,33994,33995,33997,34000,34001,34006,34009,34010,34012,34028,34030,34036,34044,34047,34048,34054,34065,34067,34068,34069,34071,34072,34074,34079,34081,34086,34092,34093,34101,34109,34112,34113,34115,34120,34121,34122,34123,34126,34131,34133,34135,34136,34137,34138,34147,34152,34153,34154,34155,34157,34167,34174,34176,34180,34183,34184,34186,34192,34193,34196,34199,34201,34203,34204,34212,34214,34216,34217,34218,34219,34220,34222,34223,34224,34233,34234,34241,34249,34253,34255,34256,34261,34268,34269,34276,34277,34281,34282,34295,34297,34298,34299,34302,34306,34310,34311,34314,34315,34323,34326,34327,34330,34338,34349,34351,34352,34367,34381,34382,34384,34388,34389,34394,34396,34398,34399,34407,34411,34417,34425,34427,34442,34443,34444,34451,34453,34467,34468,34473,34474,34475,34479,34480,34486,34500,34502,34503,34505,34507,34509,34510,34516,34521,34523,34526,34527,34532,34537,34540,34541,34542,34543,34552,34553,34555,34558,34560,34562,34563,34566,34568,34569,34570,34573,34577,34578,34584,34586,34588,34597,34601,34612,34615,34619,34623,34633,34635,34636,34638,34643,34645,34647,34649,34655,34656,34659,34662,34664,34666,34670,34676,34678,34680,34687,34690,34701,34719,34722,34731,34735,34739,34746,34747,34749,34752,34756,34758,34759,34763,34768,34770,34784,34799,34802,34806,34807,34809,34811,34814,34821,34823,34829,34830,34831,34833,34837,34838,34849,34850,34851,34855,34865,34870,34873,34875,34880,34882,34884,34886,34892,34893,34898,34899,34903,34905,34907,34909,34910,34913,34914,34915,34920,34923,34928,34930,34933,34935,34941,34942,34943,34945,34946,34952,34955,34957,34962,34966,34967,34969,34974,34978,34980,34987,34990,34992,34993,34996,34997,34999,35007,35009,35010,35011,35012,35013,35023,35028,35029,35032,35033,35036,35037,35039,35041,35048,35058,35059,35060,35061,35064,35065,35068,35069,35070,35074,35076,35079,35082,35084,35088,35090,35091,35100,35101,35102,35109,35114,35115,35126,35128,35131,35137,35139,35140,35148,35149,35158,35166,35167,35168,35172,35174,35178,35181,35183,35186,35188,35191,35198,35199,35201,35203,35206,35207,35208,35210,35211,35215,35219,35222,35223,35224,35226,35233,35238,35239,35241,35242,35244,35247,35250,35251,35258,35261,35263,35264,35282,35290,35292,35293,35299,35302,35303,35316,35320,35328,35330,35331,35336,35338,35340,35342,35344,35346,35347,35350,35351,35352,35355,35357,35359,35363,35365,35370,35373,35377,35379,35380,35382,35383,35386,35387,35388,35393,35398,35400,35408,35409,35410,35412,35413,35419,35422,35424,35426,35427,35430,35433,35435,35436,35437,35438,35440,35441,35442,35443,35449,35452,35458,35460,35461,35463,35465,35468,35469,35473,35475,35477,35480,35482,35486,35488,35489,35491,35492,35493,35494,35495,35496,35500,35501,35504,35506,35513,35516,35518,35519,35522,35524,35527,35531,35532,35533,35535,35538,35542,35546,35547,35548,35550,35551,35552,35553,35554,35556,35558,35559,35563,35565,35566,35569,35571,35574,35575,35576,35578,35582,35584,35585,35586,35588,35591,35596,35598,35600,35604,35606,35607,35609,35610,35611,35613,35616,35617,35622,35624,35627,35628,35635,35641,35646,35649,35657,35660,35662,35663,35667,35670,35672,35674,35675,35676,35679,35686,35691,35692,35695,35696,35697,35698,35700,35703,35709,35711,35712,35715,35722,35724,35726,35728,35730,35731,35734,35737,35738,35895,35898,35903,35905,35910,35912,35914,35916,35918,35920,35925,35930,35937,35938,35946,35947,35948,35960,35961,35962,35964,35970,35973,35977,35978,35980,35981,35982,35988,35992,35997,35998,36000,36001,36002,36007,36008,36009,36010,36011,36012,36013,36014,36015,36016,36018,36019,36020,36022,36023,36024,36027,36028,36029,36031,36032,36033,36034,36035,36036,36039,36040,36042,36045,36046,36049,36051,36058,36059,36060,36062,36064,36066,36067,36068,36070,36074,36077,36080,36084,36090,36091,36092,36093,36100,36101,36103,36104,36106,36107,36109,36111,36112,36114,36115,36116,36118,36196,36198,36199,36203,36205,36208,36209,36211,36212,36214,36215,36225,36229,36234,36249,36259,36264,36275,36282,36286,36290,36299,36300,36303,36310,36314,36315,36317,36319,36321,36323,36328,36330,36331,36335,36339,36341,36348,36351,36360,36361,36362,36367,36368,36381,36382,36383,36394,36400,36404,36405,36418,36420,36423,36424,36425,36426,36428,36432,36437,36441,36447,36448,36451,36452,36466,36468,36470,36476,36481,36484,36485,36487,36490,36491,36493,36497,36499,36500,36505,36513,36522,36523,36524,36527,36528,36529,36542,36549,36550,36552,36554,36555,36556,36557,36559,36562,36571,36575,36578,36579,36587,36600,36603,36604,36605,36606,36611,36613,36617,36618,36620,36626,36627,36628,36629,36633,36635,36636,36637,36639,36646,36649,36650,36655,36659,36664,36665,36667,36670,36671,36674,36676,36677,36678,36681,36684,36685,36686,36695,36700,36703,36705,36706,36707,36708,36763,36764,36766,36767,36771,36775,36776,36781,36782,36783,36784,36785,36786,36791,36794,36795,36796,36799,36802,36804,36805,36814,36817,36820,36826,36834,36837,36838,36841,36842,36843,36845,36847,36848,36852,36855,36856,36857,36858,36861,36864,36865,36867,36869,36870,36875,36877,36878,36879,36880,36881,36883,36884,36885,36886,36887,36889,36890,36893,36894,36895,36896,36897,36898,36899,36903,36910,36913,36914,36917,36918,36920,36921,36924,36926,36929,36930,36933,36935,36937,36938,36939,36941,36942,36943,36944,36945,36946,36947,36948,36949,36950,36952,36953,36956,36958,36960,36961,36963,36965,36967,36968,36969,36973,36974,36975,36978,36981,36982,36983,36984,36986,36988,36989,36991,36992,36993,36994,36995,36996,36999,37001,37002,37007,37009,37027,37030,37032,37034,37039,37041,37045,37048,37057,37066,37070,37083,37086,37089,37090,37092,37096,37101,37109,37111,37117,37122,37138,37141,37145,37159,37165,37168,37170,37193,37194,37195,37196,37197,37198,37202,37204,37206,37208,37218,37219,37221,37225,37226,37228,37234,37235,37237,37239,37240,37250,37255,37257,37259,37261,37264,37266,37271,37276,37282,37284,37290,37291,37295,37300,37301,37304,37306,37312,37313,37318,37319,37320,37321,37323,37324,37325,37326,37327,37328,37329,37334,37335,37336,37338,37339,37340,37341,37342,37343,37345,37347,37348,37349,37350,37351,37357,37358,37365,37366,37372,37375,37382,37386,37389,37390,37392,37393,37396,37397,37406,37417,37420,37428,37431,37433,37434,37436,37439,37440,37444,37445,37448,37449,37451,37454,37456,37457,37463,37465,37466,37467,37470,37474,37476,37478,37479,37489,37495,37496,37502,37504,37507,37509,37512,37521,37523,37525,37526,37528,37530,37531,37532,37543,37549,37559,37561,37583,37584,37586,37587,37589,37591,37593,37600,37604,37607,37609,37610,37613,37618,37619,37624,37625,37626,37627,37628,37631,37634,37638,37647,37648,37656,37657,37658,37661,37662,37664,37665,37666,37667,37669,37670,37672,37675,37676,37678,37679,37682,37685,37690,37691,37700,37704,37707,37709,37716,37718,37719,37723,37724,37728,37740,37742,37744,37749,37756,37758,37772,37780,37782,37783,37786,37796,37799,37804,37805,37806,37808,37817,37827,37830,37832,37840,37841,37846,37847,37848,37853,37854,37857,37860,37861,37864,37880,37891,37895,37904,37907,37908,37912,37913,37914,37921,37931,37937,37941,37942,37944,37946,37953,37956,37957,37960,37969,37970,37971,37978,37979,37982,37984,37986,37994,38000,38005,38007,38012,38013,38014,38015,38017,38263,38272,38274,38275,38279,38281,38282,38283,38287,38289,38290,38291,38292,38294,38296,38297,38304,38306,38307,38308,38309,38311,38312,38317,38322,38329,38331,38332,38334,38339,38343,38346,38348,38349,38356,38357,38358,38360,38364,38369,38370,38373,38428,38433,38440,38442,38446,38447,38450,38459,38463,38464,38466,38468,38475,38476,38477,38479,38480,38491,38492,38493,38494,38495,38498,38499,38500,38501,38502,38506,38508,38512,38514,38515,38517,38518,38519,38520,38522,38525,38533,38534,38536,38538,38539,38541,38542,38543,38548,38549,38551,38552,38553,38555,38556,38557,38560,38563,38567,38568,38570,38575,38576,38577,38578,38580,38582,38583,38584,38585,38587,38588,38592,38593,38596,38597,38598,38599,38601,38603,38604,38605,38606,38609,38613,38614,38617,38619,38620,38626,38627,38632,38634,38635,38640,38642,38646,38647,38649,38651,38656,38660,38662,38663,38664,38666,38669,38670,38671,38673,38675,38678,38681,38684,38686,38692,38695,38698,38704,38706,38707,38712,38713,38715,38717,38718,38722,38723,38724,38726,38728,38729,38733,38735,38737,38738,38741,38742,38745,38748,38750,38752,38753,38754,38756,38758,38760,38761,38763,38765,38769,38772,38777,38778,38780,38785,38788,38789,38790,38795,38797,38799,38800,38808,38812,38816,38819,38822,38824,38827,38829,38835,38836,38851,38854,38856,38859,38867,38876,38893,38894,38898,38899,38901,38902,38907,38911,38913,38914,38915,38917,38918,38920,38924,38927,38928,38929,38930,38931,38935,38936,38938,38945,38948,38956,38957,38964,38967,38968,38971,38972,38973,38982,38987,38988,38989,38990,38991,38996,38997,38999,39000,39003,39006,39013,39015,39019,39023,39024,39025,39027,39028,39080,39082,39087,39089,39094,39107,39108,39110,39131,39132,39135,39138,39145,39147,39149,39150,39151,39154,39156,39164,39165,39166,39171,39173,39177,39178,39180,39184,39186,39187,39188,39192,39197,39198,39200,39201,39204,39207,39208,39212,39214,39229,39230,39234,39237,39241,39243,39244,39248,39249,39250,39253,39255,39318,39319,39320,39321,39326,39333,39336,39340,39341,39342,39347,39348,39356,39361,39364,39365,39366,39368,39376,39377,39378,39381,39384,39387,39389,39391,39394,39405,39406,39409,39410,39416,39419,39423,39425,39429,39438,39439,39442,39443,39449,39464,39467,39472,39479,39486,39488,39490,39491,39493,39501,39502,39509,39511,39514,39515,39519,39522,39524,39525,39529,39530,39531,39592,39597,39600,39608,39612,39616,39620,39631,39633,39635,39636,39640,39641,39644,39646,39647,39650,39651,39654,39658,39659,39661,39662,39663,39665,39668,39671,39675,39686,39704,39706,39711,39714,39715,39717,39719,39720,39721,39722,39726,39727,39729,39730,39739,39740,39745,39746,39747,39748,39749,39757,39758,39759,39761,39764,39768,39770,39791,39794,39796,39797,39811,39822,39823,39825,39826,39827,39830,39831,39839,39840,39848,39850,39851,39853,39854,39857,39860,39865,39867,39872,39878,39881,39882,39887,39889,39890,39892,39894,39899,39905,39906,39907,39908,39912,39920,39921,39922,39925,39936,39940,39942,39944,39945,39946,39948,39949,39952,39954,39955,39956,39957,39963,39969,39972,39973,39981,39982,39983,39984,39986,39993,39994,39995,39998,40006,40007,40008,40018,40023,40026,40032,40039,40054,40056,40165,40167,40169,40171,40172,40176,40179,40180,40182,40195,40198,40199,40200,40201,40206,40210,40213,40219,40223,40227,40230,40232,40234,40235,40236,40251,40254,40255,40257,40260,40262,40264,40272,40273,40281,40284,40285,40286,40288,40289,40292,40299,40300,40303,40304,40306,40314,40327,40329,40335,40346,40356,40361,40363,40367,40370,40372,40376,40378,40379,40385,40386,40388,40390,40399,40403,40409,40422,40429,40431,40434,40440,40441,40442,40445,40473,40474,40475,40478,40565,40568,40569,40573,40575,40577,40584,40587,40588,40593,40594,40595,40597,40599,40605,40607,40613,40614,40617,40618,40621,40632,40633,40634,40635,40636,40638,40639,40644,40652,40653,40654,40655,40656,40657,40658,40660,40664,40665,40667,40668,40669,40670,40672,40677,40680,40687,40692,40694,40695,40697,40699,40700,40701,40711,40712,40718,40723,40725,40736,40737,40748,40763,40766,40778,40779,40782,40783,40786,40788,40799,40800,40801,40802,40803,40806,40807,40810,40812,40818,40822,40823,40845,40853,40860,40861,40864,63785,63964,64014,64015,64016,64017,64018,64019,64020,64021,64022,64023,64024,64025,64026,64027,64028,64029,64030,64031,64032,64033,64034,64035,64036,64037,64038,64039,64040,64041,64042,64043,64044,64045,65281,65282,65283,65284,65285,65286,65287,65288,65289,65290,65291,65292,65293,65294,65295,65296,65297,65298,65299,65300,65301,65302,65303,65304,65305,65306,65307,65308,65309,65310,65311,65312,65313,65314,65315,65316,65317,65318,65319,65320,65321,65322,65323,65324,65325,65326,65327,65328,65329,65330,65331,65332,65333,65334,65335,65336,65337,65338,65339,65340,65341,65342,65343,65344,65345,65346,65347,65348,65349,65350,65351,65352,65353,65354,65355,65356,65357,65358,65359,65360,65361,65362,65363,65364,65365,65366,65367,65368,65369,65370,65371,65372,65373,65374,65377,65378,65379,65380,65381,65382,65383,65384,65385,65386,65387,65388,65389,65390,65391,65392,65393,65394,65395,65396,65397,65398,65399,65400,65401,65402,65403,65404,65405,65406,65407,65408,65409,65410,65411,65412,65413,65414,65415,65416,65417,65418,65419,65420,65421,65422,65423,65424,65425,65426,65427,65428,65429,65430,65431,65432,65433,65434,65435,65436,65437,65438,65439,65504,65505,65506,65507,65508,65509,728,711,184,729,733,175,731,730,65374,900,901,161,166,191,186,170,169,174,8482,164,8470,902,904,905,906,938,908,910,939,911,940,941,942,943,970,912,972,962,973,971,944,974,1026,1027,1028,1029,1030,1031,1032,1033,1034,1035,1036,1038,1039,1106,1107,1108,1109,1110,1111,1112,1113,1114,1115,1116,1118,1119,198,272,294,306,321,319,330,216,338,358,222,230,273,240,295,305,307,312,322,320,329,331,248,339,223,359,254,193,192,196,194,258,461,256,260,197,195,262,264,268,199,266,270,201,200,203,202,282,278,274,280,284,286,290,288,292,205,204,207,206,463,304,298,302,296,308,310,313,317,315,323,327,325,209,211,210,214,212,465,336,332,213,340,344,342,346,348,352,350,356,354,218,217,220,219,364,467,368,362,370,366,360,471,475,473,469,372,221,376,374,377,381,379,225,224,228,226,259,462,257,261,229,227,263,265,269,231,267,271,233,232,235,234,283,279,275,281,501,285,287,289,293,237,236,239,238,464,299,303,297,309,311,314,318,316,324,328,326,241,243,242,246,244,466,337,333,245,341,345,343,347,349,353,351,357,355,250,249,252,251,365,468,369,363,371,367,361,472,476,474,470,373,253,255,375,378,382,380,19970,19972,19973,19980,19986,19999,20003,20004,20008,20011,20014,20015,20016,20021,20032,20033,20036,20039,20049,20058,20060,20067,20072,20073,20084,20085,20089,20095,20109,20118,20119,20125,20143,20153,20163,20176,20186,20187,20192,20193,20194,20200,20207,20209,20211,20213,20221,20222,20223,20224,20226,20227,20232,20235,20236,20242,20245,20246,20247,20249,20270,20273,20320,20275,20277,20279,20281,20283,20286,20288,20290,20296,20297,20299,20300,20306,20308,20310,20312,20319,20323,20330,20332,20334,20337,20343,20344,20345,20346,20349,20350,20353,20354,20356,20357,20361,20362,20364,20366,20368,20370,20371,20372,20375,20377,20378,20382,20383,20402,20407,20409,20411,20412,20413,20414,20416,20417,20421,20422,20424,20425,20427,20428,20429,20431,20434,20444,20448,20450,20464,20466,20476,20477,20479,20480,20481,20484,20487,20490,20492,20494,20496,20499,20503,20504,20507,20508,20509,20510,20514,20519,20526,20528,20530,20531,20533,20544,20545,20546,20549,20550,20554,20556,20558,20561,20562,20563,20567,20569,20575,20576,20578,20579,20582,20583,20586,20589,20592,20593,20539,20609,20611,20612,20614,20618,20622,20623,20624,20626,20627,20628,20630,20635,20636,20638,20639,20640,20641,20642,20650,20655,20656,20665,20666,20669,20672,20675,20676,20679,20684,20686,20688,20691,20692,20696,20700,20701,20703,20706,20708,20710,20712,20713,20719,20721,20726,20730,20734,20739,20742,20743,20744,20747,20748,20749,20750,20722,20752,20759,20761,20763,20764,20765,20766,20771,20775,20776,20780,20781,20783,20785,20787,20788,20789,20792,20793,20802,20810,20815,20819,20821,20823,20824,20831,20836,20838,20862,20867,20868,20875,20878,20888,20893,20897,20899,20909,20920,20922,20924,20926,20927,20930,20936,20943,20945,20946,20947,20949,20952,20958,20962,20965,20974,20978,20979,20980,20983,20993,20994,20997,21010,21011,21013,21014,21016,21026,21032,21041,21042,21045,21052,21061,21065,21077,21079,21080,21082,21084,21087,21088,21089,21094,21102,21111,21112,21113,21120,21122,21125,21130,21132,21139,21141,21142,21143,21144,21146,21148,21156,21157,21158,21159,21167,21168,21174,21175,21176,21178,21179,21181,21184,21188,21190,21192,21196,21199,21201,21204,21206,21211,21212,21217,21221,21224,21225,21226,21228,21232,21233,21236,21238,21239,21248,21251,21258,21259,21260,21265,21267,21272,21275,21276,21278,21279,21285,21287,21288,21289,21291,21292,21293,21296,21298,21301,21308,21309,21310,21314,21324,21323,21337,21339,21345,21347,21349,21356,21357,21362,21369,21374,21379,21383,21384,21390,21395,21396,21401,21405,21409,21412,21418,21419,21423,21426,21428,21429,21431,21432,21434,21437,21440,21445,21455,21458,21459,21461,21466,21469,21470,21472,21478,21479,21493,21506,21523,21530,21537,21543,21544,21546,21551,21553,21556,21557,21571,21572,21575,21581,21583,21598,21602,21604,21606,21607,21609,21611,21613,21614,21620,21631,21633,21635,21637,21640,21641,21645,21649,21653,21654,21660,21663,21665,21670,21671,21673,21674,21677,21678,21681,21687,21689,21690,21691,21695,21702,21706,21709,21710,21728,21738,21740,21743,21750,21756,21758,21759,21760,21761,21765,21768,21769,21772,21773,21774,21781,21802,21803,21810,21813,21814,21819,21820,21821,21825,21831,21833,21834,21837,21840,21841,21848,21850,21851,21854,21856,21857,21860,21862,21887,21889,21890,21894,21896,21902,21903,21905,21906,21907,21908,21911,21923,21924,21933,21938,21951,21953,21955,21958,21961,21963,21964,21966,21969,21970,21971,21975,21976,21979,21982,21986,21993,22006,22015,22021,22024,22026,22029,22030,22031,22032,22033,22034,22041,22060,22064,22067,22069,22071,22073,22075,22076,22077,22079,22080,22081,22083,22084,22086,22089,22091,22093,22095,22100,22110,22112,22113,22114,22115,22118,22121,22125,22127,22129,22130,22133,22148,22149,22152,22155,22156,22165,22169,22170,22173,22174,22175,22182,22183,22184,22185,22187,22188,22189,22193,22195,22199,22206,22213,22217,22218,22219,22223,22224,22220,22221,22233,22236,22237,22239,22241,22244,22245,22246,22247,22248,22257,22251,22253,22262,22263,22273,22274,22279,22282,22284,22289,22293,22298,22299,22301,22304,22306,22307,22308,22309,22313,22314,22316,22318,22319,22323,22324,22333,22334,22335,22341,22342,22348,22349,22354,22370,22373,22375,22376,22379,22381,22382,22383,22384,22385,22387,22388,22389,22391,22393,22394,22395,22396,22398,22401,22403,22412,22420,22423,22425,22426,22428,22429,22430,22431,22433,22421,22439,22440,22441,22444,22456,22461,22471,22472,22476,22479,22485,22493,22494,22500,22502,22503,22505,22509,22512,22517,22518,22520,22525,22526,22527,22531,22532,22536,22537,22497,22540,22541,22555,22558,22559,22560,22566,22567,22573,22578,22585,22591,22601,22604,22605,22607,22608,22613,22623,22625,22628,22631,22632,22648,22652,22655,22656,22657,22663,22664,22665,22666,22668,22669,22671,22672,22676,22678,22685,22688,22689,22690,22694,22697,22705,22706,22724,22716,22722,22728,22733,22734,22736,22738,22740,22742,22746,22749,22753,22754,22761,22771,22789,22790,22795,22796,22802,22803,22804,34369,22813,22817,22819,22820,22824,22831,22832,22835,22837,22838,22847,22851,22854,22866,22867,22873,22875,22877,22878,22879,22881,22883,22891,22893,22895,22898,22901,22902,22905,22907,22908,22923,22924,22926,22930,22933,22935,22943,22948,22951,22957,22958,22959,22960,22963,22967,22970,22972,22977,22979,22980,22984,22986,22989,22994,23005,23006,23007,23011,23012,23015,23022,23023,23025,23026,23028,23031,23040,23044,23052,23053,23054,23058,23059,23070,23075,23076,23079,23080,23082,23085,23088,23108,23109,23111,23112,23116,23120,23125,23134,23139,23141,23143,23149,23159,23162,23163,23166,23179,23184,23187,23190,23193,23196,23198,23199,23200,23202,23207,23212,23217,23218,23219,23221,23224,23226,23227,23231,23236,23238,23240,23247,23258,23260,23264,23269,23274,23278,23285,23286,23293,23296,23297,23304,23319,23348,23321,23323,23325,23329,23333,23341,23352,23361,23371,23372,23378,23382,23390,23400,23406,23407,23420,23421,23422,23423,23425,23428,23430,23434,23438,23440,23441,23443,23444,23446,23464,23465,23468,23469,23471,23473,23474,23479,23482,23484,23488,23489,23501,23503,23510,23511,23512,23513,23514,23520,23535,23537,23540,23549,23564,23575,23582,23583,23587,23590,23593,23595,23596,23598,23600,23602,23605,23606,23641,23642,23644,23650,23651,23655,23656,23657,23661,23664,23668,23669,23674,23675,23676,23677,23687,23688,23690,23695,23698,23709,23711,23712,23714,23715,23718,23722,23730,23732,23733,23738,23753,23755,23762,23773,23767,23790,23793,23794,23796,23809,23814,23821,23826,23851,23843,23844,23846,23847,23857,23860,23865,23869,23871,23874,23875,23878,23880,23893,23889,23897,23882,23903,23904,23905,23906,23908,23914,23917,23920,23929,23930,23934,23935,23937,23939,23944,23946,23954,23955,23956,23957,23961,23963,23967,23968,23975,23979,23984,23988,23992,23993,24003,24007,24011,24016,24014,24024,24025,24032,24036,24041,24056,24057,24064,24071,24077,24082,24084,24085,24088,24095,24096,24110,24104,24114,24117,24126,24139,24144,24137,24145,24150,24152,24155,24156,24158,24168,24170,24171,24172,24173,24174,24176,24192,24203,24206,24226,24228,24229,24232,24234,24236,24241,24243,24253,24254,24255,24262,24268,24267,24270,24273,24274,24276,24277,24284,24286,24293,24299,24322,24326,24327,24328,24334,24345,24348,24349,24353,24354,24355,24356,24360,24363,24364,24366,24368,24372,24374,24379,24381,24383,24384,24388,24389,24391,24397,24400,24404,24408,24411,24416,24419,24420,24423,24431,24434,24436,24437,24440,24442,24445,24446,24457,24461,24463,24470,24476,24477,24482,24487,24491,24484,24492,24495,24496,24497,24504,24516,24519,24520,24521,24523,24528,24529,24530,24531,24532,24542,24545,24546,24552,24553,24554,24556,24557,24558,24559,24562,24563,24566,24570,24572,24583,24586,24589,24595,24596,24599,24600,24602,24607,24612,24621,24627,24629,24640,24647,24648,24649,24652,24657,24660,24662,24663,24669,24673,24679,24689,24702,24703,24706,24710,24712,24714,24718,24721,24723,24725,24728,24733,24734,24738,24740,24741,24744,24752,24753,24759,24763,24766,24770,24772,24776,24777,24778,24779,24782,24783,24788,24789,24793,24795,24797,24798,24802,24805,24818,24821,24824,24828,24829,24834,24839,24842,24844,24848,24849,24850,24851,24852,24854,24855,24857,24860,24862,24866,24874,24875,24880,24881,24885,24886,24887,24889,24897,24901,24902,24905,24926,24928,24940,24946,24952,24955,24956,24959,24960,24961,24963,24964,24971,24973,24978,24979,24983,24984,24988,24989,24991,24992,24997,25000,25002,25005,25016,25017,25020,25024,25025,25026,25038,25039,25045,25052,25053,25054,25055,25057,25058,25063,25065,25061,25068,25069,25071,25089,25091,25092,25095,25107,25109,25116,25120,25122,25123,25127,25129,25131,25145,25149,25154,25155,25156,25158,25164,25168,25169,25170,25172,25174,25178,25180,25188,25197,25199,25203,25210,25213,25229,25230,25231,25232,25254,25256,25267,25270,25271,25274,25278,25279,25284,25294,25301,25302,25306,25322,25330,25332,25340,25341,25347,25348,25354,25355,25357,25360,25363,25366,25368,25385,25386,25389,25397,25398,25401,25404,25409,25410,25411,25412,25414,25418,25419,25422,25426,25427,25428,25432,25435,25445,25446,25452,25453,25457,25460,25461,25464,25468,25469,25471,25474,25476,25479,25482,25488,25492,25493,25497,25498,25502,25508,25510,25517,25518,25519,25533,25537,25541,25544,25550,25553,25555,25556,25557,25564,25568,25573,25578,25580,25586,25587,25589,25592,25593,25609,25610,25616,25618,25620,25624,25630,25632,25634,25636,25637,25641,25642,25647,25648,25653,25661,25663,25675,25679,25681,25682,25683,25684,25690,25691,25692,25693,25695,25696,25697,25699,25709,25715,25716,25723,25725,25733,25735,25743,25744,25745,25752,25753,25755,25757,25759,25761,25763,25766,25768,25772,25779,25789,25790,25791,25796,25801,25802,25803,25804,25806,25808,25809,25813,25815,25828,25829,25833,25834,25837,25840,25845,25847,25851,25855,25857,25860,25864,25865,25866,25871,25875,25876,25878,25881,25883,25886,25887,25890,25894,25897,25902,25905,25914,25916,25917,25923,25927,25929,25936,25938,25940,25951,25952,25959,25963,25978,25981,25985,25989,25994,26002,26005,26008,26013,26016,26019,26022,26030,26034,26035,26036,26047,26050,26056,26057,26062,26064,26068,26070,26072,26079,26096,26098,26100,26101,26105,26110,26111,26112,26116,26120,26121,26125,26129,26130,26133,26134,26141,26142,26145,26146,26147,26148,26150,26153,26154,26155,26156,26158,26160,26161,26163,26169,26167,26176,26181,26182,26186,26188,26193,26190,26199,26200,26201,26203,26204,26208,26209,26363,26218,26219,26220,26238,26227,26229,26239,26231,26232,26233,26235,26240,26236,26251,26252,26253,26256,26258,26265,26266,26267,26268,26271,26272,26276,26285,26289,26290,26293,26299,26303,26304,26306,26307,26312,26316,26318,26319,26324,26331,26335,26344,26347,26348,26350,26362,26373,26375,26382,26387,26393,26396,26400,26402,26419,26430,26437,26439,26440,26444,26452,26453,26461,26470,26476,26478,26484,26486,26491,26497,26500,26510,26511,26513,26515,26518,26520,26521,26523,26544,26545,26546,26549,26555,26556,26557,26617,26560,26562,26563,26565,26568,26569,26578,26583,26585,26588,26593,26598,26608,26610,26614,26615,26706,26644,26649,26653,26655,26664,26663,26668,26669,26671,26672,26673,26675,26683,26687,26692,26693,26698,26700,26709,26711,26712,26715,26731,26734,26735,26736,26737,26738,26741,26745,26746,26747,26748,26754,26756,26758,26760,26774,26776,26778,26780,26785,26787,26789,26793,26794,26798,26802,26811,26821,26824,26828,26831,26832,26833,26835,26838,26841,26844,26845,26853,26856,26858,26859,26860,26861,26864,26865,26869,26870,26875,26876,26877,26886,26889,26890,26896,26897,26899,26902,26903,26929,26931,26933,26936,26939,26946,26949,26953,26958,26967,26971,26979,26980,26981,26982,26984,26985,26988,26992,26993,26994,27002,27003,27007,27008,27021,27026,27030,27032,27041,27045,27046,27048,27051,27053,27055,27063,27064,27066,27068,27077,27080,27089,27094,27095,27106,27109,27118,27119,27121,27123,27125,27134,27136,27137,27139,27151,27153,27157,27162,27165,27168,27172,27176,27184,27186,27188,27191,27195,27198,27199,27205,27206,27209,27210,27214,27216,27217,27218,27221,27222,27227,27236,27239,27242,27249,27251,27262,27265,27267,27270,27271,27273,27275,27281,27291,27293,27294,27295,27301,27307,27311,27312,27313,27316,27325,27326,27327,27334,27337,27336,27340,27344,27348,27349,27350,27356,27357,27364,27367,27372,27376,27377,27378,27388,27389,27394,27395,27398,27399,27401,27407,27408,27409,27415,27419,27422,27428,27432,27435,27436,27439,27445,27446,27451,27455,27462,27466,27469,27474,27478,27480,27485,27488,27495,27499,27502,27504,27509,27517,27518,27522,27525,27543,27547,27551,27552,27554,27555,27560,27561,27564,27565,27566,27568,27576,27577,27581,27582,27587,27588,27593,27596,27606,27610,27617,27619,27622,27623,27630,27633,27639,27641,27647,27650,27652,27653,27657,27661,27662,27664,27666,27673,27679,27686,27687,27688,27692,27694,27699,27701,27702,27706,27707,27711,27722,27723,27725,27727,27730,27732,27737,27739,27740,27755,27757,27759,27764,27766,27768,27769,27771,27781,27782,27783,27785,27796,27797,27799,27800,27804,27807,27824,27826,27828,27842,27846,27853,27855,27856,27857,27858,27860,27862,27866,27868,27872,27879,27881,27883,27884,27886,27890,27892,27908,27911,27914,27918,27919,27921,27923,27930,27942,27943,27944,27751,27950,27951,27953,27961,27964,27967,27991,27998,27999,28001,28005,28007,28015,28016,28028,28034,28039,28049,28050,28052,28054,28055,28056,28074,28076,28084,28087,28089,28093,28095,28100,28104,28106,28110,28111,28118,28123,28125,28127,28128,28130,28133,28137,28143,28144,28148,28150,28156,28160,28164,28190,28194,28199,28210,28214,28217,28219,28220,28228,28229,28232,28233,28235,28239,28241,28242,28243,28244,28247,28252,28253,28254,28258,28259,28264,28275,28283,28285,28301,28307,28313,28320,28327,28333,28334,28337,28339,28347,28351,28352,28353,28355,28359,28360,28362,28365,28366,28367,28395,28397,28398,28409,28411,28413,28420,28424,28426,28428,28429,28438,28440,28442,28443,28454,28457,28458,28463,28464,28467,28470,28475,28476,28461,28495,28497,28498,28499,28503,28505,28506,28509,28510,28513,28514,28520,28524,28541,28542,28547,28551,28552,28555,28556,28557,28560,28562,28563,28564,28566,28570,28575,28576,28581,28582,28583,28584,28590,28591,28592,28597,28598,28604,28613,28615,28616,28618,28634,28638,28648,28649,28656,28661,28665,28668,28669,28672,28677,28678,28679,28685,28695,28704,28707,28719,28724,28727,28729,28732,28739,28740,28744,28745,28746,28747,28756,28757,28765,28766,28750,28772,28773,28780,28782,28789,28790,28798,28801,28805,28806,28820,28821,28822,28823,28824,28827,28836,28843,28848,28849,28852,28855,28874,28881,28883,28884,28885,28886,28888,28892,28900,28922,28931,28932,28933,28934,28935,28939,28940,28943,28958,28960,28971,28973,28975,28976,28977,28984,28993,28997,28998,28999,29002,29003,29008,29010,29015,29018,29020,29022,29024,29032,29049,29056,29061,29063,29068,29074,29082,29083,29088,29090,29103,29104,29106,29107,29114,29119,29120,29121,29124,29131,29132,29139,29142,29145,29146,29148,29176,29182,29184,29191,29192,29193,29203,29207,29210,29213,29215,29220,29227,29231,29236,29240,29241,29249,29250,29251,29253,29262,29263,29264,29267,29269,29270,29274,29276,29278,29280,29283,29288,29291,29294,29295,29297,29303,29304,29307,29308,29311,29316,29321,29325,29326,29331,29339,29352,29357,29358,29361,29364,29374,29377,29383,29385,29388,29397,29398,29400,29407,29413,29427,29428,29434,29435,29438,29442,29444,29445,29447,29451,29453,29458,29459,29464,29465,29470,29474,29476,29479,29480,29484,29489,29490,29493,29498,29499,29501,29507,29517,29520,29522,29526,29528,29533,29534,29535,29536,29542,29543,29545,29547,29548,29550,29551,29553,29559,29561,29564,29568,29569,29571,29573,29574,29582,29584,29587,29589,29591,29592,29596,29598,29599,29600,29602,29605,29606,29610,29611,29613,29621,29623,29625,29628,29629,29631,29637,29638,29641,29643,29644,29647,29650,29651,29654,29657,29661,29665,29667,29670,29671,29673,29684,29685,29687,29689,29690,29691,29693,29695,29696,29697,29700,29703,29706,29713,29722,29723,29732,29734,29736,29737,29738,29739,29740,29741,29742,29743,29744,29745,29753,29760,29763,29764,29766,29767,29771,29773,29777,29778,29783,29789,29794,29798,29799,29800,29803,29805,29806,29809,29810,29824,29825,29829,29830,29831,29833,29839,29840,29841,29842,29848,29849,29850,29852,29855,29856,29857,29859,29862,29864,29865,29866,29867,29870,29871,29873,29874,29877,29881,29883,29887,29896,29897,29900,29904,29907,29912,29914,29915,29918,29919,29924,29928,29930,29931,29935,29940,29946,29947,29948,29951,29958,29970,29974,29975,29984,29985,29988,29991,29993,29994,29999,30006,30009,30013,30014,30015,30016,30019,30023,30024,30030,30032,30034,30039,30046,30047,30049,30063,30065,30073,30074,30075,30076,30077,30078,30081,30085,30096,30098,30099,30101,30105,30108,30114,30116,30132,30138,30143,30144,30145,30148,30150,30156,30158,30159,30167,30172,30175,30176,30177,30180,30183,30188,30190,30191,30193,30201,30208,30210,30211,30212,30215,30216,30218,30220,30223,30226,30227,30229,30230,30233,30235,30236,30237,30238,30243,30245,30246,30249,30253,30258,30259,30261,30264,30265,30266,30268,30282,30272,30273,30275,30276,30277,30281,30283,30293,30297,30303,30308,30309,30317,30318,30319,30321,30324,30337,30341,30348,30349,30357,30363,30364,30365,30367,30368,30370,30371,30372,30373,30374,30375,30376,30378,30381,30397,30401,30405,30409,30411,30412,30414,30420,30425,30432,30438,30440,30444,30448,30449,30454,30457,30460,30464,30470,30474,30478,30482,30484,30485,30487,30489,30490,30492,30498,30504,30509,30510,30511,30516,30517,30518,30521,30525,30526,30530,30533,30534,30538,30541,30542,30543,30546,30550,30551,30556,30558,30559,30560,30562,30564,30567,30570,30572,30576,30578,30579,30580,30586,30589,30592,30596,30604,30605,30612,30613,30614,30618,30623,30626,30631,30634,30638,30639,30641,30645,30654,30659,30665,30673,30674,30677,30681,30686,30687,30688,30692,30694,30698,30700,30704,30705,30708,30712,30715,30725,30726,30729,30733,30734,30737,30749,30753,30754,30755,30765,30766,30768,30773,30775,30787,30788,30791,30792,30796,30798,30802,30812,30814,30816,30817,30819,30820,30824,30826,30830,30842,30846,30858,30863,30868,30872,30881,30877,30878,30879,30884,30888,30892,30893,30896,30897,30898,30899,30907,30909,30911,30919,30920,30921,30924,30926,30930,30931,30933,30934,30948,30939,30943,30944,30945,30950,30954,30962,30963,30976,30966,30967,30970,30971,30975,30982,30988,30992,31002,31004,31006,31007,31008,31013,31015,31017,31021,31025,31028,31029,31035,31037,31039,31044,31045,31046,31050,31051,31055,31057,31060,31064,31067,31068,31079,31081,31083,31090,31097,31099,31100,31102,31115,31116,31121,31123,31124,31125,31126,31128,31131,31132,31137,31144,31145,31147,31151,31153,31156,31160,31163,31170,31172,31175,31176,31178,31183,31188,31190,31194,31197,31198,31200,31202,31205,31210,31211,31213,31217,31224,31228,31234,31235,31239,31241,31242,31244,31249,31253,31259,31262,31265,31271,31275,31277,31279,31280,31284,31285,31288,31289,31290,31300,31301,31303,31304,31308,31317,31318,31321,31324,31325,31327,31328,31333,31335,31338,31341,31349,31352,31358,31360,31362,31365,31366,31370,31371,31376,31377,31380,31390,31392,31395,31404,31411,31413,31417,31419,31420,31430,31433,31436,31438,31441,31451,31464,31465,31467,31468,31473,31476,31483,31485,31486,31495,31508,31519,31523,31527,31529,31530,31531,31533,31534,31535,31536,31537,31540,31549,31551,31552,31553,31559,31566,31573,31584,31588,31590,31593,31594,31597,31599,31602,31603,31607,31620,31625,31630,31632,31633,31638,31643,31646,31648,31653,31660,31663,31664,31666,31669,31670,31674,31675,31676,31677,31682,31685,31688,31690,31700,31702,31703,31705,31706,31707,31720,31722,31730,31732,31733,31736,31737,31738,31740,31742,31745,31746,31747,31748,31750,31753,31755,31756,31758,31759,31769,31771,31776,31781,31782,31784,31788,31793,31795,31796,31798,31801,31802,31814,31818,31829,31825,31826,31827,31833,31834,31835,31836,31837,31838,31841,31843,31847,31849,31853,31854,31856,31858,31865,31868,31869,31878,31879,31887,31892,31902,31904,31910,31920,31926,31927,31930,31931,31932,31935,31940,31943,31944,31945,31949,31951,31955,31956,31957,31959,31961,31962,31965,31974,31977,31979,31989,32003,32007,32008,32009,32015,32017,32018,32019,32022,32029,32030,32035,32038,32042,32045,32049,32060,32061,32062,32064,32065,32071,32072,32077,32081,32083,32087,32089,32090,32092,32093,32101,32103,32106,32112,32120,32122,32123,32127,32129,32130,32131,32133,32134,32136,32139,32140,32141,32145,32150,32151,32157,32158,32166,32167,32170,32179,32182,32183,32185,32194,32195,32196,32197,32198,32204,32205,32206,32215,32217,32256,32226,32229,32230,32234,32235,32237,32241,32245,32246,32249,32250,32264,32272,32273,32277,32279,32284,32285,32288,32295,32296,32300,32301,32303,32307,32310,32319,32324,32325,32327,32334,32336,32338,32344,32351,32353,32354,32357,32363,32366,32367,32371,32376,32382,32385,32390,32391,32394,32397,32401,32405,32408,32410,32413,32414,32572,32571,32573,32574,32575,32579,32580,32583,32591,32594,32595,32603,32604,32605,32609,32611,32612,32613,32614,32621,32625,32637,32638,32639,32640,32651,32653,32655,32656,32657,32662,32663,32668,32673,32674,32678,32682,32685,32692,32700,32703,32704,32707,32712,32718,32719,32731,32735,32739,32741,32744,32748,32750,32751,32754,32762,32765,32766,32767,32775,32776,32778,32781,32782,32783,32785,32787,32788,32790,32797,32798,32799,32800,32804,32806,32812,32814,32816,32820,32821,32823,32825,32826,32828,32830,32832,32836,32864,32868,32870,32877,32881,32885,32897,32904,32910,32924,32926,32934,32935,32939,32952,32953,32968,32973,32975,32978,32980,32981,32983,32984,32992,33005,33006,33008,33010,33011,33014,33017,33018,33022,33027,33035,33046,33047,33048,33052,33054,33056,33060,33063,33068,33072,33077,33082,33084,33093,33095,33098,33100,33106,33111,33120,33121,33127,33128,33129,33133,33135,33143,33153,33168,33156,33157,33158,33163,33166,33174,33176,33179,33182,33186,33198,33202,33204,33211,33227,33219,33221,33226,33230,33231,33237,33239,33243,33245,33246,33249,33252,33259,33260,33264,33265,33266,33269,33270,33272,33273,33277,33279,33280,33283,33295,33299,33300,33305,33306,33309,33313,33314,33320,33330,33332,33338,33347,33348,33349,33350,33355,33358,33359,33361,33366,33372,33376,33379,33383,33389,33396,33403,33405,33407,33408,33409,33411,33412,33415,33417,33418,33422,33425,33428,33430,33432,33434,33435,33440,33441,33443,33444,33447,33448,33449,33450,33454,33456,33458,33460,33463,33466,33468,33470,33471,33478,33488,33493,33498,33504,33506,33508,33512,33514,33517,33519,33526,33527,33533,33534,33536,33537,33543,33544,33546,33547,33620,33563,33565,33566,33567,33569,33570,33580,33581,33582,33584,33587,33591,33594,33596,33597,33602,33603,33604,33607,33613,33614,33617,33621,33622,33623,33648,33656,33661,33663,33664,33666,33668,33670,33677,33682,33684,33685,33688,33689,33691,33692,33693,33702,33703,33705,33708,33726,33727,33728,33735,33737,33743,33744,33745,33748,33757,33619,33768,33770,33782,33784,33785,33788,33793,33798,33802,33807,33809,33813,33817,33709,33839,33849,33861,33863,33864,33866,33869,33871,33873,33874,33878,33880,33881,33882,33884,33888,33892,33893,33895,33898,33904,33907,33908,33910,33912,33916,33917,33921,33925,33938,33939,33941,33950,33958,33960,33961,33962,33967,33969,33972,33978,33981,33982,33984,33986,33991,33992,33996,33999,34003,34012,34023,34026,34031,34032,34033,34034,34039,34098,34042,34043,34045,34050,34051,34055,34060,34062,34064,34076,34078,34082,34083,34084,34085,34087,34090,34091,34095,34099,34100,34102,34111,34118,34127,34128,34129,34130,34131,34134,34137,34140,34141,34142,34143,34144,34145,34146,34148,34155,34159,34169,34170,34171,34173,34175,34177,34181,34182,34185,34187,34188,34191,34195,34200,34205,34207,34208,34210,34213,34215,34228,34230,34231,34232,34236,34237,34238,34239,34242,34247,34250,34251,34254,34221,34264,34266,34271,34272,34278,34280,34285,34291,34294,34300,34303,34304,34308,34309,34317,34318,34320,34321,34322,34328,34329,34331,34334,34337,34343,34345,34358,34360,34362,34364,34365,34368,34370,34374,34386,34387,34390,34391,34392,34393,34397,34400,34401,34402,34403,34404,34409,34412,34415,34421,34422,34423,34426,34445,34449,34454,34456,34458,34460,34465,34470,34471,34472,34477,34481,34483,34484,34485,34487,34488,34489,34495,34496,34497,34499,34501,34513,34514,34517,34519,34522,34524,34528,34531,34533,34535,34440,34554,34556,34557,34564,34565,34567,34571,34574,34575,34576,34579,34580,34585,34590,34591,34593,34595,34600,34606,34607,34609,34610,34617,34618,34620,34621,34622,34624,34627,34629,34637,34648,34653,34657,34660,34661,34671,34673,34674,34683,34691,34692,34693,34694,34695,34696,34697,34699,34700,34704,34707,34709,34711,34712,34713,34718,34720,34723,34727,34732,34733,34734,34737,34741,34750,34751,34753,34760,34761,34762,34766,34773,34774,34777,34778,34780,34783,34786,34787,34788,34794,34795,34797,34801,34803,34808,34810,34815,34817,34819,34822,34825,34826,34827,34832,34841,34834,34835,34836,34840,34842,34843,34844,34846,34847,34856,34861,34862,34864,34866,34869,34874,34876,34881,34883,34885,34888,34889,34890,34891,34894,34897,34901,34902,34904,34906,34908,34911,34912,34916,34921,34929,34937,34939,34944,34968,34970,34971,34972,34975,34976,34984,34986,35002,35005,35006,35008,35018,35019,35020,35021,35022,35025,35026,35027,35035,35038,35047,35055,35056,35057,35061,35063,35073,35078,35085,35086,35087,35093,35094,35096,35097,35098,35100,35104,35110,35111,35112,35120,35121,35122,35125,35129,35130,35134,35136,35138,35141,35142,35145,35151,35154,35159,35162,35163,35164,35169,35170,35171,35179,35182,35184,35187,35189,35194,35195,35196,35197,35209,35213,35216,35220,35221,35227,35228,35231,35232,35237,35248,35252,35253,35254,35255,35260,35284,35285,35286,35287,35288,35301,35305,35307,35309,35313,35315,35318,35321,35325,35327,35332,35333,35335,35343,35345,35346,35348,35349,35358,35360,35362,35364,35366,35371,35372,35375,35381,35383,35389,35390,35392,35395,35397,35399,35401,35405,35406,35411,35414,35415,35416,35420,35421,35425,35429,35431,35445,35446,35447,35449,35450,35451,35454,35455,35456,35459,35462,35467,35471,35472,35474,35478,35479,35481,35487,35495,35497,35502,35503,35507,35510,35511,35515,35518,35523,35526,35528,35529,35530,35537,35539,35540,35541,35543,35549,35551,35564,35568,35572,35573,35574,35580,35583,35589,35590,35595,35601,35612,35614,35615,35594,35629,35632,35639,35644,35650,35651,35652,35653,35654,35656,35666,35667,35668,35673,35661,35678,35683,35693,35702,35704,35705,35708,35710,35713,35716,35717,35723,35725,35727,35732,35733,35740,35742,35743,35896,35897,35901,35902,35909,35911,35913,35915,35919,35921,35923,35924,35927,35928,35931,35933,35929,35939,35940,35942,35944,35945,35949,35955,35957,35958,35963,35966,35974,35975,35979,35984,35986,35987,35993,35995,35996,36004,36025,36026,36037,36038,36041,36043,36047,36054,36053,36057,36061,36065,36072,36076,36079,36080,36082,36085,36087,36088,36094,36095,36097,36099,36105,36114,36119,36123,36197,36201,36204,36206,36223,36226,36228,36232,36237,36240,36241,36245,36254,36255,36256,36262,36267,36268,36271,36274,36277,36279,36281,36283,36288,36293,36294,36295,36296,36298,36302,36305,36308,36309,36311,36313,36324,36325,36327,36332,36336,36284,36337,36338,36340,36349,36353,36356,36357,36358,36363,36369,36372,36374,36384,36385,36386,36387,36390,36391,36401,36403,36406,36407,36408,36409,36413,36416,36417,36427,36429,36430,36431,36436,36443,36444,36445,36446,36449,36450,36457,36460,36461,36463,36464,36465,36473,36474,36475,36482,36483,36489,36496,36498,36501,36506,36507,36509,36510,36514,36519,36521,36525,36526,36531,36533,36538,36539,36544,36545,36547,36548,36551,36559,36561,36564,36572,36584,36590,36592,36593,36599,36601,36602,36589,36608,36610,36615,36616,36623,36624,36630,36631,36632,36638,36640,36641,36643,36645,36647,36648,36652,36653,36654,36660,36661,36662,36663,36666,36672,36673,36675,36679,36687,36689,36690,36691,36692,36693,36696,36701,36702,36709,36765,36768,36769,36772,36773,36774,36789,36790,36792,36798,36800,36801,36806,36810,36811,36813,36816,36818,36819,36821,36832,36835,36836,36840,36846,36849,36853,36854,36859,36862,36866,36868,36872,36876,36888,36891,36904,36905,36911,36906,36908,36909,36915,36916,36919,36927,36931,36932,36940,36955,36957,36962,36966,36967,36972,36976,36980,36985,36997,37000,37003,37004,37006,37008,37013,37015,37016,37017,37019,37024,37025,37026,37029,37040,37042,37043,37044,37046,37053,37068,37054,37059,37060,37061,37063,37064,37077,37079,37080,37081,37084,37085,37087,37093,37074,37110,37099,37103,37104,37108,37118,37119,37120,37124,37125,37126,37128,37133,37136,37140,37142,37143,37144,37146,37148,37150,37152,37157,37154,37155,37159,37161,37166,37167,37169,37172,37174,37175,37177,37178,37180,37181,37187,37191,37192,37199,37203,37207,37209,37210,37211,37217,37220,37223,37229,37236,37241,37242,37243,37249,37251,37253,37254,37258,37262,37265,37267,37268,37269,37272,37278,37281,37286,37288,37292,37293,37294,37296,37297,37298,37299,37302,37307,37308,37309,37311,37314,37315,37317,37331,37332,37335,37337,37338,37342,37348,37349,37353,37354,37356,37357,37358,37359,37360,37361,37367,37369,37371,37373,37376,37377,37380,37381,37382,37383,37385,37386,37388,37392,37394,37395,37398,37400,37404,37405,37411,37412,37413,37414,37416,37422,37423,37424,37427,37429,37430,37432,37433,37434,37436,37438,37440,37442,37443,37446,37447,37450,37453,37454,37455,37457,37464,37465,37468,37469,37472,37473,37477,37479,37480,37481,37486,37487,37488,37493,37494,37495,37496,37497,37499,37500,37501,37503,37512,37513,37514,37517,37518,37522,37527,37529,37535,37536,37540,37541,37543,37544,37547,37551,37554,37558,37560,37562,37563,37564,37565,37567,37568,37569,37570,37571,37573,37574,37575,37576,37579,37580,37581,37582,37584,37587,37589,37591,37592,37593,37596,37597,37599,37600,37601,37603,37605,37607,37608,37612,37614,37616,37625,37627,37631,37632,37634,37640,37645,37649,37652,37653,37660,37661,37662,37663,37665,37668,37669,37671,37673,37674,37683,37684,37686,37687,37703,37704,37705,37712,37713,37714,37717,37719,37720,37722,37726,37732,37733,37735,37737,37738,37741,37743,37744,37745,37747,37748,37750,37754,37757,37759,37760,37761,37762,37768,37770,37771,37773,37775,37778,37781,37784,37787,37790,37793,37795,37796,37798,37800,37803,37812,37813,37814,37818,37801,37825,37828,37829,37830,37831,37833,37834,37835,37836,37837,37843,37849,37852,37854,37855,37858,37862,37863,37881,37879,37880,37882,37883,37885,37889,37890,37892,37896,37897,37901,37902,37903,37909,37910,37911,37919,37934,37935,37937,37938,37939,37940,37947,37951,37949,37955,37957,37960,37962,37964,37973,37977,37980,37983,37985,37987,37992,37995,37997,37998,37999,38001,38002,38020,38019,38264,38265,38270,38276,38280,38284,38285,38286,38301,38302,38303,38305,38310,38313,38315,38316,38324,38326,38330,38333,38335,38342,38344,38345,38347,38352,38353,38354,38355,38361,38362,38365,38366,38367,38368,38372,38374,38429,38430,38434,38436,38437,38438,38444,38449,38451,38455,38456,38457,38458,38460,38461,38465,38482,38484,38486,38487,38488,38497,38510,38516,38523,38524,38526,38527,38529,38530,38531,38532,38537,38545,38550,38554,38557,38559,38564,38565,38566,38569,38574,38575,38579,38586,38602,38610,23986,38616,38618,38621,38622,38623,38633,38639,38641,38650,38658,38659,38661,38665,38682,38683,38685,38689,38690,38691,38696,38705,38707,38721,38723,38730,38734,38735,38741,38743,38744,38746,38747,38755,38759,38762,38766,38771,38774,38775,38776,38779,38781,38783,38784,38793,38805,38806,38807,38809,38810,38814,38815,38818,38828,38830,38833,38834,38837,38838,38840,38841,38842,38844,38846,38847,38849,38852,38853,38855,38857,38858,38860,38861,38862,38864,38865,38868,38871,38872,38873,38877,38878,38880,38875,38881,38884,38895,38897,38900,38903,38904,38906,38919,38922,38937,38925,38926,38932,38934,38940,38942,38944,38947,38950,38955,38958,38959,38960,38962,38963,38965,38949,38974,38980,38983,38986,38993,38994,38995,38998,38999,39001,39002,39010,39011,39013,39014,39018,39020,39083,39085,39086,39088,39092,39095,39096,39098,39099,39103,39106,39109,39112,39116,39137,39139,39141,39142,39143,39146,39155,39158,39170,39175,39176,39185,39189,39190,39191,39194,39195,39196,39199,39202,39206,39207,39211,39217,39218,39219,39220,39221,39225,39226,39227,39228,39232,39233,39238,39239,39240,39245,39246,39252,39256,39257,39259,39260,39262,39263,39264,39323,39325,39327,39334,39344,39345,39346,39349,39353,39354,39357,39359,39363,39369,39379,39380,39385,39386,39388,39390,39399,39402,39403,39404,39408,39412,39413,39417,39421,39422,39426,39427,39428,39435,39436,39440,39441,39446,39454,39456,39458,39459,39460,39463,39469,39470,39475,39477,39478,39480,39495,39489,39492,39498,39499,39500,39502,39505,39508,39510,39517,39594,39596,39598,39599,39602,39604,39605,39606,39609,39611,39614,39615,39617,39619,39622,39624,39630,39632,39634,39637,39638,39639,39643,39644,39648,39652,39653,39655,39657,39660,39666,39667,39669,39673,39674,39677,39679,39680,39681,39682,39683,39684,39685,39688,39689,39691,39692,39693,39694,39696,39698,39702,39705,39707,39708,39712,39718,39723,39725,39731,39732,39733,39735,39737,39738,39741,39752,39755,39756,39765,39766,39767,39771,39774,39777,39779,39781,39782,39784,39786,39787,39788,39789,39790,39795,39797,39799,39800,39801,39807,39808,39812,39813,39814,39815,39817,39818,39819,39821,39823,39824,39828,39834,39837,39838,39846,39847,39849,39852,39856,39857,39858,39863,39864,39867,39868,39870,39871,39873,39879,39880,39886,39888,39895,39896,39901,39903,39909,39911,39914,39915,39919,39923,39927,39928,39929,39930,39933,39935,39936,39938,39947,39951,39953,39958,39960,39961,39962,39964,39966,39970,39971,39974,39975,39976,39977,39978,39985,39989,39990,39991,39997,40001,40003,40004,40005,40009,40010,40014,40015,40016,40019,40020,40022,40024,40027,40029,40030,40031,40035,40041,40042,40028,40043,40040,40046,40048,40050,40053,40055,40059,40166,40178,40183,40185,40203,40194,40209,40215,40216,40220,40221,40222,40239,40240,40242,40243,40244,40250,40252,40261,40253,40258,40259,40263,40266,40275,40276,40287,40291,40290,40293,40297,40298,40299,40304,40310,40311,40315,40316,40318,40323,40324,40326,40330,40333,40334,40338,40339,40341,40342,40343,40344,40353,40362,40364,40366,40369,40373,40377,40380,40383,40387,40391,40393,40394,40404,40405,40406,40407,40410,40414,40415,40416,40421,40423,40425,40427,40430,40432,40435,40436,40446,40458,40450,40455,40462,40464,40465,40466,40469,40470,40473,40476,40477,40570,40571,40572,40576,40578,40579,40580,40581,40583,40590,40591,40598,40600,40603,40606,40612,40616,40620,40622,40623,40624,40627,40628,40629,40646,40648,40651,40661,40671,40676,40679,40684,40685,40686,40688,40689,40690,40693,40696,40703,40706,40707,40713,40719,40720,40721,40722,40724,40726,40727,40729,40730,40731,40735,40738,40742,40746,40747,40751,40753,40754,40756,40759,40761,40762,40764,40765,40767,40769,40771,40772,40773,40774,40775,40787,40789,40790,40791,40792,40794,40797,40798,40808,40809,40813,40814,40815,40816,40817,40819,40821,40826,40829,40847,40848,40849,40850,40852,40854,40855,40862,40865,40866,40867,40869]],
        ];
        foreach ($series as $test) {
            foreach ($test[0] as $a => $input) {
                $class = $this->testedClass;
                $char = hex2bin($input);
                $exp = $test[1][$a];
                $s = new $class($char);
                $this->assertSame($exp, $s->nextCode(), "Sequence $input did not decode to $exp.");
                $this->assertFalse($s->nextCode(), "Sequence $input did not end after one character");
            }
        }
    }
}
