<?php
/** @license MIT
 * Copyright 2018 J. King et al.
 * See LICENSE and AUTHORS files for details */

declare(strict_types=1);
namespace MensBeam\Intl\TestCase\Encoding;

use MensBeam\Intl\Encoding\ShiftJIS;
use MensBeam\Intl\Encoding\Encoding;
use MensBeam\Intl\Encoding\EncoderException;

class TestShiftJIS extends \MensBeam\Intl\Test\CoderDecoderTest {
    protected $testedClass = ShiftJIS::class;
    /*
        Char 0 U+007A   (1 byte)  Offset 0
        Char 1 U+FF96   (1 byte)  Offset 1
        Char 2 U+3088   (2 bytes) Offset 2
        Char 3 U+FF0D   (2 bytes) Offset 4
        Char 4 U+005C   (1 byte)  Offset 6
        Char 5 U+FF9B   (1 byte)  Offset 7
        Char 6 U+E000   (2 bytes) Offset 8
        End of string at char 7, offset 10
    */
    protected $seekString = "7A D6 82E6 817C 5C DB F040";
    protected $seekCodes = [0x7A, 0xFF96, 0x3088, 0xFF0D, 0x5C, 0xFF9B, 0xE000];
    protected $seekOffsets = [0, 1, 2, 4, 6, 7, 8, 10];
    /* This string contains an invalid character sequence sandwiched between two null characters */
    protected $brokenChar = "00 FF 00";

    public function provideCodePoints() {
        return [
            'U+0064 (HTML)'  => [false, 0x64, "64"],
            'U+0064 (fatal)' => [true,  0x64, "64"],
            'U+00A5 (HTML)'  => [false, 0xA5, "5C"],
            'U+00A5 (fatal)' => [true,  0xA5, "5C"],
            'U+203E (HTML)'  => [false, 0x203E, "7E"],
            'U+203E (fatal)' => [true,  0x203E, "7E"],
            'U+3088 (HTML)'  => [false, 0x3088, "82 E6"],
            'U+3088 (fatal)' => [true,  0x3088, "82 E6"],
            'U+FF96 (HTML)'  => [false, 0xFF96, "D6"],
            'U+FF96 (fatal)' => [true,  0xFF96, "D6"],
            'U+2212 (HTML)'  => [false, 0x2212, "81 7C"],
            'U+2212 (fatal)' => [true,  0x2212, "81 7C"],
            'U+00E6 (HTML)'  => [false, 0xE6, bin2hex("&#230;")],
            'U+00E6 (fatal)' => [true,  0xE6, new EncoderException("", Encoding::E_UNAVAILABLE_CODE_POINT)],
            'U+FFE2 (HTML)'  => [false, 0xFFE2, "81 CA"],
            'U+FFE2 (fatal)' => [true,  0xFFE2, "81 CA"],
            'U+2116 (HTML)'  => [false, 0x2116, "87 82"],
            'U+2116 (fatal)' => [true,  0x2116, "87 82"],
            'U+E000 (HTML)'  => [false, 0xE000, bin2hex("&#57344;")],
            'U+E000 (fatal)' => [true,  0xE000, new EncoderException("", Encoding::E_UNAVAILABLE_CODE_POINT)],
            '-1 (HTML)'  => [false, -1, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
            '-1 (fatal)' => [true,  -1, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
            '0x110000 (HTML)'  => [false, 0x110000, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
            '0x110000 (fatal)' => [true,  0x110000, new EncoderException("", Encoding::E_INVALID_CODE_POINT)],
        ];
    }

    public function provideStrings() {
        return [
            'empty string' => ["", []],
            'sanity check' => ["40", [64]],
            'invalid byte' => ["FF", [65533]],
            'former ASCII deviations' => ["5C 7E", [92, 126]],
            'JIS X 0201 range' => ["A1 DF", [65377, 65439]],
            'EUDC range' => ["F040 F9FC", [57344, 59223]],
            'JIS X 0208 assigned range' => ["8140 FC4B", [12288, 40657]],
            'JIS X 0208 total range' => ["8140 FCFC", [12288, 65533]],
            'JIS X 0208 truncated character 1' => ["81", [65533]],
            'JIS X 0208 truncated character 2' => ["81 20", [65533, 32]],
            'JIS X 0208 truncated character 3' => ["81 FF", [65533]],
        ];
    }

    /**
     * @dataProvider provideCodePoints
     * @covers MensBeam\Intl\Encoding\Encoder
     * @covers MensBeam\Intl\Encoding\ShiftJIS::encode
     * @covers MensBeam\Intl\Encoding\ShiftJIS::errEnc
     */
    public function testEncodeCodePoints(bool $fatal, $input, $exp) {
        return parent::testEncodeCodePoints($fatal, $input, $exp);
    }

    /**
     * @dataProvider provideCodePoints
     * @covers MensBeam\Intl\Encoding\ShiftJIS::encode
     * @covers MensBeam\Intl\Encoding\ShiftJIS::errEnc
     */
    public function testEncodeCodePointsStatically(bool $fatal, $input, $exp) {
        return parent::testEncodeCodePointsStatically($fatal, $input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\ShiftJIS::__construct
     * @covers MensBeam\Intl\Encoding\ShiftJIS::nextCode
     */
    public function testDecodeMultipleCharactersAsCodePoints(string $input, array $exp) {
        return parent::testDecodeMultipleCharactersAsCodePoints($input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\ShiftJIS::__construct
     * @covers MensBeam\Intl\Encoding\ShiftJIS::nextChar
     */
    public function testDecodeMultipleCharactersAsStrings(string $input, array $exp) {
        return parent::testDecodeMultipleCharactersAsStrings($input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\ShiftJIS::seekBack
     */
    public function testSTepBackThroughAString(string $input, array $exp) {
        return parent::testSTepBackThroughAString($input, $exp);
    }

    /**
     * @covers MensBeam\Intl\Encoding\ShiftJIS::seek
     * @covers MensBeam\Intl\Encoding\ShiftJIS::posChar
     * @covers MensBeam\Intl\Encoding\ShiftJIS::posByte
     * @covers MensBeam\Intl\Encoding\ShiftJIS::rewind
     */
    public function testSeekThroughAString() {
        return parent::testSeekThroughAString();
    }

    /**
     * @covers MensBeam\Intl\Encoding\ShiftJIS::posChar
     * @covers MensBeam\Intl\Encoding\ShiftJIS::posByte
     * @covers MensBeam\Intl\Encoding\ShiftJIS::eof
     */
    public function testTraversePastTheEndOfAString() {
        return parent::testTraversePastTheEndOfAString();
    }

    /**
     * @covers MensBeam\Intl\Encoding\ShiftJIS::peekChar
     * @covers MensBeam\Intl\Encoding\ShiftJIS::stateSave
     * @covers MensBeam\Intl\Encoding\ShiftJIS::stateApply
     */
    public function testPeekAtCharacters() {
        return parent::testPeekAtCharacters();
    }

    /**
     * @covers MensBeam\Intl\Encoding\ShiftJIS::peekCode
     * @covers MensBeam\Intl\Encoding\ShiftJIS::stateSave
     * @covers MensBeam\Intl\Encoding\ShiftJIS::stateApply
     */
    public function testPeekAtCodePoints() {
        return parent::testPeekAtCodePoints();
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\ShiftJIS::lenChar
     * @covers MensBeam\Intl\Encoding\ShiftJIS::lenByte
     * @covers MensBeam\Intl\Encoding\ShiftJIS::stateSave
     * @covers MensBeam\Intl\Encoding\ShiftJIS::stateApply
     */
    public function testGetStringLength(string $input, array $points) {
        return parent::testGetStringLength($input, $points);
    }

    /**
     * @covers MensBeam\Intl\Encoding\ShiftJIS::errDec
     */
    public function testReplacementModes() {
        return parent::testReplacementModes();
    }

    /**
     * @dataProvider provideStrings
     * @covers MensBeam\Intl\Encoding\ShiftJIS::rewind
     * @covers MensBeam\Intl\Encoding\ShiftJIS::chars
     * @covers MensBeam\Intl\Encoding\ShiftJIS::codes
     */
    public function testIterateThroughAString(string $input, array $exp) {
        return parent::testIterateThroughAString($input, $exp);
    }

    /**
     * @dataProvider provideStrings
     * @coversNothing
     */
    public function testIterateThroughAStringAllowingSurrogates(string $input, array $strictExp, array $relaxedExp = null) {
        return parent::testIterateThroughAStringAllowingSurrogates($input, $strictExp, $relaxedExp);
    }

    /**
     * @covers MensBeam\Intl\Encoding\ShiftJIS::seekBack
     */
    public function testSeekBackOverRandomData() {
        return parent::testSeekBackOverRandomData();
    }

    /**
     * @group optional
     */
    public function testPedanticallyDecodeSingleCharactersAsCodePoint() {
        $series = [
            'characters' => [["80","5C","8198","814E","818B","817D","814C","81F7","817E","8180","839F","83A0","83A1","83A2","83A3","83A4","83A5","83A6","83A7","83A8","83A9","83AA","83AB","83AC","83AD","83AE","83AF","83B0","83B1","83B2","83B3","83B4","83B5","83B6","83BF","83C0","83C1","83C2","83C3","83C4","83C5","83C6","83C7","83C8","83C9","83CA","83CB","83CC","83CD","83CE","83CF","83D0","83D1","83D2","83D3","83D4","83D5","83D6","8446","8440","8441","8442","8443","8444","8445","8447","8448","8449","844A","844B","844C","844D","844E","844F","8450","8451","8452","8453","8454","8455","8456","8457","8458","8459","845A","845B","845C","845D","845E","845F","8460","8470","8471","8472","8473","8474","8475","8477","8478","8479","847A","847B","847C","847D","847E","8480","8481","8482","8483","8484","8485","8486","8487","8488","8489","848A","848B","848C","848D","848E","848F","8490","8491","8476","815D","815C","8165","8166","8167","8168","81F5","81F6","8164","8163","81F1","818C","818D","81A6","7E","818E","8782","8784","81F0","8754","8755","8756","8757","8758","8759","875A","875B","875C","875D","FA40","FA41","FA42","FA43","FA44","FA45","FA46","FA47","FA48","FA49","81A9","81AA","81A8","81AB","81CB","81CC","81CD","81DD","81CE","81DE","81B8","81B9","8794","817C","81E3","81E5","8187","8798","81DA","8161","81C8","81C9","81BF","81BE","81E7","81E8","8793","8188","81E6","81E4","81E0","8182","81DF","8185","8186","81E1","81E2","81BC","81BD","81BA","81BB","81DB","8799","81DC","8740","8741","8742","8743","8744","8745","8746","8747","8748","8749","874A","874B","874C","874D","874E","874F","8750","8751","8752","8753","849F","84AA","84A0","84AB","84A1","84AC","84A2","84AD","84A4","84AF","84A3","84AE","84A5","84BA","84B5","84B0","84A7","84BC","84B7","84B2","84A6","84B6","84BB","84B1","84A8","84B8","84BD","84B3","84A9","84B9","84BE","84B4","81A1","81A0","81A3","81A2","81A5","81A4","819F","819E","819B","819D","819C","81FC","819A","8199","818A","8189","81F4","81F3","81F2","8140","8141","8142","8156","8158","8159","815A","8171","8172","8173","8174","8175","8176","8177","8178","8179","817A","81A7","81AC","816B","816C","8780","8781","829F","82A0","82A1","82A2","82A3","82A4","82A5","82A6","82A7","82A8","82A9","82AA","82AB","82AC","82AD","82AE","82AF","82B0","82B1","82B2","82B3","82B4","82B5","82B6","82B7","82B8","82B9","82BA","82BB","82BC","82BD","82BE","82BF","82C0","82C1","82C2","82C3","82C4","82C5","82C6","82C7","82C8","82C9","82CA","82CB","82CC","82CD","82CE","82CF","82D0","82D1","82D2","82D3","82D4","82D5","82D6","82D7","82D8","82D9","82DA","82DB","82DC","82DD","82DE","82DF","82E0","82E1","82E2","82E3","82E4","82E5","82E6","82E7","82E8","82E9","82EA","82EB","82EC","82ED","82EE","82EF","82F0","82F1","814A","814B","8154","8155","8340","8341","8342","8343","8344","8345","8346","8347","8348","8349","834A","834B","834C","834D","834E","834F","8350","8351","8352","8353","8354","8355","8356","8357","8358","8359","835A","835B","835C","835D","835E","835F","8360","8361","8362","8363","8364","8365","8366","8367","8368","8369","836A","836B","836C","836D","836E","836F","8370","8371","8372","8373","8374","8375","8376","8377","8378","8379","837A","837B","837C","837D","837E","8380","8381","8382","8383","8384","8385","8386","8387","8388","8389","838A","838B","838C","838D","838E","838F","8390","8391","8392","8393","8394","8395","8396","8145","815B","8152","8153","878A","878B","878C","8785","8786","8787","8788","8789","8765","8769","8760","8763","8761","876B","876A","8764","876C","8766","876E","875F","876D","8762","8767","8768","877E","878F","878E","878D","8772","8773","876F","8770","8771","8775","8774","8783","88EA","929A","8EB5","969C","8FE4","8E4F","8FE3","89BA","9573","975E","98A0","894E","8A8E","98A1","90A2","99C0","8B75","95B8","8FE5","97BC","95C0","FA68","98A2","9286","98A3","8BF8","98A4","8ADB","924F","8EE5","98A5","98A6","98A7","9454","8B76","9456","93E1","8CC1","9652","E568","98A8","8FE6","98A9","89B3","8BE3","8CEE","96E7","9BA4","9790","93FB","8AA3","8B54","98AA","98AB","97B9","975C","9188","98AD","8E96","93F1","98B0","895D","8CDD","8CDC","88E4","986A","9869","8DB1","889F","98B1","98B2","98B3","9653","98B4","8CF0","88E5","9692","8B9C","8B9D","8B9E","92E0","97BA","98B5","98B6","98B7","906C","8F59","906D","98BC","98BA","98BB","8B77","8DA1","89EE","98B9","98B8","95A7","8E65","8E64","91BC","98BD","9574","90E5","8157","98BE","98C0","FA69","91E3","97DF","88C8","98BF","89BC","8BC2","9287","8C8F","98C1","9443","FA6A","FA6B","8AE9","FA6C","98C2","88C9","8CDE","8AEA","959A","94B0","8B78","89EF","98E5","9360","948C","98C4","94BA","97E0","904C","FA6D","8E66","8E97","89BE","92CF","9241","98C8","88CA","92E1","8F5A","8DB2","9743","91CC","89BD","FA6E","98C7","975D","98C3","98C5","8DEC","98C6","9B43","98CE","98D1","98CF","89C0","95B9","98C9","98CD","8CF1","8E67","8AA4","98D2","98CA","FA70","97E1","8E98","98CB","98D0","FA6F","FA72","98D3","98CC","FA71","8B9F","88CB","8BA0","89BF","9B44","9699","958E","8CF2","904E","97B5","95D6","8C57","91A3","89E2","FA61","8F72","FA73","98D7","98DC","98DA","98D5","91AD","98D8","98DB","98D9","95DB","98D6","904D","9693","98DD","98DE","8F43","98EB","946F","9555","98E6","95EE","89B4","98EA","FA76","98E4","98ED","9171","8CC2","947B","E0C5","98EC","937C","98E1","8CF4","8CF3","98DF","FA77","8ED8","98E7","FA75","95ED","926C","98E3","8C91","98E0","98E8","98E2","97CF","98E9","9860","8BE4","8C90","FA74","FA7A","98EE","FA78","98EF","98F3","88CC","95CE","98F2","98F1","98F5","98F4","92E2","8C92","98F6","FA79","8EC3","91A4","92E3","8BF4","98F7","8B55","98F8","98FA","9654","8C86","FA7B","8E50","94F5","98F9","8DC3","9762","98FC","9942","98FB","8DC2","8F9D","8C58","9943","8BCD","9940","9941","93AD","919C","8BA1","966C","9944","FA7D","97BB","9945","9948","9946","916D","9947","9949","FA7C","994B","994A","95C6","8B56","994D","994E","89AD","994C","8EF2","9951","9950","994F","98D4","9952","8F9E","9953","9744","96D7","9955","9954","9957","9956","9958","9959","88F2","8CB3","8C5A","8F5B","929B","8BA2","90E6","8CF5","FA7E","8D8E","995B","96C6","9365","8E99","995A","995C","937D","8A95","995D","FA80","93FC","9153","995F","9960","94AA","8CF6","985A","9961","8BA4","95BA","91B4","8BEF","9354","8C93","9962","9963","93E0","897E","9966","8DFB","9965","8DC4","9967","E3EC","9968","9660","9969","996A","996B","8FE7","8ECA","FA81","8AA5","996E","996C","96BB","996D","9579","996F","9970","9971","937E","9975","9973","9974","9972","8DE1","9976","96E8","97E2","9977","FA82","90A6","9978","8F79","9979","929C","97BD","9380","99C3","997A","EAA3","8BC3","997B","967D","8F88","91FA","997D","93E2","FA83","997E","9980","8A4D","9981","8BA5","93CA","899A","8F6F","949F","9982","9381","906E","9983","95AA","90D8","8AA0","8AA7","9984","9986","8C59","9985","FA84","97F1","8F89","94BB","95CA","9987","9798","9988","9989","939E","998A","90A7","8DFC","8C94","998B","8E68","8D8F","92E4","998D","91A5","8DED","998E","998F","914F","998C","9991","9655","8D84","9990","8C95","8DDC","948D","9994","9992","959B","8FE8","999B","8A84","9995","9993","916E","9997","9996","8A63","8C80","999C","97AB","9998","999D","999A","9999","97CD","FA85","8CF7","89C1","97F2","FA86","8F95","9377","8D85","99A0","99A1","FB77","97E3","984A","99A3","8CF8","99A2","8A4E","FA87","99A4","9675","92BA","9745","95D7","99A5","E8D3","93AE","99A6","8AA8","96B1","FA88","8F9F","99A7","95E5","99AB","90A8","99A8","8BCE","99A9","8AA9","8C4D","99AC","99AD","99AE","99AF","8ED9","8CF9","96DC","FA89","96E6","93F5","95EF","99B0","FA8A","99B1","99B3","99B5","99B4","99B6","89BB","966B","8DFA","99B7","9178","8FA0","8BA7","99B8","FA8B","94D9","99B9","99BA","99BB","99BC","9543","8BE6","88E3","93BD","99BD","8F5C","90E7","99BF","99BE","8FA1","8CDF","99C1","94BC","99C2","94DA","91B2","91EC","8BA6","93EC","9250","948E","966D","99C4","90E8","8C54","99C5","99C6","894B","88F3","8AEB","FA8C","91A6","8B70","9791","99C9","89B5","99C8","8BA8","99CA","96EF","FA8D","99CB","97D0","8CFA","8CB4","99CC","99CE","99CD","907E","8958","897D","99CF","99D0","FA8E","8CB5","99D1","8B8E","8E51","99D2","9694","8DB3","8B79","9746","916F","94BD","8EFB","8F66","8EE6","8EF3","8F96","94BE","FA8F","99D5","8962","9170","8CFB","8CC3","8BE5","99D9","9240","91FC","8BA9","8FA2","99DA","99D8","89C2","91E4","8EB6","8E6A","8945","8A90","8D86","8E69","99DB","99DC","8B68","8A65","8D87","8B67","92DD","8944","93AF","96BC","8D40","9799","9366","8CFC","8C4E","99E5","8BE1","9669","94DB","99E4","8ADC","99DF","99E0","99E2","99E3","8B7A","9081","95AB","99E1","99DD","8CE1","99DE","9843","95F0","92E6","8CE0","8D90","99E6","93DB","99EA","8EFC","8EF4","99ED","99EB","96A1","99E8","99F1","99EC","99EF","8CC4","96BD","99F0","99F2","99F4","FA92","8DEE","9861","99E9","99E7","99F3","99EE","FA91","99F6","9A42","99F8","99FC","FA93","9A40","99F9","9A5D","8DE7","8A50","99F7","9A44","88F4","9A43","88A3","9569","9A41","99FA","99F5","99FB","8DC6","9A45","88F5","9A4E","9A46","9A47","8FA3","9689","9A4C","9A4B","934E","9A4D","9A4A","FA94","8953","8DB4","904F","9A48","9382","9A49","88A0","9A53","9742","8FA5","9A59","9A58","9A4F","91C1","9A50","91ED","9A55","8FA4","9A52","96E2","8C5B","9A56","9A57","9A54","9A5A","9A51","9A60","9A65","9A61","9A5C","9A66","9150","FA95","9A68","8D41","9A5E","929D","9A62","9A5B","8AAB","8AEC","8A85","9A63","9A5F","8C96","9A69","9A67","9172","8B69","8BAA","9A64","8BF2","8963","9A6D","9A6B","9AA5","9A70","9A6A","9A6E","9A6C","8E6B","9A6F","9A72","9A77","9A75","9A74","9251","89C3","9A71","9A73","8FA6","8952","9A76","89DC","9A82","8FFA","9A7D","9A7B","9A7C","9A7E","895C","9158","9A78","9A79","8A9A","9A81","8AED","9A84","9A80","9A83","95AC","93D3","94B6","9A86","9A85","8A64","9A87","9A8A","9A89","9A88","9458","9A8B","9A8C","9A8E","9A8D","9A90","9A93","9A91","9A8F","9A92","9A94","9A95","9A96","9A97","9A98","9964","8EFA","8E6C","89F1","88F6","9263","9A99","8DA2","88CD","907D","9A9A","8CC5","8D91","9A9C","9A9B","95DE","9A9D","9A9F","9A9E","9AA0","9AA1","8C97","8980","9AA2","9AA4","9AA3","9AA6","9379","9AA7","88B3","8DDD","8C5C","926E","9AA8","9AA9","9AAB","9AAC","8DE2","8BCF","9656","9AAA","9AAD","8DBF","8D42","FA96","9AB1","8DA3","FA97","9252","9AAE","92D8","9AB2","9082","9AB0","9AB3","8C5E","9AB4","9AB5","8D43","8A5F","9AB7","9AB8","FA98","9AB9","9AB6","9AAF","9ABA","9ABB","FA9A","FA99","9684","8FE9","9ABD","9ABE","9ABC","9AC0","9457","88E6","9575","9AC1","8FFB","8EB7","947C","8AEE","8DE9","9678","93B0","8C98","91CD","9ABF","9AC2","91C2","9AC3","9AC4","9AC6","92E7","8AAC","EA9F","8981","95F1","8FEA","9367","8DE4","9ACC","95BB","97DB","89F2","9AC8","9159","9ACB","9383","9368","9384","94B7","92CB","8DC7","9AC7","8996","9355","9AC9","9AC5","906F","9ACD","8F6D","8BAB","9ACE","95E6","919D","92C4","FA9D","9AD0","966E","9AD1","9AD6","FA9E","95AD","9AD5","9ACF","9AD2","9AD4","8DA4","95C7","9AD7","9264","89F3","8FEB","9AD9","9AD8","8D88","9ADA","9ADC","9ADB","9ADE","9AD3","9AE0","9ADF","9ADD","8E6D","9070","9173","9AE1","90BA","88EB","9484","92D9","9AE3","9AE2","9AE4","9AE5","9AE6","9AE7","95CF","9AE8","FA9F","89C4","9AE9","975B","8A4F","99C7","8F67","91BD","9AEA","96E9","96B2","9AEC","91E5","9356","91BE","9576","9AED","9AEE","899B","8EB8","9AEF","88CE","9AF0","9AF1","8982","8AEF","93DE","95F2","9AF5","9174","9AF4","8C5F","FAA0","967A","9AF3","9385","9AF7","9AF6","FAA1","FAA2","9AF9","9AF8","FAA3","899C","9AFA","8FA7","9AFC","9244","9AFB","95B1","8F97","937A","9B40","8D44","9B41","9440","94DC","96CF","9444","9B4A","8B57","9764","96AD","9BAA","9B42","9B45","FAA4","91C3","9657","9369","9B46","9685","FAA5","8DC8","8FA8","9B47","8E6F","8E6E","88B7","8CC6","90A9","88CF","9B4B","9B4C","9B49","8957","8AAD","9B48","96C3","9550","88A6","88F7","8E70","88D0","88A1","9B51","9B4F","96BA","9B52","9B50","9B4E","9050","9B4D","95D8","8CE2","9B56","9B57","8FA9","9B53","984B","946B","9B55","8DA5","9B58","9577","9B59","9B54","96B9","947D","9B5A","9551","9B5B","9B5F","9B5C","89C5","9B5E","8EB9","9B5D","8C99","9B6B","9B64","9B61","9284","9B60","9B62","9B63","9B65","9B66","8AF0","9B68","9B67","9B69","8FEC","9B6C","92DA","8964","9B6A","9B6D","9B6E","9B71","9B6F","9B70","8E71","9B72","8D45","9B73","FAA6","8E9A","91B6","9B74","9B75","8E79","8D46","96D0","8B47","8CC7","9B76","8A77","9B77","91B7","9B78","9BA1","9B79","9B7A","9B7B","9B7D","9B7E","9B80","91EE","8946","8EE7","88C0","9176","8AAE","8EB3","8D47","9386","8F40","8AAF","9288","92E8","88B6","8B58","95F3","8EC0","8B71","90E9","8EBA","9747","9B81","8B7B","8DC9","8A51","8983","8FAA","89C6","9B82","9765","8F68","FAA7","8EE2","9B83","8AF1","93D0","96A7","9B84","9B85","9578","9B87","8AA6","8BF5","9B86","FAA9","8AB0","9051","9B8B","8E40","89C7","9B8A","9B88","9B8C","9B89","944A","9ECB","9052","9B8D","FAAA","97BE","9B8E","9B90","929E","9B8F","90A1","8E9B","91CE","8EF5","9595","90EA","8ECB","9B91","8FAB","9B92","9B93","88D1","91B8","9071","9B94","93B1","8FAC","8FAD","9B95","90EB","8FAE","FAAB","9B96","9B97","96DE","9B98","8BC4","8F41","9B99","9B9A","8EDA","904B","93F2","9073","94F6","9441","8BC7","9B9B","8B8F","9B9C","8BFC","93CD","89AE","8E72","9B9D","9BA0","9B9F","8BFB","9B9E","9357","91AE","936A","8EC6","9177","979A","9BA2","9BA3","93D4","8E52","9BA5","9BA6","9BA7","8AF2","9BA8","9BA9","89AA","FAAC","915A","8AE2","9BAB","96A6","91D0","8A78","9BAD","9BAF","8ADD","FAAD","9BAC","9BAE","9BB1","9BB0","9BB2","9BB3","93BB","8BAC","89E3","9BB4","9BB9","9BB7","95F5","95F4","FAAE","9387","9BB6","8F73","9BB5","9092","9BBA","8DE8","9BC0","9BC1","9BBB","8A52","9BBC","9BC5","9BC4","9BC3","9BBF","9BBE","9BC2","FAAF","95F6","FAB2","9BC9","9BC6","9BC8","9792","9BC7","FAB0","9BBD","9093","9BCA","FAB3","8DB5","9BCB","9BCC","9BCF","9BCE","9BCD","9388","9BB8","9BD5","9BD1","9BD0","9BD2","9BD3","9BD6","FAB4","FAB5","97E4","9BD7","9BD4","9BD8","8ADE","9BD9","FAB6","9BDB","9BDA","9BDC","9BDD","90EC","8F42","8F84","9183","8D48","8DB6","8D49","8B90","9BDE","8DB7","8CC8","9BDF","96A4","9462","9BE0","8D4A","8AAA","9246","8BD0","8E73","957A","94BF","9BE1","8AF3","9BE4","929F","9BE3","9BE2","9BE5","92E9","9083","8E74","90C8","91D1","8B41","92A0","9BE6","9BE7","8FED","9658","9BEA","9BE9","9BE8","959D","9BF1","9679","9BEB","9BED","968B","9BEC","9BEE","94A6","9BEF","95BC","9BF0","8AB1","95BD","944E","9BF2","9BF3","8D4B","8AB2","9BF4","8CB6","9763","9748","8AF4","9BF6","92A1","8D4C","8FAF","94DD","8FB0","8F98","92EA","95F7","9358","8D4D","957B","9BF7","9378","8DC0","8CC9","92EB","88C1","8F8E","8D4E","9766","9BF8","9BF9","9470","9BFA","97F5","984C","9BFC","9BFB","8A66","9C40","9C43","9C44","9C42","955F","8FB1","9C46","9C45","9C41","9C47","9C48","9C49","9C4C","9C4A","9C4B","9C4D","8984","92EC","9C4E","8C9A","89F4","9455","9C4F","93F9","95D9","9C50","984D","9C51","95BE","9C54","989F","98AF","8EAE","93F3","9C55","8B7C","92A2","88F8","9C56","95A4","8D4F","926F","92ED","FAB7","96ED","8CB7","8CCA","9C57","9C58","9C5E","8EE3","FAB8","92A3","8BAD","9C59","954A","9265","9C5A","FA67","9C5B","8BAE","9C5C","9C5D","9C5F","9396","9C60","9C61","9C62","9C53","9C52","9C63","8C60","9546","FAB9","8DCA","9556","92A4","956A","9C64","8FB2","8965","9C65","9C66","96F0","94DE","9C69","899D","90AA","9C68","9C67","8C61","91D2","9C6D","9C6B","9C6A","97A5","8CE3","8F99","9C6C","936B","8F5D","93BE","9C70","9C6F","9C6E","9C71","8CE4","9C72","959C","8F7A","9C73","94F7","93BF","92A5","FABA","934F","9C74","8B4A","9053","954B","8AF5","9445","9C75","8E75","9659","965A","899E","9C7A","FABB","9289","9C77","89F5","9CAB","9C79","944F","9C78","9C76","8D9A","9C7C","9C83","9C89","9C81","937B","9C86","957C","9C80","9C85","97E5","8E76","91D3","9C7D","8B7D","9C88","90AB","8985","9C82","89F6","9C87","8BAF","9C84","9C8A","9C8C","9C96","9C94","9C91","9C90","97F6","9C92","8BB0","8D50","8F9A","9C99","9C8B","FABC","9C8F","9C7E","89F8","9C93","9C95","9270","8DA6","89B6","9C8D","9C98","9C97","8BB1","91A7","8A86","8C62","9C8E","9C9A","9C9D","9C9F","FABD","8EBB","FABE","9CA5","92EE","9C9B","9CA3","89F7","9CA1","9CA2","9C9E","9CA0","8CE5","9749","8AB3","8978","9CA4","9459","88AB","94DF","9C7B","9CAA","9CAE","96E3","9CA7","9389","9CAC","8FEE","9CAD","93D5","9866","9CA9","FAC0","9CAF","8D9B","90C9","FABF","88D2","9CA8","9CA6","9179","9C9C","8E53","91C4","9CBB","FAC2","917A","9CB6","9CB3","9CB4","8EE4","9CB7","9CBA","9CB5","8F44","9CB8","9CB2","96FA","96F9","9CBC","9CBD","88D3","FAC3","9CB1","8BF0","88A4","8AB4","FAC1","9CB9","9CC1","9CC0","9CC5","FAC5","9CC6","FAC4","9CC4","9CC7","9CBF","9CC3","9CC8","9CC9","9CBE","8E9C","9CC2","91D4","8D51","9CB0","9054","9CD6","95E7","9CCC","9CCD","9CCE","9CD5","9CD4","969D","8AB5","9CD2","8C64","8A53","9CCF","97B6","9CD1","88D4","9CD3","9CCA","9CD0","9CD7","8C63","9CCB","977C","974A","9CDA","9CDE","919E","97F7","9CDF","9CDC","9CD9","FAC6","9CD8","9CDD","95AE","93B2","8C65","9CE0","9CDB","9CE1","8C9B","89AF","9CE9","8AB6","9CE7","9CE8","8DA7","9CE6","9CE4","9CE3","9CEA","9CE2","9CEC","89F9","9CEE","9CED","92A6","9CF1","9CEF","9CE5","8C9C","9CF0","9CF4","9CF3","9CF5","9CF2","9CF6","9CF7","9CF8","95E8","9CFA","9CF9","8F5E","90AC","89E4","89FA","FAC7","9CFB","88BD","90CA","9CFC","E6C1","9D40","8C81","9D41","90ED","9D42","9D43","8B59","9D44","9D45","9D46","91D5","8CCB","96DF","965B","8F8A","9D47","90EE","E7BB","94E0","8EE8","8DCB","9D48","91C5","95A5","91EF","9D4B","9D49","9D4C","9D4A","9D4D","95AF","88B5","957D","94E1","9D4E","9D51","8FB3","8B5A","9D4F","9D56","8FB4","9D50","9463","977D","9D52","9D53","9D57","938A","9D54","8D52","90DC","9D65","94B2","91F0","FAC8","94E2","9DAB","95F8","92EF","9695","9D5A","899F","928A","9D63","9253","9D5D","9D64","9D5F","9D66","9D62","9D61","948F","9D5B","89FB","9D59","8B91","91F1","9D55","9D58","8D53","90D9","8FB5","9D60","9471","8B92","8A67","8A87","9040","9D68","9D6D","9D69","8C9D","9D6E","8E41","8D89","8F45","9D5C","8E9D","9D6B","8E77","9D6C","88C2","9D67","92A7","8B93","8BB2","9D6A","88A5","8DC1","9055","92F0","94D2","9D70","917D","91A8","8E4A","9D71","9D73","9D6F","95DF","92BB","917B","95F9","8ECC","9D80","9D7E","9098","8C9E","9D78","8FB7","93E6","9450","9D76","917C","8EF6","9D7B","8FB6","9D75","9D7A","9472","9D74","8C40","8A7C","9D7C","97A9","8DCC","9254","9D79","90DA","8D54","9084","8986","915B","9D77","8B64","8C66","92CD","9D7D","917E","9D81","9D83","91B5","9D89","9D84","9D86","9560","92F1","9D87","974B","9767","8AB7","88AC","9D85","9D82","8AF6","8987","FAC9","9D88","9768","9D8C","91B9","9D93","9D8D","9D8A","9D91","9D72","9D8E","9D92","94C0","938B","9D8B","9D8F","8C67","8DEF","90DB","9D97","9345","FACA","9D94","9680","9D95","9D96","96CC","90A0","8C82","9D9D","8E54","9D9A","9D99","9451","FACB","93B3","9350","9D9B","9D9C","958F","9464","8E42","90EF","966F","8A68","9DA3","9D9E","9769","9DA5","9DA1","9DA2","9180","FACC","9DA0","9D5E","9DA4","9D9F","9DA9","9DAA","9346","9DAC","8E43","9DA7","8B5B","9DAD","9DA6","9DB1","9DB0","9DAF","9DB2","9DB4","8FEF","9DB3","9DB7","9DB5","9DB6","9D90","9DB9","9DB8","9D98","9DBA","9DAE","8E78","9DBB","9DBC","9DBE","9DBD","9DBF","89FC","8D55","95FA","90AD","8CCC","9DC1","9DC4","FACD","9571","8B7E","9DC3","9DC2","9473","9DC5","8BB3","9DC7","9DC6","8AB8","8E55","93D6","8C68","9094","9DC8","90AE","9347","957E","9DC9","9DCA","9DCB","95B6","9B7C","90C4","956B","8DD6","94E3","94C1","936C","97BF","9DCD","8ECE","9DCE","88B4","8BD2","90CB","9580","9DCF","8E61","9266","8E7A","9056","9DD0","95FB","8997","8E7B","9DD3","9DD1","9DD4","97B7","9DD2","90F9","9DD5","91B0","9DD6","8AF8","9DD8","9DD7","9DD9","9DDA","8AF9","93FA","9255","8B8C","8E7C","9181","8F7B","88AE","9DDB","89A0","9DDF","FACE","8D56","9DDE","8DA9","8FB8","FAD1","9DDD","8FB9","96BE","8DA8","88D5","90CC","FACF","9DE4","FAD3","90AF","8966","FAD4","8F74","9686","8DF0","8FBA","FAD2","90A5","FA63","9DE3","9DE1","9DE2","FAD0","928B","9E45","9DE8","8E9E","8D57","9DE6","9DE7","9057","9DE5","8E4E","FAD6","FAD7","9DEA","9DE9","9DEE","9DEF","9DEB","FAD5","8A41","9DEC","9DED","94D3","9581","8C69","9DF0","FAD9","90B0","8FBB","9271","8BC5","9DF1","9DF5","89C9","9DF2","9DF4","9DF3","8F8B","9267","88C3","9DF6","FADA","9DF7","FADB","92A8","97EF","8E62","95E9","FADC","965C","9E41","9DF9","9DFC","9DFB","FADD","9DF8","9E40","93DC","9DFA","9E42","8F8C","9E43","976A","9498","9E44","9E46","9E47","9E48","8BC8","8967","8D58","9E49","9E4A","8F91","9182","FADE","FA66","99D6","915D","915C","91D6","8DC5","98F0","8C8E","974C","95FC","959E","FADF","9E4B","8DF1","92BD","9E4C","984E","965D","92A9","9E4D","8AFA","9E4E","9E4F","96D8","96A2","9696","967B","8E44","9E51","8EE9","9670","9E53","9E56","9E55","8AF7","8B80","9E52","9E54","9E57","9099","979B","88C7","8DDE","91BA","8EDB","8FF1","9E5A","936D","9E58","91A9","9E59","8FF0","96DB","9E5B","9E5C","9788","FAE1","9E61","8D59","9474","9E5E","938C","9DDC","9DE0","8B6E","9466","9E60","8FBC","94C2","9E66","94F8","9E5D","9E63","9E62","90CD","968D","97D1","9687","89CA","8E7D","9867","9E65","9095","9E64","9E5F","8CCD","9E6B","9E69","89CB","9E67","9E6D","9E73","FAE2","FAE4","91C6","95BF","9E75","9541","9E74","9490","965E","8AB9","90F5","8F5F","92D1","974D","9E70","9E6F","9E71","9E6E","9E76","9E6C","9E6A","9E72","9E68","928C","96F6","8EC4","8DF2","8DB8","968F","8A60","FAE5","92CC","93C8","8968","90F0","90B2","8C49","9E78","8D5A","8A9C","9E7A","8A94","9E81","9E7D","90F1","8A6A","8DAA","8A69","8DCD","9E7B","8C85","8C6A","938D","FAE6","9E79","88C4","9E7C","9E7E","8BCB","8C4B","FAE3","8ABA","8B6A","9E82","8DF7","9691","8E56","9E83","954F","9E8F","89B1","9E84","9E95","9E85","97C0","9E8C","947E","9E94","9E87","88B2","9E89","8D5B","9E8B","9E8A","9E86","9E91","8FBD","9AEB","8CE6","979C","9E88","92F2","8A42","8DAB","9E80","9E90","8A81","9E8E","9E92","938E","8AFC","9EB0","FA64","96C7","9E97","8AFB","9E9E","FAE7","965F","9E9F","9EA1","9EA5","9E99","9249","938F","9EA9","9E9C","9EA6","9EA0","9058","9EAA","90B1","9EA8","8ABB","986F","9E96","9EA4","88D6","9E98","96B8","9E9D","9041","92C5","9E93","9EA3","909A","9EAD","8A91","8C9F","9EAF","9E9A","9EAE","9EA7","9E9B","9EAB","9EAC","9EBD","93CC","9EA2","9EB9","9EBB","92D6","976B","9596","9EB6","91C8","9EBC","915E","9EB3","9EC0","9EBF","93ED","9EBE","93E8","FAE9","9EC2","9EB5","8BC6","9EB8","8F7C","9480","9EBA","8BC9","9EB2","9EB4","9EB1","984F","8A79","9EB7","9EC1","8A54","8DE5","897C","9ED2","9850","9ED5","FAEB","9059","9ED4","9ED3","9ED0","9EC4","9EE1","9EC3","9ED6","9ECE","9EC9","9EC6","9EC7","9ECF","EAA0","9ECC","8D5C","92C6","9184","9ECA","9EC5","9EC8","976C","968A","9ECD","9ED7","FAEC","9EDF","9ED8","9EE5","9EE3","9EDE","9EDD","92CE","9185","9EDB","9ED9","9EE0","9EE6","94F3","9EEC","9EE7","9EEA","9EE4","9294","9557","9EDA","9EE2","8FBE","96CD","9EF6","9EE9","8CA0","89A1","8A7E","9ED1","FAED","8FBF","9EEE","9EF5","8EF7","8A92","924D","9EEB","FAEF","9EF0","9EF4","8BB4","8B6B","9EF2","8B40","93C9","9EF1","9EF3","FAEE","9EED","FAF0","9EEF","FAF1","8A80","9268","9EFA","9EF8","8CE7","9EF7","9F40","9E77","9EF9","9EFB","9EFC","9F4B","9F47","9E8D","9F46","9F45","9F42","9EE8","9F44","9F43","9F49","9845","9F4C","8BF9","9F48","9F4A","FAF2","FAF3","94A5","9F4D","9F51","9F4E","9793","9F4F","9EDC","9F52","9F53","8954","9F55","8C87","8E9F","8BD3","89A2","977E","9F57","9F56","9F59","8B5C","8BD4","8ABC","9F5C","9F5B","9F5D","89CC","9256","9F5E","8ABD","9F60","9F5F","9F61","9F62","9F63","8E7E","90B3","8D9F","9590","95E0","9863","8E95","8DCE","97F0","9F64","9F65","8E80","9F66","9F67","9F69","9F68","9677","8F7D","8EEA","8E63","9F6A","9F6C","9042","9F6B","9F6D","9F6E","9F6F","9F70","9F71","9F73","9F72","9F74","89A3","9269","9F75","8E45","8A6B","9F76","9361","9ACA","8B42","9F77","9F78","95EA","9688","93C5","9F79","94E4","FAF4","94F9","96D1","9F7A","9F7C","9F7B","9F7E","9F7D","9F81","8E81","96AF","9F82","9F83","8B43","9F84","9F86","9F85","9085","9558","8969","94C3","FAF5","92F3","8F60","8B81","94C4","8EAC","9F88","8ABE","8998","FAF6","93F0","9F87","8D5D","9272","9F89","9F91","9F8A","FAF8","91BF","8B82","9F92","8C88","8B44","9F90","9F8E","9F8B","9780","FAF7","92BE","93D7","9F8C","9F94","9F93","8C42","89AB","8DB9","9F8D","9F8F","9676","91F2","9697","9F9C","9F9D","89CD","95A6","96FB","9F9F","8EA1","8FC0","9F98","9F9E","8988","8BB5","9F95","9F9A","90F2","9491","94E5","9F97","9640","9F99","9FA2","FAF9","9FA0","9F9B","9641","9467","8B83","9344","928D","9FA3","9FA1","91D7","9F96","896A","FAFA","976D","9FAE","9FAD","90F4","9FAA","978C","93B4","9FA4","92C3","896B","8D5E","9FA7","8F46","9FAC","9FAB","9FA6","9FA9","8A88","9FA8","9468","97AC","8FF2","90F3","9FB4","9FB2","956C","9FAF","9FB1","8959","8D5F","9851","8A5C","9582","FAFC","9781","8A43","905A","9FB3","9FB8","FAFB","8FC1","974F","9FB5","9FB0","9FB6","FB40","97DC","9393","93C0","FB41","8A55","8974","9FBC","9FBF","97C1","9784","9FC6","9FC0","9FBD","97D2","9FC3","FB42","8F69","9FC5","9FCA","9391","9FC8","9FC2","9257","9FC9","9FBE","9FC4","9FCB","88FA","9FC1","9FCC","905B","FB44","8F7E","95A3","8DAC","FB43","9FB9","9FC7","9359","FB45","90B4","8A89","8DCF","8FC2","9FBB","8F61","8C6B","9FBA","9FD0","8F8D","8CB8","9FDF","9FD9","8B94","936E","9FD4","9FDD","88AD","8951","FB48","89B7","9FD6","91AA","9FCD","9FCF","8D60","9FE0","FB46","9FDB","FB49","9FD3","9FDA","96A9","9FD8","9FDC","8CCE","8FC3","9258","FB47","9FD2","974E","9FD5","9FCE","9392","9FD1","9FD7","9870","8EBC","969E","9FE1","94AC","9FED","8CB9","8F80","9FE3","97AD","8D61","9FF0","88EC","9FEE","9FE2","9FE8","9FEA","976E","9FE5","934D","9FE7","FB4A","9FEF","9FE9","96C5","9FE4","8EA0","9FFC","8A8A","9FE6","9FEB","9FEC","91EA","91D8","9FF4","9FFA","9FF8","9348","E042","9FF5","9FF6","9FDE","8B99","9559","8EBD","8D97","9852","9FF2","E041","8989","9186","9499","8ABF","97F8","969F","92D0","9FF9","9FFB","9151","E040","9FF7","9FF1","8AC1","8C89","E04E","E049","90F6","8A83","8F81","E052","E04B","92AA","E048","92D7","E06B","E045","E044","E04D","E047","E046","E04C","909F","E043","FB4B","E04F","E050","8AC0","E055","E054","E056","E059","9362","E053","FB4C","E057","8C83","91F7","E051","945A","E058","E05D","E05B","E05E","E061","E05A","8D8A","9447","9FB7","9794","E05C","E060","91F3","E05F","E04A","FB4D","E889","E064","E068","E066","FB4E","FB4F","E062","E063","E067","E065","956D","E06D","E06A","E069","E06C","93D2","E06E","9295","91EB","FB50","90A3","E06F","E071","E070","9FF3","E072","93E5","E073","89CE","9394","8A44","8B84","8EDC","8DD0","FB51","9846","9086","898A","E075","E074","FB52","E078","9259","E07B","E076","E07A","E079","935F","88D7","FA62","97F3","E07D","8947","E080","E07E","E07C","E077","9642","E082","FB54","E081","FB53","898B","E084","95B0","E083","96B3","8FC5","9152","8FC4","FB56","FB57","97F9","E08A","90F7","E086","E08B","898C","FB55","E089","9481","E085","E088","8FC6","94CF","E08C","8ECF","90F8","E08F","E087","8C46","E08D","976F","E090","EAA4","8F6E","E091","E092","944D","E094","E095","FB59","9452","9395","E097","E099","97D3","E096","E098","898D","E093","9A7A","E09A","9187","8E57","E09C","E09B","9043","99D7","E09D","E09F","E08E","E09E","FB5A","E0A0","949A","E0A1","E0A2","E0A3","E0A4","92DC","E0A6","E0A5","E0A7","E0A8","8EDD","9583","96EA","E0A9","E0AA","9175","8EA2","E0AB","E0AC","E0AD","95D0","94C5","E0AE","9476","92AB","E0AF","89E5","8B8D","96C4","96B4","89B2","9853","9671","95A8","90B5","E0B0","93C1","8CA1","E0B1","8DD2","E0B3","E0B2","E0B4","E0B5","E0B6","8B5D","E0B7","E0B8","8CA2","94C6","FB5B","E0BA","8FF3","E0B9","FB5C","8BB6","E0BB","E0BD","E0BC","E0BE","8CCF","E0BF","8BE7","915F","8D9D","E0C1","E0C2","E0C0","8EEB","93C6","8BB7","E0C4","924B","E0C3","9854","9482","E0C7","E0C9","E0C6","96D2","E0C8","E0CA","97C2","FB5D","E0CE","E0CD","9296","944C","8CA3","E0CC","E0CB","9750","9751","E0CF","898E","8D96","8E82","E0D0","E0D1","E0D3","8F62","E0D5","E0D4","E0D6","8A6C","E0D8","FB5F","E0D7","E0DA","E0D9","8CBA","97A6","8BCA","89A4","8BE8","8ADF","97E6","E0DC","E0DE","FB60","E0DF","89CF","E0DB","FB61","8E58","92BF","E0DD","FB64","FB62","E0E2","8EEC","FB63","E0E0","8C5D","94C7","E0E1","E0FC","FB66","E0E7","8CBB","8B85","E0E4","979D","FB65","97AE","91F4","E0E6","FB67","FB69","FB68","FB6A","E0E8","97D4","8BD5","94FA","9469","E0E9","E0EB","E0EE","E0EA","E0ED","8CE8","896C","E0EF","9090","E0EC","97DA","FB6B","E0F2","EAA2","E0F0","E0F3","E0E5","E0F1","8DBA","E0F4","E0F5","979E","FB6C","E0F6","E0F7","FB6D","E0E3","E0F8","8AC2","8EA3","E0F9","E0FA","E0FB","895A","E140","955A","E141","8AA2","E142","E143","E144","E146","E147","E145","9572","E149","E148","FB6E","E14B","E14A","E14C","E14D","E14F","E14E","8D99","E151","E150","8AC3","9072","935B","E152","90B6","8E59","8999","E153","9770","95E1","E154","FAA8","9363","9752","8D62","905C","926A","99B2","92AC","89E6","E155","E156","E15B","E159","E158","9DC0","8A45","E157","88D8","94A8","94C8","97AF","E15C","E15A","927B","90A4","94A9","954C","E15E","97AA","8C6C","E15F","E15D","94D4","E160","E161","FB6F","88D9","8FF4","E166","E163","93EB","E162","8B45","E169","E164","E165","E168","E167","9544","9161","9160","8B5E","E16A","E16B","E16C","E16E","E16D","8975","E176","94E6","E170","E172","E174","905D","E175","E173","8EBE","E16F","E171","9561","8FC7","E178","E177","E179","8EA4","8DAD","9397","E17A","92C9","E17C","979F","E17B","9189","E182","E184","E185","9273","E183","E180","E17D","E17E","E181","E188","E186","E187","E189","E18B","E18C","E18D","E18E","E18A","E190","E18F","E191","97C3","E194","E192","E193","8AE0","96FC","95C8","E196","E195","E197","E198","E19C","E199","E19A","E19B","E19D","E19E","E19F","E1A0","E1A1","94AD","936F","E1A2","9492","9553","E1A3","FB70","E1A4","9349","8A46","8D63","E1A5","E1A6","E1A7","8E48","E1A9","E1A8","E1AA","E1AB","FB73","FB71","FB72","FB74","94E7","E1AC","E1AD","EA89","E1AE","E1AF","E1B0","8E4D","E1B1","9475","967E","896D","8976","E1B2","E1B4","E1B3","9390","90B7","9F58","E1B5","96BF","E1B6","8AC4","94D5","E1B7","E1B8","E1B9","96DA","96D3","92BC","918A","E1BB","8F82","8FC8","E1BE","E1BD","E1BC","94FB","8AC5","8CA7","E1C4","E1C1","905E","96B0","E1C0","E1C2","E1C3","E1BF","E1C5","E1C6","92AD","8AE1","9285","FB76","E1C7","E1C8","E1CB","9087","93C2","E1CC","9672","E1C9","E1CA","E1CF","E1CE","E1CD","E1D1","E1D0","E1D2","E1D4","E1D3","95CB","8F75","97C4","E1D5","93B5","E1D6","E1D7","E1DB","E1D9","E1DA","E1D8","E1DC","E1DD","E1DE","E1DF","96B5","E1E0","96EE","E1E1","926D","948A","8BE9","925A","E1E2","8BB8","90CE","E1E3","8DBB","E1E4","E1E5","8CA4","8DD3","E1E7","FB78","9375","8DD4","8B6D","9643","946A","9376","8D7B","E1E9","FB79","8FC9","FB7A","97B0","8D64","8CA5","94A1","E1EB","FB7B","E1ED","8CE9","E1EC","92F4","E1EF","8A56","E1EA","94E8","894F","8DEA","9871","E1EE","E1F0","95C9","90D7","E1F2","E1F3","E1F1","8A6D","E1F9","E1F8","8EA5","E1FA","E1F5","E1FB","E1F6","94D6","E1F4","E1F7","E241","E240","9681","E1FC","88E9","E243","E242","8FCA","E244","9162","E246","E245","E247","E1E6","E1E8","E249","E248","FB7C","8EA6","97E7","8ED0","E24A","8C56","8B5F","8B46","8E83","9753","E250","E24F","9163","E24C","E24E","8F6A","905F","E24D","E24B","9449","8FCB","955B","8DD5","9398","E251","E252","E268","8BD6","985C","9154","E253","89D0","92F5","959F","FB81","FB83","E254","8B9A","E255","E257","E258","9448","E259","E25A","E25B","8BD7","89D1","93C3","8F47","8E84","E25C","8F48","89C8","9562","E25D","94E9","9164","E260","E261","9489","9060","E25E","9281","E25F","8FCC","88DA","8B48","E262","92F6","E263","90C5","96AB","9542","E264","E265","9274","97C5","E267","E266","8EED","E269","88EE","E26C","E26A","89D2","8C6D","E26B","8D65","8D92","95E4","E26D","9673","E26F","90CF","896E","89B8","88AA","E26E","E270","E271","8FF5","E272","8A6E","E274","8C8A","8B86","E275","8BF3","E276","90FA","93CB","90DE","8DF3","E277","9282","918B","E279","E27B","E278","E27A","8C41","E27C","8C45","8B87","9771","E27E","E280","894D","E283","8A96","E282","E281","E285","E27D","E286","97A7","E287","E288","FB84","9AF2","E28A","E289","E28B","E28C","97B3","E28D","E8ED","8FCD","E28E","E28F","8F76","93B6","E290","FB85","9247","FB87","E291","925B","E292","8BA3","995E","927C","8EB1","8AC6","E293","E2A0","E296","8B88","E295","E2A2","E294","8FCE","E298","E299","934A","E29A","8A7D","9079","9584","E29C","91E6","E297","E29B","E29D","8DF9","E2A4","954D","94A4","9399","8BD8","E2A3","E2A1","94B3","E29E","927D","939B","939A","8DF4","E2B6","E2A6","E2A8","E2AB","E2AC","E2A9","E2AA","E2A7","E2A5","E29F","95CD","89D3","E2B3","E2B0","E2B5","E2B4","9493","96A5","8E5A","E2AE","E2B7","E2B2","E2B1","E2AD","FB88","E2AF","8AC7","925C","90FB","94A0","E2BC","94A2","90DF","E2B9","94CD","E2BD","95D1","927A","E2B8","E2BA","E2BB","E2BE","8EC2","93C4","E2C3","E2C2","E2BF","9855","E2C8","E2CC","E2C9","E2C5","E2C6","E2CB","E2C0","99D3","E2C7","E2C1","E2CA","E2D0","8AC8","E2CD","E2CE","E2CF","E2D2","E2D1","94F4","E2D3","97FA","95EB","E2D8","E2D5","E2D4","90D0","E2D7","E2D9","E2D6","E2DD","E2DA","E2DB","E2C4","E2DC","E2DE","E2DF","95C4","E2E0","96E0","8BCC","8C48","E2E1","95B2","9088","96AE","E2E2","97B1","9494","9165","9453","8F6C","88BE","E2E7","E2E5","E2E3","8A9F","8FCF","E2E8","E2E6","E2E4","E2EC","E2EB","E2EA","E2E9","E2ED","E2EE","90B8","E2EF","E2F1","E2F0","8CD0","9157","E2F3","939C","E2F2","E2F4","95B3","918C","8D66","E2F5","97C6","E2F7","E2F8","E2F9","E2FA","8E85","E2FB","8C6E","8B8A","8B49","E340","96F1","8D67","E2FC","E343","96E4","945B","9552","8F83","E342","8ED1","8D68","8E86","8B89","95B4","E341","9166","9661","8DF5","8E87","92DB","E346","97DD","8DD7","E347","9061","E349","8FD0","8DAE","E348","8F49","8CBC","9167","E344","E34A","FB8A","E345","8C6F","E34D","E351","8C8B","E34C","E355","FB8B","8D69","978D","88BA","E352","8B8B","E34F","E350","939D","E34E","E34B","8A47","90E2","8CA6","E357","E354","E356","E353","8C70","91B1","E358","918E","E365","FB8D","E361","E35B","E35F","8EF8","88DB","E35A","E362","E366","8D6A","96D4","92D4","E35C","FB8C","E364","E359","925D","E35E","88BB","96C8","E35D","8BD9","94EA","918D","97CE","8F8F","E38E","FB8E","E367","90FC","E363","E368","E36A","92F7","E36D","E369","95D2","8AC9","96C9","88DC","E36C","97FB","E36B","898F","93EA","E36E","E375","E36F","E376","E372","949B","8EC8","E374","E371","E377","E370","8F63","9644","8F6B","E373","E380","E37B","E37E","E37C","E381","E37A","E360","90D1","94C9","E37D","E378","9140","8C71","8F4A","FB8F","9044","9155","E384","E386","E387","E383","E385","E379","E382","E38A","E389","969A","8C4A","E388","E38C","E38B","E38F","E391","8E5B","E38D","E392","E393","FA5C","E394","E39A","935A","E396","E395","E397","E398","E399","E39B","E39C","8ACA","E39D","E39E","E39F","FB90","E3A0","E3A1","E3A2","E3A3","E3A4","E3A6","E3A5","E3A7","E3A8","E3A9","E3AC","E3AA","E3AB","8DDF","8C72","9275","94B1","8F90","946C","94EB","E3AD","9CEB","E3AE","E3B0","9785","E3AF","E3B2","E3B1","9772","E3B3","94FC","E3B4","E3B7","E3B6","E3B5","FB91","E3B8","8C51","9141","8B60","E3BC","E3B9","E3BA","E3BD","E3BE","E3BB","8948","89A5","E3C0","E3C1","E3C2","9782","8F4B","E3C4","E3C3","9089","E3C5","E3C6","E3C7","8AE3","8ACB","E3C8","E3C9","967C","9783","9773","9856","8D6C","E3CC","8ED2","E3CB","E3CD","8EA7","91CF","E3CE","8D6B","96D5","E3CF","E3D0","E3D1","E3D2","E3D3","8EA8","96EB","E3D5","925E","E3D4","E3D7","E3D6","E3D8","90B9","E3D9","E3DA","95B7","E3DB","918F","E3DC","E3DD","97FC","E3E0","E3DF","E3DE","92AE","E3E1","9045","E3E2","E3E3","9857","E3E4","E3E5","E3E7","E3E6","94A3","93F7","985D","94A7","E3E9","8FD1","9549","E3EA","E3E8","8ACC","8CD2","8E88","94EC","8CA8","9662","E3ED","E3EB","8D6D","8D6E","88E7","8DE6","9478","88DD","E3F2","925F","9477","91D9","E3F4","E3F0","E3F3","E3EE","E3F1","9645","8CD3","88FB","E3EF","E3F6","E3F7","93B7","8BB9","E445","945C","8E89","8BBA","90C6","9865","96AC","E3F5","90D2","8B72","E3F8","E3FA","E3F9","E3FB","9245","945D","92AF","E442","E441","E3FC","9074","9585","E444","E443","8D6F","9872","E454","E448","E449","8EEE","E447","8D98","E446","E44A","92B0","95A0","9142","91DA","E44E","E44F","E44B","E44C","E44D","8D70","E455","E451","9586","968C","9547","E450","E453","E452","9663","E456","E457","9156","E458","E45A","E45E","E45B","E459","945E","E45C","E45D","89B0","E464","E45F","E460","E461","919F","E463","E462","E465","E466","E467","9062","89E7","E468","97D5","8EA9","8F4C","8E8A","9276","E469","E46A","8950","E46B","E46C","E46D","E46E","E46F","8BBB","9DA8","E470","90E3","E471","8EC9","E472","98AE","E473","95DC","8ADA","9143","8F77","9591","8F4D","E474","8D71","E475","94CA","E484","E477","91C7","9495","8CBD","E476","9144","E478","92F8","E47A","E479","E47C","E47B","E47D","E480","E47E","8ACD","E481","E482","E483","8DAF","97C7","E485","9046","8990","E486","E487","E488","88F0","E489","E48A","9587","8EC5","E48C","8A48","88B0","E48B","E48E","946D","9063","89D4","9646","8C7C","8BDA","E48D","89E8","8AA1","8991","E492","97E8","91DB","9563","E49E","89D5","E49C","E49A","E491","E48F","E490","8EE1","8BEA","9297","93CF","8970","E494","E493","E499","E495","E498","FB93","96CE","E497","89D6","8A9D","E49B","E49D","8C73","E4A1","E4AA","E4AB","88A9","E4B2","88EF","E4A9","E4A8","E4A3","E4A2","E4A0","E49F","9283","91F9","E4A5","E4A4","E4A7","9190","8C74","8960","E4A6","8D72","9191","FB94","E4B8","E4B9","89D7","89AC","E4B6","FB95","E4AC","E4B4","E4BB","E4B5","E4B3","E496","E4B1","E4AD","8ACE","E4AF","E4BA","E4B0","E4BC","E4AE","949C","9789","E4B7","E4CD","E4C5","909B","FB96","8B65","8BDB","E4C0","89D9","8FD2","E4C3","8DD8","9370","E4C8","95EC","E4BF","89D8","8CD4","9548","E4C9","E4BD","FB97","E4C6","E4D0","E4C1","E4C2","93B8","E4C7","E4C4","9647","E4CA","88DE","E4BE","E4CC","E4CB","948B","E4D2","E4DD","8A9E","E4E0","E4CE","E4D3","978E","E4DC","FB98","9774","97A8","9298","8A8B","9592","E4E2","939F","88AF","E4DB","E4D7","9192","E4D1","E4D9","E4DE","944B","88A8","E4D6","E4DF","9598","E4DA","E4D5","8FD3","8F4E","8EAA","96D6","9566","E4E5","E4EE","E4D8","8A97","FB99","8FF6","E4E3","E4E8","9193","E4E4","E4EB","927E","E4EC","9775","E4E1","8A57","E4E7","E4EA","96AA","E4ED","E4E6","E4E9","FA60","9648","9840","E4F1","E4F8","E4F0","8EC1","E4CF","95CC","96A0","E4F7","E4F6","E4F2","E4F3","8955","E4F5","E4EF","92D3","E4F4","88FC","91A0","95C1","E4F9","E540","94D7","E4FC","8FD4","8EC7","E542","8BBC","FB9A","E543","9599","E4FB","FB9B","E4D4","E4FA","986E","93A0","9593","FB9C","E54A","E550","E551","E544","9496","E54E","E546","E548","E552","E547","E54B","8992","93E3","E54C","E54F","E545","9145","E549","8E46","9064","8C4F","96F2","96F7","8F92","FB9E","E556","E554","986D","E553","9795","E555","E557","E558","E55B","E559","93A1","E55A","94CB","E54D","8F93","E55C","E561","9194","E560","E541","E562","9168","E55D","E55F","E55E","9F50","9F41","E564","E563","9796","E1BA","E565","E566","E567","8CD5","8B73","E569","997C","8B95","97B8","8BF1","E56A","E56B","928E","E56C","93F8","88B8","89E1","E571","E572","E56D","8E5C","E56E","9461","E56F","E570","E57A","E574","E577","E573","E575","E576","8ED6","E578","9260","8C75","8A61","E57B","8A5E","E581","E57C","E580","94B8","E57D","E57E","9567","94D8","E582","91FB","E58C","E588","89E9","E586","9649","E587","E584","E585","E58A","E58D","E58B","E589","E583","9277","E594","96A8","E592","E593","E58E","E590","E591","E58F","90E4","9858","E598","E599","E59F","9049","E59B","E59E","E596","E595","E5A0","89DA","E59C","E5A1","E59D","E59A","92B1","E597","9488","E5A5","975A","E5A4","E5A3","E5AC","E5A6","E5AE","9786","E5B1","E5A8","E5A9","E5AD","E5B0","E5AF","E5A7","E5AA","E5BB","E5B4","E5B2","E5B3","E5B8","E5B9","8A49","8B61","E5B7","E5A2","FBA1","E5B6","E5BA","E5B5","E5BC","E5BE","E5BD","E5C0","E5BF","E579","E5C4","E5C1","E5C2","E5C3","E5C5","8C8C","E5C7","E5C6","8F4F","8D73","9FA5","E5C8","8F70","8A58","E5C9","8971","8FD5","E5CA","8D74","E5CB","88DF","955C","E5CC","908A","E5D3","E5D0","928F","E5D1","E5CE","8BDC","E5CD","E5D4","8C55","91DC","E5DA","E5D6","91B3","E5D5","E5D8","E5CF","E5D9","E5DB","94ED","E5D7","E5DC","E5DE","8CD1","E5D2","88BF","E5DD","8DD9","97F4","E5DF","E5E0","9195","97A0","E5E1","9754","E5E2","E5E3","95E2","E5E4","8DBE","97A1","E5E9","E5EA","8FD6","E5E8","FBA2","9787","E5E5","E5E7","90BB","909E","E5E6","E5EB","95A1","E5ED","E5EC","8A8C","964A","E5EE","FA5D","E5FA","E5F0","E5F1","E5F2","E5F3","E5F7","E5F8","E5F6","E5F4","E5EF","E5F5","E5F9","E8B5","89A6","E5FC","8BDD","E5FB","E641","E640","E643","E642","E644","8F50","E645","E646","E647","90BC","9776","E648","95A2","9465","E649","E64A","8CA9","8B4B","E64B","8E8B","9460","E64C","8A6F","E64D","E64F","9797","E64E","9065","E650","E651","E652","8ACF","E653","E654","E655","E656","8A70","E657","E658","E659","89F0","9047","E65A","E65B","E65C","8CBE","92F9","E65D","8C76","9075","E660","93A2","E65F","FBA3","8C50","E65E","91F5","8B4C","E661","E662","8FD7","8C8D","E663","964B","90DD","8B96","96F3","9169","E664","FBA4","9066","9290","8FD8","E665","E668","E669","8DBC","91C0","E667","8FD9","955D","E666","8E8C","8972","E66D","8C77","8E8E","8E8D","986C","E66C","E66B","9146","8B6C","9862","8A59","8FDA","FBA5","E66A","E66F","E670","E66E","8CD6","975F","8E8F","9446","E673","90BE","9261","9755","E676","8CEA","90BD","E672","E677","8CEB","E674","E675","FBA6","E671","90E0","93C7","924E","89DB","94EE","8B62","FBA7","92B2","E67A","E678","926B","90BF","8AD0","E679","907A","97C8","985F","E67B","E687","92B3","E686","FBA8","E683","E68B","E684","E680","92FA","E67E","E67C","9740","8E90","E681","E67D","FBAA","E685","8F94","8CBF","91F8","9664","8979","88E0","93A3","E689","E688","93E4","E68D","E682","E68C","E68E","8CAA","E68A","8D75","8ED3","E68F","9777","E692","E695","E693","9554","E690","8BDE","E694","E696","E69A","E697","E699","E698","FBAB","E69B","8EAF","E69D","E69C","9588","E69F","8C78","E69E","E6A0","E6A1","8B63","E3BF","8FF7","E6A2","8CEC","E6A3","FBAC","E6A4","8E5D","9DCC","E6A5","E6A6","8F51","E6A7","E6A8","E6A9","E6AA","E6AB","924A","E6AC","E6AE","E6AD","93A4","E6AF","964C","E6B0","E6B1","E6B2","E6B3","93D8","8FDB","E6B4","8D8B","98AC","E6B5","E6B6","955E","E6B7","E6BF","E6B8","E6BA","E6B9","E6BB","9665","E6BC","E6BD","E6BE","E6C0","8A4C","92E5","9589","8DE0","8D76","956E","89DD","94CC","E6C3","8AD1","90D3","E6C2","E6C7","9299","96E1","E6C5","E6C6","8B4D","E6C8","9483","91DD","94EF","935C","E6C4","9666","89EA","E6CA","9847","92C0","9864","8E91","E6C9","91AF","E6DA","9147","93F6","956F","E6CD","8E5E","8E92","8FDC","9485","8CAB","E6CC","E6CB","958A","8EBF","9371","FBAD","FBAE","E6CF","E6D0","8D77","E6CE","E6D1","E6D2","E6D4","91A1","E6D3","8AE4","E6D6","E6D5","E6D7","FBAF","E6D9","E6DB","E6DC","90D4","8ECD","E6DD","8A71","E6DE","9196","E6DF","E6E0","958B","FBB0","8B4E","E6E1","92B4","897A","E6E2","8EEF","9096","91AB","E6E5","E6E4","E6E3","E6EB","E6E9","E6E6","E6E8","E6E7","E6EA","8B97","E6EE","90D5","E6EF","8CD7","E6EC","E6ED","9848","92B5","9148","E6F0","E6F3","E6F1","E6F2","9778","93A5","E6F6","E6F4","E6F5","E6F7","E748","E6FA","E6FB","E6F9","E6F8","92FB","E740","E744","E741","E6FC","E742","E743","E74A","E745","90D6","E747","E749","E746","E74C","8F52","E74B","E74D","E74E","E751","E750","E74F","E753","E752","96F4","E755","E754","E756","E757","E759","E758","9067","E75A","8BEB","E75B","E75D","E75E","E75F","E75C","E760","8ED4","E761","8B4F","8C52","FBB2","8CAC","E762","93EE","935D","E763","E766","8EB2","E765","E764","8C79","E767","8A72","E769","8DDA","E768","E771","E76B","E76D","95E3","E76A","E76C","E770","E76E","8B50","E76F","E772","9479","97D6","8F53","E773","9741","E775","E774","E778","9760","E777","8A8D","E776","E77B","E77A","E779","9351","E77C","E77D","E77E","8D8C","8C44","E780","E781","E782","9068","E783","8EAB","E784","E785","999F","999E","E786","E390","E787","9243","904A","945F","E788","95D3","92D2","8D9E","9248","8949","9698","9076","8C7D","8BDF","95D4","E789","E78B","E78A","89DE","93F4","E78C","9497","9352","E78D","8F71","E78F","96C0","E79E","E791","E792","92C7","91DE","9197","93A6","E790","8B74","E799","E796","E7A3","93A7","9280","E793","92FC","9372","E794","E798","9080","9487","92CA","90C0","E797","91AC","91A2","E795","88A7","9841","E79A","91DF","8F54","9069","E79C","E79B","88ED","E79D","954E","E7A5","93D9","908B","9278","8BF6","E7A4","9756","895E","95D5","89DF","E79F","E7A0","E7A1","E7A2","93B9","9242","88E1","E7A6","E7A7","EAA1","91BB","E7A8","8993","916B","8CAD","9779","FBB5","E7A9","934B","9198","8ED5","E7AA","E7AD","8F85","E7AB","914A","9149","88E2","97C9","E7AF","94F0","E7B1","E7B0","E7AE","E284","8AD2","E78E","E7B3","E7B2","E7B4","9757","93DF","964D","E7B5","8ED7","E7B6","E7B7","E7B8","9340","88E8","8D78","9859","E7BC","FBB6","8C53","E7B9","E7BA","9594","8A73","9758","8BBD","9373","E7BD","E7BE","FBB8","E7BF","FBB9","9341","E7C1","E7C0","93D1","E7C2","8F55","8EDE","947A","9291","8EF0","908C","E7C3","E7C4","907C","E7C5","E7C6","E7C7","978F","8F56","E7C9","E7C8","8D79","8D93","8E5F","E7CC","8F86","E7CB","E7CA","91E7","8CED","90C1","94AE","8F58","E7CD","8FDD","E7D0","E7CE","E7CF","E7D2","E7D1","8FF8","E7D3","E7D4","E7D5","94CE","8DD1","8EDF","E7D6","E7D7","97A2","8F64","96EC","97CA","E7D8","8BE0","E7D9","FBBB","9342","FBBA","E7DC","8A98","906A","FBBC","E7DA","E7DB","92DE","FBBF","FBC0","9674","8BFA","FBBD","FBBE","E7DE","E7DF","E7DD","E7E1","FBC1","FBC3","93DD","8A62","FBC2","E7E5","E7E2","E7E4","E7E0","E86E","E7E3","97E9","8CD8","FBCA","FBC4","FBC6","E7ED","FBC5","9353","E7E8","E7EB","E7E9","E7EE","FBC7","E7EF","FBC9","E7E7","FBC8","E7F4","8994","E7E6","94AB","E7EA","8FDE","FBCB","8D7A","FBCD","FBCE","9667","8BE2","8F65","93BA","FA5F","914C","E7F2","E7EC","E7F1","96C1","92B6","E7F3","E7F0","FBCC","914B","E7F7","E7F6","E7F5","FBD2","964E","FBD6","FBD4","FBD0","FBD1","FBD5","8F9B","FBCF","E7F8","95DD","8973","9565","9292","8B98","FA65","E7FA","FBD9","8D7C","FBDC","FBDE","8E4B","E7F9","908D","908E","E840","E842","FBDD","FBDB","8FF9","FBD8","E841","E843","FBD7","8BD1","9564","8EE0","9842","E7FC","8DF6","985E","E845","E844","E846","E7FB","FA5E","93E7","9374","92D5","E84B","FBE0","9262","E847","E848","8C4C","E84A","FBDF","8CAE","E849","8FDF","8A99","E84F","8DBD","9199","92C8","FBE1","8A5A","E84D","E84E","92C1","E84C","E850","E856","FBE2","E859","E858","934C","E851","E852","E855","E857","FBE3","8BBE","E85A","E854","E853","FBE4","E85E","E85F","E860","E85D","E85C","8FE0","93A8","E85B","E864","E862","FBE5","E863","E861","91F6","E865","E866","E868","FBE6","FBE7","8AD3","E867","96F8","E873","E869","E86C","E86A","E86B","E86D","E86F","E870","E871","E874","E872","E875","E877","E876","92B7","96E5","E878","914D","E879","95C2","E87A","8A4A","895B","8AD5","FBE8","8AD4","E87B","E87C","E87D","E87E","E880","8AD6","8A74","8D7D","94B4","E882","E881","E883","897B","E886","E885","E884","E887","E88A","88C5","E888","E88C","E88B","E88E","E88D","E88F","93AC","E890","E891","E893","E892","958C","E894","E895","8DE3","E896","E897","9668","916A","88A2","91C9","E898","958D","E89B","E899","8D7E","E89A","8CC0","95C3","E89D","E89F","E89E","E8A0","8940","9077","8F9C","8AD7","E8A1","9486","E8A3","8941","E8A2","92C2","97CB","93A9","E89C","97A4","8CAF","977A","8BF7","97B2","8C47","91E0","E440","E8A4","8A4B","908F","8A75","E8A6","E8A7","E8A5","8C84","8DDB","8FE1","FBEB","8942","97D7","E8A9","E7AC","E8A8","FBEC","E8AC","E8AA","E8AB","E8AD","E8AE","97EA","E8AF","E8B0","90C7","94B9","909D","8AE5","9759","89EB","8F57","8CD9","E8B3","E8B2","8E93","E8B4","E8B1","8E47","E8B8","E5AB","99D4","9097","E8B6","97A3","93EF","894A","90E1","8EB4","95B5","895F","97EB","978B","E8B9","9364","8EF9","E8BA","E8BB","906B","E8BC","97EC","E8B7","E8BE","E8C0","E8BF","E8BD","E8C1","E8C2","919A","89E0","E8C3","96B6","E8C4","E8C5","9849","FBED","9E50","E8C6","FBEE","E8C7","E8C8","E8CC","FBEF","E8C9","E8CA","E8CB","E8CD","FBF0","FBF1","FBF2","90C2","FBF3","96F5","90C3","E8CE","94F1","E8CF","EA72","96CA","E8D0","E8D1","E8D2","8A76","E8D4","9078","E8D5","8C43","E8D6","E8DA","E8D8","E8D9","8A93","E8D7","E8DB","E8DC","88C6","E8DD","E8DE","8FE2","E8DF","8B66","E8E2","E8E1","E8E0","E691","95DA","E8E3","E8E4","E8E5","E8E6","E8E7","E8E8","8AD8","E8E9","E8EA","9442","E8EC","89B9","E8EF","E8EE","8943","8BBF","95C5","92B8","8DA0","8D80","8F87","907B","E8F1","E8F0","9761","8AE6","94D0","93DA","909C","97CC","8C7A","E8F4","E8F3","966A","93AA","896F","E8F5","E8F2","9570","978A","E8F6","E8F7","E8F9","91E8","8A7A","8A7B","E8F8","8AE7","8CB0","FBF4","8AE8","935E","97DE","FBF5","8CDA","E8FA","E8FB","E8FC","E940","E942","E941","9597","E943","E944","E945","E946","E948","E947","E949","94F2","E3CA","9048","8B51","E94A","E94B","99AA","9F5A","94D1","88F9","88B9","8E94","964F","8FFC","E94C","96DD","E94D","977B","8961","8E60","E94E","89EC","E94F","E950","E952","E953","E955","E951","E954","FBF8","8AD9","E956","E957","E958","E959","E95A","E95C","E95B","E95E","E961","E95D","E95F","E960","E962","8BC0","8EF1","E963","E964","8D81","FBFA","E965","8A5D","946E","E966","E967","9279","93E9","E968","949D","91CA","8977","8BEC","8BED","9293","E96D","8BEE","89ED","E96C","E96A","E96B","E969","E977","E96E","E96F","E970","E971","E973","E972","8F78","E974","E976","8B52","E975","919B","8CB1","E978","91CB","E979","93AB","E97A","E980","E97D","E97C","E97E","E97B","E982","FBFB","E981","E984","8BC1","E983","E985","E986","E988","E987","E989","E98B","E98A","8D9C","E98C","E98D","8A5B","E98E","E98F","9091","E990","E991","E992","E993","8D82","FBFC","FC40","E994","E995","E996","E997","E998","94AF","E99A","9545","E99B","E999","E99D","E99C","E99E","E99F","E9A0","E9A1","E9A2","E9A3","E9A4","E9A5","E9A6","E9A7","E9A8","E9A9","E9AA","E9AB","E9AC","9F54","E9AD","E2F6","8B53","8A40","8DB0","E9AF","E9AE","96A3","E9B1","E9B2","E9B0","E9B3","9682","E9B4","8B9B","9844","FC42","E9B5","FC41","E9B7","88BC","FC43","E9B8","95A9","E9B6","E9B9","E9BA","E9BB","E9BC","E9BD","968E","8E4C","8DF8","914E","FC44","E9BE","E9C1","FC45","E9BF","E9C2","8CEF","E9C0","E9C3","E9C4","E9C5","E9C9","8E49","91E2","E9CA","E9C7","E9C6","E9C8","8C7E","E9CE","E9CD","E9CC","88B1","FC46","E9D8","E9D4","E9D5","E9D1","E9D7","E9D3","8A82","986B","E9D6","E9D2","E9D0","E9CF","E9DA","E9DD","E9DC","E9DB","9568","E9D9","88F1","E9DE","E9E0","8A8F","E9CB","8956","E9E2","E9E1","E9DF","924C","9690","97D8","E9E3","E9E4","E9E5","E9E6","E9E7","92B9","E9E8","94B5","E9ED","E9E9","E9EA","9650","96C2","93CE","E9EE","E9EF","93BC","E9EC","E9EB","89A8","E9F7","E9F6","8995","E9F4","E9F3","E9F1","8A9B","E9F0","8EB0","89A7","8D83","E9FA","E9F9","E9F8","E9F5","E9FB","E9FC","EA44","EA43","EA45","894C","EA40","EA41","8D94","96B7","EA42","FC48","9651","EA4A","FC47","EA46","EA4B","EA48","EA47","8C7B","EA4C","EA4D","EA4E","EA49","E9F2","EA4F","92DF","EA53","EA54","EA52","EA51","EA57","EA50","EA55","EA56","EA59","EA58","EA5B","EA5C","EA5D","9868","EA5A","91E9","8DEB","EA5E","FC4A","EA5F","EA60","EA61","EA62","8CB2","EA63","EA64","8EAD","EA65","EA66","EA67","EA68","EA6B","EA69","985B","EA6A","97ED","EA6C","97D9","EA6D","949E","EA6E","EA70","EA71","EA6F","8D8D","96CB","9683","9BF5","9F80","969B","89A9","EA73","8B6F","EA74","EA75","EA76","FC4B","8D95","EA77","E0D2","96D9","91E1","EA78","EA7A","EA79","EA7B","EA7C","EA7D","EA7E","EA80","EA81","EA82","EA83","EA84","EA85","EA86","EA87","EA88","9343","8CDB","EA8A","916C","EA8B","EA8C","9540","EA8D","EA8E","E256","E6D8","E8EB","EA8F","EA90","EA92","EA93","EA94","97EE","EA91","EA95","EA96","EA98","EA97","EA9A","EA9B","EA99","97B4","EA9C","EA9D","E273","EA9E","FAE0","FBE9","FA90","FA9B","FA9C","FAB1","FAD8","FAE8","FAEA","FB58","FB5E","FB75","FB7D","FB7E","FB80","FB82","FB86","FB89","FB92","FB9D","FB9F","FBA0","FBA9","FBB1","FBB3","FBB4","FBB7","FBD3","FBDA","FBEA","FBF6","FBF7","FBF9","FC49","8149","FA57","8194","8190","8193","8195","FA56","8169","816A","8196","817B","8143","817C","8144","815E","824F","8250","8251","8252","8253","8254","8255","8256","8257","8258","8146","8147","8183","8181","8184","8148","8197","8260","8261","8262","8263","8264","8265","8266","8267","8268","8269","826A","826B","826C","826D","826E","826F","8270","8271","8272","8273","8274","8275","8276","8277","8278","8279","816D","815F","816E","814F","8151","814D","8281","8282","8283","8284","8285","8286","8287","8288","8289","828A","828B","828C","828D","828E","828F","8290","8291","8292","8293","8294","8295","8296","8297","8298","8299","829A","816F","8162","8170","8160","A1","A2","A3","A4","A5","A6","A7","A8","A9","AA","AB","AC","AD","AE","AF","B0","B1","B2","B3","B4","B5","B6","B7","B8","B9","BA","BB","BC","BD","BE","BF","C0","C1","C2","C3","C4","C5","C6","C7","C8","C9","CA","CB","CC","CD","CE","CF","D0","D1","D2","D3","D4","D5","D6","D7","D8","D9","DA","DB","DC","DD","DE","DF","8191","8192","81CA","8150","FA55","818F"], [128,92,167,168,176,177,180,182,215,247,913,914,915,916,917,918,919,920,921,922,923,924,925,926,927,928,929,931,932,933,934,935,936,937,945,946,947,948,949,950,951,952,953,954,955,956,957,958,959,960,961,963,964,965,966,967,968,969,1025,1040,1041,1042,1043,1044,1045,1046,1047,1048,1049,1050,1051,1052,1053,1054,1055,1056,1057,1058,1059,1060,1061,1062,1063,1064,1065,1066,1067,1068,1069,1070,1071,1072,1073,1074,1075,1076,1077,1078,1079,1080,1081,1082,1083,1084,1085,1086,1087,1088,1089,1090,1091,1092,1093,1094,1095,1096,1097,1098,1099,1100,1101,1102,1103,1105,8208,8213,8216,8217,8220,8221,8224,8225,8229,8230,8240,8242,8243,8251,126,8451,8470,8481,8491,8544,8545,8546,8547,8548,8549,8550,8551,8552,8553,8560,8561,8562,8563,8564,8565,8566,8567,8568,8569,8592,8593,8594,8595,8658,8660,8704,8706,8707,8711,8712,8715,8721,65293,8730,8733,8734,8735,8736,8741,8743,8744,8745,8746,8747,8748,8750,8756,8757,8765,8786,8800,8801,8806,8807,8810,8811,8834,8835,8838,8839,8869,8895,8978,9312,9313,9314,9315,9316,9317,9318,9319,9320,9321,9322,9323,9324,9325,9326,9327,9328,9329,9330,9331,9472,9473,9474,9475,9484,9487,9488,9491,9492,9495,9496,9499,9500,9501,9504,9507,9508,9509,9512,9515,9516,9519,9520,9523,9524,9527,9528,9531,9532,9535,9538,9547,9632,9633,9650,9651,9660,9661,9670,9671,9675,9678,9679,9711,9733,9734,9792,9794,9834,9837,9839,12288,12289,12290,12291,12293,12294,12295,12296,12297,12298,12299,12300,12301,12302,12303,12304,12305,12306,12307,12308,12309,12317,12319,12353,12354,12355,12356,12357,12358,12359,12360,12361,12362,12363,12364,12365,12366,12367,12368,12369,12370,12371,12372,12373,12374,12375,12376,12377,12378,12379,12380,12381,12382,12383,12384,12385,12386,12387,12388,12389,12390,12391,12392,12393,12394,12395,12396,12397,12398,12399,12400,12401,12402,12403,12404,12405,12406,12407,12408,12409,12410,12411,12412,12413,12414,12415,12416,12417,12418,12419,12420,12421,12422,12423,12424,12425,12426,12427,12428,12429,12430,12431,12432,12433,12434,12435,12443,12444,12445,12446,12449,12450,12451,12452,12453,12454,12455,12456,12457,12458,12459,12460,12461,12462,12463,12464,12465,12466,12467,12468,12469,12470,12471,12472,12473,12474,12475,12476,12477,12478,12479,12480,12481,12482,12483,12484,12485,12486,12487,12488,12489,12490,12491,12492,12493,12494,12495,12496,12497,12498,12499,12500,12501,12502,12503,12504,12505,12506,12507,12508,12509,12510,12511,12512,12513,12514,12515,12516,12517,12518,12519,12520,12521,12522,12523,12524,12525,12526,12527,12528,12529,12530,12531,12532,12533,12534,12539,12540,12541,12542,12849,12850,12857,12964,12965,12966,12967,12968,13059,13069,13076,13080,13090,13091,13094,13095,13099,13110,13115,13129,13130,13133,13137,13143,13179,13180,13181,13182,13198,13199,13212,13213,13214,13217,13252,13261,19968,19969,19971,19975,19976,19977,19978,19979,19981,19982,19984,19985,19988,19989,19990,19991,19992,19993,19998,20001,20006,20008,20010,20013,20017,20018,20022,20024,20025,20027,20028,20031,20034,20035,20037,20043,20045,20046,20047,20053,20054,20055,20056,20057,20061,20062,20063,20066,20081,20083,20094,20096,20098,20101,20102,20104,20105,20106,20107,20108,20110,20113,20114,20116,20117,20120,20121,20123,20124,20126,20127,20128,20129,20130,20132,20133,20134,20136,20139,20140,20141,20142,20144,20147,20150,20154,20160,20161,20162,20164,20166,20167,20170,20171,20173,20174,20175,20180,20181,20182,20183,20184,20185,20189,20190,20191,20193,20195,20196,20197,20205,20206,20208,20210,20214,20215,20219,20220,20224,20225,20227,20233,20234,20237,20238,20239,20240,20241,20250,20252,20253,20271,20272,20276,20278,20280,20281,20282,20284,20285,20291,20294,20295,20301,20302,20303,20304,20305,20307,20309,20310,20311,20313,20314,20315,20316,20317,20318,20329,20335,20336,20339,20341,20342,20347,20348,20351,20355,20358,20360,20362,20363,20365,20367,20369,20370,20372,20374,20376,20378,20379,20381,20384,20385,20395,20397,20398,20399,20405,20406,20415,20418,20419,20420,20425,20426,20429,20430,20432,20433,20436,20439,20440,20442,20443,20445,20447,20449,20451,20452,20453,20462,20463,20467,20469,20470,20472,20474,20478,20479,20485,20486,20489,20491,20493,20495,20497,20498,20500,20502,20505,20506,20510,20511,20513,20514,20515,20516,20517,20518,20520,20521,20522,20523,20524,20525,20534,20537,20544,20546,20547,20550,20551,20552,20553,20559,20560,20565,20566,20570,20572,20581,20588,20592,20594,20596,20597,20598,20600,20605,20608,20613,20621,20625,20628,20632,20633,20634,20652,20653,20658,20659,20660,20661,20663,20670,20674,20677,20681,20682,20685,20687,20689,20693,20694,20696,20698,20702,20707,20709,20711,20717,20718,20724,20725,20729,20731,20736,20737,20738,20740,20745,20754,20756,20757,20758,20760,20762,20767,20769,20778,20786,20791,20794,20795,20796,20799,20800,20801,20803,20804,20805,20806,20807,20808,20809,20810,20811,20812,20813,20814,20816,20818,20820,20826,20828,20834,20836,20837,20840,20841,20842,20843,20844,20845,20846,20849,20853,20854,20855,20856,20860,20864,20866,20869,20870,20873,20874,20876,20877,20879,20880,20881,20882,20883,20885,20886,20887,20889,20893,20896,20898,20900,20901,20902,20904,20905,20906,20907,20908,20912,20913,20914,20915,20916,20917,20918,20919,20925,20926,20932,20933,20934,20937,20939,20940,20941,20950,20955,20956,20957,20960,20961,20966,20967,20969,20970,20972,20973,20976,20977,20981,20982,20984,20985,20986,20989,20990,20992,20995,20996,20998,20999,21000,21002,21003,21006,21009,21012,21013,21015,21021,21028,21029,21031,21033,21034,21038,21040,21043,21046,21047,21048,21049,21050,21051,21059,21060,21063,21066,21067,21068,21069,21071,21076,21078,21083,21086,21091,21092,21093,21097,21098,21103,21104,21105,21106,21107,21108,21109,21117,21119,21123,21127,21128,21129,21133,21137,21138,21140,21147,21148,21151,21152,21155,21158,21161,21162,21163,21164,21165,21167,21169,21172,21173,21177,21180,21182,21184,21185,21187,21189,21191,21193,21197,21202,21205,21207,21208,21209,21211,21213,21214,21215,21216,21218,21219,21220,21222,21223,21234,21235,21237,21240,21241,21242,21246,21247,21248,21249,21250,21253,21254,21255,21256,21261,21263,21264,21269,21270,21271,21273,21274,21277,21280,21281,21283,21284,21290,21295,21297,21299,21304,21305,21306,21307,21311,21312,21313,21315,21317,21318,21319,21320,21321,21322,21325,21329,21330,21331,21332,21335,21336,21338,21340,21342,21344,21350,21353,21358,21359,21360,21361,21362,21363,21364,21365,21367,21368,21371,21375,21378,21380,21395,21398,21400,21402,21407,21408,21413,21414,21416,21417,21421,21422,21424,21426,21427,21430,21435,21442,21443,21448,21449,21450,21451,21452,21453,21454,21460,21462,21463,21465,21467,21469,21471,21473,21474,21475,21476,21477,21480,21481,21482,21483,21484,21485,21486,21487,21488,21489,21490,21491,21494,21495,21496,21498,21505,21507,21508,21512,21513,21514,21515,21516,21517,21518,21519,21520,21521,21531,21533,21535,21536,21542,21545,21547,21548,21549,21550,21558,21560,21561,21563,21564,21565,21566,21568,21570,21574,21576,21577,21578,21582,21585,21599,21608,21610,21616,21617,21619,21621,21622,21623,21627,21628,21629,21632,21636,21638,21642,21643,21644,21646,21647,21648,21650,21660,21666,21668,21669,21672,21673,21675,21676,21679,21682,21683,21688,21692,21693,21694,21696,21697,21698,21700,21703,21704,21705,21720,21729,21730,21733,21734,21736,21737,21741,21742,21746,21754,21757,21759,21764,21766,21767,21775,21776,21780,21782,21806,21807,21809,21811,21816,21817,21822,21824,21828,21829,21830,21836,21839,21843,21846,21847,21852,21853,21859,21883,21884,21886,21888,21891,21892,21894,21895,21897,21898,21899,21912,21913,21914,21916,21917,21918,21919,21927,21928,21929,21930,21931,21932,21934,21936,21942,21956,21957,21959,21972,21978,21980,21983,21987,21988,22007,22009,22013,22014,22022,22025,22036,22038,22039,22040,22043,22057,22063,22065,22066,22068,22070,22072,22082,22092,22094,22096,22107,22116,22120,22122,22123,22124,22132,22136,22138,22144,22150,22151,22154,22159,22164,22176,22178,22181,22190,22196,22198,22204,22208,22209,22210,22211,22216,22222,22225,22227,22231,22232,22234,22235,22238,22240,22243,22254,22256,22258,22259,22265,22266,22269,22271,22272,22275,22276,22280,22281,22283,22285,22287,22290,22291,22294,22296,22300,22303,22310,22311,22312,22317,22320,22327,22328,22331,22336,22338,22343,22346,22350,22351,22352,22353,22361,22369,22372,22373,22374,22377,22378,22399,22402,22408,22409,22411,22419,22432,22434,22435,22436,22442,22444,22448,22451,22464,22467,22470,22471,22472,22475,22478,22482,22483,22484,22486,22492,22495,22496,22499,22516,22519,22521,22522,22524,22528,22530,22533,22534,22538,22539,22549,22553,22557,22561,22564,22570,22575,22576,22577,22580,22581,22586,22589,22592,22593,22602,22603,22609,22610,22612,22615,22616,22617,22618,22622,22626,22633,22635,22640,22642,22645,22649,22654,22659,22661,22675,22679,22684,22686,22687,22696,22699,22702,22706,22707,22712,22713,22714,22715,22718,22721,22725,22727,22730,22732,22737,22739,22741,22743,22744,22745,22748,22750,22751,22756,22757,22763,22764,22766,22767,22768,22769,22770,22775,22777,22778,22779,22780,22781,22786,22793,22794,22795,22799,22800,22805,22806,22808,22809,22810,22811,22812,22818,22821,22823,22825,22826,22827,22828,22829,22830,22833,22834,22839,22840,22846,22852,22855,22856,22857,22862,22863,22864,22865,22867,22868,22869,22871,22872,22874,22875,22877,22880,22882,22883,22885,22887,22888,22889,22890,22892,22894,22899,22900,22904,22909,22913,22914,22915,22916,22922,22925,22931,22934,22937,22939,22941,22947,22948,22949,22952,22956,22962,22969,22970,22971,22974,22982,22985,22987,22992,22993,22995,22996,23001,23002,23004,23013,23014,23016,23018,23019,23030,23035,23039,23041,23043,23049,23057,23064,23066,23068,23071,23072,23077,23081,23087,23093,23094,23100,23104,23105,23110,23113,23130,23138,23142,23146,23148,23167,23186,23194,23195,23228,23229,23230,23233,23234,23241,23243,23244,23248,23254,23255,23265,23267,23270,23273,23290,23291,23305,23307,23308,23318,23330,23338,23340,23344,23346,23350,23358,23360,23363,23365,23376,23377,23380,23381,23382,23383,23384,23386,23387,23388,23389,23391,23395,23396,23397,23398,23401,23403,23408,23409,23411,23413,23416,23418,23424,23427,23429,23431,23432,23433,23435,23436,23437,23439,23445,23447,23448,23449,23450,23451,23452,23453,23455,23458,23459,23460,23461,23462,23470,23472,23475,23476,23477,23478,23480,23481,23487,23488,23490,23491,23492,23493,23494,23495,23497,23500,23504,23506,23507,23508,23512,23515,23517,23518,23519,23521,23522,23524,23525,23526,23527,23528,23529,23531,23532,23534,23536,23539,23541,23542,23544,23546,23550,23551,23553,23554,23556,23557,23558,23559,23560,23561,23562,23563,23565,23566,23567,23569,23571,23574,23578,23582,23584,23586,23588,23592,23597,23601,23608,23609,23610,23611,23612,23613,23614,23615,23616,23617,23621,23622,23624,23626,23627,23629,23630,23631,23632,23633,23635,23637,23646,23648,23649,23652,23653,23660,23662,23663,23665,23670,23673,23692,23696,23697,23700,23713,23718,23720,23721,23723,23724,23729,23731,23734,23735,23736,23738,23739,23740,23742,23749,23751,23769,23776,23777,23784,23785,23786,23789,23791,23792,23797,23798,23802,23803,23805,23815,23819,23822,23825,23828,23829,23830,23831,23832,23833,23834,23835,23839,23842,23847,23849,23874,23883,23884,23886,23888,23890,23891,23900,23913,23916,23917,23919,23923,23926,23938,23940,23943,23947,23948,23952,23965,23970,23980,23982,23991,23992,23993,23994,23996,23997,24009,24012,24013,24016,24018,24019,24022,24027,24029,24030,24033,24035,24037,24038,24039,24040,24043,24046,24049,24050,24051,24052,24053,24055,24059,24061,24062,24066,24067,24070,24075,24076,24081,24086,24089,24090,24091,24093,24101,24107,24109,24111,24112,24115,24118,24119,24120,24125,24128,24131,24132,24133,24135,24140,24142,24148,24149,24151,24159,24161,24162,24163,24164,24178,24179,24180,24181,24182,24184,24185,24186,24187,24188,24189,24190,24191,24193,24195,24196,24199,24202,24207,24213,24214,24215,24218,24220,24224,24230,24231,24235,24237,24245,24246,24247,24248,24257,24258,24259,24264,24265,24266,24271,24272,24275,24278,24282,24283,24285,24287,24288,24289,24290,24291,24296,24297,24300,24304,24305,24307,24308,24310,24311,24312,24314,24315,24316,24318,24319,24321,24323,24324,24329,24330,24331,24332,24333,24335,24336,24337,24339,24340,24341,24342,24343,24344,24347,24351,24353,24357,24358,24359,24361,24365,24367,24369,24372,24373,24375,24376,24380,24382,24385,24389,24392,24394,24396,24398,24401,24403,24406,24407,24409,24412,24413,24417,24418,24422,24423,24425,24426,24427,24428,24429,24432,24433,24435,24439,24441,24444,24447,24448,24449,24450,24451,24452,24453,24455,24456,24458,24459,24460,24464,24465,24466,24467,24471,24472,24473,24478,24480,24481,24488,24489,24490,24493,24494,24499,24500,24503,24505,24508,24509,24515,24517,24524,24525,24534,24535,24536,24537,24540,24541,24542,24544,24548,24555,24560,24561,24565,24568,24571,24573,24575,24590,24591,24592,24594,24597,24598,24601,24603,24604,24605,24608,24609,24613,24614,24615,24616,24617,24618,24619,24623,24625,24634,24641,24642,24643,24646,24650,24651,24653,24656,24658,24661,24665,24666,24669,24671,24672,24674,24675,24676,24677,24680,24681,24682,24683,24684,24685,24687,24688,24693,24695,24705,24707,24708,24709,24713,24714,24715,24716,24717,24722,24724,24726,24727,24730,24731,24735,24736,24739,24742,24743,24745,24746,24754,24755,24756,24757,24758,24760,24764,24765,24773,24774,24775,24785,24787,24789,24792,24794,24796,24798,24799,24800,24801,24803,24807,24808,24816,24817,24818,24819,24820,24822,24823,24825,24826,24827,24832,24833,24835,24838,24840,24841,24845,24846,24847,24849,24853,24858,24859,24863,24864,24865,24871,24872,24876,24880,24884,24887,24892,24893,24894,24895,24898,24900,24903,24904,24906,24907,24908,24909,24910,24915,24917,24920,24921,24922,24925,24927,24930,24931,24933,24935,24936,24939,24942,24943,24944,24945,24947,24948,24949,24950,24951,24958,24962,24967,24970,24974,24976,24977,24980,24982,24984,24985,24986,24996,24999,25001,25003,25004,25006,25010,25014,25018,25022,25027,25030,25031,25032,25033,25034,25035,25036,25037,25040,25059,25062,25074,25076,25078,25079,25080,25082,25084,25085,25086,25087,25088,25096,25097,25098,25100,25101,25102,25104,25105,25106,25107,25108,25110,25114,25115,25117,25118,25119,25121,25126,25130,25134,25135,25136,25138,25139,25140,25144,25147,25151,25152,25153,25159,25160,25161,25163,25165,25166,25171,25173,25176,25179,25182,25184,25187,25192,25198,25201,25206,25209,25212,25214,25215,25216,25218,25219,25220,25225,25226,25233,25234,25235,25236,25237,25238,25239,25240,25243,25244,25246,25254,25259,25260,25265,25269,25273,25275,25276,25277,25282,25285,25286,25287,25288,25289,25290,25292,25293,25295,25296,25297,25298,25299,25300,25303,25304,25305,25307,25308,25309,25312,25313,25324,25325,25326,25327,25329,25331,25333,25334,25335,25342,25343,25345,25346,25351,25352,25353,25356,25361,25369,25375,25383,25384,25387,25391,25402,25405,25406,25407,25417,25420,25421,25423,25424,25429,25431,25436,25447,25448,25449,25451,25454,25458,25462,25463,25466,25467,25472,25475,25480,25481,25484,25486,25487,25490,25494,25496,25499,25503,25504,25505,25506,25507,25509,25511,25512,25513,25514,25515,25516,25522,25524,25525,25531,25534,25536,25539,25540,25542,25545,25551,25552,25554,25558,25562,25563,25569,25571,25577,25582,25588,25589,25590,25594,25606,25613,25615,25619,25622,25623,25628,25638,25640,25644,25645,25652,25654,25658,25662,25666,25678,25688,25696,25703,25705,25711,25718,25720,25722,25731,25736,25746,25747,25749,25754,25757,25758,25764,25765,25769,25771,25773,25774,25776,25778,25785,25787,25788,25793,25794,25797,25799,25805,25806,25810,25812,25816,25818,25824,25825,25826,25827,25830,25831,25836,25839,25841,25842,25844,25846,25850,25853,25854,25856,25861,25880,25884,25885,25891,25892,25898,25899,25900,25903,25908,25909,25910,25911,25912,25913,25915,25918,25919,25925,25928,25933,25934,25935,25937,25941,25942,25943,25944,25945,25949,25950,25954,25955,25958,25964,25968,25970,25972,25973,25975,25976,25986,25987,25991,25992,25993,25996,25998,26000,26001,26007,26009,26011,26012,26015,26017,26020,26021,26023,26027,26028,26029,26031,26032,26039,26041,26044,26045,26049,26051,26052,26053,26054,26059,26060,26063,26066,26071,26073,26075,26080,26081,26082,26085,26086,26087,26088,26089,26092,26093,26097,26106,26107,26112,26114,26115,26118,26119,26121,26122,26124,26126,26127,26131,26132,26133,26140,26142,26143,26144,26148,26149,26151,26152,26157,26158,26159,26161,26164,26165,26166,26171,26172,26175,26177,26178,26179,26180,26185,26187,26191,26194,26199,26201,26205,26206,26207,26210,26212,26213,26214,26215,26216,26217,26222,26223,26224,26227,26228,26230,26234,26241,26243,26244,26247,26248,26249,26254,26257,26262,26263,26264,26265,26269,26272,26274,26278,26283,26286,26290,26292,26296,26297,26300,26302,26303,26305,26308,26311,26313,26326,26329,26330,26332,26333,26336,26342,26345,26352,26354,26355,26356,26357,26359,26360,26361,26362,26363,26364,26365,26366,26367,26368,26371,26376,26377,26379,26381,26382,26383,26388,26389,26390,26391,26395,26397,26398,26399,26406,26407,26408,26410,26411,26412,26413,26414,26417,26420,26422,26423,26424,26426,26429,26431,26433,26438,26441,26446,26447,26448,26449,26451,26454,26457,26460,26462,26463,26464,26465,26466,26467,26468,26469,26470,26474,26477,26479,26480,26481,26482,26483,26485,26487,26492,26494,26495,26501,26503,26505,26507,26508,26512,26517,26519,26522,26524,26525,26528,26529,26530,26534,26537,26543,26547,26548,26550,26551,26552,26553,26555,26560,26561,26564,26566,26570,26574,26575,26576,26577,26579,26580,26584,26586,26589,26590,26594,26596,26599,26601,26604,26606,26607,26609,26611,26612,26613,26619,26622,26623,26625,26626,26627,26628,26643,26646,26647,26654,26657,26658,26665,26666,26667,26674,26676,26680,26681,26684,26685,26688,26689,26690,26691,26692,26694,26696,26701,26702,26704,26705,26706,26707,26708,26713,26716,26717,26719,26723,26727,26740,26742,26743,26750,26751,26753,26755,26757,26765,26767,26771,26772,26775,26779,26781,26783,26784,26786,26790,26791,26792,26797,26799,26800,26801,26803,26805,26806,26809,26810,26812,26820,26822,26824,26825,26826,26827,26829,26831,26834,26836,26837,26839,26840,26842,26847,26848,26849,26851,26855,26862,26863,26866,26873,26874,26880,26881,26884,26885,26888,26891,26892,26893,26894,26895,26898,26905,26906,26907,26908,26913,26914,26915,26917,26918,26920,26922,26928,26932,26934,26937,26941,26943,26954,26963,26964,26965,26969,26970,26972,26973,26974,26976,26977,26978,26984,26986,26987,26989,26990,26991,26995,26996,26997,26999,27000,27001,27004,27005,27006,27009,27010,27018,27022,27025,27028,27029,27032,27035,27036,27040,27047,27054,27057,27058,27060,27067,27070,27071,27073,27075,27079,27082,27083,27084,27085,27086,27088,27091,27096,27097,27101,27102,27106,27111,27112,27115,27117,27122,27129,27131,27133,27135,27138,27141,27146,27147,27148,27154,27155,27156,27159,27161,27163,27166,27167,27169,27170,27171,27177,27178,27179,27182,27184,27189,27190,27192,27193,27194,27197,27204,27206,27207,27208,27211,27224,27225,27231,27233,27234,27238,27243,27250,27251,27256,27262,27263,27264,27268,27277,27278,27280,27287,27292,27296,27298,27299,27306,27308,27310,27315,27320,27323,27329,27330,27331,27345,27347,27354,27355,27358,27359,27362,27364,27368,27370,27386,27387,27396,27397,27402,27410,27414,27421,27423,27424,27425,27427,27431,27442,27447,27448,27449,27450,27453,27454,27459,27463,27465,27468,27470,27472,27475,27476,27481,27483,27487,27489,27490,27491,27492,27494,27497,27498,27503,27507,27508,27512,27513,27515,27519,27520,27523,27524,27526,27529,27530,27531,27533,27541,27542,27544,27550,27556,27562,27563,27567,27569,27570,27571,27572,27573,27575,27578,27579,27580,27583,27584,27589,27590,27595,27597,27598,27602,27603,27604,27606,27608,27611,27615,27627,27628,27631,27635,27656,27663,27665,27667,27668,27671,27675,27683,27684,27700,27703,27704,27710,27711,27712,27713,27714,27726,27728,27733,27735,27738,27740,27741,27742,27743,27744,27746,27752,27754,27759,27760,27762,27763,27770,27773,27774,27777,27778,27779,27782,27784,27788,27789,27792,27794,27795,27798,27801,27802,27803,27809,27810,27819,27822,27825,27827,27832,27833,27834,27835,27836,27837,27838,27839,27841,27844,27845,27849,27850,27852,27859,27861,27863,27865,27866,27867,27869,27873,27874,27875,27877,27880,27882,27887,27888,27889,27891,27908,27915,27916,27922,27927,27929,27931,27934,27935,27941,27945,27946,27947,27954,27955,27957,27958,27960,27963,27965,27966,27969,27972,27973,27993,27994,27996,28003,28004,28006,28009,28010,28012,28014,28015,28020,28023,28024,28025,28037,28039,28040,28044,28046,28051,28053,28054,28057,28059,28060,28076,28079,28082,28085,28088,28092,28096,28101,28102,28103,28107,28108,28111,28113,28114,28117,28120,28121,28126,28129,28132,28134,28136,28138,28139,28140,28142,28145,28146,28147,28149,28151,28152,28153,28154,28155,28156,28165,28167,28168,28169,28170,28171,28179,28181,28185,28186,28187,28189,28191,28192,28193,28195,28196,28197,28198,28199,28201,28203,28204,28205,28206,28207,28216,28217,28218,28220,28222,28227,28234,28237,28238,28246,28248,28251,28252,28255,28263,28267,28270,28271,28274,28278,28286,28287,28288,28290,28300,28303,28304,28310,28312,28316,28317,28319,28322,28325,28330,28335,28338,28342,28343,28346,28349,28351,28354,28356,28357,28361,28363,28364,28369,28371,28372,28373,28381,28382,28396,28399,28402,28404,28407,28408,28414,28415,28417,28418,28422,28425,28431,28433,28435,28436,28437,28448,28450,28451,28459,28460,28465,28466,28472,28478,28479,28481,28485,28500,28504,28507,28508,28511,28516,28518,28525,28526,28527,28528,28532,28536,28538,28540,28544,28545,28546,28548,28550,28552,28558,28561,28567,28577,28579,28580,28586,28593,28595,28597,28601,28608,28609,28610,28611,28614,28628,28629,28632,28635,28639,28640,28641,28644,28651,28652,28654,28655,28657,28659,28661,28662,28666,28670,28673,28677,28679,28681,28683,28687,28689,28693,28696,28698,28699,28701,28702,28703,28710,28711,28712,28716,28720,28722,28734,28748,28753,28760,28771,28779,28783,28784,28792,28796,28797,28805,28809,28810,28814,28818,28825,28843,28844,28845,28846,28847,28851,28856,28857,28858,28859,28872,28875,28879,28889,28893,28895,28913,28921,28925,28932,28937,28943,28948,28953,28954,28956,28961,28966,28982,28988,28998,28999,29001,29004,29006,29013,29014,29017,29020,29026,29028,29029,29030,29031,29033,29036,29038,29053,29060,29064,29066,29071,29076,29077,29081,29087,29096,29100,29105,29113,29118,29121,29123,29128,29129,29134,29136,29138,29140,29141,29143,29151,29152,29157,29158,29159,29164,29165,29166,29173,29177,29179,29180,29182,29183,29190,29197,29200,29211,29224,29226,29228,29229,29232,29234,29237,29238,29242,29243,29244,29245,29246,29247,29248,29254,29255,29256,29259,29260,29266,29272,29273,29275,29277,29279,29281,29282,29287,29289,29298,29300,29305,29309,29310,29312,29313,29314,29319,29330,29334,29344,29346,29351,29356,29359,29361,29362,29366,29369,29374,29378,29379,29380,29382,29390,29392,29394,29399,29401,29403,29408,29409,29410,29417,29420,29421,29431,29432,29433,29436,29437,29450,29462,29463,29467,29468,29469,29471,29476,29477,29481,29482,29483,29486,29487,29492,29494,29495,29502,29503,29508,29509,29518,29519,29527,29539,29544,29546,29552,29554,29557,29559,29560,29562,29563,29572,29575,29577,29579,29590,29609,29618,29619,29627,29629,29632,29634,29640,29641,29642,29645,29646,29650,29654,29662,29664,29667,29669,29674,29677,29678,29681,29685,29688,29694,29699,29701,29702,29703,29705,29730,29733,29734,29737,29738,29742,29746,29747,29748,29749,29750,29754,29759,29761,29781,29785,29786,29787,29788,29790,29791,29792,29794,29795,29796,29801,29802,29807,29808,29811,29814,29822,29827,29833,29835,29854,29855,29858,29863,29872,29885,29898,29903,29908,29916,29920,29922,29923,29926,29927,29929,29934,29936,29937,29938,29942,29943,29944,29953,29955,29956,29957,29964,29965,29966,29969,29971,29973,29976,29978,29980,29982,29983,29987,29989,29990,29992,29995,29996,29999,30000,30001,30002,30003,30007,30008,30010,30011,30012,30020,30022,30025,30026,30027,30028,30029,30031,30033,30036,30041,30042,30043,30044,30045,30048,30050,30052,30053,30054,30055,30057,30058,30059,30061,30063,30064,30067,30068,30070,30071,30072,30079,30082,30086,30087,30089,30090,30091,30094,30095,30097,30100,30106,30109,30115,30117,30123,30129,30130,30131,30133,30136,30137,30140,30141,30142,30146,30147,30149,30151,30154,30157,30162,30164,30165,30168,30169,30171,30174,30178,30179,30185,30192,30194,30195,30196,30202,30204,30206,30207,30209,30217,30219,30221,30239,30240,30241,30242,30244,30247,30256,30260,30267,30274,30278,30279,30280,30284,30290,30294,30296,30300,30305,30306,30311,30312,30313,30314,30316,30320,30322,30326,30328,30330,30331,30332,30333,30334,30336,30338,30339,30340,30342,30343,30344,30347,30350,30352,30355,30358,30361,30362,30363,30364,30366,30374,30382,30384,30388,30391,30392,30393,30394,30399,30402,30403,30406,30408,30410,30413,30418,30422,30423,30427,30428,30430,30431,30433,30435,30436,30437,30439,30442,30446,30450,30452,30456,30459,30462,30465,30468,30471,30472,30473,30475,30476,30491,30494,30495,30496,30500,30501,30502,30505,30519,30520,30522,30524,30528,30534,30535,30554,30555,30561,30563,30565,30566,30568,30571,30585,30590,30591,30603,30606,30609,30622,30624,30629,30636,30637,30640,30643,30646,30649,30651,30652,30653,30655,30663,30669,30679,30682,30683,30684,30690,30691,30693,30695,30697,30701,30702,30703,30707,30716,30722,30732,30738,30740,30741,30752,30753,30757,30758,30759,30770,30772,30778,30783,30789,30798,30813,30820,30827,30828,30831,30834,30836,30842,30844,30849,30854,30855,30860,30861,30862,30865,30867,30869,30871,30874,30883,30887,30889,30890,30895,30901,30906,30908,30910,30913,30917,30918,30922,30923,30928,30929,30932,30938,30951,30952,30956,30959,30964,30973,30977,30983,30990,30993,30994,31001,31014,31018,31019,31020,31024,31034,31036,31038,31040,31041,31047,31048,31049,31056,31059,31061,31062,31063,31066,31069,31070,31071,31072,31074,31077,31080,31085,31095,31098,31103,31104,31105,31108,31109,31114,31117,31118,31119,31124,31131,31133,31142,31143,31146,31150,31152,31155,31161,31162,31165,31166,31167,31168,31169,31177,31179,31185,31186,31189,31192,31199,31201,31203,31204,31206,31207,31209,31212,31216,31227,31232,31240,31243,31245,31246,31252,31255,31256,31257,31258,31260,31263,31264,31278,31281,31282,31287,31291,31292,31293,31294,31295,31296,31298,31299,31302,31305,31309,31310,31311,31312,31319,31329,31330,31331,31337,31339,31344,31348,31350,31353,31354,31357,31359,31361,31363,31364,31368,31378,31379,31381,31382,31383,31384,31391,31401,31402,31406,31407,31408,31414,31418,31423,31427,31428,31429,31431,31432,31434,31435,31437,31439,31441,31442,31443,31445,31449,31450,31452,31453,31455,31456,31457,31458,31459,31461,31462,31463,31466,31467,31469,31471,31472,31478,31480,31481,31482,31487,31490,31492,31494,31496,31498,31499,31503,31505,31512,31513,31515,31518,31520,31525,31526,31528,31532,31539,31541,31542,31545,31557,31558,31560,31561,31563,31564,31565,31567,31568,31569,31570,31572,31574,31581,31589,31591,31596,31598,31600,31601,31604,31605,31610,31622,31623,31627,31629,31631,31634,31636,31637,31639,31640,31641,31642,31644,31645,31646,31647,31649,31658,31661,31665,31668,31672,31680,31681,31684,31686,31687,31689,31691,31692,31695,31709,31712,31716,31717,31718,31721,31725,31731,31734,31735,31744,31751,31757,31761,31762,31763,31764,31767,31775,31777,31779,31783,31786,31787,31799,31800,31805,31806,31807,31808,31811,31820,31821,31823,31824,31828,31830,31832,31839,31840,31844,31845,31852,31859,31861,31870,31873,31874,31875,31881,31883,31885,31888,31890,31893,31895,31896,31899,31903,31905,31906,31908,31909,31911,31912,31915,31917,31918,31921,31922,31923,31929,31933,31934,31936,31938,31941,31946,31950,31954,31958,31960,31964,31966,31967,31968,31970,31975,31983,31986,31988,31990,31992,31994,31995,31998,32000,32002,32004,32005,32006,32010,32011,32013,32016,32020,32021,32023,32024,32025,32026,32027,32028,32032,32033,32034,32043,32044,32046,32047,32048,32050,32051,32053,32057,32058,32063,32066,32067,32068,32069,32070,32072,32075,32076,32078,32079,32080,32086,32091,32092,32094,32097,32098,32099,32102,32104,32110,32113,32114,32115,32117,32118,32121,32125,32137,32143,32147,32153,32154,32155,32156,32159,32160,32162,32163,32171,32172,32173,32174,32175,32176,32177,32178,32180,32181,32183,32184,32186,32187,32189,32190,32191,32199,32202,32203,32207,32209,32210,32213,32214,32216,32218,32220,32221,32222,32224,32225,32228,32232,32233,32236,32239,32242,32244,32251,32257,32260,32261,32265,32266,32267,32274,32283,32286,32287,32289,32290,32291,32294,32299,32302,32305,32306,32309,32311,32313,32314,32315,32317,32318,32321,32323,32326,32330,32331,32333,32338,32340,32341,32342,32345,32346,32349,32350,32358,32359,32361,32362,32365,32368,32377,32379,32380,32381,32383,32386,32387,32392,32393,32394,32396,32398,32399,32400,32402,32403,32404,32406,32411,32412,32566,32568,32570,32581,32583,32588,32589,32590,32592,32593,32596,32597,32600,32607,32608,32615,32616,32617,32618,32619,32622,32624,32626,32629,32631,32632,32633,32642,32643,32645,32646,32647,32648,32650,32652,32654,32660,32666,32669,32670,32673,32675,32676,32680,32681,32686,32687,32690,32694,32696,32697,32701,32705,32709,32710,32714,32716,32722,32724,32725,32736,32737,32742,32745,32747,32752,32755,32761,32763,32764,32768,32769,32771,32772,32773,32774,32779,32780,32784,32786,32789,32791,32792,32793,32796,32801,32808,32819,32822,32827,32829,32831,32838,32842,32850,32854,32856,32858,32862,32863,32865,32866,32872,32879,32880,32882,32883,32884,32886,32887,32889,32893,32894,32895,32900,32901,32902,32903,32905,32907,32908,32915,32918,32920,32922,32923,32925,32929,32930,32933,32937,32938,32940,32941,32943,32945,32946,32948,32954,32963,32964,32966,32972,32974,32982,32985,32986,32987,32989,32990,32993,32996,32997,33007,33009,33012,33016,33020,33021,33026,33029,33030,33031,33032,33033,33034,33050,33051,33059,33065,33071,33073,33075,33081,33086,33094,33099,33102,33104,33105,33107,33108,33109,33119,33125,33126,33131,33134,33136,33137,33140,33144,33145,33146,33151,33152,33154,33155,33160,33162,33167,33171,33173,33178,33180,33181,33184,33187,33188,33192,33193,33200,33203,33205,33208,33210,33213,33214,33215,33216,33218,33222,33224,33225,33229,33233,33235,33240,33241,33242,33247,33248,33251,33253,33255,33256,33258,33261,33267,33268,33274,33275,33276,33278,33281,33282,33285,33287,33288,33289,33290,33292,33293,33294,33296,33298,33302,33303,33304,33307,33308,33310,33311,33321,33322,33323,33324,33326,33331,33333,33334,33335,33336,33337,33344,33351,33368,33369,33370,33373,33375,33378,33380,33382,33384,33386,33387,33390,33391,33393,33394,33398,33399,33400,33406,33419,33421,33426,33433,33437,33439,33445,33446,33451,33452,33453,33455,33457,33459,33464,33465,33467,33469,33477,33489,33490,33491,33492,33495,33497,33499,33500,33502,33503,33505,33507,33509,33510,33511,33515,33521,33523,33524,33529,33530,33531,33537,33538,33539,33540,33541,33542,33545,33550,33558,33559,33560,33564,33571,33576,33579,33583,33585,33586,33588,33589,33590,33592,33593,33600,33605,33609,33610,33615,33616,33618,33624,33634,33651,33653,33655,33659,33660,33663,33669,33671,33673,33674,33678,33683,33686,33690,33694,33695,33696,33698,33704,33706,33707,33713,33717,33725,33729,33733,33735,33738,33740,33742,33747,33750,33752,33756,33759,33760,33769,33771,33775,33776,33777,33778,33780,33782,33783,33787,33789,33795,33796,33799,33803,33804,33805,33806,33811,33824,33826,33833,33834,33836,33841,33845,33848,33852,33853,33862,33864,33865,33870,33879,33883,33889,33890,33891,33894,33897,33899,33900,33901,33902,33903,33905,33909,33911,33913,33914,33922,33924,33931,33936,33940,33945,33948,33951,33953,33965,33970,33972,33976,33977,33979,33980,33983,33985,33988,33990,33993,33994,33995,33997,34000,34001,34006,34009,34010,34012,34028,34030,34036,34044,34047,34048,34054,34065,34067,34068,34069,34071,34072,34074,34079,34081,34086,34092,34093,34101,34109,34112,34113,34115,34120,34121,34122,34123,34126,34131,34133,34135,34136,34137,34138,34147,34152,34153,34154,34155,34157,34167,34174,34176,34180,34183,34184,34186,34192,34193,34196,34199,34201,34203,34204,34212,34214,34216,34217,34218,34219,34220,34222,34223,34224,34233,34234,34241,34249,34253,34255,34256,34261,34268,34269,34276,34277,34281,34282,34295,34297,34298,34299,34302,34306,34310,34311,34314,34315,34323,34326,34327,34330,34338,34349,34351,34352,34367,34381,34382,34384,34388,34389,34394,34396,34398,34399,34407,34411,34417,34425,34427,34442,34443,34444,34451,34453,34467,34468,34473,34474,34475,34479,34480,34486,34500,34502,34503,34505,34507,34509,34510,34516,34521,34523,34526,34527,34532,34537,34540,34541,34542,34543,34552,34553,34555,34558,34560,34562,34563,34566,34568,34569,34570,34573,34577,34578,34584,34586,34588,34597,34601,34612,34615,34619,34623,34633,34635,34636,34638,34643,34645,34647,34649,34655,34656,34659,34662,34664,34666,34670,34676,34678,34680,34687,34690,34701,34719,34722,34731,34735,34739,34746,34747,34749,34752,34756,34758,34759,34763,34768,34770,34784,34799,34802,34806,34807,34809,34811,34814,34821,34823,34829,34830,34831,34833,34837,34838,34849,34850,34851,34855,34865,34870,34873,34875,34880,34882,34884,34886,34892,34893,34898,34899,34903,34905,34907,34909,34910,34913,34914,34915,34920,34923,34928,34930,34933,34935,34941,34942,34943,34945,34946,34952,34955,34957,34962,34966,34967,34969,34974,34978,34980,34987,34990,34992,34993,34996,34997,34999,35007,35009,35010,35011,35012,35013,35023,35028,35029,35032,35033,35036,35037,35039,35041,35048,35058,35059,35060,35061,35064,35065,35068,35069,35070,35074,35076,35079,35082,35084,35088,35090,35091,35100,35101,35102,35109,35114,35115,35126,35128,35131,35137,35139,35140,35148,35149,35158,35166,35167,35168,35172,35174,35178,35181,35183,35186,35188,35191,35198,35199,35201,35203,35206,35207,35208,35210,35211,35215,35219,35222,35223,35224,35226,35233,35238,35239,35241,35242,35244,35247,35250,35251,35258,35261,35263,35264,35282,35290,35292,35293,35299,35302,35303,35316,35320,35328,35330,35331,35336,35338,35340,35342,35344,35346,35347,35350,35351,35352,35355,35357,35359,35363,35365,35370,35373,35377,35379,35380,35382,35383,35386,35387,35388,35393,35398,35400,35408,35409,35410,35412,35413,35419,35422,35424,35426,35427,35430,35433,35435,35436,35437,35438,35440,35441,35442,35443,35449,35452,35458,35460,35461,35463,35465,35468,35469,35473,35475,35477,35480,35482,35486,35488,35489,35491,35492,35493,35494,35495,35496,35500,35501,35504,35506,35513,35516,35518,35519,35522,35524,35527,35531,35532,35533,35535,35538,35542,35546,35547,35548,35550,35551,35552,35553,35554,35556,35558,35559,35563,35565,35566,35569,35571,35574,35575,35576,35578,35582,35584,35585,35586,35588,35591,35596,35598,35600,35604,35606,35607,35609,35610,35611,35613,35616,35617,35622,35624,35627,35628,35635,35641,35646,35649,35657,35660,35662,35663,35667,35670,35672,35674,35675,35676,35679,35686,35691,35692,35695,35696,35697,35698,35700,35703,35709,35711,35712,35715,35722,35724,35726,35728,35730,35731,35734,35737,35738,35895,35898,35903,35905,35910,35912,35914,35916,35918,35920,35925,35930,35937,35938,35946,35947,35948,35960,35961,35962,35964,35970,35973,35977,35978,35980,35981,35982,35988,35992,35997,35998,36000,36001,36002,36007,36008,36009,36010,36011,36012,36013,36014,36015,36016,36018,36019,36020,36022,36023,36024,36027,36028,36029,36031,36032,36033,36034,36035,36036,36039,36040,36042,36045,36046,36049,36051,36058,36059,36060,36062,36064,36066,36067,36068,36070,36074,36077,36080,36084,36090,36091,36092,36093,36100,36101,36103,36104,36106,36107,36109,36111,36112,36114,36115,36116,36118,36196,36198,36199,36203,36205,36208,36209,36211,36212,36214,36215,36225,36229,36234,36249,36259,36264,36275,36282,36286,36290,36299,36300,36303,36310,36314,36315,36317,36319,36321,36323,36328,36330,36331,36335,36339,36341,36348,36351,36360,36361,36362,36367,36368,36381,36382,36383,36394,36400,36404,36405,36418,36420,36423,36424,36425,36426,36428,36432,36437,36441,36447,36448,36451,36452,36466,36468,36470,36476,36481,36484,36485,36487,36490,36491,36493,36497,36499,36500,36505,36513,36522,36523,36524,36527,36528,36529,36542,36549,36550,36552,36554,36555,36556,36557,36559,36562,36571,36575,36578,36579,36587,36600,36603,36604,36605,36606,36611,36613,36617,36618,36620,36626,36627,36628,36629,36633,36635,36636,36637,36639,36646,36649,36650,36655,36659,36664,36665,36667,36670,36671,36674,36676,36677,36678,36681,36684,36685,36686,36695,36700,36703,36705,36706,36707,36708,36763,36764,36766,36767,36771,36775,36776,36781,36782,36783,36784,36785,36786,36791,36794,36795,36796,36799,36802,36804,36805,36814,36817,36820,36826,36834,36837,36838,36841,36842,36843,36845,36847,36848,36852,36855,36856,36857,36858,36861,36864,36865,36867,36869,36870,36875,36877,36878,36879,36880,36881,36883,36884,36885,36886,36887,36889,36890,36893,36894,36895,36896,36897,36898,36899,36903,36910,36913,36914,36917,36918,36920,36921,36924,36926,36929,36930,36933,36935,36937,36938,36939,36941,36942,36943,36944,36945,36946,36947,36948,36949,36950,36952,36953,36956,36958,36960,36961,36963,36965,36967,36968,36969,36973,36974,36975,36978,36981,36982,36983,36984,36986,36988,36989,36991,36992,36993,36994,36995,36996,36999,37001,37002,37007,37009,37027,37030,37032,37034,37039,37041,37045,37048,37057,37066,37070,37083,37086,37089,37090,37092,37096,37101,37109,37111,37117,37122,37138,37141,37145,37159,37165,37168,37170,37193,37194,37195,37196,37197,37198,37202,37204,37206,37208,37218,37219,37221,37225,37226,37228,37234,37235,37237,37239,37240,37250,37255,37257,37259,37261,37264,37266,37271,37276,37282,37284,37290,37291,37295,37300,37301,37304,37306,37312,37313,37318,37319,37320,37321,37323,37324,37325,37326,37327,37328,37329,37334,37335,37336,37338,37339,37340,37341,37342,37343,37345,37347,37348,37349,37350,37351,37357,37358,37365,37366,37372,37375,37382,37386,37389,37390,37392,37393,37396,37397,37406,37417,37420,37428,37431,37433,37434,37436,37439,37440,37444,37445,37448,37449,37451,37454,37456,37457,37463,37465,37466,37467,37470,37474,37476,37478,37479,37489,37495,37496,37502,37504,37507,37509,37512,37521,37523,37525,37526,37528,37530,37531,37532,37543,37549,37559,37561,37583,37584,37586,37587,37589,37591,37593,37600,37604,37607,37609,37610,37613,37618,37619,37624,37625,37626,37627,37628,37631,37634,37638,37647,37648,37656,37657,37658,37661,37662,37664,37665,37666,37667,37669,37670,37672,37675,37676,37678,37679,37682,37685,37690,37691,37700,37704,37707,37709,37716,37718,37719,37723,37724,37728,37740,37742,37744,37749,37756,37758,37772,37780,37782,37783,37786,37796,37799,37804,37805,37806,37808,37817,37827,37830,37832,37840,37841,37846,37847,37848,37853,37854,37857,37860,37861,37864,37880,37891,37895,37904,37907,37908,37912,37913,37914,37921,37931,37937,37941,37942,37944,37946,37953,37956,37957,37960,37969,37970,37971,37978,37979,37982,37984,37986,37994,38000,38005,38007,38012,38013,38014,38015,38017,38263,38272,38274,38275,38279,38281,38282,38283,38287,38289,38290,38291,38292,38294,38296,38297,38304,38306,38307,38308,38309,38311,38312,38317,38322,38329,38331,38332,38334,38339,38343,38346,38348,38349,38356,38357,38358,38360,38364,38369,38370,38373,38428,38433,38440,38442,38446,38447,38450,38459,38463,38464,38466,38468,38475,38476,38477,38479,38480,38491,38492,38493,38494,38495,38498,38499,38500,38501,38502,38506,38508,38512,38514,38515,38517,38518,38519,38520,38522,38525,38533,38534,38536,38538,38539,38541,38542,38543,38548,38549,38551,38552,38553,38555,38556,38557,38560,38563,38567,38568,38570,38575,38576,38577,38578,38580,38582,38583,38584,38585,38587,38588,38592,38593,38596,38597,38598,38599,38601,38603,38604,38605,38606,38609,38613,38614,38617,38619,38620,38626,38627,38632,38634,38635,38640,38642,38646,38647,38649,38651,38656,38660,38662,38663,38664,38666,38669,38670,38671,38673,38675,38678,38681,38684,38686,38692,38695,38698,38704,38706,38707,38712,38713,38715,38717,38718,38722,38723,38724,38726,38728,38729,38733,38735,38737,38738,38741,38742,38745,38748,38750,38752,38753,38754,38756,38758,38760,38761,38763,38765,38769,38772,38777,38778,38780,38785,38788,38789,38790,38795,38797,38799,38800,38808,38812,38816,38819,38822,38824,38827,38829,38835,38836,38851,38854,38856,38859,38867,38876,38893,38894,38898,38899,38901,38902,38907,38911,38913,38914,38915,38917,38918,38920,38924,38927,38928,38929,38930,38931,38935,38936,38938,38945,38948,38956,38957,38964,38967,38968,38971,38972,38973,38982,38987,38988,38989,38990,38991,38996,38997,38999,39000,39003,39006,39013,39015,39019,39023,39024,39025,39027,39028,39080,39082,39087,39089,39094,39107,39108,39110,39131,39132,39135,39138,39145,39147,39149,39150,39151,39154,39156,39164,39165,39166,39171,39173,39177,39178,39180,39184,39186,39187,39188,39192,39197,39198,39200,39201,39204,39207,39208,39212,39214,39229,39230,39234,39237,39241,39243,39244,39248,39249,39250,39253,39255,39318,39319,39320,39321,39326,39333,39336,39340,39341,39342,39347,39348,39356,39361,39364,39365,39366,39368,39376,39377,39378,39381,39384,39387,39389,39391,39394,39405,39406,39409,39410,39416,39419,39423,39425,39429,39438,39439,39442,39443,39449,39464,39467,39472,39479,39486,39488,39490,39491,39493,39501,39502,39509,39511,39514,39515,39519,39522,39524,39525,39529,39530,39531,39592,39597,39600,39608,39612,39616,39620,39631,39633,39635,39636,39640,39641,39644,39646,39647,39650,39651,39654,39658,39659,39661,39662,39663,39665,39668,39671,39675,39686,39704,39706,39711,39714,39715,39717,39719,39720,39721,39722,39726,39727,39729,39730,39739,39740,39745,39746,39747,39748,39749,39757,39758,39759,39761,39764,39768,39770,39791,39794,39796,39797,39811,39822,39823,39825,39826,39827,39830,39831,39839,39840,39848,39850,39851,39853,39854,39857,39860,39865,39867,39872,39878,39881,39882,39887,39889,39890,39892,39894,39899,39905,39906,39907,39908,39912,39920,39921,39922,39925,39936,39940,39942,39944,39945,39946,39948,39949,39952,39954,39955,39956,39957,39963,39969,39972,39973,39981,39982,39983,39984,39986,39993,39994,39995,39998,40006,40007,40008,40018,40023,40026,40032,40039,40054,40056,40165,40167,40169,40171,40172,40176,40179,40180,40182,40195,40198,40199,40200,40201,40206,40210,40213,40219,40223,40227,40230,40232,40234,40235,40236,40251,40254,40255,40257,40260,40262,40264,40272,40273,40281,40284,40285,40286,40288,40289,40292,40299,40300,40303,40304,40306,40314,40327,40329,40335,40346,40356,40361,40363,40367,40370,40372,40376,40378,40379,40385,40386,40388,40390,40399,40403,40409,40422,40429,40431,40434,40440,40441,40442,40445,40473,40474,40475,40478,40565,40568,40569,40573,40575,40577,40584,40587,40588,40593,40594,40595,40597,40599,40605,40607,40613,40614,40617,40618,40621,40632,40633,40634,40635,40636,40638,40639,40644,40652,40653,40654,40655,40656,40657,40658,40660,40664,40665,40667,40668,40669,40670,40672,40677,40680,40687,40692,40694,40695,40697,40699,40700,40701,40711,40712,40718,40723,40725,40736,40737,40748,40763,40766,40778,40779,40782,40783,40786,40788,40799,40800,40801,40802,40803,40806,40807,40810,40812,40818,40822,40823,40845,40853,40860,40861,40864,63785,63964,64014,64015,64016,64017,64018,64019,64020,64021,64022,64023,64024,64025,64026,64027,64028,64029,64030,64031,64032,64033,64034,64035,64036,64037,64038,64039,64040,64041,64042,64043,64044,64045,65281,65282,65283,65284,65285,65286,65287,65288,65289,65290,65291,65292,65293,65294,65295,65296,65297,65298,65299,65300,65301,65302,65303,65304,65305,65306,65307,65308,65309,65310,65311,65312,65313,65314,65315,65316,65317,65318,65319,65320,65321,65322,65323,65324,65325,65326,65327,65328,65329,65330,65331,65332,65333,65334,65335,65336,65337,65338,65339,65340,65341,65342,65343,65344,65345,65346,65347,65348,65349,65350,65351,65352,65353,65354,65355,65356,65357,65358,65359,65360,65361,65362,65363,65364,65365,65366,65367,65368,65369,65370,65371,65372,65373,65374,65377,65378,65379,65380,65381,65382,65383,65384,65385,65386,65387,65388,65389,65390,65391,65392,65393,65394,65395,65396,65397,65398,65399,65400,65401,65402,65403,65404,65405,65406,65407,65408,65409,65410,65411,65412,65413,65414,65415,65416,65417,65418,65419,65420,65421,65422,65423,65424,65425,65426,65427,65428,65429,65430,65431,65432,65433,65434,65435,65436,65437,65438,65439,65504,65505,65506,65507,65508,65509]],
        ];
        foreach ($series as $test) {
            foreach ($test[0] as $a => $input) {
                $class = $this->testedClass;
                $char = hex2bin($input);
                $exp = $test[1][$a];
                $s = new $class($char);
                $this->assertSame($exp, $s->nextCode(), "Sequence $input did not decode to $exp.");
                $this->assertFalse($s->nextCode(), "Sequence $input did not end after one character");
            }
        }
    }
}
